unit f_htmllistview;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, rImprovedComps, rHTMLLabel, ImgList, ComCtrls, rHintWindow,
  rHTMLListView, ExtCtrls, AppEvnts, DB, DBClient, RosiCompGlobalCfg;

type
  THTMLListViewForm = class(TForm)
    rHTMLLabel6: TrHTMLLabel;
    rGroupBox1: TrGroupBox;
    rHTMLListView1: TrHTMLListView;
    rHTMLHint1: TrHTMLHint;
    ApplicationEvents1: TApplicationEvents;
    TimerLoadFiles: TTimer;
    btAddItems: TButton;
    ClientDataSet: TClientDataSet;
    ClientDataSetName: TWideStringField;
    ClientDataSetImage: TGraphicField;
    TimerLoadDB: TTimer;
    btSort: TButton;
    ImgFav: TImage;
    ClientDataSetID: TIntegerField;
    ClientDataSetPrice: TFloatField;
    ClientDataSetDescription: TWideMemoField;
    ClientDataSetImageFileName: TWideStringField;
    ImageList1: TImageList;
    GroupBox1: TGroupBox;
    rbIcon: TRadioButton;
    rbList: TRadioButton;
    btStyle1: TButton;
    btStyle2: TButton;
    btStyle3: TButton;
    btStyle4: TButton;
    GroupBox2: TGroupBox;
    btLoadFiles: TrButton;
    btLoadDB: TrButton;
    rbImed: TRadioButton;
    rbIdle: TRadioButton;
    rbTimer: TRadioButton;
    procedure FormShow(Sender: TObject);
    procedure btLoadFilesClick(Sender: TObject);
    procedure btStyle1Click(Sender: TObject);
    procedure btStyle2Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure TimerLoadFilesTimer(Sender: TObject);
    procedure btAddItemsClick(Sender: TObject);
    procedure rHTMLListView1SelectItem(Sender: TObject; Item: TListItem; Selected: Boolean);
    procedure btLoadDBClick(Sender: TObject);
    procedure TimerLoadDBTimer(Sender: TObject);
    procedure rHTMLListView1GetItemParams(Sender: TObject; Item: TrHTMLListItem; var ItemBkColor: TColor; var ItemFrameType: TFrameType;
      var ItemFrameColor: TColor; TextFont: TFont; var TextBkColor: TColor; var TextFrameType: TFrameType; var TextFrameColor: TColor);
    procedure btSortClick(Sender: TObject);
    procedure rHTMLListView1GetItemContent(Sender: TObject; Item: TrHTMLListItem; var HTMLText: string; Image: TBitmap);
    procedure rHTMLListView1CustomDrawHTMLItemEnd(ACanvas: TCanvas; Item: TrHTMLListItem; ItemRect: TRect);
    procedure rbIconClick(Sender: TObject);
    procedure rbListClick(Sender: TObject);
    procedure btStyle3Click(Sender: TObject);
    procedure btStyle4Click(Sender: TObject);
  private
    OnIdle_LoadFileList: Boolean;
    OnIdle_LoadDBList: Boolean;
    procedure CreateDatabase;
    procedure LoadDataset;
    procedure ResetAllEvents;
  public
  end;

var
  HTMLListViewForm: THTMLListViewForm;

implementation

{$R *.dfm}
uses rtool, rdlg, rpictool;

procedure THTMLListViewForm.FormCreate(Sender: TObject);
begin
  OnIdle_LoadFileList:=false;
  OnIdle_LoadDBList:=false;
end;

procedure THTMLListViewForm.FormShow(Sender: TObject);
begin
  btStyle1.Click;
end;

procedure THTMLListViewForm.CreateDatabase;
begin
  {
    ClientDataSet.CreateDataSet;
    ClientDataSet.InsertRecord([1, 'Waterfall', 'Color Wallpaper', 120, 'FAV_0053809264.jpg']);
    ClientDataSet.InsertRecord([2, 'Mountains', 'Black and White', 56, 'file451264266022.jpg']);
    ClientDataSet.InsertRecord([3, 'Castle', 'Castle with water', 86, 'file000132701536.jpg']);
    ClientDataSet.InsertRecord([4, 'Flower', 'White flower in water', 110, 'file621250696198.jpg']);
    ClientDataSet.SaveToFile(GetAppPath + 'database.xml', dfXML);
    ClientDataSet.Close;
  }
end;

procedure THTMLListViewForm.LoadDataset;
var
  P: TPicture;
  GrType: TrGraphicType;
  fn: string;
begin
  P:=TPicture.Create;
  try
    if ClientDataSet.Active then
      ClientDataSet.Close;
    // load records from file
    ClientDataSet.CreateDataSet;
    ClientDataSet.LoadFromFile(GetAppDir + 'database.xml');
    // add images to dataset
    ClientDataSet.First;
    while not ClientDataSet.Eof do
    begin
      fn:=GetAppDir + 'Images\' + self.ClientDataSetImageFileName.AsString;
      if FileExists(fn) then
        try
          ClientDataSet.Edit;
          GrType:=rpictool.LoadPictureFromFile(P, fn);
          rpictool.WritePictureToField(self.ClientDataSetImage, P, GrType);
          ClientDataSet.Post;
        finally
          ClientDataSet.Cancel;
        end;
      ClientDataSet.Next;
    end;
  finally
    P.Free;
  end;
end;

procedure THTMLListViewForm.ResetAllEvents;
begin
  OnIdle_LoadFileList:=false;
  OnIdle_LoadDBList:=false;
  TimerLoadFiles.Enabled:=false;
  TimerLoadDB.Enabled:=false;
end;

procedure THTMLListViewForm.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
begin
  if OnIdle_LoadFileList then
  begin
    // load next file image
    Done:=rHTMLListView1.LoadFileImages;
    if Done then
      OnIdle_LoadFileList:=false;
  end;
  if OnIdle_LoadDBList then
  begin
    // load next db image
    Done:=rHTMLListView1.LoadDBImages(ClientDataSet, ClientDataSetID, ClientDataSetImage);
    if Done then
      OnIdle_LoadDBList:=false;
  end;
end;

procedure THTMLListViewForm.TimerLoadFilesTimer(Sender: TObject);
begin
  // load next file image
  rHTMLListView1.LoadFileImages;
end;

procedure THTMLListViewForm.TimerLoadDBTimer(Sender: TObject);
begin
  // load next db image
  rHTMLListView1.LoadDBImages(ClientDataSet, ClientDataSetID, ClientDataSetImage);
end;

procedure THTMLListViewForm.rbIconClick(Sender: TObject);
begin
  ResetAllEvents;
  rHTMLListView1.ViewStyle:=vsIcon;
  rHTMLListView1.ItemOptionsEx.FullItemWidth:=150;
  rHTMLListView1.ItemOptionsEx.FullItemHeight:=130;
  rHTMLListView1.ItemOptionsEx.TextLines:=2;
  rHTMLListView1.ItemOptionsEx.TextWidth:=0;
end;

procedure THTMLListViewForm.rbListClick(Sender: TObject);
begin
  ResetAllEvents;
  rHTMLListView1.ViewStyle:=vsList;
  rHTMLListView1.ItemOptionsEx.FullItemWidth:=270;
  rHTMLListView1.ItemOptionsEx.FullItemHeight:=80;
  rHTMLListView1.ItemOptionsEx.TextLines:=0;
  rHTMLListView1.ItemOptionsEx.TextWidth:=100;
end;

procedure THTMLListViewForm.rHTMLListView1CustomDrawHTMLItemEnd(ACanvas: TCanvas; Item: TrHTMLListItem; ItemRect: TRect);
begin
  // draw favorite icon to top right corner if filename starts with FAV
  if SameText(copy(ExtractFileName(VarToStrDef(Item.ItemID, '')), 1, 3), 'FAV') then
    ACanvas.Draw(ItemRect.Right - ImgFav.Picture.Bitmap.Width - 3, 2, ImgFav.Picture.Bitmap);
end;

procedure THTMLListViewForm.rHTMLListView1GetItemContent(Sender: TObject; Item: TrHTMLListItem; var HTMLText: string; Image: TBitmap);
begin
  // draw gray rectangle directly to Image for PNG files
  if not Item.ImageNotLoaded and SameText(ExtractFileExt(VarToStrDef(Item.ItemID, '')), '.png') then
  begin
    Image.Canvas.Brush.Color:=clGray;
    Image.Canvas.FillRect(Rect(5, 5, Image.Width - 5, 15));
  end;
end;

procedure THTMLListViewForm.rHTMLListView1GetItemParams(Sender: TObject; Item: TrHTMLListItem; var ItemBkColor: TColor; var ItemFrameType: TFrameType;
  var ItemFrameColor: TColor; TextFont: TFont; var TextBkColor: TColor; var TextFrameType: TFrameType; var TextFrameColor: TColor);
begin
  // change text color and frame type for PNG files
  if SameText(ExtractFileExt(VarToStrDef(Item.ItemID, '')), '.png') then
  begin
    TextFrameType:=ftSingleLine;
    TextFrameColor:=clMaroon;
    TextFont.Color:=clMaroon;
  end;
end;

procedure THTMLListViewForm.rHTMLListView1SelectItem(Sender: TObject; Item: TListItem; Selected: Boolean);
begin
  // load selected file immediatelly
  if TimerLoadFiles.Enabled then
    rHTMLListView1.LoadFileImages(false, true);
  if TimerLoadDB.Enabled then
    rHTMLListView1.LoadDBImages(ClientDataSet, ClientDataSetID, ClientDataSetImage,false,true);
end;

procedure THTMLListViewForm.btStyle1Click(Sender: TObject);
begin
  rHTMLListView1.Color:=clWindow;
  rHTMLListView1.Font.Color:=clBlack;
  rHTMLListView1.ItemOptionsEx.ItemFrameColorFocused:=clHotLight;
  rHTMLListView1.ItemOptionsEx.ItemBkColorSelected:=clSkyBlue;

  rHTMLListView1.ItemOptionsEx.ItemBkColor:=$00EBEBEB;
  rHTMLListView1.ItemOptionsEx.ItemFrameType:=ft3D;

  rHTMLListView1.ItemOptionsEx.TextFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.TextBkColor:=clInfoBk;
end;

procedure THTMLListViewForm.btStyle2Click(Sender: TObject);
begin
  rHTMLListView1.Color:=clMedGray;
  rHTMLListView1.Font.Color:=clBlack;
  rHTMLListView1.ItemOptionsEx.ItemFrameColorFocused:=clRed;
  rHTMLListView1.ItemOptionsEx.ItemBkColorSelected:=clHotLight;

  rHTMLListView1.ItemOptionsEx.ItemBkColor:=clMoneyGreen;
  rHTMLListView1.ItemOptionsEx.ItemFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.ItemFrameColor:=clGreen;

  rHTMLListView1.ItemOptionsEx.TextFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.TextBkColor:=clSilver;
end;

procedure THTMLListViewForm.btStyle3Click(Sender: TObject);
begin
  rHTMLListView1.Color:=clSilver;
  rHTMLListView1.Font.Color:=clNavy;
  rHTMLListView1.ItemOptionsEx.ItemFrameColorFocused:=clHotLight;
  rHTMLListView1.ItemOptionsEx.ItemBkColorSelected:=clNone;

  rHTMLListView1.ItemOptionsEx.ItemBkColor:=$00AAFFFF;
  rHTMLListView1.ItemOptionsEx.ItemFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.ItemFrameColor:=clWindowFrame;

  rHTMLListView1.ItemOptionsEx.TextFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.TextBkColor:=clWhite;

end;

procedure THTMLListViewForm.btStyle4Click(Sender: TObject);
begin
  rHTMLListView1.Color:=clWindow;
  rHTMLListView1.Font.Color:=clBlack;
  rHTMLListView1.ItemOptionsEx.ItemFrameColorFocused:=clRed;
  rHTMLListView1.ItemOptionsEx.ItemBkColorSelected:=clNone;

  rHTMLListView1.ItemOptionsEx.ItemBkColor:=clNone;
  rHTMLListView1.ItemOptionsEx.ItemFrameType:=ftSingleLine;
  rHTMLListView1.ItemOptionsEx.ItemFrameColor:=clSilver;

  rHTMLListView1.ItemOptionsEx.TextFrameType:=ftNone;
  rHTMLListView1.ItemOptionsEx.TextBkColor:=clNone;
end;

procedure THTMLListViewForm.btLoadFilesClick(Sender: TObject);
var
  LoadAll: Boolean;
begin
  ResetAllEvents;
  rHTMLListView1.InitLoadingItems(true);
  LoadAll:=rbImed.Checked;
  rHTMLListView1.LoadFileList(GetAppDir + 'Images', '*.png', '<B><<File>></B><BR>Size: <<W>>x<<H>>', true, false);
  rHTMLListView1.LoadFileList(GetAppDir + 'Images', '*.jpg', '<B><<File>></B><BR>Size: <<W>>x<<H>>', true, LoadAll);
  OnIdle_LoadFileList:=rbIdle.Checked;
  TimerLoadFiles.Enabled:=rbTimer.Checked;
end;

procedure THTMLListViewForm.btLoadDBClick(Sender: TObject);
var
  LoadAll: Boolean;
begin
  ResetAllEvents;
  LoadDataset;

  rHTMLListView1.InitLoadingItems(true);
  LoadAll:=rbImed.Checked;
  rHTMLListView1.LoadDBList(ClientDataSet, ClientDataSetID, ClientDataSetImage, '<B><<Name>></B><BR>Price: $ <<Price>>',
    '<B><<Name>></B><BR><<Description>><BR>Size: <<W>>x<<H>>', LoadAll);
  OnIdle_LoadDBList:=rbIdle.Checked;
  self.TimerLoadDB.Enabled:=rbTimer.Checked;
end;

procedure THTMLListViewForm.btAddItemsClick(Sender: TObject);
var
  s: string;
  B: TBitmap;
begin
  ResetAllEvents;
  // add text item and set ImageIndex from ImageList
  rHTMLListView1.InitLoadingItems(true);
  s:='Any <B>HTML</B> text with <COLOR=Blue>colors</COLOR> and <I>formats</I>';
  rHTMLListView1.AddHTMLItem(s, '<U>Formatted hint</U>:<BR>' + s, 1);

  // add item with image (not from ImageList)
  B:=TBitmap.Create;
  try
    rpictool.LoadBitmapFromFile(B, GetAppDir + 'Petals.jpg');
    rHTMLListView1.AddHTMLItemData(s, '<U>Formatted hint</U>:<BR>' + s, B, null, null, nil);
  finally
    B.Free;
  end;
end;

procedure THTMLListViewForm.btSortClick(Sender: TObject);
begin
  ResetAllEvents;
  rHTMLListView1.SortType:=stText;
  rHTMLListView1.SortType:=stNone;
end;

end.
