unit PropSaveWinXComps;

{<
  PropSaveWinXComps contains TWinXCompPropSave components for re/storing
  properties of WinX components.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(March 2020)
  @lastmod(March 2020)
  }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, PropSaveMain, RosiCompGlobalCfg, WinXCtrls;

type
  { descendant of TCustomPropSave component and
    provides re/storing of TAdv component properties }
  TWinXCompPropSave = class(TCustomPropSave)
  private
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    { Return whether re/storing of Components from parameter
      is supported by TTMSColGridsPropSave }
    function CanSave(Comp: TComponent): boolean; override;
  published
    { Define whether columns properties (fonts, color, alignmetn...) is re/stored }
  end;


procedure Register;

implementation

{$R *.res}
uses rtool;

const
  sState = 'State';

procedure Register;
begin
  RegisterComponents('Rosi PS', [TWinXCompPropSave]);
end;

// ==============================================================================
// DBGridsPropSave functions
function TWinXCompPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TToggleSwitch);
end;

procedure TWinXCompPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);

  procedure RSToggleSwitch(TS: TToggleSwitch);
  begin
    case A of
      psaSave:
        begin
          PSFile.WriteInteger(Key, sState, ord(TS.State));
        end;
      psaRead:
        begin
          TS.State:=TToggleSwitchState(PSFile.ReadInteger(Key, sState, ord(TS.State)));
        end;
      psaDelete:
        begin
        end;
    end;
  end;

begin
  if (Comp is TToggleSwitch) then
    RSToggleSwitch(Comp as TToggleSwitch);
end;


end.
