unit rDBGridSorter_DBX;

{<
  rDBGridSorter_DBX unit contains TrDBGridSorter_DBX - sorter for TSimpleDataSet
  used by rDBGrid to sort DBX datasets.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2006 - Jun 2013)
  @lastmod(Dec 2015)
}

interface

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus, Dialogs,
  StdCtrls, Forms, DB, Variants, rDBGrid,
  SqlExpr, DBClient, SimpleDS;

type
  { Sorter for TSimpleDataSet - used by rDBGrid to sort DBX datasets }
  TrDBGridSorter_DBX = class(TrDBGridCustomSorter)
  public
    function SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string): boolean; override;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}
uses RosiCompGlobalCfg;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBGridSorter_DBX]);
end;

function TrDBGridSorter_DBX.SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string)
  : boolean;
var
  DS: TSimpleDataSet;
  idxName: string;
begin
  if (DataSet is TSimpleDataSet) then
  begin
    DS:=DataSet as TSimpleDataSet;
    if Trim(SortFields)='' then
      DS.IndexName:=''
    else
    begin
      // indexname can be max 31 chars long - use hash from fields names
      idxName:=GetIndexNameHash(SortFields + '_' + DescendingFields);
      if DS.IndexDefs.IndexOf(idxName) < 0 then
        DS.AddIndex(idxName, SortFields, [], DescendingFields);
      DS.IndexDefs.Update;
      DS.IndexName:=idxName;
    end;
    Result:=true;
  end
  else
    Result:= inherited;
end;

end.
