unit rDBGrid_FilterCondForm;

{<
  rDBGrid_FilterCondForm form shows possible filter conditions for selected field type
  and allows selection of filtered values or range.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(July 2013)
  @lastmod(May 2018)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, rdlg_baseform, PropSaveMain, ImgList, rImprovedComps, StdCtrls,
  ExtCtrls, db, ComCtrls, Menus, ToolWin, DBGrids, dateutils, strutils;

type
  { form shows possible filter conditions for selected field type
    and allows selection of filtered values or range }
  TrDBGridFilterCondForm = class(TrDlgBaseForm)
    pc: TPageControl;
    tsNum: TTabSheet;
    tsText: TTabSheet;
    tsDate: TTabSheet;
    lbSignNum_nostore: TListBox;
    eValNum1: TrEditNum;
    eValNum2: TrEditNum;
    Label1: TLabel;
    lblValNum1: TLabel;
    lblValNum2: TLabel;
    Label4: TLabel;
    lbSignText_nostore: TListBox;
    Label5: TLabel;
    cbValText1_nostore: TComboBox;
    Label6: TLabel;
    lbSignDate_nostore: TListBox;
    lblValDate1: TLabel;
    eValDate1: TDateTimePicker;
    eValDate2: TDateTimePicker;
    lblValDate2: TLabel;
    ToolBarPeriod: TToolBar;
    btPeriod: TToolButton;
    pmPeriod: TPopupMenu;
    mnDay: TMenuItem;
    rDay1: TMenuItem;
    rDay2: TMenuItem;
    rDay3: TMenuItem;
    N2: TMenuItem;
    aDay1: TMenuItem;
    aDay2: TMenuItem;
    mnWeek: TMenuItem;
    rWeek1: TMenuItem;
    rWeek2: TMenuItem;
    rWeek3: TMenuItem;
    mnMonth: TMenuItem;
    rMonth1: TMenuItem;
    rMonth2: TMenuItem;
    rMonth3: TMenuItem;
    N1: TMenuItem;
    aMonth1: TMenuItem;
    aMonth2: TMenuItem;
    aMonth3: TMenuItem;
    aMonth4: TMenuItem;
    aMonth5: TMenuItem;
    aMonth6: TMenuItem;
    aMonth7: TMenuItem;
    aMonth8: TMenuItem;
    aMonth9: TMenuItem;
    aMonth10: TMenuItem;
    aMonth11: TMenuItem;
    aMonth12: TMenuItem;
    mnQuarter: TMenuItem;
    rQuarter1: TMenuItem;
    rQuarter2: TMenuItem;
    rQuarter3: TMenuItem;
    N4: TMenuItem;
    aQuarter1: TMenuItem;
    aQuarter2: TMenuItem;
    aQuarter3: TMenuItem;
    aQuarter4: TMenuItem;
    mnHalfYear: TMenuItem;
    rHalf1: TMenuItem;
    rHalf2: TMenuItem;
    rHalf3: TMenuItem;
    N6: TMenuItem;
    aHalf1: TMenuItem;
    aHalf2: TMenuItem;
    mnYear: TMenuItem;
    rYear1: TMenuItem;
    rYear2: TMenuItem;
    rYear3: TMenuItem;
    N3: TMenuItem;
    aYear1: TMenuItem;
    aYear2: TMenuItem;
    aYear3: TMenuItem;
    aYear4: TMenuItem;
    aYear5: TMenuItem;
    aYear6: TMenuItem;
    aYear7: TMenuItem;
    aYear8: TMenuItem;
    aYear9: TMenuItem;
    aYear10: TMenuItem;
    aYear11: TMenuItem;
    mnOther: TMenuItem;
    rOther1: TMenuItem;
    rOther2: TMenuItem;
    rOther3: TMenuItem;
    rOther4: TMenuItem;
    N5: TMenuItem;
    rOther5: TMenuItem;
    rOther6: TMenuItem;
    rOther7: TMenuItem;
    rOther8: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure lbSignNum_nostoreClick(Sender: TObject);
    procedure lbSignDate_nostoreClick(Sender: TObject);
    procedure mnPeriodClick(Sender: TObject);
    procedure eValNum1Change(Sender: TObject);
    procedure cbValText1_nostoreChange(Sender: TObject);
  private
    Gr: TDBGrid;
    Fld: TField;
    procedure InitMenu;
    procedure RefAll;
  public
    FilterString: string;
    procedure InitCond(DBGrid: TDBGrid; DisplayField: TField);
  end;

var
  rDBGridFilterCondForm: TrDBGridFilterCondForm;

implementation

{$R *.dfm}

uses RosiCompGlobalCfg, rstring, rtool, rdlg, rdbtool, rDBGrid;

procedure TrDBGridFilterCondForm.FormCreate(Sender: TObject);
begin
  inherited;
  InitMenu;
  eValDate1.Date:=Date;
  eValDate2.Date:=Date;
end;

procedure TrDBGridFilterCondForm.InitMenu;
var
  i, j: integer;
  mi: TMenuItem;
begin
  for i:=0 to self.pmPeriod.Items.Count - 1 do
    for j:=0 to pmPeriod.Items[i].Count - 1 do
    begin
      mi:=pmPeriod.Items[i].Items[j];
      mi.OnClick:=mnPeriodClick;
      if PosEx('aYear', mi.Name) = 1 then
        mi.Caption:=IntToStr(dateutils.YearOf(Date) + mi.Tag);
      if PosEx('aMonth', mi.Name) = 1 then
      begin
{$IFDEF DXEP}
        mi.Caption:=FormatSettings.LongMonthNames[mi.Tag];
{$ELSE}
        mi.Caption:=LongMonthNames[mi.Tag];
{$ENDIF}
      end;
    end;
end;

procedure TrDBGridFilterCondForm.FormShow(Sender: TObject);
begin
  inherited;
  RefAll;
  case pc.ActivePageIndex of
    0:
      self.eValNum1.SetFocus;
    1:
      self.cbValText1_nostore.SetFocus;
    2:
      self.eValDate1.SetFocus;
  end;
end;

procedure TrDBGridFilterCondForm.InitCond(DBGrid: TDBGrid; DisplayField: TField);
var
  i: integer;
begin
  Gr:=DBGrid;
  Fld:=DisplayField;
  Caption:=Fld.DisplayLabel;
  DlgPropSave.FormKey:=Fld.FieldName;
  FilterString:='';

  lbSignNum_nostore.ItemIndex:=0;
  lbSignText_nostore.ItemIndex:=0;
  lbSignDate_nostore.ItemIndex:=0;

  for i:=0 to pc.PageCount - 1 do
    pc.Pages[i].TabVisible:=false;
  if Fld is TNumericField then
    pc.ActivePage:=tsNum
  else
    if IsDateTimeField(Fld) then
      pc.ActivePage:=tsDate
    else
      pc.ActivePage:=tsText;
end;

procedure TrDBGridFilterCondForm.RefAll;
begin
  lblValNum2.Visible:=lbSignNum_nostore.ItemIndex = 6;
  eValNum2.Visible:=lblValNum2.Visible;
  lblValDate2.Visible:=lbSignDate_nostore.ItemIndex = 3;
  eValDate2.Visible:=lblValDate2.Visible;
  btPeriod.Visible:=lblValDate2.Visible;
end;

procedure TrDBGridFilterCondForm.lbSignNum_nostoreClick(Sender: TObject);
begin
  RefAll;
end;

procedure TrDBGridFilterCondForm.lbSignDate_nostoreClick(Sender: TObject);
begin
  RefAll;
end;

procedure TrDBGridFilterCondForm.cbValText1_nostoreChange(Sender: TObject);
begin
  if pc.ActivePage = tsText then
    btOk.Enabled:=not((lbSignText_nostore.ItemIndex > 1) and (cbValText1_nostore.Text = ''));
end;

procedure TrDBGridFilterCondForm.eValNum1Change(Sender: TObject);
begin
  if pc.ActivePage = tsNum then
    btOk.Enabled:=not((eValNum1.Text = '') or (eValNum2.Visible and (eValNum2.Text = '')));
end;

procedure TrDBGridFilterCondForm.mnPeriodClick(Sender: TObject);
var
  mi: TMenuItem;
  act: TDateTime;
begin
  mi:=Sender as TMenuItem;
  act:=Trunc(dateutils.Today);
  if PosEx('rDay', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=act + mi.Tag;
    eValDate2.DateTime:=eValDate1.DateTime;
    Exit;
  end;
  if PosEx('aDay', mi.Name) = 1 then
  begin
    if mi.Tag = 1 then
      eValDate1.DateTime:=dateutils.StartOfTheMonth(act)
    else
      eValDate1.DateTime:=dateutils.EndOfTheMonth(act);
    eValDate2.DateTime:=eValDate1.DateTime;
    Exit;
  end;
  if PosEx('rWeek', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=dateutils.StartOfTheWeek(act) + mi.Tag * 7;
    eValDate2.DateTime:=eValDate1.DateTime + 6;
    Exit;
  end;
  if PosEx('rMonth', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(dateutils.StartOfTheMonth(act), mi.Tag);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('aMonth', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), mi.Tag);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('rQuarter', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 3 * 3 + 1)), mi.Tag * 3);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 2));
    Exit;
  end;
  if PosEx('aQuarter', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(YearOf(act), (mi.Tag - 1) * 3 + 1, 1);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 2));
    Exit;
  end;
  if PosEx('rHalf', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 6 * 6 + 1)), mi.Tag * 6);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 5));
    Exit;
  end;
  if PosEx('aHalf', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(YearOf(act), (mi.Tag - 1) * 6 + 1, 1);
    eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 5));
    Exit;
  end;
  if PosEx('rYear', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncYear(dateutils.StartOfTheYear(act), mi.Tag);
    eValDate2.DateTime:=dateutils.EndOfTheYear(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('aYear', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(dateutils.YearOf(act) + mi.Tag, 1, 1);
    eValDate2.DateTime:=dateutils.EndOfTheYear(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('rOther', mi.Name) = 1 then
  begin
    case mi.Tag of
      - 1:
        eValDate1.DateTime:=dateutils.StartOfTheMonth(act);
      -2:
        eValDate1.DateTime:=dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 3 * 3 + 1));
      -3:
        eValDate1.DateTime:=dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 6 * 6 + 1));
      -4:
        eValDate1.DateTime:=dateutils.StartOfTheYear(act);
      1:
        eValDate2.DateTime:=dateutils.EndOfTheMonth(act);
      2:
        eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 3 * 3 + 1))
              , 2));
      3:
        eValDate2.DateTime:=dateutils.EndOfTheMonth(IncMonth(dateutils.RecodeMonth(dateutils.StartOfTheMonth(act), ((dateutils.MonthOf(act) - 1) div 6 * 6 + 1))
              , 5));
      4:
        eValDate2.DateTime:=dateutils.EndOfTheYear(act);
    end;
    if mi.Tag < 0 then
      eValDate2.DateTime:=act
    else
      eValDate1.DateTime:=act;
    Exit;
  end;
end;

procedure TrDBGridFilterCondForm.btOkClick(Sender: TObject);
const
  NumCond: array [0 .. 5] of string = ('=', '<>', '>', '>=', '<', '<=');
  TextCond: array [0 .. 1] of string = ('=', '<>');
  DateCond: array [0 .. 2] of string = ('=', '<', '>');
var
  FldName, s, LikeCh, Oper: string;
begin
  FldName:=rdbtool.GetQuotedFieldName(Fld.FieldName);
  if pc.ActivePage = tsNum then
  begin
    if lbSignNum_nostore.ItemIndex = 6 then
      FilterString:=MakeFilterBetweenCond(FldName, GetNumericFieldFilterValue(eValNum1.Value), GetNumericFieldFilterValue(eValNum2.Value))
    else
      FilterString:=MakeFilterCond(FldName, NumCond[lbSignNum_nostore.ItemIndex], GetNumericFieldFilterValue(eValNum1.Value));
  end;
  if pc.ActivePage = tsText then
  begin
    s:=cbValText1_nostore.Text;
    LikeCh:=TrDBGrid(Gr).OptionsEx2.Filters.GetTextFilterLikeChar;
    Oper:=TrDBGrid(Gr).OptionsEx2.Filters.GetTextFilterLikeOperator;
    case lbSignText_nostore.ItemIndex of
      2:
        s:=s + LikeCh;
      3:
        s:=LikeCh + s + LikeCh;
      else
        Oper:=TextCond[lbSignText_nostore.ItemIndex];
    end;
    if IsClientDataset(Fld.DataSet) and rdbtool.IsMemoField(Fld) then
    begin
      // use "LIKE" or "NOT FieldName LIKE" instead of "=" or "<>" for ClientDataSet and memo field
      if lbSignText_nostore.ItemIndex = 1 then
        FldName:='NOT ' + FldName;
      FilterString:=MakeFilterCond(FldName, 'LIKE', s, '''');
    end
    else
      FilterString:=MakeFilterCond(FldName, Oper, s, '''');
  end;
  if pc.ActivePage = tsDate then
  begin
    if lbSignDate_nostore.ItemIndex = 3 then
      FilterString:=MakeFilterBetweenCond(FldName, GetDateFieldFilterValue(eValDate1.Date), GetDateFieldFilterValue(eValDate2.Date), '''')
    else
      FilterString:=MakeFilterCond(FldName, DateCond[lbSignDate_nostore.ItemIndex], GetDateFieldFilterValue(eValDate1.Date), '''');
  end;
end;

end.
