unit rDBGrid_GraphicForm;

{<
  rDBGrid_GraphicForm form shows content of any graphic fields
  inside of TrDBGrid with BMP, JPG and PNG support.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Jan 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DB, DBCtrls, ComCtrls, ToolWin, StdCtrls, PropSaveMain, ImgList,
  ExtCtrls, Buttons, rpictool, rImageZoom, rDBComponents,
  PropSaveOtherComps, rImprovedComps, rdlg_baseform;

type
  { form shows content of any graphic fields
    inside of TrDBGrid with BMP, JPG and PNG support }
  TrDBGridGraphicForm = class(TrDlgBaseForm)
    ToolBar1: TToolBar;
    tbLoad: TToolButton;
    tbSave: TToolButton;
    tbCopy: TToolButton;
    tbPaste: TToolButton;
    ToolButton5: TToolButton;
    tbClear: TToolButton;
    ToolButton3: TToolButton;
    ToolButton1: TToolButton;
    DBImg: TrDBImage;
    RosiCompsPropSave1: TRosiCompsPropSave;
    procedure tbCopyClick(Sender: TObject);
    procedure tbPasteClick(Sender: TObject);
    procedure tbClearClick(Sender: TObject);
    procedure tbLoadClick(Sender: TObject);
    procedure tbSaveClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormShow(Sender: TObject);
  private
    procedure SetAll;
  public
    IsEdit: boolean;
    GrType: TrGraphicType;
  end;

var
  rDBGridGraphicForm: TrDBGridGraphicForm;

implementation

{$R *.dfm}

uses rdlg, rstring, rdbtool, clipbrd;

procedure TrDBGridGraphicForm.FormShow(Sender: TObject);
begin
  inherited;
  SetAll;
end;

procedure TrDBGridGraphicForm.SetAll;
begin
  btOk.Enabled:=IsEdit;
  tbLoad.Enabled:=IsEdit;
  tbPaste.Enabled:=IsEdit;
  tbClear.Enabled:=IsEdit;
  DBImg.ReadOnly:=not IsEdit;
end;

procedure TrDBGridGraphicForm.tbCopyClick(Sender: TObject);
begin
  DBImg.CopyImageToClipboard;
end;

procedure TrDBGridGraphicForm.tbPasteClick(Sender: TObject);
begin
  DBImg.PasteImageFromClipboard;
  SetAll;
end;

procedure TrDBGridGraphicForm.tbClearClick(Sender: TObject);
begin
  DBImg.CutImageToClipboard;
  SetAll;
end;

procedure TrDBGridGraphicForm.tbLoadClick(Sender: TObject);
begin
  DBImg.LoadFromFile('');
  SetAll;
end;

procedure TrDBGridGraphicForm.tbSaveClick(Sender: TObject);
begin
  DBImg.SaveAsFile('');
end;

procedure TrDBGridGraphicForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #27 then
    Close;
end;

end.
