unit rDBGrid_MemoForm;

{<
  rDBGrid_MemoForm form shows content of memo or long string fields inside of TrDBGrid.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Jan 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Menus, Dialogs, StdCtrls, PropSaveMain, Buttons, ExtCtrls, ImgList,
  rImprovedComps, rdlg_baseform, ComCtrls;

type
  { form shows content of memo or long string fields inside of TrDBGrid }
  TrDBGridMemoForm = class(TrDlgBaseForm)
    MainMenu: TMainMenu;
    mnOptions: TMenuItem;
    mnFont: TMenuItem;
    mnEnter: TMenuItem;
    mnTab: TMenuItem;
    mnWrap: TMenuItem;
    FontDlg: TFontDialog;
    mnWindows: TMenuItem;
    Memo: TrMemoEx;
    rRichEdit: TrRichEdit;
    procedure mnFontClick(Sender: TObject);
    procedure mnEnterClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormShow(Sender: TObject);
    procedure DlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure DlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure btOkClick(Sender: TObject);
  private
    procedure SetAll;
  public
    IsEdit: boolean;
    IsRichEdit: boolean;
    Text: string;
  end;

var
  rDBGridMemoForm: TrDBGridMemoForm;

implementation

{$R *.dfm}

uses rstring, rdlg;

procedure TrDBGridMemoForm.SetAll;
begin
  btOk.Enabled:=IsEdit;
  Memo.ReadOnly:=not IsEdit;
  Memo.WantReturns:=mnEnter.Checked;
  Memo.WantTabs:=mnTab.Checked;
  Memo.WordWrap:=mnWrap.Checked;
  if Memo.WordWrap then
    Memo.ScrollBars:=ssVertical
  else
    Memo.ScrollBars:=ssBoth;
end;

procedure TrDBGridMemoForm.FormShow(Sender: TObject);
begin
  inherited;
  rRichEdit.Visible:=IsRichEdit;
  Memo.Visible:=not IsRichEdit;
  mnOptions.Visible:=Memo.Visible;
  mnWindows.Visible:=Memo.Visible;
  if IsRichEdit then
  begin
    rRichEdit.RtfText:=Text;
    rRichEdit.SelStart:=Length(Text);
    rRichEdit.SelLength:=1;
  end
  else
  begin
    Memo.Lines.Text:=Text;
    Memo.SelStart:=Length(Text);
    Memo.SelLength:=1;
  end;
  SetAll;
end;

procedure TrDBGridMemoForm.mnFontClick(Sender: TObject);
begin
  FontDlg.Font:=Memo.Font;
  if FontDlg.Execute then
    Memo.Font:=FontDlg.Font;
end;

procedure TrDBGridMemoForm.mnEnterClick(Sender: TObject);
begin
  SetAll;
end;

procedure TrDBGridMemoForm.btOkClick(Sender: TObject);
begin
  if IsRichEdit then
    Text:=rRichEdit.RtfText
  else
    Text:=Memo.Lines.Text;
  ModalResult:=mrOk;
end;

procedure TrDBGridMemoForm.DlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.ReadFont(Key + '.Memo', Memo.Font);
end;

procedure TrDBGridMemoForm.DlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteFont(Key + '.Memo', Memo.Font);
end;

procedure TrDBGridMemoForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #27 then
    Close;
end;

end.
