unit rDBMemoGrid;

{<
  rDBMemoGrid unit contains TrDBMemoGrid - enhanced TrStringGridEd component which provides
  function for editing content of db memo field in grid form.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(May 2008)
  @lastmod(Oct 2014)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, StdCtrls, Grids,
  Forms, Menus, DB, DBCtrls, rDBComponents, Clipbrd, ComCtrls, Math,
  rStringGridEd, RosiCompGlobalCfg;

type
  { bound TrDBMemo used by TrDBMemoGrid }
  TBoundDBMemo = class(TrDBMemo)
  private
    CanChangeBounds: boolean;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    procedure SetBounds(ALeft: Integer; ATop: Integer; AWidth: Integer; AHeight: Integer); override; {< @exclude }
    procedure SetBoundsEx(ALeft: Integer; ATop: Integer; AWidth: Integer; AHeight: Integer); {< @exclude }
  end;

  { display method used by TrDBMemoGrid - as grid, as memo, ar automatic selection according content }
  TMemoGridOpt = (mgoGrid, mgoMemo, mgoAuto);
  { additional option used by TrDBMemoGrid }
  TMemoGridOptEx = (mgeForceDeleteRow, mgeForceMoveRow, mgeForceStartEdit, mgeAutoPost, mgeFirstLineAsTitles);
  { additional options used by TrDBMemoGrid }
  TMemoGridOptsEx = set of TMemoGridOptEx;
  { cell selection option used by TrDBMemoGrid }
  TMemoGridSelection = (mgsRows, mgsColumns, mgsCells);
  { cell selection options used by TrDBMemoGrid }
  TMemoGridSelections = set of TMemoGridSelection;
  { event used for custom definition of display method }
  TMemoCheckFormat = procedure(Sender: TObject; Lines: TStrings; var ShowGrid: boolean) of object;

  { enhanced TrStringGridEd component which provides
    function for editing content of db memo field in grid form }
  TrDBMemoGrid = class(TrStringGridEd)
  private
    FDataLink: TFieldDataLink;
    FDBEditLabel: TDBBoundLabel;
    FDBMemo: TBoundDBMemo;
    FLines: TStrings;
    FReadOnlyColor: TColor;
    FVisibleAll: boolean;
    FMemoGridOpt: TMemoGridOpt;
    FMemoGridOptsEx: TMemoGridOptsEx;
    FMemoGridSelections: TMemoGridSelections;
    FMemoCheckFormat: TMemoCheckFormat;
    FColTitles: TStrings;
    FColSeparator: string;
    FEditInfoChar: string;
    FPopupMenu: TPopupMenu;
    FFocused: boolean;
    isChanging: boolean;
    isSelLock: boolean;
    FSep: string;
    SelectedRows: array of boolean;
    SelectedCols: array of boolean;
    SelectedCells: array of array of boolean;
    FSelectedColor: TColor;
    FReadOnly: boolean;
    FNoNewLines: boolean;
    FRowFilterColumn: Integer;
    FRowFilterValues: TStrings;
    procedure SetColTitles(Value: TStrings);
    procedure SetColSeparator(const Value: string);
    procedure SetRowFilterColumn(Value: Integer);
    procedure SetRowFilterValues(Value: TStrings);
    procedure SetMemoGridOpt(Value: TMemoGridOpt);
    procedure UpdateColTitles;
    procedure UpdateRowTitles;
    procedure OnPopupMenu(Sender: TObject);
    procedure PopupClick(Sender: TObject);
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure SetSubComp;
    function CanEdit(StartEd: boolean; force: boolean): boolean;
    function StartEdit(force: boolean): boolean;
    function GetStartDataRow: Integer;
  protected
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetEditText(ACol, ARow: Longint; const Value: string); override;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetParent(AParent: TWinControl); override;
    procedure SetFocused(Value: boolean);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure SetVisibleAll(Value: boolean);
    function CheckFormat(Lines: TStrings): boolean;
    procedure GetCellParams(Col, Row: Longint; AFont: TFont; var Background: TColor; var Highlight: boolean; var DataImageDef: TDataImageDef); override;
    procedure EditorKeyPress(var Key: Char); override;
    function CheckLastRow: boolean; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Loaded; override; {< @exclude }
    procedure SetBounds(ALeft: Integer; ATop: Integer; AWidth: Integer; AHeight: Integer); override; {< @exclude }
    { reload values from database memo field to grid }
    procedure LoadFromField;
    { store values from grid to database memo field }
    procedure StoreToField;
    { set one value of grid, dataset has to be in editing mode }
    function SetValue(Col, Row: Integer; const Value: string): boolean;
    { get one value of grid }
    function GetValue(Col, Row: Integer): string;
    { delete one line, dataset has to be in editing mode }
    function DeleteLine(Index: Integer): boolean;
    { move one line, dataset has to be in editing mode }
    function MoveLine(Index: Integer; Delta: Integer): boolean;
    { copy whole table, result can be inserted to clipboard or to text }
    function CopyTable(const Separator: string): string;
    { insert whole table, values are separated by separator }
    function InsertTable(const Values, Separator: string): boolean;
    { append whole table, values are separated by separator }
    function AppendTable(const Values, Separator: string): boolean;
    { select or deselect one row }
    procedure SelectRow(ARow: Integer; Selected: boolean);
    { indicate whether row is selected }
    function IsRowSelected(ARow: Integer): boolean;
    { returns count of selected rows }
    function SelectedRowCount: Integer;
    { select or deselect one column }
    procedure SelectCol(ACol: Integer; Selected: boolean);
    { indicate whether column is selected }
    function IsColSelected(ACol: Integer): boolean;
    { returns count of selected columns }
    function SelectedColCount: Integer;
    { select or deselect one cell }
    procedure SelectOneCell(ARow, ACol: Integer; Selected: boolean);
    { indicate whether cell is selected }
    function IsCellSelected(ARow, ACol: Integer): boolean;
    { returns count of selected cells }
    function SelectedCellCount: Integer;
    { returns content of selection }
    function GetSelection(const ValueSeparator, LineSeparator: string): string;
    { returns content of line }
    function GetActiveLine(const ValueSeparator: string): string;
    { indicates whether selection exists }
    function SelectionExists: boolean;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { defines whether data is in read only mode - no modification can be done }
    property ReadOnly: boolean read FReadOnly write FReadOnly default false;
    { defines bounded label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { defines titles of columns }
    property ColTitles: TStrings read FColTitles write SetColTitles;
    { defines separator for values in one row }
    property ColSeparator: string read FColSeparator write SetColSeparator;
    { defines char in cell 0,0 if editing }
    property EditInfoChar: string read FEditInfoChar write FEditInfoChar;
    { defines visibility of all component including bound label,
      used instead of property Visible }
    property VisibleAll: boolean read FVisibleAll write SetVisibleAll;
    { defines color of memo when dataset is not editing }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor;
    { defines showing of grid, memo or auto according data structure }
    property MemoGridOpt: TMemoGridOpt read FMemoGridOpt write SetMemoGridOpt;
    { defines what can be selected - rows, columns, cells }
    property MemoGridSelections: TMemoGridSelections read FMemoGridSelections write FMemoGridSelections;
    { event for user defined check of data format }
    property OnMemoCheckFormat: TMemoCheckFormat read FMemoCheckFormat write FMemoCheckFormat;
    { defines background color for selected cells }
    property SelectedColor: TColor read FSelectedColor write FSelectedColor;
    { prohibit adding of new lines }
    property NoNewLines: boolean read FNoNewLines write FNoNewLines;
    { set of extended options - enable deleting of rows if readonly }
    property MemoGridOptsEx: TMemoGridOptsEx read FMemoGridOptsEx write FMemoGridOptsEx;
    { column index which is compared for filtered rows (filter is working only for ReadOnly memogrid) }
    property RowFilterColumn: Integer read FRowFilterColumn write SetRowFilterColumn default 0;
    { displayed values for filtered rows (filter is working only for ReadOnly memogrid) }
    property RowFilterValues: TStrings read FRowFilterValues write SetRowFilterValues;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rdlg, rtool, rdbtool;

const
  cDelRow = 1;
  cMoveUpRow = 2;
  cMoveDownRow = 3;
  cCopyCell = 11;
  cPasteCell = 12;
  cCopyTable = 21;
  cPasteTable = 22;
  cAppendTable = 23;
  cClearTable = 24;
  cSelectRow = 31;
  cSelectAll = 32;
  cSelectCol = 33;
  cSelectCell = 34;
  cDeselectAll = 35;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBMemoGrid]);
end;

// ==============================================================================
// TBoundDBMemo functions
constructor TBoundDBMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  CanChangeBounds:=false;
end;

procedure TBoundDBMemo.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if CanChangeBounds then
    inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

procedure TBoundDBMemo.SetBoundsEx(ALeft, ATop, AWidth, AHeight: Integer);
begin
  CanChangeBounds:=true;
  try
    SetBounds(ALeft, ATop, AWidth, AHeight);
  finally
    CanChangeBounds:=false;
  end;
end;

// ==============================================================================
// TrDBMemoGrid functions
constructor TrDBMemoGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  // set StringGrid props
  isChanging:=false;
  isSelLock:=false;
  Options:=[goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goColSizing];
  FMemoGridSelections:=[mgsRows, mgsColumns, mgsCells];
  // DefaultDrawing:=false;
  ShowHint:=true;
  FSelectedColor:=clOlive;
  DefaultRowHeight:=17;
  DefaultColWidth:=120;
  self.ColWidths[0]:=16;

  // set DataLink props and events
  FDataLink:=TFieldDataLink.Create;
  FDataLink.Control:=self;
  FDataLink.OnActiveChange:=ActiveChange;
  FDataLink.OnDataChange:=DataChange;
  FDataLink.OnEditingChange:=EditingChange;
  FDataLink.OnUpdateData:=UpdateData;

  // set EditLabel
  FDBEditLabel:=TDBBoundLabel.Create(self);
  FDBEditLabel.FreeNotification(self);

  // set DBMemo
  if not(csDesigning in ComponentState) then
  begin
    FDBMemo:=TBoundDBMemo.Create(self);
    FDBMemo.FreeNotification(self);
    FDBMemo.Visible:=false;
  end;

  // set other props
  FLines:=TStringList.Create;
  FColTitles:=TStringList.Create;
  FColSeparator:='<TAB>';
  FEditInfoChar:='^';
  FVisibleAll:=true;
  FReadOnlyColor:=clBtnFace;

  FRowFilterValues:=TStringList.Create;
  FRowFilterColumn:=0;

  // create Popup
  FPopupMenu:=TPopupMenu.Create(self);
  FPopupMenu.Name:='DBMemoTablePopup';
  FPopupMenu.OnPopup:=OnPopupMenu;
  PopupMenu:=FPopupMenu;
  // popup items
  FPopupMenu.Items.Add(CreateMenuItem(self, mnDelRow, cDelRow, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnMoveUpRow, cMoveUpRow, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnMoveDownRow, cMoveDownRow, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, '-', 0, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnCopyCell, cCopyCell, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnPasteCell, cPasteCell, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, '-', 0, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnCopyTable, cCopyTable, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnPasteTable, cPasteTable, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnAppendTable, cAppendTable, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnClearTable, cClearTable, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, '-', 0, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnSelectRow, '', ShortCut(ord('R'), [ssCtrl]), cSelectRow, -1, PopupClick, nil));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnSelectAll, cSelectAll, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnSelectCol, '', ShortCut(ord('T'), [ssCtrl]), cSelectCol, -1, PopupClick, nil));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnSelectCell, '', ShortCut(ord('E'), [ssCtrl]), cSelectCell, -1, PopupClick, nil));
  FPopupMenu.Items.Add(CreateMenuItem(self, mnDeselectAll, cDeselectAll, PopupClick));

  // Refresh
  UpdateColTitles;
  UpdateRowTitles;
  Columns[0].Alignment:=taCenter;
  Columns[0].TitleAlignment:=taCenter;
  FNoNewLines:=false;
end;

destructor TrDBMemoGrid.Destroy;
begin
  // free datalink
  FreeAndNil(FDataLink);
  // free all other
  FreeAndNil(FLines);
  FreeAndNil(FColTitles);
  FreeAndNil(FRowFilterValues);
  inherited Destroy;
end;

procedure TrDBMemoGrid.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource);
  FDBEditLabel.StoreVisible;
  SetVisibleAll(FVisibleAll);
  UpdateColTitles;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBMemoGrid.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FDBMemo) then
    FDBMemo.Enabled:=Enabled;
  Invalidate;
end;

procedure TrDBMemoGrid.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  // if Assigned(FDBEditLabel) then FDBEditLabel.ChangeVisible(Visible);
  // if Assigned(FDBMemo) then FDBMemo.Visible:=Visible;
end;

procedure TrDBMemoGrid.SetVisibleAll(Value: boolean);
begin
  FVisibleAll:=Value;
  if csLoading in ComponentState then
    Exit;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(FVisibleAll);
  DataChange(self);
end;

procedure TrDBMemoGrid.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and (AComponent = DataSource) then
    DataSource:=nil;
  if (AComponent = FDBEditLabel) and (Operation = opRemove) then
    FDBEditLabel:=nil;
  if (AComponent = FDBMemo) and (Operation = opRemove) then
    FDBMemo:=nil;
end;

procedure TrDBMemoGrid.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if Assigned(FDBMemo) then
    FDBMemo.SetBoundsEx(Left, Top, Width, Height);
end;

procedure TrDBMemoGrid.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FDBMemo) then
    FDBMemo.Parent:=AParent;
end;

function TrDBMemoGrid.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBMemoGrid.SetDataSource(Value: TDataSource);
begin
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLink.DataSource:=Value;
    SetSubComp;
    if not(csDesigning in ComponentState) then
    begin
      EditingChange(self);
      DataChange(self);
    end;
  end;
  if Value <> nil then
    Value.FreeNotification(self);

  if not(csDesigning in ComponentState) then
  begin
    // set DBMemo instead Grid
    FDBMemo.Parent:=Parent;
    FDBMemo.SetBoundsEx(Left, Top, Width, Height);
    FDBMemo.SetDataSource(DataSource, DataField);
    FDBMemo.Color:=Color;
    FDBMemo.ReadOnlyColor:=ReadOnlyColor;
    FDBMemo.TabStop:=TabStop;
    FDBMemo.TabOrder:=TabOrder;
    FDBMemo.ParentFont:=ParentFont;
    FDBMemo.ParentColor:=ParentColor;
    FDBMemo.ParentShowHint:=ParentShowHint;
    FDBMemo.Hint:=Hint;
    FDBMemo.Font.Assign(Font);
    FDBMemo.DBEditLabel.ChangeVisible(false);
  end;
end;

function TrDBMemoGrid.GetDataField: string;
begin
  Result:=FDataLink.FieldName;
end;

procedure TrDBMemoGrid.SetDataField(const Value: string);
begin
  FDataLink.FieldName:=Value;
  SetSubComp;
end;

function TrDBMemoGrid.GetStartDataRow: Integer;
begin
  if mgeFirstLineAsTitles in FMemoGridOptsEx then
    Result:=0
  else
    Result:=1;
end;

function TrDBMemoGrid.CanEdit(StartEd: boolean; force: boolean): boolean;
begin
  // check if dataset can be edited
  Result:=false;
  if (not ReadOnly or force) and Assigned(FDataLink.Field) then
    if not FDataLink.Field.ReadOnly then
      if IsEditing(DataSource) then
        Result:=true
      else
        if DataSource.AutoEdit or (mgeForceStartEdit in FMemoGridOptsEx) then
          try
            isSelLock:=true;
            try
              if StartEd then
              begin
                DataSource.DataSet.Edit;
                Result:=IsEditing(DataSource);
              end
              else
                Result:=true;
            finally
              isSelLock:=false;
            end;
          except
          end;
  if Result and not force then
    Options:=Options + [goEditing]
  else
    Options:=Options - [goEditing];
end;

function TrDBMemoGrid.StartEdit(force: boolean): boolean;
begin
  Result:=CanEdit(true, force);
  if not Result then
    DlgW(sEditRecordErr);
end;

procedure TrDBMemoGrid.ActiveChange(Sender: TObject);
begin
  SetSubComp;
  CanEdit(false, false);
  LoadFromField;
end;

procedure TrDBMemoGrid.DataChange(Sender: TObject);
begin
  if isChanging then
    Exit;
  if not isSelLock then
  begin
    SetLength(SelectedRows, 0);
    SetLength(SelectedCols, 0);
    SetLength(SelectedCells, 0, 0);
  end;
  if FDataLink.Field <> nil then
  begin
    LoadFromField;
  end;
end;

procedure TrDBMemoGrid.EditingChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  FDataLink.Reset;
  CanEdit(false, false);
  if IsEditing(DataSource) then
    Cells[0, 0]:=FEditInfoChar
  else
    Cells[0, 0]:='';
end;

procedure TrDBMemoGrid.UpdateData(Sender: TObject);
begin
  if FDataLink.Field <> nil then
  begin
    StoreToField;
  end;
end;

procedure TrDBMemoGrid.SetFocused(Value: boolean);
begin
  if FFocused <> Value then
  begin
    FFocused:=Value;
    isSelLock:=true;
    try
      FDataLink.Reset;
    finally
      isSelLock:=false;
    end;
  end;
end;

procedure TrDBMemoGrid.CMEnter(var Message: TCMEnter);
begin
  SetFocused(true);
  inherited;
end;

procedure TrDBMemoGrid.CMExit(var Message: TCMExit);
begin
  try
    if IsEditing(DataSource) then
      FDataLink.UpdateRecord;
  except
    SetFocus;
    raise ;
  end;
  SetFocused(false);
  DoExit;
end;

procedure TrDBMemoGrid.SetEditText(ACol, ARow: Longint; const Value: string);
begin
  isChanging:=Cells[ACol, ARow] <> Value;
  try
    inherited SetEditText(ACol, ARow, Value);
    if isChanging then
    begin
      CanEdit(true, false);
      FDataLink.Modified;
    end;
  finally
    isChanging:=false;
  end;
end;

function TrDBMemoGrid.SetValue(Col, Row: Integer; const Value: string): boolean;
begin
  Result:=false;
  if IsEditing(DataSource) then
  begin
    if (Col < FixedCols) or (Row < FixedRows) or (Col > ColCount - 1) or (Row > RowCount - 1) then
      Exit;
    FDataLink.Modified;
    SetEditText(Col, Row, Value);
    FDataLink.UpdateRecord;
    Result:=true;
  end;
end;

function TrDBMemoGrid.GetValue(Col, Row: Integer): string;
begin
  if (Col < 1) or (Row < 1) or (Col > ColCount - 1) or (Row > RowCount - 1) then
    Result:=''
  else
    Result:=Cells[Col, Row];
end;

procedure TrDBMemoGrid.SetSubComp;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
end;

procedure TrDBMemoGrid.SetColTitles(Value: TStrings);
begin
  if Value.Count = 0 then
    Value.Add(' ');
  FColTitles.Assign(Value);
  ColCount:=Value.Count + 1;
  UpdateColTitles;
end;

procedure TrDBMemoGrid.SetColSeparator(const Value: string);
begin
  if Value = '' then
    Exit;
  FColSeparator:=Value;
  FSep:=rtool.ReplaceStr('<TAB>', #9, FColSeparator, true, false);
  LoadFromField;
end;

procedure TrDBMemoGrid.SetMemoGridOpt(Value: TMemoGridOpt);
begin
  FMemoGridOpt:=Value;
  LoadFromField;
end;

procedure TrDBMemoGrid.SetRowFilterColumn(Value: Integer);
begin
  FRowFilterColumn:=Value;
  LoadFromField;
end;

procedure TrDBMemoGrid.SetRowFilterValues(Value: TStrings);
begin
  FRowFilterValues.Assign(Value);
  LoadFromField;
end;

procedure TrDBMemoGrid.UpdateColTitles;
var
  i: Integer;
begin
  for i:=1 to ColCount - 1 do
  begin
    if FColTitles.Count < i then
      FColTitles.Add(' ');
    Cells[i, 0]:=FColTitles[i - 1];
  end;
  Invalidate;
end;

procedure TrDBMemoGrid.UpdateRowTitles;
var
  i: Integer;
begin
  for i:=1 to RowCount - 1 do
    Cells[0, i]:=IntToStr(i);
  if not FNoNewLines then
    Cells[0, RowCount - 1]:='*';
  Invalidate;
end;

function TrDBMemoGrid.CheckLastRow: boolean;
begin
  Result:= inherited CheckLastRow;
  if Length(SelectedRows) > RowCount - 1 then
    SetLength(SelectedRows, RowCount - 1);
  if Length(SelectedCells) > RowCount - 1 then
    SetLength(SelectedCells, RowCount - 1, ColCount);
  UpdateRowTitles;
end;

function TrDBMemoGrid.CheckFormat(Lines: TStrings): boolean;
var
  i: Integer;
begin
  if FMemoGridOpt = mgoGrid then
    Result:=true
  else
    if FMemoGridOpt = mgoMemo then
      Result:=false
    else
    begin
      Result:=true;
      for i:=0 to Lines.Count - 1 do
        Result:=Result and (rtool.AnsiPosEx(FSep, Lines[i], 1, false) > 0);
      if Assigned(FMemoCheckFormat) then
        FMemoCheckFormat(self, Lines, Result);
    end;
end;

procedure TrDBMemoGrid.LoadFromField;
var
  i, j, r: Integer;
  s, p: string;
  b, filtered: boolean;
begin
  if csLoading in ComponentState then
    Exit;
  for i:=1 to RowCount - 1 do
    Rows[i].Clear;
  if not FDataLink.Active then
  begin
    RowCount:=2;
    UpdateRowTitles;
    for j:=1 to ColCount - 1 do
      Cells[j, 1]:='';
    Exit;
  end;

  FLines.Text:=FDataLink.Field.AsString;
  if not(csDesigning in ComponentState) then
  begin
    b:=CheckFormat(FLines);
    Visible:=FVisibleAll and b;
    FDBMemo.Visible:=FVisibleAll and not b;
  end;

  if (mgeFirstLineAsTitles in FMemoGridOptsEx) then
  begin
    FColTitles.Clear;
    if FLines.Count > 0 then
    begin
      s:=FLines[0];
      FLines.Delete(0);
      while s <> '' do
      begin
        i:=rtool.AnsiPosEx(FSep, s);
        if i > 0 then
        begin
          p:=copy(s, 1, i - 1);
          s:=copy(s, i + Length(FSep), Length(s));
        end
        else
        begin
          p:=s;
          s:='';
        end;
        FColTitles.Add(p);
      end;
    end;
    ColCount:=Max(FColTitles.Count + 1, 2);
    UpdateColTitles;
  end;

  if (FNoNewLines) and (FLines.Count > 0) then
    RowCount:=FLines.Count + 1
  else
    RowCount:=FLines.Count + 2;
  UpdateRowTitles;
  r:=0;
  for i:=0 to FLines.Count - 1 do
  begin
    // check filter
    if FReadOnly and (FRowFilterColumn in [1 .. ColCount - 1]) and (FRowFilterValues.Count > 0) then
    begin
      s:=rtool.FindStrPart(FLines[i], FSep, FRowFilterColumn, true);
      filtered:=FRowFilterValues.IndexOf(s) >= 0;
    end
    else
      filtered:=true;
    if filtered then
    begin
      Inc(r);
      for j:=1 to ColCount - 1 do
      begin
        s:=rtool.FindStrPart(FLines[i], FSep, j, true);
        if Cells[j, r] <> s then
          Cells[j, r]:=s;
      end;
    end
    else
      if RowCount > 2 then
        RowCount:=RowCount - 1;
  end;
  if not FNoNewLines then
    for j:=1 to ColCount - 1 do
      Cells[j, RowCount - 1]:='';
end;

procedure TrDBMemoGrid.StoreToField;
var
  i, j: Integer;
  s: string;
begin
  FLines.Clear;
  for i:=GetStartDataRow to RowCount - 1 do
  begin
    if (not FNoNewLines) and (i = RowCount - 1) then
      Continue;
    s:='';
    for j:=1 to ColCount - 1 do
    begin
      if j > 1 then
        s:=s + FSep;
      s:=s + Cells[j, i];
    end;
    FLines.Add(s);
  end;
  s:=FLines.Text;
  if copy(s, Length(s) - 1, 2) = cStrEOL then
    s:=copy(s, 1, Length(s) - 2);
  isSelLock:=true;
  try
    FDataLink.Field.AsString:=s;
  finally
    isSelLock:=false;
  end;
end;

procedure TrDBMemoGrid.GetCellParams(Col, Row: Longint; AFont: TFont; var Background: TColor; var Highlight: boolean; var DataImageDef: TDataImageDef);
begin
  // show colored selection
  if (Col = 0) and (Row <= High(SelectedRows)) and (SelectedRows[Row]) then
    Background:=FSelectedColor;
  if (Row = 0) and (Col <= High(SelectedCols)) and (SelectedCols[Col]) then
    Background:=FSelectedColor;
  if (Row > 0) and (Col > 0) and (Row <= High(SelectedCells)) and (Col <= High(SelectedCells[Row])) and (SelectedCells[Row, Col]) then
    Background:=FSelectedColor;
  inherited GetCellParams(Col, Row, AFont, Background, Highlight, DataImageDef);
end;

procedure TrDBMemoGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // if Button = mbRight then
  // Button:=mbLeft;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TrDBMemoGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell: TGridCoord;
begin
  inherited MouseUp(Button, Shift, X, Y);
  Cell:=MouseCoord(X, Y);
  if (Button = mbLeft) and (Cell.X = 0) and (Cell.Y > 0) then
  begin
    SelectRow(Cell.Y, not IsRowSelected(Cell.Y));
    self.Row:=Cell.Y;
  end;
  if (Button = mbLeft) and (Cell.Y = 0) and (Cell.X > 0) then
  begin
    SelectCol(Cell.X, not IsColSelected(Cell.X));
    self.Col:=Cell.X;
  end;
  if (Shift = [ssCtrl]) and (Button = mbLeft) and (Cell.Y > 0) and (Cell.X > 0) then
    SelectOneCell(Cell.Y, Cell.X, not IsCellSelected(Cell.Y, Cell.X));
end;

procedure TrDBMemoGrid.OnPopupMenu(Sender: TObject);
var
  i: Integer;
  ena: boolean;
begin
  ena:=CanEdit(false, false);
  for i:=0 to FPopupMenu.Items.Count - 1 do
  begin
    case FPopupMenu.Items[i].Tag of
      cDelRow:
        FPopupMenu.Items[i].Enabled:=ena or (mgeForceDeleteRow in FMemoGridOptsEx);
      cMoveUpRow, cMoveDownRow:
        FPopupMenu.Items[i].Enabled:=ena or (mgeForceMoveRow in FMemoGridOptsEx);
      cPasteCell, cPasteTable, cAppendTable, cClearTable:
        FPopupMenu.Items[i].Enabled:=ena;
      cSelectRow, cSelectAll:
        FPopupMenu.Items[i].Visible:=mgsRows in FMemoGridSelections;
      cSelectCol:
        FPopupMenu.Items[i].Visible:=mgsColumns in FMemoGridSelections;
      cSelectCell:
        FPopupMenu.Items[i].Visible:=mgsCells in FMemoGridSelections;
      cDeselectAll:
        FPopupMenu.Items[i].Visible:=([mgsRows, mgsColumns, mgsCells] * FMemoGridSelections) <> [];
    end;
  end;
end;

procedure TrDBMemoGrid.PopupClick(Sender: TObject);
var
  i: Integer;
begin
  case TMenuItem(Sender).Tag of
    cDelRow:
      if (not FNoNewLines or (mgeForceDeleteRow in FMemoGridOptsEx)) and StartEdit(mgeForceStartEdit in FMemoGridOptsEx) and (DlgQ(sDelRecordQ) = mrYes) then
        DeleteLine(Row - GetStartDataRow);
    cMoveUpRow:
      if StartEdit(mgeForceMoveRow in FMemoGridOptsEx) then
        MoveLine(Row - GetStartDataRow, -1);
    cMoveDownRow:
      if StartEdit(mgeForceMoveRow in FMemoGridOptsEx) then
        MoveLine(Row - GetStartDataRow, 1);
    cCopyCell:
      Clipboard.AsText:=Cells[Col, Row];
    cPasteCell:
      begin
        if StartEdit(false) and Clipboard.HasFormat(CF_TEXT) then
        begin
          if Assigned(InplaceEditor) and InplaceEditor.Visible then
          begin
            InplaceEditor.PasteFromClipboard;
            InplaceEditor.SelStart:=Length(InplaceEditor.Text);
            SetEditText(Col, Row, InplaceEditor.Text);
          end
          else
            SetValue(Col, Row, Clipboard.AsText);
        end;
      end;
    cCopyTable:
      Clipboard.AsText:=CopyTable(#9);
    cPasteTable:
      if (not FNoNewLines) and StartEdit(false) and Clipboard.HasFormat(CF_TEXT) then
      begin
        if Assigned(InplaceEditor) and InplaceEditor.Visible then
          InplaceEditor.Hide;
        InsertTable(Clipboard.AsText, #9);
      end;
    cAppendTable:
      if (not FNoNewLines) and StartEdit(false) and Clipboard.HasFormat(CF_TEXT) then
      begin
        if Assigned(InplaceEditor) and InplaceEditor.Visible then
          InplaceEditor.Hide;
        AppendTable(Clipboard.AsText, #9);
      end;
    cClearTable:
      if (not FNoNewLines) and StartEdit(false) then
        InsertTable('', #9);
    cSelectRow:
      SelectRow(Row, not IsRowSelected(Row));
    cSelectAll:
      for i:=1 to RowCount - 1 do
        SelectRow(i, true);
    cSelectCol:
      SelectCol(Col, not IsColSelected(Col));
    cSelectCell:
      SelectOneCell(Row, Col, not IsCellSelected(Row, Col));
    cDeselectAll:
      begin
        SetLength(SelectedRows, 0);
        SetLength(SelectedCols, 0);
        SetLength(SelectedCells, 0, 0);
        Invalidate;
      end;
  end;
end;

function TrDBMemoGrid.DeleteLine(Index: Integer): boolean;
begin
  Result:=false;
  if IsEditing(DataSource) then
  begin
    StoreToField;
    FLines.Text:=FDataLink.Field.AsString;
    if (Index < 0) or (Index >= FLines.Count) then
      Exit;
    FLines.Delete(Index);
    FDataLink.Field.AsString:=FLines.Text;
    LoadFromField;
    Result:=true;
  end;
  if (mgeAutoPost in FMemoGridOptsEx) then
    DataSource.DataSet.Post;
end;

function TrDBMemoGrid.MoveLine(Index: Integer; Delta: Integer): boolean;
begin
  Result:=false;
  if IsEditing(DataSource) then
  begin
    StoreToField;
    FLines.Text:=FDataLink.Field.AsString;
    if (Index < 0) or (Index >= FLines.Count) or (Index + Delta < 0) or (Index + Delta >= FLines.Count) then
      Exit;
    FLines.Move(Index, Index + Delta);
    Row:=Index + Delta + GetStartDataRow;
    FDataLink.Field.AsString:=FLines.Text;
    LoadFromField;
    Result:=true;
  end;
  if (mgeAutoPost in FMemoGridOptsEx) then
    DataSource.DataSet.Post;
end;

function TrDBMemoGrid.CopyTable(const Separator: string): string;
begin
  if IsEditing(DataSource) then
    StoreToField;
  Result:=rtool.ReplaceStr(FSep, Separator, FDataLink.Field.AsString, true, false);
end;

function TrDBMemoGrid.InsertTable(const Values, Separator: string): boolean;
begin
  Result:=false;
  if IsEditing(DataSource) then
  begin
    FDataLink.Field.AsString:=rtool.ReplaceStr(Separator, FSep, Values, true, false);
    Row:=1;
    Col:=1;
    LoadFromField;
    Result:=true;
  end;
end;

function TrDBMemoGrid.AppendTable(const Values, Separator: string): boolean;
var
  imp: TStrings;
begin
  Result:=false;
  if IsEditing(DataSource) then
  begin
    StoreToField;
    FLines.Text:=FDataLink.Field.AsString;
    imp:=TStringList.Create;
    try
      imp.Text:=rtool.ReplaceStr(Separator, FSep, Values, true, false);
      FLines.AddStrings(imp);
    finally
      FreeAndNil(imp);
    end;
    FDataLink.Field.AsString:=FLines.Text;
    Row:=1;
    Col:=1;
    LoadFromField;
    Result:=true;
  end;
end;

procedure TrDBMemoGrid.SelectRow(ARow: Integer; Selected: boolean);
var
  i, oldmax: Integer;
begin
  if (ARow < 1) or (ARow >= RowCount) or not(mgsRows in FMemoGridSelections) then
    Exit;
  if ARow > High(SelectedRows) then
  begin
    oldmax:= High(SelectedRows);
    SetLength(SelectedRows, RowCount);
    for i:=oldmax + 1 to High(SelectedRows) do
      SelectedRows[i]:=false;
  end;
  SelectedRows[ARow]:=Selected;
  // deselect cells
  SetLength(SelectedCells, 0, 0);
  Invalidate;
end;

procedure TrDBMemoGrid.SelectCol(ACol: Integer; Selected: boolean);
var
  i, oldmax: Integer;
begin
  if (ACol < 1) or (ACol >= ColCount) or not(mgsColumns in FMemoGridSelections) then
    Exit;
  if ACol > High(SelectedCols) then
  begin
    oldmax:= High(SelectedCols);
    SetLength(SelectedCols, ColCount);
    for i:=oldmax + 1 to High(SelectedCols) do
      SelectedCols[i]:=false;
  end;
  SelectedCols[ACol]:=Selected;
  // deselect cells
  SetLength(SelectedCells, 0, 0);
  Invalidate;
end;

procedure TrDBMemoGrid.SelectOneCell(ARow, ACol: Integer; Selected: boolean);
var
  i, j, oldrmax, oldcmax: Integer;
begin
  if (ARow < 1) or (ARow >= RowCount) or (ACol < 1) or (ACol >= ColCount) or not(mgsCells in FMemoGridSelections) then
    Exit;
  if ARow > High(SelectedCells) then
  begin
    oldrmax:= High(SelectedCells);
    SetLength(SelectedCells, RowCount);
    for i:=0 to oldrmax do
    begin
      oldcmax:= High(SelectedCells[i]);
      SetLength(SelectedCells[i], ColCount);
      for j:=oldcmax + 1 to High(SelectedCells[i]) do
        SelectedCells[i, j]:=false;
    end;
    for i:=oldrmax + 1 to High(SelectedCells) do
    begin
      SetLength(SelectedCells[i], ColCount);
      for j:=0 to High(SelectedCells[i]) do
        SelectedCells[i, j]:=false;
    end;
  end;
  SelectedCells[ARow, ACol]:=Selected;
  // deselect rows and cols
  SetLength(SelectedRows, 0);
  SetLength(SelectedCols, 0);
  Invalidate;
end;

function TrDBMemoGrid.IsRowSelected(ARow: Integer): boolean;
begin
  Result:=false;
  if (ARow >= 1) and (ARow <= High(SelectedRows)) then
    Result:=SelectedRows[ARow];
end;

function TrDBMemoGrid.IsColSelected(ACol: Integer): boolean;
begin
  Result:=false;
  if (ACol >= 1) and (ACol <= High(SelectedCols)) then
    Result:=SelectedCols[ACol];
end;

function TrDBMemoGrid.IsCellSelected(ARow, ACol: Integer): boolean;
begin
  Result:=false;
  if (ARow >= 1) and (ARow <= High(SelectedCells)) and (ACol >= 1) and (ACol <= High(SelectedCells[ARow])) then
    Result:=SelectedCells[ARow, ACol];
end;

function TrDBMemoGrid.SelectedRowCount: Integer;
var
  i: Integer;
begin
  Result:=0;
  for i:=1 to High(SelectedRows) do
    if SelectedRows[i] then
      Inc(Result);
end;

function TrDBMemoGrid.SelectedColCount: Integer;
var
  i: Integer;
begin
  Result:=0;
  for i:=1 to High(SelectedCols) do
    if SelectedCols[i] then
      Inc(Result);
end;

function TrDBMemoGrid.SelectedCellCount: Integer;
var
  i, j: Integer;
begin
  Result:=0;
  for i:=1 to High(SelectedCells) do
    for j:=1 to High(SelectedCells[i]) do
      if SelectedCells[i, j] then
        Inc(Result);
end;

function TrDBMemoGrid.GetSelection(const ValueSeparator, LineSeparator: string): string;
var
  i, j: Integer;
  s: string;
  sepval, sepline: boolean;
begin
  Result:='';
  sepline:=false;
  if (SelectedRowCount > 0) or (SelectedColCount > 0) then
  begin
    for i:=1 to RowCount - 1 do
      if IsRowSelected(i) or (SelectedRowCount = 0) then
      begin
        s:='';
        sepval:=false;
        for j:=1 to ColCount - 1 do
          if IsColSelected(j) or (SelectedColCount = 0) then
          begin
            if sepval then
              s:=s + ValueSeparator;
            s:=s + Cells[j, i];
            sepval:=true;
          end;
        if sepline then
          Result:=Result + LineSeparator;
        Result:=Result + s;
        sepline:=true;
      end;
  end
  else
  begin
    for i:=1 to RowCount - 1 do
    begin
      s:='';
      sepval:=false;
      for j:=1 to ColCount - 1 do
        if IsCellSelected(i, j) then
        begin
          if sepval then
            s:=s + ValueSeparator;
          s:=s + Cells[j, i];
          sepval:=true;
        end;
      if sepval then
      begin
        if sepline then
          Result:=Result + LineSeparator;
        Result:=Result + s;
        sepline:=true;
      end;
    end;
  end;
end;

function TrDBMemoGrid.GetActiveLine(const ValueSeparator: string): string;
var
  j: Integer;
begin
  Result:='';
  if (Row < 1) or (Row >= RowCount) then
    Exit;
  for j:=1 to ColCount - 1 do
  begin
    if j > 1 then
      Result:=Result + ValueSeparator;
    Result:=Result + Cells[j, Row];
  end;
end;

function TrDBMemoGrid.SelectionExists: boolean;
begin
  Result:=(SelectedRowCount > 0) or (SelectedColCount > 0) or (SelectedCellCount > 0);
end;

procedure TrDBMemoGrid.EditorKeyPress(var Key: Char);
begin
  if Key = #10 then
    Key:=#0; // prevent Ctrl+Enter
  inherited KeyPress(Key);
end;

end.
