unit rDBRecView;

{<
  rDBrecView unit contains TrDBRecView - enhanced TCustomDrawGrid component
  which provides functionality for displaying and editing of one dataset record.
  TrDBrecView can display all type of fields including memo and grapics fields.
  Each field can be displayed with diferent font style and color.
  TrDBRecView can be linked to DBGrid or to DataSource.

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2007)
  @lastmod(May 2018)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Grids, StdCtrls,
  Forms, ComCtrls, DB, DBGrids, DBConsts, IniFiles, Dialogs, Clipbrd,
{$IFDEF DXE2P} UITypes, Types, Themes, {$ENDIF}
{$IFDEF RDBGRID} rDBGrid, {$ENDIF} RosiCompGlobalCfg;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;
  TrDBRecView = class;

  { list of OptionsEx used by TrDBRecView }
  TGridOptionEx = (goDataRight, goNumberRight, goAddColon, goAutoHeight, goUseGridPickList, goUseGridEllipsis, goOnlyPickListValues, goRTFToPlainText, goEnterNewLineMemo);
  { set of OptionsEx used by TrDBRecView }
  TGridOptionsEx = set of TGridOptionEx;

  { list of dialog tab visibility used by design time dialog for TrDBRecView }
  TGridDlgVisibleTab = (vtFieldsSetup, vtSpecFields, vtFieldsAsLink, vtPickLists, vtEllipsisFields);
  { set of dialog tab visibility used by design time dialog for TrDBRecView }
  TGridDlgVisibleTabs = set of TGridDlgVisibleTab;
  { event type for TrStringGridEd cell parameters used by TrDBRecView }
  TGetCellParamsEvent = procedure(Sender: TObject; Col, Row: longint; AFont: TFont; var Background: TColor; var Highlight: Boolean) of object;

  { @exclude }
  TrDBRecordLink = class(TDataLink)
  private
    FRecView: TrDBRecView;
  protected
    procedure ActiveChanged; override;
    procedure DataSetChanged; override;
    procedure EditingChanged; override;
    procedure UpdateData; override;
  public
    constructor Create(RecView: TrDBRecView);
  end;

  { @exclude }
  TrDBRecViewInplaceEdit = class(TInplaceEditList)
  protected
    procedure DblClick; override;
  public
    property Font;
  end;

  { Enhanced TCustomDrawGrid component which provides functionality
    for displaying and editing of one dataset record.
    can display all type of fields including memo and grapics fields.
    Each field can be displayed with diferent font style and color.
    Can be linked to DBGrid or to DataSource }
  TrDBRecView = class(TCustomDrawGrid)
  private
    FDBGrid: TCustomDBGrid;
    FCustomGlyphs: TrCustomGlyphs;
    FFldList: TStringList;
    FOnlyVisibleField: Boolean;
    FOnlySimpleField: Boolean;
    FTitleAlignment: TAlignment;
    FOptionsEx: TGridOptionsEx;
    FDateTimePicker: TDateTimePicker;
    FFields: TStringList;
    FSpecFields: TStringList;
    FBooleanFields: TStringList;
    FEllipsisFields: TStrings;
    FSpecFont: TFont;
    FFieldsAsLink: TStringList;
    isLink: Boolean;
    FOnLinkClick: TLinkClickEvent;
    MoveCoord: TGridCoord;
    LastEditedField: TField;
    LastEditedValue: string;
    GridPickList: TStringList;
    FLookupText: string;
    FPickLists: TStringList;
    FPickListsIni: TMemIniFile;
    FGridDlgVisibleTabs: TGridDlgVisibleTabs;
    FOnGetCellParams: TGetCellParamsEvent;
    FOnEditButtonClick: TNotifyEvent;
    procedure SetCustomGlyphs(const Value: TrCustomGlyphs);
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure SetDBGrid(Value: TCustomDBGrid);
    procedure SetOnlyVisibleField(Value: Boolean);
    procedure SetOnlySimpleField(Value: Boolean);
    procedure SetTitleAlignment(Value: TAlignment);
    procedure SetOptionsEx(Value: TGridOptionsEx);
    procedure SetFields(Value: TStringList);
    procedure SetBooleanFields(Value: TStringList);
    procedure SetEllipsisFields(Value: TStrings);
    procedure SetSpecFields(Value: TStringList);
    procedure SetSpecFont(Value: TFont);
    procedure SetFieldsAsLink(Value: TStringList);
    procedure SetPickLists(Value: TStringList);
    function HasParentNormalState: Boolean;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure CMMouseLeave(var AMsg: TMessage); message CM_MOUSELEAVE;
    function GetRowField(ARow: integer): TField;
    function GetFocusedField: TField;
    function GetRowCaption(ARow: integer): string;
    function GetRowDisplayText(ARow: integer): string;
    function GetFieldText(F: TField): string;
    function GetFocusedText: string;
    function GetAlignment(F: TField): integer;
    procedure GetPickList(ACol, ARow: integer; Items: TStrings);
    procedure SetFont(const FieldName: string; AFont: TFont);
    procedure EditButtonClick(Sender: TObject);
    procedure UpdateData;
    procedure CMExit(var Message: TMessage); message CM_EXIT;
    procedure DateTimePickerClose(Sender: TObject);
    procedure DateTimePickerChange(Sender: TObject);
    procedure ResetLastEdited;
    function IsBoolEx(F: TField): Boolean;
    function IsMemoEx(F: TField; TextLenAsMemo: integer): Boolean;
    function IsGraphicEx(F: TField): Boolean;
    function IsDateEx(F: TField): Boolean;
    function IsEllipsisEx(F: TField): Boolean;
    function ChangeBool(Reset: Boolean): Boolean;
    function BoolIsTrue(F: TField): Boolean;
  protected
    FDataLink: TrDBRecordLink;
    FScaleRatio: double;
    procedure Loaded; override; {< @exclude }
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure ColWidthsChanged; override;
    procedure Paint; override;
    procedure DrawCell(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    function SelectCell(ACol, ARow: longint): Boolean; override;
    procedure Click; override;
    procedure DblClick; override;
    function CreateEditor: TInplaceEdit; override;
    function CanEdit(StartEdit: Boolean): Boolean;
    function CanEditShow: Boolean; override;
    function CanEditAcceptKey(Key: Char): Boolean; override;
    function GetEditLimit: integer; override;
    function GetEditMask(ACol, ARow: longint): string; override;
    function GetEditStyle(ACol, ARow: longint): TEditStyle; override;
    function GetEditText(ACol, ARow: longint): string; override;
    procedure SetEditText(ACol, ARow: longint; const Value: string); override;
    procedure GetCellParams(ACol, ARow: longint; AFont: TFont; var Background: TColor; var Highlight: Boolean); virtual;
    procedure ShowDateTimePicker(F: TField; Rect: TRect); virtual;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: integer); override; {< @exclude }
    procedure SetCols; {< @exclude }
    procedure SetRows; {< @exclude }
    procedure RefreshData; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { indicates whether Grid is linked to any Dataset }
    function IsDatasetLinked: Boolean;
    { indicates whether linked Dataset is active }
    function IsActive: Boolean;
    { returns count of fields }
    function GetFieldCount: integer;
    { returns count of rows = GetFieldCount }
    function GetRowCount: integer;
    { returns field object according row index }
    function GetField(Idx: integer): TField;
    { returns field caption according row index }
    function GetCaption(Idx: integer): string;
    { returns field value according row index }
    function GetValue(Idx: integer): string;
    { returns used font according row index }
    procedure GetFontStyle(Idx: integer; AFont: TFont; KeepFontNameSize: Boolean);
    { returns data alignment according row index }
    function GetDataAlignment(Idx: integer): integer;
    { returns fields from selected row }
    property FocusedField: TField read GetFocusedField;
    { returns current text from selected row }
    property FocusedText: string read GetFocusedText;
    { returns count of visible columns = 2 }
    property VisibleColCount;
    { returns count of visible rows }
    property VisibleRowCount;
    { returns link to inplace editor }
    property InplaceEditor;
  published
    { link to datasource, if DBGrid is set then Datasource is set automatically from DBGrid }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { link to DBGrid, Rows (fields) and DataSource are loaded automatically from DBGrid columns }
    property DBGrid: TCustomDBGrid read FDBGrid write SetDBGrid;
    { define link to rCustomGlyphs component which can define all glyphs }
    property CustomGlyphs: TrCustomGlyphs read FCustomGlyphs write SetCustomGlyphs;
    { define if not visible field is hiden }
    property OnlyVisibleField: Boolean read FOnlyVisibleField write SetOnlyVisibleField default false;
    { define if not simple field (memo and graphics) is hiden }
    property OnlySimpleField: Boolean read FOnlySimpleField write SetOnlySimpleField default false;
    { Define list of fields as link, can be set easily by design time editor }
    property FieldsAsLink: TStringList read FFieldsAsLink write SetFieldsAsLink;
    { define allingment of title column }
    property TitleAlignment: TAlignment read FTitleAlignment write SetTitleAlignment default taLeftJustify;
    { define extension options }
    property OptionsEx: TGridOptionsEx read FOptionsEx write SetOptionsEx default[];
    { define witch tabs from design time dialog is visible in run time }
    property GridDlgVisibleTabs: TGridDlgVisibleTabs read FGridDlgVisibleTabs write FGridDlgVisibleTabs default[vtFieldsSetup, vtSpecFields, vtFieldsAsLink,
      vtPickLists, vtEllipsisFields];
    { define list of fields and captions }
    property Fields: TStringList read FFields write SetFields;
    { define fields which is shown as boolean }
    property BooleanFields: TStringList read FBooleanFields write SetBooleanFields;
    { define fields with ellipsis button }
    property EllipsisFields: TStrings read FEllipsisFields write SetEllipsisFields;
    { list of fields (fieldname) which is shown with SpecFont;
      for modification of font can be used following format:
      FieldName=list of format strings separated by ;
      format string are:  SF (standard font) N (normal) B (bold) I (italic) U (underline) colorname fontsize
      e.g.   FirstName=B;Blue;10   LastName=B;U;Red;8 }
    property SpecFields: TStringList read FSpecFields write SetSpecFields;
    { define font for fields listed in property SpecFields }
    property SpecFont: TFont read FSpecFont write SetSpecFont;
    { define fields and their picklist items, can be set easily by design time editor }
    property PickLists: TStringList read FPickLists write SetPickLists;
    { Returns TDateTimePicker for setting of their props }
    property DateTimePicker: TDateTimePicker read FDateTimePicker;
    { event is called after click to link }
    property OnLinkClick: TLinkClickEvent read FOnLinkClick write FOnLinkClick;
    { event for user defined setting of individual cell params - font, color }
    property OnGetCellParams: TGetCellParamsEvent read FOnGetCellParams write FOnGetCellParams;
    { event called after click to Ellipsis button }
    property OnEditButtonClick: TNotifyEvent read FOnEditButtonClick write FOnEditButtonClick;

    property Align;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelKind;
    property BevelOuter;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property Color;
    // property ColCount;
    property Constraints;
    property Ctl3D;
    property DefaultColWidth;
    property DefaultRowHeight;
    property DefaultDrawing;
    property DoubleBuffered;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DrawingStyle;
    property Enabled;
    property FixedColor;
    // property FixedCols;
    // property RowCount;
    // property FixedRows;
    property Font;
    property GradientEndColor;
    property GradientStartColor;
    property GridLineWidth;
    property Options;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentDoubleBuffered;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property Touch;
    property Visible;
    property OnClick;
    property OnColumnMoved;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawCell;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnFixedCellClick;
    property OnGesture;
    property OnGetEditMask;
    property OnGetEditText;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseActivate;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnRowMoved;
    property OnSelectCell;
    property OnSetEditText;
    property OnStartDock;
    property OnStartDrag;
    property OnTopLeftChanged;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rdbtool, rpictool;

const
  DateTimePickerMinWidth = 84;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 50 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBRecView]);
end;

// ==============================================================================
// TrDBRecordLink functions
constructor TrDBRecordLink.Create(RecView: TrDBRecView);
begin
  inherited Create;
  FRecView:=RecView;
  VisualControl:=true;
end;

procedure TrDBRecordLink.ActiveChanged;
begin
  inherited ActiveChanged;
  FRecView.SetRows;
  FRecView.RefreshData;
end;

procedure TrDBRecordLink.DataSetChanged;
begin
  inherited DataSetChanged;
  FRecView.RefreshData;
end;

procedure TrDBRecordLink.EditingChanged;
begin
  inherited EditingChanged;
  FRecView.RefreshData;
end;

procedure TrDBRecordLink.UpdateData;
begin
  inherited UpdateData;
  FRecView.UpdateData;
end;

// ==============================================================================
// TrDBRecViewInplaceEdit functions
procedure TrDBRecViewInplaceEdit.DblClick;
begin
  if (EditStyle = esPickList) then
    DropDown
  else
    inherited DblClick;
end;

// ==============================================================================
// TrDBRecView functions
constructor TrDBRecView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  FDataLink:=TrDBRecordLink.Create(Self);
  FFldList:=TStringList.Create;
  FFields:=TStringList.Create;
  FBooleanFields:=TStringList.Create;
  FEllipsisFields:=TStringList.Create;
  FSpecFields:=TStringList.Create;
  FSpecFields.CaseSensitive:=false;
  FSpecFont:=TFont.Create;
  FSpecFont.Assign(Font);
  FFieldsAsLink:=TStringList.Create;
  GridPickList:=TStringList.Create;
  FPickLists:=TStringList.Create;
  FPickListsIni:=TMemIniFile.Create('');

  Options:=[goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine];
  DefaultDrawing:=false;
  ScrollBars:=ssVertical;
  FGridDlgVisibleTabs:=[vtFieldsSetup, vtSpecFields, vtFieldsAsLink, vtPickLists, vtEllipsisFields];

  ColCount:=2;
  FixedCols:=1;
  FixedRows:=0;
  DefaultRowHeight:=16;
  DefaultColWidth:=90;
  FSaveCellExtents:=false;
  ResetLastEdited;

  if not(csDesigning in ComponentState) then
  begin
    // Init DateTimePicker
    FDateTimePicker:=TDateTimePicker.Create(Self);
    FDateTimePicker.SetSubComponent(true);
    FDateTimePicker.Visible:=false;
    FDateTimePicker.OnCloseUp:=DateTimePickerClose;
    FDateTimePicker.OnChange:=DateTimePickerChange;
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBRecView.Destroy;
begin
  FreeAndNil(FDataLink);
  FreeAndNil(FFldList);
  FreeAndNil(FFields);
  FreeAndNil(FBooleanFields);
  FreeAndNil(FEllipsisFields);
  FreeAndNil(FSpecFields);
  FreeAndNil(FSpecFont);
  FreeAndNil(FFieldsAsLink);
  FreeAndNil(GridPickList);
  FreeAndNil(FPickLists);
  FreeAndNil(FPickListsIni);
  if not(csDesigning in ComponentState) then
  begin
    FreeAndNil(FDateTimePicker);
  end;
  inherited Destroy;
end;

procedure TrDBRecView.Loaded;
begin
  inherited;
  // if not(csDesigning in ComponentState) then
  // rtool.DPI_ScaleFontHeight(SpecFont, FScaleRatio);
  FPickListsIni.SetStrings(FPickLists);
  SetCols;
  SetRows;
  RefreshData;
  if (goAutoHeight in FOptionsEx) then
    Options:=Options - [goRowSizing];
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBRecView.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  rtool.DPI_ScaleFontHeight(SpecFont, M, D);
end;

function TrDBRecView.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

function TrDBRecView.HasParentNormalState: Boolean;
begin
  Result:=Assigned(Parent) and (([csLoading, csDestroying] * ComponentState) = []);
end;

function TrDBRecView.IsDatasetLinked: Boolean;
begin
  Result:=Assigned(DataSource) and Assigned(DataSource.DataSet);
end;

function TrDBRecView.IsActive: Boolean;
begin
  Result:=IsDatasetLinked and DataSource.DataSet.Active;
end;

procedure TrDBRecView.ResetLastEdited;
begin
  LastEditedField:=nil;
  LastEditedValue:='';
  FLookupText:='';
end;

function TrDBRecView.IsBoolEx(F: TField): Boolean;
begin
  Result:=Assigned(F) and ((F is TBooleanField) or (FBooleanFields.IndexOfName(F.FieldName) > -1));
end;

function TrDBRecView.IsMemoEx(F: TField; TextLenAsMemo: integer): Boolean;
begin
  Result:=IsMemoField(F);
  if (F is TStringField) or (F is TWideStringField) then
    if (TextLenAsMemo > 0) and (TextLenAsMemo < F.Size) then
      Result:=true;
end;

function TrDBRecView.IsGraphicEx(F: TField): Boolean;
begin
  Result:=Assigned(F) and IsGraphicField(F);
end;

function TrDBRecView.IsDateEx(F: TField): Boolean;
begin
  Result:=Assigned(F) and IsDateField(F);
end;

function TrDBRecView.IsEllipsisEx(F: TField): Boolean;
begin
  Result:=Assigned(F) and (FEllipsisFields.IndexOf(F.FieldName) > -1);
end;

function TrDBRecView.ChangeBool(Reset: Boolean): Boolean;
var
  v, v1, v2: string;
  F: TField;
begin
  Result:=false;
  if ([goRowSelect, goEditing] * Options = [goEditing]) and CanEdit(true) then
  begin
    F:=FocusedField;
    if F is TBooleanField then
    begin
      if Reset then
        F.AsBoolean:=false
      else
        F.AsBoolean:=not F.AsBoolean;
    end
    else
    begin
      // find index and bool value
      v:=BooleanFields.Values[F.FieldName];
      if v = '' then
        v:=rDefBoolValue;
      v1:=Trim(rtool.FindStrPart(v, ';', 1, false));
      v2:=Trim(rtool.FindStrPart(v, ';', 2, false));
      if Reset then
        F.AsString:=v2
      else
        if CompareText(F.AsString, v1) = 0 then
          F.AsString:=v2
        else
          F.AsString:=v1;
    end;
    Result:=true;
    HideEditor;
    // UpdateData;
    RefreshData;
    // refresh selected cell - Invalidate event not works (BUG?) - not need if EditorMode set to false
    // DrawCell(Col, Row, CellRect(Col, Row), [gdFocused]);
  end;
end;

function TrDBRecView.BoolIsTrue(F: TField): Boolean;
var
  v, v1: string;
begin
  if F is TBooleanField then
  begin
    Result:=F.AsBoolean;
  end
  else
  begin
    // find index and bool value
    v:=BooleanFields.Values[F.FieldName];
    if v = '' then
      v:=rDefBoolValue;
    v1:=Trim(rtool.FindStrPart(v, ';', 1, false));
    Result:=CompareText(F.AsString, v1) = 0;
  end;
end;

procedure TrDBRecView.ShowDateTimePicker(F: TField; Rect: TRect);
begin
  if FDateTimePicker = nil then
    Exit;
  if not Assigned(F) then
    Exit;
  if Rect.Right - Rect.Left < UseScale(DateTimePickerMinWidth) then
    Rect.Right:=Rect.Left + UseScale(DateTimePickerMinWidth);
  FDateTimePicker.BoundsRect:=Rect;
  try
    if F.IsNull or (F.AsDateTime = 0) then
      FDateTimePicker.Date:=Date
    else
      FDateTimePicker.Date:=F.AsDateTime;
  except
    FDateTimePicker.Date:=Date;
  end;
  FDateTimePicker.Show;
  try
    FDateTimePicker.SetFocus;
    // open picker
{$IFDEF DXE2P}
    if TStyleManager.IsCustomStyleActive then
    begin
      // refresh height of picker to get correct calendar size for style
      SendMessage(FDateTimePicker.Handle, CM_FONTCHANGED, 0, 0);
      SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
      // resize back to cell size
      FDateTimePicker.BoundsRect:=Rect;
    end
    else
{$ENDIF}
      SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
  except
    FDateTimePicker.Hide;
  end;
end;

procedure TrDBRecView.SetCols;
begin
  ColWidths[1]:=Width - ColWidths[0] - GridLineWidth * 3 - GetSystemMetrics(sm_CXVScroll) - 3;
end;

procedure TrDBRecView.SetRows;
var
  a, n, h: integer;
  F: TField;
  GD: TObject;
  L: TStringList;
  fld: string;
begin
  if not HasParentNormalState then
    Exit;
  FFldList.Clear;
  GD:=nil;
  if Assigned(FDBGrid) then
    GD:=FDBGrid
  else
    if Assigned(DataSource) then
      GD:=DataSource.DataSet;
  // load fields according only visible and onlysimple
  RowCount:=GetFldList(GD, FFldList, GetFldDefs(false, FOnlyVisibleField, true, not FOnlySimpleField, not FOnlySimpleField, not FOnlySimpleField, false));
  if FFields.Count > 0 then
  begin
    // limit fields to only selected in Fields
    L:=TStringList.Create;
    try
      for a:=0 to FFields.Count - 1 do
      begin
        fld:=FFields.Names[a];
        if fld = '' then
          fld:=FFields[a];
        n:=FFldList.IndexOf(fld);
        if n >= 0 then
          L.Add(FFldList[n]);
      end;
      FFldList.Assign(L);
    finally
      FreeAndNil(L);
    end;
    RowCount:=FFldList.Count;
  end;
  if RowCount = 0 then
    RowCount:=1;
  // set height of rows
  for a:=0 to FFldList.Count - 1 do
  begin
    SetFont(FFldList[a], Canvas.Font);
    h:=Canvas.TextHeight('Wg') + 4;
    if h > DefaultRowHeight then
      RowHeights[a]:=h
    else
      RowHeights[a]:=DefaultRowHeight;
    try
      if IsDatasetLinked then
      begin
        F:=DataSource.DataSet.FindField(FFldList[a]);
        if Assigned(F) and IsMemoEx(F, 0) then
          RowHeights[a]:=4 * Canvas.TextHeight('Wg') + 4;
        if Assigned(F) and IsGraphicEx(F) then
          RowHeights[a]:=6 * Canvas.TextHeight('Wg') + 4;
      end;
    except
    end;
  end;
  Invalidate;
end;

procedure TrDBRecView.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FDBGrid) then
      FDBGrid:=nil;
    if (AComponent = FCustomGlyphs) then
    begin
      FCustomGlyphs:=nil;
      Invalidate;
    end;
  end;
end;

procedure TrDBRecView.SetCustomGlyphs(const Value: TrCustomGlyphs);
begin
  FCustomGlyphs:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
  Invalidate;
end;

procedure TrDBRecView.SetParent(AParent: TWinControl);
begin
  inherited;
  if Assigned(AParent) and Assigned(FDateTimePicker) then
  begin
    FDateTimePicker.Parent:=Self;
    FDateTimePicker.BoundsRect:=Rect(0, 0, 0, 0);
  end;
end;

procedure TrDBRecView.ColWidthsChanged;
begin
  inherited;
  SetCols;
end;

procedure TrDBRecView.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited;
  SetCols;
end;

procedure TrDBRecView.RefreshData;
begin
  if not HasParentNormalState then
    Exit;
  if not IsEditing(DataSource) then
    HideEditor;
  ResetLastEdited;
  Invalidate;
end;

function TrDBRecView.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBRecView.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource:=Value;
  // if not (csLoading in ComponentState) and Assigned(FDBGrid) then
  // if FDBGrid.DataSource<>Value then FDBGrid:=nil;
  SetCols;
  SetRows;
  RefreshData;
end;

procedure TrDBRecView.SetDBGrid(Value: TCustomDBGrid);
begin
  FDBGrid:=Value;
  if Assigned(FDBGrid) then
  begin
    if Assigned(FDBGrid.DataSource) then
      DataSource:=FDBGrid.DataSource;
{$IFDEF RDBGRID}
    if not(csLoading in ComponentState) and (Value is TrDBGrid) then
    begin
      FFieldsAsLink.Assign(TrDBGrid(Value).FieldsAsLink);
      FBooleanFields.Assign(TrDBGrid(Value).FieldsAsBoolean);
    end;
{$ENDIF}
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TrDBRecView.SetOnlyVisibleField(Value: Boolean);
begin
  FOnlyVisibleField:=Value;
  SetRows;
  RefreshData;
end;

procedure TrDBRecView.SetOnlySimpleField(Value: Boolean);
begin
  FOnlySimpleField:=Value;
  SetRows;
  RefreshData;
end;

procedure TrDBRecView.SetTitleAlignment(Value: TAlignment);
begin
  FTitleAlignment:=Value;
  RefreshData;
end;

procedure TrDBRecView.SetOptionsEx(Value: TGridOptionsEx);
begin
  FOptionsEx:=Value;
  if (goAutoHeight in FOptionsEx) then
    Options:=Options - [goRowSizing]
  else
    Options:=Options + [goRowSizing];
  RefreshData;
end;

procedure TrDBRecView.SetFields(Value: TStringList);
begin
  FFields.Assign(Value);
  SetRows;
  RefreshData;
end;

procedure TrDBRecView.SetBooleanFields(Value: TStringList);
begin
  FBooleanFields.Assign(Value);
  RefreshData;
end;

procedure TrDBRecView.SetEllipsisFields(Value: TStrings);
begin
  FEllipsisFields.Assign(Value);
end;

procedure TrDBRecView.SetSpecFields(Value: TStringList);
begin
  FSpecFields.Assign(Value);
  RefreshData;
end;

procedure TrDBRecView.SetSpecFont(Value: TFont);
begin
  FSpecFont.Assign(Value);
  RefreshData;
end;

procedure TrDBRecView.SetFieldsAsLink(Value: TStringList);
begin
  FFieldsAsLink.Assign(Value);
end;

procedure TrDBRecView.SetPickLists(Value: TStringList);
begin
  FPickLists.Assign(Value);
end;

function TrDBRecView.GetRowField(ARow: integer): TField;
begin
  Result:=nil;
  if not Assigned(FFldList) or (ARow < 0) or (ARow > FFldList.Count - 1) then
    Exit;
  if IsDatasetLinked then
    Result:=DataSource.DataSet.FindField(FFldList[ARow]);
end;

function TrDBRecView.GetFocusedField;
begin
  Result:=GetRowField(Row);
end;

function TrDBRecView.GetAlignment(F: TField): integer;
begin
  Result:=DT_LEFT;
  if goDataRight in FOptionsEx then
    Result:=DT_RIGHT;
  case F.DataType of
    ftSmallint, ftInteger, ftWord, ftBCD, ftAutoInc, ftFloat, ftCurrency:
      if goNumberRight in FOptionsEx then
        Result:=DT_RIGHT;
  end;
end;

function TrDBRecView.GetFieldText(F: TField): string;
begin
  if IsMemoField(F) or IsSimpleField(F) then
    Result:=F.AsString
  else
    Result:='';
end;

function TrDBRecView.GetFocusedText;
begin
  if Assigned(LastEditedField) then
    Result:=LastEditedValue
  else
    Result:=GetFieldText(FocusedField);
end;

function TrDBRecView.GetRowCaption(ARow: integer): string;
begin
  try
    Result:=FFields.Values[FFldList[ARow]];
    if Result = '' then
      Result:=GetRowField(ARow).DisplayLabel;
  except
    Result:='';
  end;
end;

function TrDBRecView.GetRowDisplayText(ARow: integer): string;
var
  F: TField;
  REdit: TRichEdit;
  Stream: TStringStream;
begin
  F:=GetRowField(ARow);
  if IsMemoEx(F, 0) then
  begin
    Result:=F.AsString;
    if goRTFToPlainText in FOptionsEx then
    begin
      // convert rtf to plaintext
      REdit:=TRichEdit.CreateParented(HWND_MESSAGE);
      try
        REdit.WordWrap:=false;
        REdit.Width:=8000;
        REdit.PlainText:=false;
        Stream:=TStringStream.Create(Result);
        try
          REdit.Lines.LoadFromStream(Stream);
        finally
          FreeAndNil(Stream);
        end;
        REdit.PlainText:=true;
        Result:=REdit.Lines.Text;
      finally
        FreeAndNil(REdit);
      end;
    end;
  end
  else
    Result:=F.DisplayText;
end;

procedure TrDBRecView.SetFont(const FieldName: string; AFont: TFont);
var
  param, Frm: string;
begin
  AFont.Assign(Font);
  if FSpecFields.IndexOfName(FieldName) >= 0 then
  begin
    AFont.Assign(FSpecFont);
    Frm:=FSpecFields.Values[FieldName];
    while Frm <> '' do
    begin
      param:=AnsiUpperCase(Trim(rtool.RemoveStrPart(Frm, ';')));
      // sf=standard font (not special font), should be at first possition
      if param = 'SF' then
        AFont.Assign(Font)
      else
        UseFontFormatParams(AFont, param);
    end;
  end;
end;

procedure TrDBRecView.GetCellParams(ACol, ARow: longint; AFont: TFont; var Background: TColor; var Highlight: Boolean);
begin
  if Assigned(FOnGetCellParams) then
    FOnGetCellParams(Self, ACol, ARow, AFont, Background, Highlight);
end;

procedure TrDBRecView.Paint;
begin
  inherited;
  // draw backgroung for edited cell
  // BUG - in standard Grid is not drawn and causes black background and flickering
  if EditorMode then
    DrawCellBackground(CellRect(Col, Row), Color, [], Col, Row);
end;

procedure TrDBRecView.DrawCell(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState);
const
  Format1 = DT_VCENTER or DT_SINGLELINE or DT_NOPREFIX or DT_END_ELLIPSIS;
  Format2 = DT_VCENTER or DT_NOPREFIX or DT_WORDBREAK;
var
  F: TField;
  B: TBitmap;
  Frm: integer;
  s: string;
  FontColor, aColor: TColor;
  Highlighted: Boolean;

  procedure DrwTxt(Text: string; BasicFrm, EnhFrm: integer);
  var
    h: integer;
    CalcRec: TRect;
  begin
    BiDi_ModifyRect(Canvas, ARect);
    // if Text='' then Text:=' ';
    if (goAutoHeight in FOptionsEx) and (ACol > 0) then
    begin
      CalcRec:=ARect;

      DrawText(Canvas.Handle, PChar(Text), Length(Text), CalcRec, Format2 or DT_CALCRECT);
      h:=CalcRec.Bottom - CalcRec.Top + 4;
      if h < DefaultRowHeight then
        h:=DefaultRowHeight;
      if h <> RowHeights[ARow] then
        RowHeights[ARow]:=h;
    end;
    if RowHeights[ARow] > DefaultRowHeight then
      BasicFrm:=Format2 or EnhFrm;

    if UseRightToLeftAlignment then
    begin
      BasicFrm:=DrawTextBiDiModeFlags(BasicFrm); // or DT_RTLREADING;
      BasicFrm:=BasicFrm and not DT_RIGHT;
    end;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), ARect, BasicFrm);
  end;

  procedure DrwBmp(Bmp: TBitmap; Bool: Boolean);
  var
    h, w, rh, rw: integer;
  begin
    if (goAutoHeight in FOptionsEx) and (ACol > 0) and not Bool then
    begin
      h:=Bmp.Height;
      if h < DefaultRowHeight then
        h:=DefaultRowHeight;
      if h <> RowHeights[ARow] then
        RowHeights[ARow]:=h;
    end;
    h:=ARect.Bottom - ARect.Top;
    w:=ARect.Right - ARect.Left;
    if Bool then
    begin
      ARect.Right:=ARect.Left + h + 4;
      rpictool.DrawBmpCenter(Canvas, ARect, Bmp);
    end
    else
    begin
      if (Bmp.Height < h) and (Bmp.Width < w) then
        rpictool.DrawBmpCenter(Canvas, ARect, Bmp)
      else
      begin
        if h / Bmp.Height < w / Bmp.Width then
        begin
          rh:=h;
          rw:=Round(Bmp.Width * h / Bmp.Height);
        end
        else
        begin
          rh:=Round(Bmp.Height * w / Bmp.Width);
          rw:=w;
        end;
        ARect.Top:=ARect.Top + (h - rh) div 2;
        ARect.Left:=ARect.Left + (w - rw) div 2;
        ARect.Bottom:=ARect.Top + rh;
        ARect.Right:=ARect.Left + rw;
        rpictool.DrawBmpCenter(Canvas, ARect, Bmp, true);
      end;
    end;
  end;

begin
  if not HasParentNormalState then
    Exit;

  try
    F:=GetRowField(ARow);
  except
    F:=nil;
  end;
  // set default font and color
  FontColor:=Canvas.Font.Color;
  Canvas.Font.Assign(Font);
  if Assigned(F) and (ACol > 0) then
  begin
    SetFont(F.FieldName, Canvas.Font);
    if gdFocused in AState then
      Canvas.Font.Color:=FontColor
    else
      if FFieldsAsLink.IndexOfName(F.FieldName) >= 0 then
      begin
        Canvas.Font.Style:=Canvas.Font.Style + [fsUnderline];
        Canvas.Font.Color:=clHighLight;
      end;
  end;

  // get cell params
  Highlighted:=((gdSelected in AState) and (gdFocused in AState));
  aColor:=Canvas.Brush.Color;
  GetCellParams(ACol, ARow, Canvas.Font, aColor, Highlighted);
  Canvas.Brush.Color:=aColor;

  // draw background
  if Highlighted then
    DrawCellHighlight(ARect, AState, ACol, ARow)
  else
    DrawCellBackground(ARect, Canvas.Brush.Color, AState, ACol, ARow);

  // draw focus
  if gdFocused in AState then
  begin
    Canvas.Brush.Style:=bsSolid;
    BiDi_FrameRect(Canvas, ARect, clWindow);
    BiDi_DrawFocusRect(Canvas, ARect);
  end;
  Canvas.Brush.Style:=bsClear;

  if not Assigned(F) then
    Exit;
  InflateRect(ARect, -2, -2);
  // draw bitmap or text
  if ACol = 0 then
  begin
    case FTitleAlignment of
      taRightJustify:
        Frm:=DT_RIGHT;
      taCenter:
        Frm:=DT_CENTER
      else
        Frm:=DT_LEFT;
    end;
    s:=GetRowCaption(ARow);
    if goAddColon in FOptionsEx then
      s:=s + ':';
    DrwTxt(s, Format1, Frm);
  end
  else
    if IsActive then
    begin
      s:=GetRowDisplayText(ARow);
      if IsMemoEx(F, 0) then
        DrwTxt(s, Format2, 0)
      else
        if IsBoolEx(F) then
        begin
          DrwBmp(rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(BoolIsTrue(F)), FScaleRatio, FCustomGlyphs), true);
        end
        else
          if IsGraphicEx(F) then
          begin
            B:=TBitmap.Create;
            try
              try
                ReadBitmapFromField(F, B);
                DrwBmp(B, false);
              except
                DrwTxt(sPictureCorrupted, Format1, DT_CENTER)
              end;
            finally
              FreeAndNil(B);
            end;
          end
          else
          begin
            Frm:=GetAlignment(F);
            DrwTxt(s, Format1, Frm);
          end;
    end;
end;

procedure TrDBRecView.CMExit(var Message: TMessage);
begin
  try
    UpdateData;
  except
    SetFocus;
    raise ;
  end;
  inherited;
end;

procedure TrDBRecView.KeyDown(var Key: Word; Shift: TShiftState);
var
  F: TField;

  procedure DoCopy;
  begin
    try
      if Assigned(F) then
        Clipboard.AsText:=GetFieldText(F);
      Key:=0;
    except
    end;
  end;
  procedure DoDelete;
  begin
    if Assigned(F) and CanEdit(true) then
    begin
      F.Clear;
      UpdateData;
      RefreshData;
      Key:=0;
    end;
  end;

begin
  F:=FocusedField;
  // cancel changes
  if Key = VK_ESCAPE then
  begin
    FDataLink.RecordChanged(nil);
    HideEditor;
    ResetLastEdited;
    Key:=0;
  end;
  if IsBoolEx(F) and (Key = VK_SPACE) then
  begin
    ChangeBool(false);
    Key:=0;
  end;
  // Del key - delete field
  if not IsBoolEx(F) and (Key = VK_DELETE) then
    DoDelete;
  // Ctrl+C or Ctrl+Insert - copy to clipboard
  if (CharInSet(Chr(Key), ['c', 'C']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssCtrl])) then
    DoCopy;
  // Ctrl+X - Cut to clipboard
  if (Shift = [ssCtrl]) and CharInSet(Chr(Key), ['x', 'X']) then
  begin
    DoCopy;
    DoDelete;
  end;
  // Ctrl+V or Shift+Insert- paste from clipboard
  if (CharInSet(Chr(Key), ['v', 'V']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssShift])) then
  begin
    if Assigned(F) and CanEdit(true) then
    begin
      F.AsString:=Clipboard.AsText;
      UpdateData;
      RefreshData;
      Key:=0;
    end;
  end;
  inherited KeyDown(Key, Shift);
end;

procedure TrDBRecView.KeyPress(var Key: Char);
var
  F: TField;
  found: Boolean;
  I: integer;
begin
  F:=FocusedField;
  if (Key = #13) then
  begin
    if (goEnterNewLineMemo in FOptionsEx) and IsMemoEx(FocusedField, 0) and Assigned(InplaceEditor) and InplaceEditor.Visible then
    begin
      // add new line manually
      I:=InplaceEditor.SelStart;
      InplaceEditor.Text:=copy(InplaceEditor.Text, 1, I) + cStrEOL + copy(InplaceEditor.Text, I + InplaceEditor.SelLength + 1, High(integer));
      InplaceEditor.SelStart:=I + 2;
      // update grid content (UpdateText is not called in this case)
      SetEditText(Col, Row, InplaceEditor.Text);
      // reset key
      Key:=#0;
    end
    else
      UpdateData;
  end;
  if Assigned(InplaceEditor) and (TInplaceEditList(InplaceEditor).EditStyle = esPickList) then
    try
      if (goOnlyPickListValues in FOptionsEx) then
        Key:=#0;
      if Assigned(F) and CharInSet(Key, [#8, #28 .. #255]) then
      begin
        if CanEdit(true) then
        begin
          if Self.InplaceEditor.SelStart = Length(FLookupText) then
          begin
            if Key = #8 then
            begin
              FLookupText:=copy(FLookupText, 1, Length(FLookupText) - 1);
              if (F.FieldKind <> fkLookup) then
                Self.InplaceEditor.Text:=FLookupText;
            end
            else
            begin
              if (F.FieldKind = fkLookup) then
              begin
                // select and set lookup field
                if F.LookupDataSet.Locate(F.LookupResultField, FLookupText + Key, [loCaseInsensitive, loPartialKey]) then
                begin
                  FLookupText:=FLookupText + Key;
                  Self.InplaceEditor.Text:=F.LookupDataSet.FieldByName(F.LookupResultField).AsString;
                  F.DataSet.FieldByName(F.KeyFields).AsVariant:=F.LookupDataSet.FieldByName(F.LookupKeyFields).AsVariant;
                end
                else
                  Beep;
              end
              else
              begin
                // select and set field according PickList
                found:=false;
                for I:=0 to GridPickList.Count - 1 do
                  if AnsiPosEx(FLookupText + Key, GridPickList.Strings[I], 1, false) = 1 then
                  begin
                    Self.InplaceEditor.Text:=GridPickList.Strings[I];
                    found:=true;
                    Break;
                  end;
                FLookupText:=FLookupText + Key;
                if not found then
                  Self.InplaceEditor.Text:=FLookupText;
              end;
            end;
            Self.InplaceEditor.SelStart:=Length(FLookupText);
            Self.InplaceEditor.SelLength:=Self.InplaceEditor.GetTextLen;
            Key:=#0;
          end
          else
            FLookupText:='';
          LastEditedField:=F;
          LastEditedValue:=Self.InplaceEditor.Text;
        end;
      end;
    except
    end;
  inherited KeyPress(Key);
end;

procedure TrDBRecView.CMHintShow(var Message: TCMHintShow);
var
  F: TField;
  s: string;
begin
  if IsActive then
    try
      F:=nil;
      s:='';
      if (MoveCoord.Y >= 0) and (MoveCoord.Y < FFldList.Count) then
        F:=Self.GetRowField(MoveCoord.Y);
      if Assigned(F) then
      begin
        if MoveCoord.X = 0 then
        begin
          s:=FFields.Values[FFldList[MoveCoord.Y]];
          if s = '' then
            s:=F.DisplayLabel;
          if Canvas.TextWidth(s) + 5 < ColWidths[0] then
            s:='';
        end;
        if MoveCoord.X = 1 then
        begin
          if IsGraphicEx(F) then
            s:=''
          else
            s:=GetRowDisplayText(MoveCoord.Y);
          if (Canvas.TextWidth(s) + 5 < ColWidths[1]) then
            s:='';
        end;
        // Set hint string in message
        Message.HintInfo^.HintStr:=s;
      end;
    except
    end;
  inherited;
end;

procedure TrDBRecView.MouseMove(Shift: TShiftState; X, Y: integer);
var
  NewCoord: TGridCoord;
  Reshow: Boolean;
  F: TField;
  s: string;
  R: TRect;
begin
  NewCoord:=MouseCoord(X, Y);
  Reshow:=(NewCoord.X <> MoveCoord.X) or (NewCoord.Y <> MoveCoord.Y);
  MoveCoord:=NewCoord;
  // Activate hint if mouse enter new cell
  if Reshow then
    Application.ActivateHint(Mouse.CursorPos);
  if IsActive then
  begin
    // change cursor
    isLink:=false;
    if (MoveCoord.X = 1) and (FFldList.Count > MoveCoord.Y) and (FFieldsAsLink.IndexOfName(FFldList[MoveCoord.Y]) >= 0) then
    begin
      F:=GetRowField(MoveCoord.Y);
      if Assigned(F) then
      begin
        s:=F.DisplayText;
        R:=CellRect(NewCoord.X, NewCoord.Y);
        InflateRect(R, -3, -3);
        OffsetRect(R, 0, 2);
        if Self.GetAlignment(F) = DT_LEFT then
          R.Right:=R.Left + Canvas.TextWidth(s)
        else
          R.Left:=R.Right - Canvas.TextWidth(s);
        if PtInRect(R, Point(X, Y)) then
          isLink:=true;
      end;
    end;
    if isLink then
      Screen.Cursor:=crHandPoint
    else
      Screen.Cursor:=Cursor;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TrDBRecView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  s: string;
  F: TField;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if isLink then
  begin
    F:=GetRowField(MoveCoord.Y);
    if not Assigned(F) or (F.DisplayText = '') then
      Exit;
    s:=FFieldsAsLink.Values[FFldList[MoveCoord.Y]];
    if SameText(s, rRunHTTP) or SameText(s, rRunMailTo) then
    begin
      if AnsiPosEx(s, F.DisplayText, 1, false) = 0 then
        s:=s + F.DisplayText
      else
        s:=F.DisplayText;
      RunFileSimple(s);
    end
    else
      if Assigned(FOnLinkClick) then
        FOnLinkClick(Self, DataSource.DataSet, nil, F, F.AsVariant);
  end;
end;

procedure TrDBRecView.CMMouseLeave(var AMsg: TMessage);
begin
  Screen.Cursor:=Cursor;
end;

function TrDBRecView.SelectCell(ACol, ARow: longint): Boolean;
begin
  UpdateData;
  Result:= inherited SelectCell(ACol, ARow);
end;

procedure TrDBRecView.Click;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  inherited;
end;

procedure TrDBRecView.DblClick;
var
  Inh: Boolean;
  F: TField;
begin
  Inh:=true;
  F:=GetRowField(Row);
  if IsBoolEx(F) then
  begin
    ChangeBool(false);
    Inh:=false;
  end;
  if Inh then
  begin
    inherited DblClick;
    case GetEditStyle(Col, Row) of
      esEllipsis:
        begin
          ShowEditor;
          if EditorMode then
            EditButtonClick(nil);
        end;
      esPickList:
        begin
          ShowEditor;
          if Assigned(InplaceEditor) then
            TrDBRecViewInplaceEdit(InplaceEditor).DropDown;
        end;
    else
      ShowEditor;
    end;
  end;
end;

function TrDBRecView.CreateEditor: TInplaceEdit;
begin
  Result:=TrDBRecViewInplaceEdit.Create(Self);
  with (Result as TrDBRecViewInplaceEdit) do
  begin
    OnGetPickListItems:=GetPickList;
    OnEditButtonClick:=EditButtonClick;
    DropDownRows:=12;
    Font.Assign(Font);
  end;
end;

function TrDBRecView.GetEditLimit: integer;
var
  F: TField;
begin
  Result:=0;
  F:=FocusedField;
  if Assigned(F) and not rdbtool.IsMemoField(F) then
    Result:=F.DisplayWidth;
end;

function TrDBRecView.GetEditMask(ACol, ARow: longint): string;
var
  F: TField;
begin
  Result:='';
  F:=GetRowField(ARow);
  if Assigned(F) then
    Result:=F.EditMask;
end;

function TrDBRecView.GetEditStyle(ACol, ARow: longint): TEditStyle;
var
  F: TField;
  I: integer;
begin
  // set edit style and fill picklist
  GridPickList.Clear;
  F:=GetRowField(ARow);
  Result:=esSimple;
  if Assigned(F) then
  begin
    if Assigned(InplaceEditor) then
      SetFont(F.FieldName, TrDBRecViewInplaceEdit(InplaceEditor).Font);
    if IsEllipsisEx(F) or IsDateEx(F) then
      Result:=esEllipsis
    else
      if (F.FieldKind = fkLookup) then
        try
          Result:=esPickList;
          if Assigned(F.LookupDataSet) then
          begin
            rdbtool.LoadRecordList(F.LookupDataSet.FieldByName(F.LookupResultField), nil, GridPickList);
          end;
        except
        end
      else
      begin
        // try grids column picklist
        if Assigned(FDBGrid) and ((goUseGridEllipsis in FOptionsEx) or (goUseGridPickList in FOptionsEx)) then
        begin
          with THackCustomDBGrid(FDBGrid) do
          for I:=0 to Columns.Count - 1 do
            if SameText(Columns[I].FieldName, F.FieldName) then
            begin
              if (Columns[I].ButtonStyle = cbsEllipsis) and (goUseGridEllipsis in FOptionsEx) then
              begin
                Result:=esEllipsis;
              end;
              if (Columns[I].PickList.Count > 0) and (goUseGridPickList in FOptionsEx) then
              begin
                Result:=esPickList;
                GridPickList.Assign(Columns[I].PickList);
              end;
              Break;
            end;
        end;
        // try RecView picklist
        FPickListsIni.SetStrings(FPickLists);
        if FPickListsIni.SectionExists(F.FieldName) then
        begin
          Result:=esPickList;
          FPickListsIni.ReadSectionValues(F.FieldName, GridPickList);
        end;
      end;
  end;
end;

procedure TrDBRecView.GetPickList(ACol, ARow: integer; Items: TStrings);
begin
  Items.Clear;
  Items.Assign(GridPickList);
end;

procedure TrDBRecView.UpdateData;
begin
  FLookupText:='';
  if not IsEditing(DataSource) then
    Exit;
  if Assigned(LastEditedField) then
  begin
    LastEditedField.AsString:=LastEditedValue;
    ResetLastEdited;
  end;
end;

function TrDBRecView.CanEdit(StartEdit: Boolean): Boolean;
var
  F: TField;
begin
  F:=FocusedField;
  // check if dataset can be edited
  Result:=false;
  if Assigned(F) then
    if not F.ReadOnly then
      if IsEditing(DataSource) then
        Result:=true
      else
        if DataSource.AutoEdit then
          try
            if StartEdit then
              DataSource.Edit;
            Result:=true;
          except
          end;
end;

function TrDBRecView.CanEditShow: Boolean;
var
  F: TField;
begin
  Result:= inherited CanEditShow;
  F:=FocusedField;
  Result:=Result and (IsSimpleField(F) or IsMemoField(F)) and CanEdit(true) and not IsBoolEx(F);
  if not Result then
    EditorMode:=false;
end;

function TrDBRecView.CanEditAcceptKey(Key: Char): Boolean;
var
  F: TField;
begin
  F:=FocusedField;
  Result:=Assigned(F) and F.IsValidChar(Key);
end;

function TrDBRecView.GetEditText(ACol, ARow: longint): string;
var
  F: TField;
begin
  Result:='';
  F:=GetRowField(ARow);
  Result:=Self.GetFieldText(F);
end;

procedure TrDBRecView.SetEditText(ACol, ARow: longint; const Value: string);
var
  F: TField;
begin
  if not IsEditing(DataSource) then
    Exit;
  F:=GetRowField(ARow);
  LastEditedField:=F;
  LastEditedValue:=Value;

  if Assigned(F) and (F.FieldKind = fkLookup) then
    try
      // do not update as string
      LastEditedField:=nil;
      // update lookup field
      if F.LookupDataSet.Locate(F.LookupResultField, Value, []) then
      begin
        F.DataSet.FieldByName(F.KeyFields).AsVariant:=F.LookupDataSet.FieldByName(F.LookupKeyFields).AsVariant;
      end;
    except
    end;
  // UpdateData;
end;

procedure TrDBRecView.EditButtonClick(Sender: TObject);
var
  F: TField;
begin
  F:=FocusedField;
  if IsDateEx(F) then
    ShowDateTimePicker(F, CellRect(Col, Row))
  else
    if Assigned(FOnEditButtonClick) then
      FOnEditButtonClick(Self);
end;

procedure TrDBRecView.DateTimePickerChange(Sender: TObject);
var
  F: TField;
begin
  F:=FocusedField;
  if not Assigned(F) then
    Exit;
  try
    F.AsDateTime:=Trunc(FDateTimePicker.Date);
    InplaceEditor.Text:=GetEditText(Col, Row);
    UpdateData;
  except
  end;
end;

procedure TrDBRecView.DateTimePickerClose(Sender: TObject);
begin
  FDateTimePicker.Hide;
  try
    SetFocus;
    if not Assigned(InplaceEditor) then
      Exit;
    InplaceEditor.SetFocus;
    InplaceEditor.SelectAll;
  except
  end;
end;

function TrDBRecView.GetFieldCount: integer;
begin
  Result:=FFldList.Count;
end;

function TrDBRecView.GetField(Idx: integer): TField;
begin
  Result:=GetRowField(Idx);
end;

function TrDBRecView.GetCaption(Idx: integer): string;
begin
  Result:=GetRowCaption(Idx);
end;

function TrDBRecView.GetValue(Idx: integer): string;
begin
  Result:=GetRowDisplayText(Idx);
end;

procedure TrDBRecView.GetFontStyle(Idx: integer; AFont: TFont; KeepFontNameSize: Boolean);
var
  F: TField;
  OldF: TFont;
begin
  OldF:=TFont.Create;
  try
    F:=GetField(Idx);
    if Assigned(F) then
    begin
      OldF.Assign(AFont);
      SetFont(F.FieldName, AFont);
      if KeepFontNameSize then
      begin
        AFont.Name:=OldF.Name;
        AFont.Size:=OldF.Size;
      end;
    end;
  finally
    FreeAndNil(OldF);
  end;
end;

function TrDBRecView.GetDataAlignment(Idx: integer): integer;
begin
  Result:=GetAlignment(GetField(Idx));
end;

function TrDBRecView.GetRowCount;
begin
  Result:=Self.RowCount;
end;

end.
