unit rHTMLDrawText;

{<
  rHTMLDrawText contain simple HTML parser and draw HTML content directly to Canvas.
  Used for rHTMLHint and rHTMLListView but can be used for any component,
  uses same tags as rHTMLLabel - see rHTMLLabel for details.

  Version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2015)
  @lastmod(Dec 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, Forms,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} Dialogs, ShellApi, Variants,
  RosiCompGlobalCfg;

type
  { list of HTML drawing types }
  TrHTMLMode = (hmHTML, hmHTMLSkipEnter, hmHTMLSkipLink, hmHTMLSkipEnterAndLink, hmPlainText);

  { draw HTML text to Canvas with default parameters }
function rHTML_DrawTextSimple(C: TWinControl; Canvas: TCanvas; HTMLText: string; var Rect: TRect; CalculateSize: boolean = false; Enabled: boolean = true;
  WordWrap: boolean = false; ShowBeveled: boolean = false; Alignment: TAlignment = taLeftJustify; Layout: TTextLayout = tlTop): integer;

{ draw HTML text to Canvas according parameters }
function rHTML_DrawText(C: TWinControl; Canvas: TCanvas; HTMLText: string; var Rect: TRect; CalculateSize, Enabled, WordWrap, ShowBeveled: boolean;
  Alignment: TAlignment = taLeftJustify; Layout: TTextLayout = tlTop; InnerSpace: integer = 1; TabGridWidth: integer = 30; rHTMLMode: TrHTMLMode = hmHTML;
  AnchorColor: TColor = clNone; AnchorInfoList: TList = nil; HotAnchorIndex: integer = -1): integer;

{ check mouse position over HTML link and return link index, application must store and provide AnchorInfoList defined by rHTML_DrawText }
function rHTML_MouseOverAnchor(C: TWinControl; Left, Top, X, Y: integer; AnchorInfoList: TList; Cursor: TCursor): integer;
{ run link action, application must store and provide AnchorInfoList defined by rHTML_DrawText }
function rHTML_RunGetAnchorAction(AnchorInfoList: TList; HotAnchorIndex: integer): string;
{ free all items from AnchorInfoList, application must call on destroy if defined and used in rHTML_DrawText }
procedure rHTML_FreeAnchorInfo(AnchorInfoList: TList);

implementation

uses rtool;

const
  FrmStrCount = 20;
  FrmStr: array [1 .. FrmStrCount] of string = ('<B>', '</B>', '<I>', '</I>', '<U>', '</U>', '<BIG>', '</BIG>', '<SMALL>', '</SMALL>', '<A', '</A>', '<COLOR=',
    '</COLOR>', '<FONT COLOR=', '</FONT>', '<BGCOLOR=', '</BGCOLOR>', '<TAB>', '<INDENT>');

  FrmAnchorStartNum = 11;
  FrmAnchorStopNum = 12;
  FrmColor1StartNum = 13;
  FrmColor1StopNum = 14;
  FrmColor2StartNum = 15;
  FrmColor2StopNum = 16;
  FrmBgColorStartNum = 17;
  FrmBgColorStopNum = 18;
  FrmTabNum = 19;
  FrmIndentNum = 20;

  FrmCR = '<BR>';
  FrmStop = '>';
  FrmParamStart = '<VAR>';
  FrmParamStop = '</VAR>';

  FrmNoWordWrap = '<NOWORDWRAP>';

function FindFirstFrm(const s: string): integer;
var
  a, p, m: integer;
begin
  p:=Length(s) + 1;
  for a:=1 to FrmStrCount do
  begin
    m:=AnsiPosEx(FrmStr[a], s, 1, false);
    if (m > 0) and (m < p) then
      p:=m;
  end;
  if p = Length(s) + 1 then
    Result:=0
  else
    Result:=p;
end;

function rHTML_DrawTextSimple(C: TWinControl; Canvas: TCanvas; HTMLText: string; var Rect: TRect; CalculateSize: boolean = false; Enabled: boolean = true;
  WordWrap: boolean = false; ShowBeveled: boolean = false; Alignment: TAlignment = taLeftJustify; Layout: TTextLayout = tlTop): integer;
begin
  Result:=rHTML_DrawText(C, Canvas, HTMLText, Rect, CalculateSize, Enabled, WordWrap, ShowBeveled, Alignment, Layout);
end;

function rHTML_DrawText(C: TWinControl; Canvas: TCanvas; HTMLText: string; var Rect: TRect; CalculateSize, Enabled, WordWrap, ShowBeveled: boolean;
  Alignment: TAlignment = taLeftJustify; Layout: TTextLayout = tlTop; InnerSpace: integer = 1; TabGridWidth: integer = 30; rHTMLMode: TrHTMLMode = hmHTML;
  AnchorColor: TColor = clNone; AnchorInfoList: TList = nil; HotAnchorIndex: integer = -1): integer;
const
  Deviders = ' ._-/\@#$&|';
var
  aFlags: longint;
  Text: string;
  LineCount, X, actW, actH, MaxW, Indent, NextIndent: integer;
  Line, Rest: string;
  calcRect, Rc, Rd: TRect;
  oldFont: TFontRecall;
  oldPen: TPenRecall;
  OriginalFont: TFont;
  OriginalFontColor: TColor;
  OriginalBrush: TBrush;
  // origWidth: integer;

  IsAnchorDraw: boolean;
  actAnchorInfo: TAnchorInfo;
  LastAnchorText: string;
  BeforeAnchorFont: TFontRecall;
  ColorEnabled: boolean;

  function ProcFrm(var s: string; Calc: boolean): integer;
  var
    a: integer;

    function ExtractFrmParam(FrmStr: string): string;
    var
      i, l: integer;
    begin
      Result:='';
      i:=AnsiPosEx(FrmStop, s, 1, false);
      if i = 0 then
        Exit;
      l:=Length(FrmStr);
      Result:=copy(s, l + 1, i - l - 1);
      s:=FrmStr + copy(s, i + 1, Length(s));
    end;

    procedure ProcAnchorStop;
    begin
      LastAnchorText:='';
      if Assigned(BeforeAnchorFont) then
        FreeAndNil(BeforeAnchorFont);
      if IsAnchorDraw and not Calc then
        IsAnchorDraw:=false;
    end;

    procedure ProcAnchorStart;
    var
      temp: string;
      p: integer;
    begin
      ProcAnchorStop;
      // Find anchor info
      LastAnchorText:=copy(s, 1, AnsiPosEx(FrmStop, s, 1, false) + Length(FrmStop) - 1);
      if LastAnchorText = '' then
        Exit;
      temp:=ExtractFrmParam(FrmStr[FrmAnchorStartNum]);
      // Set anchor font and color
      BeforeAnchorFont:=TFontRecall.Create(Canvas.Font);
      if not(rHTMLMode in [hmHTMLSkipLink, hmHTMLSkipEnterAndLink]) then
        Canvas.Font.Style:=Canvas.Font.Style + [fsUnderline];
      if ColorEnabled then
      begin
        if Assigned(AnchorInfoList) and (AnchorInfoList.Count = HotAnchorIndex) then
          Canvas.Font.Color:=clHotLight
        else
          if not(rHTMLMode in [hmHTMLSkipLink, hmHTMLSkipEnterAndLink]) then
            Canvas.Font.Color:=clHighLight;
        if AnchorColor <> clNone then
          Canvas.Font.Color:=AnchorColor;
      end;
      // Save anchor info
      if not Calc then
      begin
        IsAnchorDraw:=true;
        if Assigned(AnchorInfoList) then
        begin
          actAnchorInfo:=TAnchorInfo.Create;
          AnchorInfoList.Add(actAnchorInfo);
          p:=Pos('=', temp);
          if p = 0 then
          begin
            actAnchorInfo.LinkType:='';
            actAnchorInfo.LinkName:=Trim(temp);
          end
          else
          begin
            actAnchorInfo.LinkType:=AnsiUpperCase(Trim(copy(temp, 1, p - 1)));
            actAnchorInfo.LinkName:=Trim(copy(temp, p + 1, Length(temp)));
          end;
        end;
      end;
    end;

    function GetColorFromHex(HexCode: string): longint;
    var
      Red, Blue: longint;
    begin
      try
        Result:=StrToInt('$' + HexCode); { but bytes are backwards! }
        Red:=Result and $FF;
        Blue:=Result and $FF0000;
        Result:=(Result and $00FF00) + (Red shl 16) + (Blue shr 16);
      except
        Result:=0;
      end;
    end;

    procedure ProcColorStart(idx: integer);
    var
      temp: string;
      Color: longint;
    begin
      temp:=ExtractFrmParam(FrmStr[idx]);
      if ColorEnabled then
      begin
        if Pos('#', temp) = 1 then
          Canvas.Font.Color:=GetColorFromHex(copy(temp, 2, 6))
        else
          if IdentToColor('cl' + temp, Color) then
            Canvas.Font.Color:=Color;
      end;
    end;

    procedure ProcColorStop;
    begin
      Canvas.Font.Color:=OriginalFontColor;
    end;

    procedure ProcBgColorStart(idx: integer);
    var
      temp: string;
      Color: longint;
    begin
      temp:=ExtractFrmParam(FrmStr[idx]);
      if ColorEnabled then
      begin
        if Pos('#', temp) = 1 then
          Canvas.Brush.Color:=GetColorFromHex(copy(temp, 2, 6))
        else
          if IdentToColor('cl' + temp, Color) then
            Canvas.Brush.Color:=Color;
      end;
    end;

    procedure ProcBgColorStop;
    begin
      if Assigned(OriginalBrush) then
        Canvas.Brush.Assign(OriginalBrush);
    end;

  begin
    Result:=0;
    for a:=1 to FrmStrCount do
    begin
      if AnsiPosEx(FrmStr[a], s, 1, false) = 1 then
        with Canvas.Font do
        begin
          case a of
            1:
              Style:=Style + [fsBold];
            2:
              Style:=Style - [fsBold];
            3:
              Style:=Style + [fsItalic];
            4:
              Style:=Style - [fsItalic];
            5:
              Style:=Style + [fsUnderline];
            6:
              Style:=Style - [fsUnderline];
            7:
              Size:=Size + 2;
            8:
              Size:=OriginalFont.Size;
            9:
              Size:=Size - 2;
            10:
              Size:=OriginalFont.Size;
            FrmAnchorStartNum:
              ProcAnchorStart;
            FrmAnchorStopNum:
              ProcAnchorStop;
            FrmColor1StartNum, FrmColor2StartNum:
              ProcColorStart(a);
            FrmColor1StopNum, FrmColor2StopNum:
              ProcColorStop;
            FrmBgColorStartNum:
              ProcBgColorStart(a);
            FrmBgColorStopNum:
              ProcBgColorStop;
          end;
          s:=copy(s, Length(FrmStr[a]) + 1, Length(s));
          Result:=a;
          Break;
        end;
    end;
  end;

  function GetLabelText: string;
  var
    s: string;
  begin
    s:=TrimRight(HTMLText);
    // replace nowrap tag
    if AnsiPosEx(FrmNoWordWrap, s, 1, false) > 0 then
    begin
      s:=ReplaceStr(FrmNoWordWrap, '', s, true, false);
      if WordWrap then
      begin
        WordWrap:=false;
        // AdjustBounds;
        // s:=s + '<B>'; // ensure extend drawing
      end;
    end;
    // replace spec chars
    // s:=ReplaceStr(FrmTAB,#9,s,true,false);
    s:=ReplaceStr(#9, '<TAB>', s, true, false);
    s:=ReplaceStr(#10, '', s, true, false);
    if rHTMLMode in [hmHTMLSkipEnter, hmHTMLSkipEnterAndLink] then
    begin
      // proc "<br> enter" tag as <br> only
      s:=ReplaceStr(FrmCR + #13, FrmCR, s, true, false);
      // proc "enter" tag as space
      s:=ReplaceStr(#13, ' ', s, true, false);
    end;
    // change <br> to new line
    s:=ReplaceStr(FrmCR, #13, s, true, false);
    // replace empty lines with space
    if copy(s, 1, 1) = #13 then
      s:=' ' + s;
    while Pos(#13#13, s) > 0 do
      s:=ReplaceStr(#13#13, #13 + ' ' + #13, s, true, false);
    // if not (csDesigning in ComponentState) and CaptionAsHint then Hint:=s;
    Result:=s;
  end;

  function DrawTextPart(var R: TRect; F: longint; Wrap: boolean; s: string): string;
  var
    a, DivPos, RestW: integer;
  begin
    // cannot calculate empty string - add space
    if s = '' then
      s:=' ';
    // RestW:=origWidth - R.Left - 2 - 2 * InnerSpace; // rest width till end of line
    RestW:=Rect.Right - R.Left - 2 - 2 * InnerSpace; // rest width till end of line
    if Wrap and (Canvas.TextWidth(s) > RestW) then
    begin
      // find position for wrapping
      DivPos:=-1;
      a:=1;
      while a <= Length(s) do
      begin
        if Pos(s[a], Deviders) > 0 then
          DivPos:=a;
        if Canvas.TextWidth(copy(s, 1, a)) > RestW then
          Break;
        Inc(a);
      end;
      if (a = 1) and (R.Left <= InnerSpace + 1 + Indent) then
        Result:='' // too small width to wrap
      else
      begin
        if (a - 1 = Length(s)) then
          DivPos:=a // no wrap, use whole line
        else
          if (DivPos = -1) then // no separator found
            if (R.Left > calcRect.Left + Indent + 1) then
              DivPos:=0 // not first word, wrap all
            else
              DivPos:=a - 1; // wrap last letter
        Result:=copy(s, DivPos + 1, Length(s)); // wraped part
        s:=Trim(copy(s, 1, DivPos));
        // if s='' then s:=' ';
      end;
    end
    else
      Result:='';
    // Save anchor start point
    if IsAnchorDraw and Assigned(actAnchorInfo) then
      actAnchorInfo.Rect:=R;
    if s <> '' then
      DrawText(Canvas.Handle, PChar(s), Length(s), R, F);
    // Save anchor stop point
    if IsAnchorDraw and Assigned(actAnchorInfo) then
    begin
      actAnchorInfo.Rect.Right:=actAnchorInfo.Rect.Left + Canvas.TextWidth(s);
      actAnchorInfo.Rect.Bottom:=actAnchorInfo.Rect.Top + Canvas.TextHeight(s);
    end;
    Inc(actW, R.Right - R.Left);
    R.Left:=R.Left + Canvas.TextWidth(s);
  end;

  function ProcLine(var R: TRect; F: longint; Calc, Wrap: boolean): string;
  var
    p, oldLeft: integer;
    s: string;
  begin
    if Calc then
      F:=F or DT_CALCRECT;
    Result:='';
    s:=Line; // +' ';
    { for p:=1 to FrmStrCount do
      if CompareText(s, FrmStr[p]) = 0 then
      s:=s + ' ';
      // check only anchor on line
      if (AnsiPosEx('<A', s, 1, false) = 1) and (AnsiPosEx('</A>', s, 1, false) = Length(s) - 3) then
      s:=s + ' ';
      }
    repeat
      if rHTMLMode = hmPlainText then
        p:=0
      else
        p:=FindFirstFrm(s);
      if p = 0 then
      begin
        Result:=DrawTextPart(R, F, Wrap, s);
        Break;
      end;
      if p > 1 then
        Result:=DrawTextPart(R, F, Wrap, copy(s, 1, p - 1));
      s:=copy(s, p, Length(s));
      if Result <> '' then // wraped ?
      begin
        Result:=Result + s; // add rest of line
        Break;
      end;
      case ProcFrm(s, Calc) of
        FrmTabNum:
          begin
            // proc TAB tag relative to Rect start
            oldLeft:=R.Left;
            R.Left:=(((R.Left - Rect.Left) div TabGridWidth) + 1) * TabGridWidth + Rect.Left;
            Inc(actW, R.Left - oldLeft);
          end;
        FrmIndentNum:
          if Calc then
            NextIndent:=R.Left - 1;
      end;
    until false;
    // if (Result='') and (s<>'') then Result:=DrawTextPart(R,F,Wrap,s);
    // set Line for drawing
    Line:=copy(Line, 1, Length(Line) - Length(Result));
    if LastAnchorText <> '' then
    begin
      // End of line without </A>; (one anchor can not be on two lines)
      // add start tag to next line and add stop tag to this line
      if Result <> '' then
        Result:=LastAnchorText + Result;
      s:=FrmStr[FrmAnchorStopNum];
      ProcFrm(s, Calc);
    end;
  end;

begin
  Result:=0;
  BeforeAnchorFont:=nil;
  OriginalFont:=TFont.Create;
  OriginalBrush:=TBrush.Create;
  try
    OriginalFont.Assign(Canvas.Font);
    if Enabled then
      OriginalFontColor:=OriginalFont.Color
    else
      OriginalFontColor:=clBtnShadow;
    // origWidth:=Rect.Right - Rect.Left;
    // origHeight:=Rect.Bottom - Rect.Top;
    OriginalBrush.Assign(Canvas.Brush);

    Text:=GetLabelText;
    if Text = '' then
      Exit;
    ColorEnabled:=true;

    if ShowBeveled and (not CalculateSize) then
    begin
      // Draw bevel
      oldPen:=TPenRecall.Create(Canvas.Pen);
      try
        Canvas.Pen.Color:=clBtnShadow;
        Canvas.PolyLine([Point(Rect.Left, Rect.Bottom - 1), Rect.TopLeft, Point(Rect.Right - 1, Rect.Top)]);
        Canvas.Pen.Color:=clBtnHighlight;
        Canvas.PolyLine([Point(Rect.Right - 1, Rect.Top), Point(Rect.Right - 1, Rect.Bottom - 1), Point(Rect.Left, Rect.Bottom - 1)]);
      finally
        FreeAndNil(oldPen);
      end;
    end;

    // Exteded formating
    aFlags:=DT_NOPREFIX or DT_EXPANDTABS;
    if Assigned(C) then
      aFlags:=C.DrawTextBiDiModeFlags(aFlags);

    // substract space
    calcRect:=Rect;
    if not CalculateSize and (Layout <> tlTop) then
    begin
      rHTML_DrawText(C, Canvas, HTMLText, calcRect, true, Enabled, WordWrap, ShowBeveled, Alignment, Layout, InnerSpace, TabGridWidth, rHTMLMode, AnchorColor,
        nil, -1);
      actH:=Rect.Bottom - Rect.Top - (calcRect.Bottom - calcRect.Top);
      calcRect:=Rect;
      if actH > 0 then
      begin
        if Layout = tlBottom then
          calcRect.Top:=Rect.Top + actH
        else
          calcRect.Top:=Rect.Top + actH div 2;
      end;
    end;

    InflateRect(calcRect, -InnerSpace, -InnerSpace);
    // Canvas.Font:=origFont;
    Rd:=calcRect;
    LineCount:=1;
    MaxW:=0;
    actH:=0;
    rHTML_FreeAnchorInfo(AnchorInfoList);
    actAnchorInfo:=nil;
    IsAnchorDraw:=false;
    LastAnchorText:='';
    repeat
      // Find and proc one line
      Rest:=FindStrPart(Text, #13, LineCount, false);
      if Rest = '' then
        Break;
      Indent:=0;
      NextIndent:=0;
      repeat
        // calc width of the line and proccess WordWrap
        if Rest = '' then
          Break;
        Line:=Rest;
        Rc:=calcRect;
        actW:=Indent;
        Rc.Left:=Rc.Left + actW;
        if CalculateSize then
          Rest:=ProcLine(Rc, aFlags, true, WordWrap)
        else
        begin
          // Reset font back after proc line for Drawing
          oldFont:=TFontRecall.Create(Canvas.Font);
          try
            Rest:=ProcLine(Rc, aFlags, true, WordWrap);
          finally
            FreeAndNil(oldFont);
          end;
        end;
        // store maxWidth and Height
        if actW > MaxW then
          MaxW:=actW;
        // if Rc.Bottom - Rc.Top = 0 then ShowMessage('0');

        Inc(actH, Rc.Bottom - Rc.Top);

        if not CalculateSize then
        begin
          // draw the line
          case Alignment of
            taCenter:
              begin
                X:=(calcRect.Right - calcRect.Left - actW) div 2 + calcRect.Left;
                if X < calcRect.Left + 1 + Indent then
                  X:=calcRect.Left + 1 + Indent;
              end;
            taRightJustify:
              X:=calcRect.Right - actW - 1;
            taLeftJustify:
              X:=calcRect.Left + 1 + Indent;
          else
            X:=0;
          end;
          Rd.Left:=X;

          if (not Enabled) then
          begin
            ColorEnabled:=false;
            oldFont:=TFontRecall.Create(Canvas.Font);
            try
              OffsetRect(Rd, 1, 1);
              Canvas.Font.Color:=clBtnHighlight;
              ProcLine(Rd, aFlags, false, false);
            finally
              FreeAndNil(oldFont);
            end;
            OffsetRect(Rd, -1, -1);
            Rd.Left:=X;
            Canvas.Font.Color:=clBtnShadow;
          end;
          ColorEnabled:=Enabled;
          ProcLine(Rd, aFlags, false, false);
          // Set Rect for next line
          Rd.Top:=calcRect.Top + actH;
        end;
        if Alignment = taLeftJustify then
          Indent:=NextIndent;
      until false;
      Inc(LineCount);
    until false;

    if CalculateSize then
    begin
      // Set calculated bounds
      { if not WordWrap then
        calcRect.Right:=calcRect.Left + MaxW + 2 + InnerSpace;
        calcRect.Bottom:=calcRect.Top + H + 2 + InnerSpace;
        Rect:=calcRect;
        Rect.Top:=Rect.Top-InnerSpace;
        Rect.Left:=Rect.Left-InnerSpace;
        }
      if not WordWrap then
        Rect.Right:=calcRect.Left + MaxW + 2 + InnerSpace;
      Rect.Bottom:=calcRect.Top + actH + 0 + InnerSpace;
    end
    else
    begin
      if (Rect.Right - Rect.Left < MaxW + 2 * InnerSpace) or (Rect.Bottom - Rect.Top < actH - 2 * InnerSpace) then
        Result:=1;
    end;
  finally
    Canvas.Font.Assign(OriginalFont);
    FreeAndNil(OriginalFont);
    Canvas.Brush.Assign(OriginalBrush);
    FreeAndNil(OriginalBrush);
  end;
  if Assigned(BeforeAnchorFont) then
    FreeAndNil(BeforeAnchorFont);
end;

function rHTML_MouseOverAnchor(C: TWinControl; Left, Top, X, Y: integer; AnchorInfoList: TList; Cursor: TCursor): integer;
var
  i: integer;
  a: TAnchorInfo;
begin
  Result:=-1;
  if not Assigned(AnchorInfoList) then
    Exit;
  // Find active anchor
  for i:=0 to AnchorInfoList.Count - 1 do
  begin
    a:=AnchorInfoList.Items[i];
    if (X > a.Rect.Left) and (X < a.Rect.Right) and (Y > a.Rect.Top) and (Y < a.Rect.Bottom) then
    begin
      Result:=i;
      Break;
    end;
  end;
  if Result <> -1 then
    Screen.Cursor:=crHandPoint
  else
    Screen.Cursor:=Cursor;
end;

function rHTML_RunGetAnchorAction(AnchorInfoList: TList; HotAnchorIndex: integer): string;
var
  a: TAnchorInfo;
begin
  Result:='';
  if Assigned(AnchorInfoList) and (HotAnchorIndex >= 0) and (HotAnchorIndex < AnchorInfoList.Count) then
  begin
    a:=AnchorInfoList.Items[HotAnchorIndex];
    if a.LinkType = 'HREF' then
      ShellExecute(Application.Handle, 'open', PChar(a.LinkName), nil, nil, SW_NORMAL)
    else
      Result:=a.LinkName;
  end;
end;

procedure rHTML_FreeAnchorInfo(AnchorInfoList: TList);
var
  i: integer;
  a: TAnchorInfo;
begin
  if Assigned(AnchorInfoList) then
  begin
    for i:=0 to AnchorInfoList.Count - 1 do
    begin
      a:=AnchorInfoList.Items[i];
      FreeAndNil(a);
    end;
    AnchorInfoList.Clear;
  end;
end;

end.
