unit rImageEdit;

{<
  rImageEdit form is universal dialog which can show and zoom picture
  in independend window and provide function for editing of brightness and contrast.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2009)
  @lastmod(March 2011)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, ComCtrls, ToolWin, rImageZoom, StdCtrls, Buttons,
  ExtCtrls, PropSaveMain, rHTMLLabel, Menus, rdlg_baseform,
  rImprovedComps;

type
  { universal dialog which can show and zoom picture
    in independend window and provide function for editing of
    brightness and contrast }
  TrImageEditForm = class(TrDlgBaseForm)
    Img: TrImageZoom;
    pmImg: TPopupMenu;
    pmZoomIn: TMenuItem;
    pmZoomOut: TMenuItem;
    pmZoomFit: TMenuItem;
    N1: TMenuItem;
    pmReload: TMenuItem;
    Panel5: TPanel;
    Label6: TLabel;
    bt1: TSpeedButton;
    bt2: TSpeedButton;
    btReload: TrURLLabel;
    btZoomIn: TrButton;
    btZoomOut: TrButton;
    btZoomFit: TrButton;
    pcEdit: TPageControl;
    ts1: TTabSheet;
    Label9: TLabel;
    Label1: TLabel;
    trBrAll: TTrackBar;
    trContr: TTrackBar;
    ts2: TTabSheet;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    trBrR: TTrackBar;
    trBrG: TTrackBar;
    trBrB: TTrackBar;
    btClose: TrButton;
    procedure btCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btReloadClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure trBrAllChange(Sender: TObject);
    procedure btZoomInClick(Sender: TObject);
    procedure btZoomOutClick(Sender: TObject);
    procedure btZoomFitClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormResize(Sender: TObject);
    procedure bt2Click(Sender: TObject);
    procedure bt1Click(Sender: TObject);
  private
    OrigBmp: TBitmap;
    Changing: boolean;
    procedure ReloadBmp;
  public
  end;

var
  rImageEditForm: TrImageEditForm;
  LastPos: TRect;  {< @exclude }

  { show rImageEdit dialog, as image use Bitmap or Picture,
    possition of window can be saved in PSFile,
    dialog can me modal or not modal and can save changes back to image }
function ShowImageEdit(B: TBitmap; P: TPicture; PSFile: TPropSaveFile; ShowModal, CanSave: boolean): boolean;

implementation

{$R *.dfm}

uses RosiCompGlobalCfg, rstring, rtool, rdlg;

const
  center = 127;

function ShowImageEdit(B: TBitmap; P: TPicture; PSFile: TPropSaveFile; ShowModal, CanSave: boolean): boolean;
var
  F: TrImageEditForm;
begin
  Result:=false;
  F:=TrImageEditForm.Create(GetActiveOwner);
  // set PSFile and possition
  F.DlgPropSave.PropSaveFile:=PSFile;
  if (LastPos.Right > 0) and (LastPos.Bottom > 0) then
    F.BoundsRect:=LastPos;
  // assign bitmap or picture
  if Assigned(B) then
    F.OrigBmp.Assign(B)
  else
    if Assigned(P) then
      F.OrigBmp.Assign(P.Graphic);
  // set buttons
  if not ShowModal then
    CanSave:=false;
  F.btOk.Visible:=CanSave;
  F.btCancel.Visible:=F.btOk.Visible;
  F.btClose.Visible:=not F.btOk.Visible;
  if ShowModal then
  begin
    if F.ShowModal = mrOk then
      try
        // save updated image
        if Assigned(B) then
          B.Assign(F.Img.Picture.Bitmap)
        else
          if Assigned(P) then
            P.Graphic.Assign(F.Img.Picture.Graphic);
        Result:=true;
      except
      end;
    FreeAndNil(F);
  end
  else
    F.Show;
end;

procedure ChangeBrightContr(Orig, Dest: TBitmap; R, G, B, Contrast: integer);
var
  i, j: integer;
  temp: integer;
  OrigPixelPointer: PByteArray;
  DestPixelPointer: PByteArray;
begin
  WCursor;
  Dest.Assign(Orig);
  try
    for i:=0 to Orig.Height - 1 do
    begin
      OrigPixelPointer:=Orig.ScanLine[i];
      DestPixelPointer:=Dest.ScanLine[i];
      for j:=0 to Orig.Width - 1 do
      begin
        temp:=OrigPixelPointer[3 * j] + B;
        temp:=Round(((temp - center) * Exp(Contrast / 50)) + center);
        if temp < 0 then
          temp:=0;
        if temp > 255 then
          temp:=255;
        DestPixelPointer[3 * j]:=temp;
        temp:=OrigPixelPointer[3 * j + 1] + G;
        temp:=Round(((temp - center) * Exp(Contrast / 50)) + center);
        if temp < 0 then
          temp:=0;
        if temp > 255 then
          temp:=255;
        DestPixelPointer[3 * j + 1]:=temp;
        temp:=OrigPixelPointer[3 * j + 2] + R;
        temp:=Round(((temp - center) * Exp(Contrast / 50)) + center);
        if temp < 0 then
          temp:=0;
        if temp > 255 then
          temp:=255;
        DestPixelPointer[3 * j + 2]:=temp;
      end;
    end;
  except
    on E: Exception do
    begin
      ShowMessage(sPictureErr + ' ' + E.Message);
      Dest.Assign(Orig);
    end;
  end;
  NCursor;
end;

procedure TrImageEditForm.FormCreate(Sender: TObject);
begin
  inherited;
  OrigBmp:=TBitmap.Create;
  Img.ShowEditor:=ieNone;
end;

procedure TrImageEditForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(OrigBmp);
  LastPos:=BoundsRect;
end;

procedure TrImageEditForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action:=caFree;
  inherited;
end;

procedure TrImageEditForm.FormShow(Sender: TObject);
begin
  inherited;
  OrigBmp.PixelFormat:=pf24bit;
  ReloadBmp;
  Img.Fit;
end;

procedure TrImageEditForm.FormResize(Sender: TObject);
begin
  // if Img.AutoFit then Img.Fit;
end;

procedure TrImageEditForm.ReloadBmp;
begin
  Img.Picture.Bitmap.Assign(OrigBmp);
  Changing:=true;
  trBrAll.Position:=0;
  trContr.Position:=0;
  trBrR.Position:=0;
  trBrG.Position:=0;
  trBrB.Position:=0;
  Changing:=false;
end;

procedure TrImageEditForm.btCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TrImageEditForm.btReloadClick(Sender: TObject);
begin
  ReloadBmp;
  Img.Zoom:=100;
end;

procedure TrImageEditForm.btZoomInClick(Sender: TObject);
begin
  Img.ZoomIn;
end;

procedure TrImageEditForm.btZoomOutClick(Sender: TObject);
begin
  Img.ZoomOut;
end;

procedure TrImageEditForm.btZoomFitClick(Sender: TObject);
begin
  Img.Fit;
end;

procedure TrImageEditForm.trBrAllChange(Sender: TObject);
begin
  if Changing then
    Exit;
  if Sender = trBrAll then
  begin
    Changing:=true;
    trBrR.Position:=trBrAll.Position;
    trBrG.Position:=trBrAll.Position;
    trBrB.Position:=trBrAll.Position;
  end;
  try
    ChangeBrightContr(OrigBmp, Img.Picture.Bitmap, trBrR.Position, trBrG.Position, trBrB.Position, trContr.Position);
    Img.Picture.Bitmap.Assign(Img.Picture.Bitmap);
  finally
    Changing:=false;
  end;
end;

procedure TrImageEditForm.bt1Click(Sender: TObject);
begin
  pcEdit.ActivePage:=ts1;
end;

procedure TrImageEditForm.bt2Click(Sender: TObject);
begin
  pcEdit.ActivePage:=ts2;
end;

initialization

LastPos:=Rect(0, 0, 0, 0);

end.
