unit rImageZoom;

{<
  rImageZoom unit contains TrImageZoom - TCustomControl designed as replacement
  of standard TPicture component providing zoom and rotate function, scrollbars, drag and move support.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2010)
  @lastmod(Oct 2014)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Forms, Graphics, Controls, Types,
  Menus, StdCtrls, Math, Clipbrd, {$IFDEF DXE2P} UITypes, {$ENDIF}
  rpictool, RosiCompGlobalCfg;

const
  // popup menu indexes, used by rDBComponents too
  cPicNone = 0;
  cPicCut = 1;
  cPicCopy = 2;
  cPicPaste = 3;
  cPicLoadFromFile = 4;
  cPicSaveToFile = 5;
  cPicZoomFit = 11;
  cPicZoomIn = 12;
  cPicZoomOut = 13;
  cPicResize = 21;
  cPicEdit = 22;
  cPicSaveInJpgFormat = 31;
  cPicInfo = 32;

type
  { @exclude }
  TTypeScrollbar = (tsNone, tsHorz, tsVert, tsBoth);
  { @exclude }
  TRotateType = (rt0, rt090, rt180, rt270);
  { define way how the independent editing window is shown }
  TImageShowEditor = (ieShowModal, ieShowNoModal, ieEditModal, ieNone);

  { @exclude }
  TScaleInfo = record
    SrcSize: TSize;
    SrcOffset: TPoint;
    DstSize: TSize;
    DstOffset: TPoint;
  end;

  { TCustomControl designed as replacement of standard TPicture component
    providing zoom and rotate function, scrollbars, drag and move support }
  TrImageZoom = class(TCustomControl)
  private
    FUseFocus: boolean;
    FShowEditor: TImageShowEditor;
    FScrollBar: boolean;
    FTypeScrollBar: TTypeScrollbar;
    FZoomOffset: TPoint;
    FInternalZoom: double;
    FLockOriginalZoom: boolean;
    FZoom: integer;
    FBitmap: TBitmap;
    FPicture: TPicture;
    FIncrementalDisplay: boolean;
    FTransparent: boolean;
    FDrawing: boolean;
    FEnlargeToFit: boolean;
    FReduceToFit: boolean;
    FStretch: boolean;
    FRotateType: TRotateType;
    FBorderColor: TColor;
    FBackgroundColor: TColor;
    FBackgroundText: string;
    FShowBackgroundText: boolean;
    FCanDrag: boolean;
    FDragAndMove: boolean;
    FSaveInJPG: boolean;
    FShowPopupMenu: boolean;
    FirstOpenMenu: boolean;
    { events }
    FOnProgress: TProgressEvent;
    FOnPaint: TNotifyEvent;
    FOnChangeZoom: TNotifyEvent;
    FOnResize: TNotifyEvent;
    FIsDraging: boolean;
    FDragStartPoint: TPoint;
    FDragStartOffset: TPoint;
    FReadOnly: boolean;
    procedure SetShowEditor(Value: TImageShowEditor);
    procedure SetShowPopupMenu(Value: boolean);
    procedure SetInternalZoom;
    procedure PictureChanged(Sender: TObject);
    procedure SetTransparent(Value: boolean);
    procedure SetLockOriginalZoom(Value: boolean);
    procedure SetZoom(Value: integer);
    procedure SetPicture(Value: TPicture);
    procedure SetScrollBar(Value: boolean);
    procedure SetEnlargeToFit(Value: boolean);
    procedure SetReduceToFit(Value: boolean);
    function GetAutoFit: boolean;
    procedure SetAutoFit(Value: boolean);
    procedure SetStretch(Value: boolean);
    procedure SetRotateType(Value: TRotateType);
    procedure SetBorderColor(Value: TColor);
    procedure SetBackgroundColor(Value: TColor);
    procedure SetBackgroundText(const Value: string);
    procedure SetShowBackgroundText(Value: boolean);
    procedure SetReadOnly(Value: boolean);
    function ScrollMessage(SbType: integer; var Msg: TWMScroll; ClientSize: integer): integer;
    function GetPictureWidth: integer;
    function GetPictureHeight: integer;
    function GetBkColor: TColor;
    procedure CMParentColorChanged(var Message: TMessage); message CM_PARENTCOLORCHANGED;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
  protected
    FPopupMenu: TPopupMenu;
    ImgGraphicType: TrGraphicType;
    procedure Loaded; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Message: TMessage); override;
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure Paint; override;
    procedure UpdateScrollBar;
    function DoPaletteChange: boolean;
    function GetPalette: HPALETTE; override;
    procedure Progress(Sender: TObject; Stage: TProgressStage; PercentDone: Byte; RedrawNow: boolean; const R: TRect; const Msg: string); dynamic;
    procedure ZoomChanged; dynamic;
    procedure DoResize; dynamic;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure DragOver(Source: TObject; X, Y: integer; State: TDragState; var Accept: boolean); override;
    function CanModify: boolean; dynamic;
    procedure OnPopupMenu(Sender: TObject); dynamic;
    procedure PopupClick(Sender: TObject); dynamic;
    function GetPopupItem(Tag: integer): TMenuItem; dynamic;
  public
    LastFileName: string;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    function ScaleInfo: TScaleInfo; {< @exclude }
    function XSrc2Pict(X: integer): integer;
    function YSrc2Pict(Y: integer): integer;
    function XPict2Src(X: integer): integer;
    function YPict2Src(Y: integer): integer;
    function Src2Pict(P: TPoint): TPoint;
    function Pict2Src(P: TPoint): TPoint;
    { change zoom of picture to fit in window }
    procedure Fit;
    { move selection of picture inside scroll area to center }
    procedure MoveToCenter;
    { move selection of picture inside scroll area }
    procedure SetZoomOffset(Value: TPoint);
    { increase zoom by selected percent }
    procedure ZoomIn(Percent: integer = 25; Center: boolean = true);
    { decrease zoom by selected percent }
    procedure ZoomOut(Percent: integer = 25; Center: boolean = true);
    { define whether picture is automatically resized to window }
    property AutoFit: boolean read GetAutoFit write SetAutoFit default false;
    { cut image and insert to clipboard }
    function CutImageToClipboard: boolean; dynamic;
    { copy image and insert to clipboard }
    function CopyImageToClipboard: boolean; dynamic;
    { paste image from clipboard }
    function PasteImageFromClipboard: boolean; dynamic;
    { save image to file, BMP and JPG format is supported }
    function SaveAsFile(const FileName: string): boolean; dynamic;
    { load image from file, BMP and JPG format is supported }
    function LoadFromFile(const FileName: string): boolean; dynamic;
    { change resolution of image }
    function ResizePic(ratio: double): boolean; dynamic;
    { return text info with format and resolution of image }
    function PictureInfo: string;
    { show image in independent window supporting edit functions }
    function ShowEditorWindow: boolean; dynamic;
  published
    property Align;
    property Cursor;
    property DragCursor;
    property DragMode;
    property Enabled;
    property IncrementalDisplay: boolean read FIncrementalDisplay write FIncrementalDisplay default false;
    property ParentShowHint;
    property Picture: TPicture read FPicture write SetPicture;
    property PopupMenu;
    property ShowHint;
    property Transparent: boolean read FTransparent write SetTransparent default false;
    property Visible;
    property Font;
    property ParentFont;
    property ParentColor;
    property ParentBackground default false;
    property DoubleBuffered;
    property ParentDoubleBuffered;
    property TabStop;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property ZoomOffset: TPoint read FZoomOffset;
    { define whether scrollbars are visible }
    property ScrollBar: boolean read FScrollBar write SetScrollBar default false;
    { define zoom in percent }
    property LockOriginalZoom: boolean read FLockOriginalZoom write SetLockOriginalZoom default false;
    { define zoom in percent }
    property Zoom: integer read FZoom write SetZoom default 100;
    { define whether small picture is automatically zoomed to window }
    property EnlargeToFit: boolean read FEnlargeToFit write SetEnlargeToFit default false;
    { define whether big picture is automatically reduced to window }
    property ReduceToFit: boolean read FReduceToFit write SetReduceToFit default false;
    { define whether picture is automatically strech to window  if AutoFit=true }
    property Stretch: boolean read FStretch write SetStretch default false;
    { define whether picture is automatically rotated }
    property RotateType: TRotateType read FRotateType write SetRotateType default rt0;
    { define whether picture use drag operation to move picture }
    property DragAndMove: boolean read FDragAndMove write FDragAndMove default true;
    { define color of border line }
    property BorderColor: TColor read FBorderColor write SetBorderColor default clWindowFrame;
    { define background color of image panel }
    property BackgroundColor: TColor read FBackgroundColor write SetBackgroundColor default clNone;
    { define text which is shown on image panel if picture is not defined or shown }
    property BackgroundText: string read FBackgroundText write SetBackgroundText;
    { define wether background text is shown }
    property ShowBackgroundText: boolean read FShowBackgroundText write SetShowBackgroundText default true;
    { define way how the independent window is shown }
    property ShowEditor: TImageShowEditor read FShowEditor write SetShowEditor default ieShowModal;
    { define whether internal popup menu is shown }
    property ShowPopupMenu: boolean read FShowPopupMenu write SetShowPopupMenu default true;
    { define if control can have focus }
    property UseFocus: boolean read FUseFocus write FUseFocus default false;
    { define if picture is saved in JPG format }
    property SaveInJPG: boolean read FSaveInJPG write FSaveInJPG default true;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnChangeZoom: TNotifyEvent read FOnChangeZoom write FOnChangeZoom;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rdlg, rImageEdit;

type
  { @exclude }
  THackControl = class(TControl)
  end;

procedure Register;
begin
  RegisterComponents('Rosi', [TrImageZoom]);
end;

constructor TrImageZoom.Create(AOwner: TComponent);
var
  m: TMenuItem;
begin
  inherited Create(AOwner);
  ControlStyle:=ControlStyle + [csReplicatable, csParentBackground] - [csOpaque];
  FUseFocus:=false;
  FBitmap:=TBitmap.Create;
  FPicture:=TPicture.Create;
  FPicture.OnChange:=PictureChanged;
  FPicture.OnProgress:=Progress;
  Height:=105;
  Width:=105;
  FZoom:=100;
  FStretch:=false;
  FRotateType:=rt0;
  FZoomOffset:=Point(0, 0);
  SetScrollBar(false);
  FTypeScrollBar:=tsNone;
  FBorderColor:=clWindowFrame;
  FBackgroundColor:=clNone;
  ParentBackground:=false;
  FIsDraging:=false;
  FDragAndMove:=true;
  DragCursor:=crSizeAll;
  DragMode:=dmManual;
  FShowBackgroundText:=true;
  FSaveInJPG:=false;
  FShowEditor:=ieShowModal;
  FShowPopupMenu:=true;

  // create Popup
  FirstOpenMenu:=true;
  FPopupMenu:=TPopupMenu.Create(Self);
  FPopupMenu.Name:=Name + '_ImagePopup';
  FPopupMenu.OnPopup:=OnPopupMenu;
  FPopupMenu.AutoHotkeys:=maManual;

  // popup items - info, size and resize
  FPopupMenu.Items.Add(CreateMenuItem(Self, '', cPicInfo, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sPictureResize, cPicResize, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sShowPicEdit, cPicEdit, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, '-', cPicNone, PopupClick));
  // popup items - zoom
  FPopupMenu.Items.Add(CreateMenuItem(Self, sZoomAutoFit, cPicZoomFit, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sZoomIn, cPicZoomIn, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sZoomOut, cPicZoomOut, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, '-', cPicNone, PopupClick));
  // popup items - cut and copy
  m:=CreateMenuItem(Self, sSaveInJPEGFormat, cPicSaveInJpgFormat, PopupClick);
  m.AutoCheck:=true;
  m.Checked:=SaveInJPG;
  m.Visible:=rpictool.IsJPGSupported;
  FPopupMenu.Items.Add(m);
  FPopupMenu.Items.Add(CreateMenuItem(Self, sCut, cPicCut, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sCopy, cPicCopy, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sPaste, cPicPaste, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, '-', cPicNone, PopupClick));
  // popup items - save and load
  FPopupMenu.Items.Add(CreateMenuItem(Self, sSaveToFile, cPicSaveToFile, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sLoadFromFile, cPicLoadFromFile, PopupClick));
  PopupMenu:=FPopupMenu;

  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrImageZoom.Destroy;
begin
  FreeAndNil(FBitmap);
  FreeAndNil(FPicture);
  inherited Destroy;
end;

procedure TrImageZoom.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style:=Style or (WS_HSCROLL or WS_VSCROLL);
    WindowClass.Style:=WindowClass.Style and not(CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TrImageZoom.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_LBUTTONDOWN, WM_LBUTTONDBLCLK:
      if not(csDesigning in ComponentState) and not Focused and FUseFocus then
      begin
        Windows.SetFocus(Handle);
        if not Focused then
          Exit;
      end;
  end;
  inherited WndProc(Message);
end;

procedure TrImageZoom.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  if FUseFocus then
    Invalidate;
end;

procedure TrImageZoom.WMKillFocus(var Message: TWMKillFocus);
begin
  inherited;
  if FUseFocus then
    Invalidate;
end;

procedure TrImageZoom.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrImageZoom.SetShowEditor(Value: TImageShowEditor);
begin
  if Value <> FShowEditor then
  begin
    FShowEditor:=Value;
    // RefBoundButtons(nil);
  end;
end;

procedure TrImageZoom.SetShowPopupMenu(Value: boolean);
begin
  if Value <> FShowPopupMenu then
  begin
    FShowPopupMenu:=Value;
    if FShowPopupMenu then
      PopupMenu:=FPopupMenu
    else
      PopupMenu:=nil;
  end;
end;

procedure TrImageZoom.SetZoom(Value: integer);
begin
  if (Value <> FZoom) and (Value > 0) then
  begin
    FZoom:=Value;
    SetInternalZoom;
  end;
end;

procedure TrImageZoom.SetLockOriginalZoom(Value: boolean);
begin
  if (Value <> FLockOriginalZoom) then
  begin
    FLockOriginalZoom:=Value;
    SetInternalZoom;
  end;
end;

procedure TrImageZoom.MoveToCenter;
var P: TPoint;
begin
  P.X:= max(0,(Picture.Width-ScaleInfo.SrcSize.cx) div 2);
  P.Y:= max(0,(Picture.Height-ScaleInfo.SrcSize.cy) div 2);
  SetZoomOffset(P);
end;

procedure TrImageZoom.SetZoomOffset(Value: TPoint);
begin
  if (FZoomOffset.X <> Value.X) or (FZoomOffset.Y <> Value.Y) then
  begin
    FZoomOffset:=Value;
    UpdateScrollBar;
  end;
end;

procedure TrImageZoom.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TrImageZoom.SetScrollBar(Value: boolean);
begin
  if FScrollBar <> Value then
  begin
    FScrollBar:=Value;
    UpdateScrollBar;
  end;
end;

procedure TrImageZoom.SetEnlargeToFit(Value: boolean);
begin
  if FEnlargeToFit <> Value then
  begin
    FEnlargeToFit:=Value;
    SetInternalZoom;
  end;
end;

procedure TrImageZoom.SetReduceToFit(Value: boolean);
begin
  if FReduceToFit <> Value then
  begin
    FReduceToFit:=Value;
    SetInternalZoom;
  end;
end;

function TrImageZoom.GetAutoFit: boolean;
begin
  Result:=FReduceToFit and FEnlargeToFit;
end;

procedure TrImageZoom.SetAutoFit(Value: boolean);
begin
  FReduceToFit:=Value;
  FEnlargeToFit:=Value;
  SetInternalZoom;
end;

procedure TrImageZoom.SetStretch(Value: boolean);
begin
  if FStretch <> Value then
  begin
    FStretch:=Value;
    SetInternalZoom;
  end;
end;

procedure TrImageZoom.SetRotateType(Value: TRotateType);
begin
  if FRotateType <> Value then
  begin
    FRotateType:=Value;
    SetInternalZoom;
  end;
end;

procedure TrImageZoom.SetBorderColor(Value: TColor);
begin
  if FBorderColor <> Value then
  begin
    FBorderColor:=Value;
    Invalidate;
  end;
end;

procedure TrImageZoom.SetBackgroundColor(Value: TColor);
begin
  if FBackgroundColor <> Value then
  begin
    FBackgroundColor:=Value;
    if FBackgroundColor <> clNone then
    begin
      ParentColor:=false;
      ParentBackground:=false;
    end;
    Invalidate;
  end;
end;

procedure TrImageZoom.SetBackgroundText(const Value: string);
begin
  if FBackgroundText <> Value then
  begin
    FBackgroundText:=Value;
    Invalidate;
  end;
end;

procedure TrImageZoom.SetShowBackgroundText(Value: boolean);
begin
  if FShowBackgroundText <> Value then
  begin
    FShowBackgroundText:=Value;
    Invalidate;
  end;
end;

procedure TrImageZoom.CMParentColorChanged(var Message: TMessage);
begin
  inherited;
  if ParentColor then
    FBackgroundColor:=clNone;
  Invalidate;
end;

function TrImageZoom.GetPalette: HPALETTE;
begin
  Result:=0;
  if FPicture.Graphic <> nil then
    Result:=FPicture.Graphic.Palette;
end;

function TrImageZoom.XSrc2Pict(X: integer): integer;
begin
  with ScaleInfo do
    Result:=SrcOffset.X + MulDiv(X, SrcSize.cX, DstSize.cX);
end;

function TrImageZoom.YSrc2Pict(Y: integer): integer;
begin
  with ScaleInfo do
    Result:=SrcOffset.Y + MulDiv(Y, SrcSize.cY, DstSize.cY);
end;

function TrImageZoom.XPict2Src(X: integer): integer;
begin
  with ScaleInfo do
    Result:=MulDiv(X, DstSize.cX, SrcSize.cX) - SrcOffset.X;
end;

function TrImageZoom.YPict2Src(Y: integer): integer;
begin
  with ScaleInfo do
    Result:=MulDiv(Y, DstSize.cY, SrcSize.cY) - SrcOffset.Y;
end;

function TrImageZoom.Src2Pict(P: TPoint): TPoint;
begin
  Result:=Point(XSrc2Pict(P.X), YSrc2Pict(P.Y));
end;

function TrImageZoom.Pict2Src(P: TPoint): TPoint;
begin
  Result:=Point(XPict2Src(P.X), YPict2Src(P.Y));
end;

function TrImageZoom.GetPictureWidth: integer;
begin
  case FRotateType of
    rt0, rt180:
      Result:=FPicture.Width;
  else
    Result:=FPicture.Height;
  end;
end;

function TrImageZoom.GetPictureHeight: integer;
begin
  case FRotateType of
    rt0, rt180:
      Result:=FPicture.Height;
  else
    Result:=FPicture.Width;
  end;
end;

procedure TrImageZoom.SetInternalZoom;
var
  w, h: double;
begin
  w:=GetPictureWidth * FZoom / 100;
  h:=GetPictureHeight * FZoom / 100;
  FInternalZoom:=FZoom;
  if FLockOriginalZoom then
    FInternalZoom:=100
  else
    if (w > 1) and (h > 1) and ((FReduceToFit and ((w > Width) or (h > Height))) or (FEnlargeToFit and (w < Width) and (h < Height))) then
    begin
      FInternalZoom:=Min(Width * 100 / GetPictureWidth, Height * 100 / GetPictureHeight);
    end;
  // self.SetZoomOffset(Point(Round((w-ClientWidth)/2*100/FZoom),Round((h-ClientHeight)/2*100/FZoom)));
  UpdateScrollBar;
  ZoomChanged;
end;

function TrImageZoom.ScaleInfo: TScaleInfo;
var
  w, h: integer;
begin
  Result.DstOffset:=Point(0, 0);
  Result.SrcOffset:=FZoomOffset;

  w:=Round(GetPictureWidth * FInternalZoom / 100);
  h:=Round(GetPictureHeight * FInternalZoom / 100);

  Result.DstSize.cX:=w;
  Result.DstSize.cY:=h;
  if ((w <= Width) and (h <= Height)) or FStretch then
  begin
    FTypeScrollBar:=tsNone;
    FCanDrag:=false;
    ShowScrollBar(Handle, SB_BOTH, false);
    Result.SrcSize.cX:=GetPictureWidth;
    Result.SrcSize.cY:=GetPictureHeight;
    Result.SrcOffset:=Point(0, 0);

    if FStretch and (w > 0) and (h > 0) then
    begin
      Result.DstSize.cX:=Self.ClientWidth;
      Result.DstSize.cY:=Self.ClientHeight;
    end
    else
    begin
      Result.DstOffset.X:=(ClientWidth - w) div 2;
      Result.DstOffset.Y:=(ClientHeight - h) div 2;
    end;
  end
  else
  begin
    if (FScrollBar) then
    begin
      FTypeScrollBar:=tsBoth;
      ShowScrollBar(Handle, SB_BOTH, true);
    end;
    FCanDrag:=true;
    if Result.DstSize.cX > ClientWidth then
      Result.DstSize.cX:=ClientWidth
    else
    begin
      Result.SrcOffset.X:=0;
      Result.DstOffset.X:=(ClientWidth - w) div 2;
    end;
    if Result.DstSize.cY > ClientHeight then
      Result.DstSize.cY:=ClientHeight
    else
    begin
      Result.SrcOffset.Y:=0;
      Result.DstOffset.Y:=(ClientHeight - h) div 2;
    end;
    Result.SrcSize.cX:=Round(Result.DstSize.cX * 100 / FInternalZoom);
    Result.SrcSize.cY:=Round(Result.DstSize.cY * 100 / FInternalZoom);
  end;
end;

function TrImageZoom.GetBkColor: TColor;
begin
  if ParentColor or (FBackgroundColor = clNone) then
    Result:=THackControl(Parent).Color
  else
    Result:=FBackgroundColor;
end;

procedure TrImageZoom.Paint;
var
  Save: boolean;
  OldBrushStyle: TBrushStyle;
  OldBrushColor: TColor;
  R, Rt: TRect;
  SI: TScaleInfo;
  pt: TPoint;
  tempBmp: TBitmap;
begin
  Save:=FDrawing;
  FDrawing:=true;
  OldBrushStyle:=Canvas.Brush.Style;
  Canvas.Brush.Style:=bsSolid;
  OldBrushColor:=Canvas.Brush.Color;

  tempBmp:=TBitmap.Create;
  try
    { Designing mode }
    if csDesigning in ComponentState then
      with Canvas do
      begin
        Pen.Style:=psDash;
        Brush.Style:=bsClear;
        Rectangle(0, 0, ClientWidth, ClientHeight);
      end;

    try
      SI:=ScaleInfo;
      With SI do
      begin
        // calc and modify bitmap
        case FRotateType of
          rt0, rt180:
            begin
              tempBmp.Width:=SrcSize.cX;
              tempBmp.Height:=SrcSize.cY;
            end
          else
          begin
            tempBmp.Width:=SrcSize.cY;
            tempBmp.Height:=SrcSize.cX;
          end;
        end;

        // fill bitmap with bk
        tempBmp.Canvas.Brush.Style:=bsSolid;
        tempBmp.Canvas.Brush.Color:=GetBkColor;
        tempBmp.Canvas.FillRect(Rect(0, 0, tempBmp.Width, tempBmp.Height));
        // load bitmap from Picture
        GetBitmapFromPicture(tempBmp, Picture);
        // rotate
        case FRotateType of
          rt090:
            rpictool.RotateBitmap(tempBmp, 90);
          rt180:
            rpictool.RotateBitmap(tempBmp, 180);
          rt270:
            rpictool.RotateBitmap(tempBmp, 270);
        end;

        // copy bitmap part with transparency
        // set size
        FBitmap.Assign(tempBmp);
        // fill bitmap with bk
        FBitmap.Canvas.Brush.Style:=bsSolid;
        FBitmap.Canvas.Brush.Color:=GetBkColor;
        FBitmap.Canvas.FillRect(Rect(0, 0, FBitmap.Width, FBitmap.Height));
        // copy
        R:=Rect(SrcOffset.X, SrcOffset.Y, SrcOffset.X + SrcSize.cX, SrcOffset.Y + SrcSize.cY);
        if FTransparent then
          FBitmap.Canvas.BrushCopy(Rect(0, 0, FBitmap.Width, FBitmap.Height), tempBmp, R, tempBmp.TransparentColor)
        else
          FBitmap.Canvas.CopyRect(Rect(0, 0, FBitmap.Width, FBitmap.Height), tempBmp.Canvas, R);

        if not ParentBackground then
        begin
          Canvas.Brush.Style:=bsSolid;
          Canvas.Brush.Color:=GetBkColor;
          Canvas.FillRect(ClientRect);
        end;

        // show bitmap
        R:=Rect(DstOffset.X, DstOffset.Y, DstSize.cX + DstOffset.X, DstSize.cY + DstOffset.Y);
        // Canvas.StretchDraw(R, FBitmap);  - replaced by HALFTONE to get smooth result
        GetBrushOrgEx(Canvas.Handle, pt);
        SetStretchBltMode(Canvas.Handle, HALFTONE);
        SetBrushOrgEx(Canvas.Handle, pt.X, pt.Y, @pt);
        StretchBlt(Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top, FBitmap.Canvas.Handle, 0, 0, FBitmap.Width, FBitmap.Height, SRCCOPY);

        // show frame
        R:=ClientRect;
        if FBorderColor <> clNone then
        begin
          BiDi_FrameRect(Canvas, R, FBorderColor);
          InflateRect(R, -1, -1);
          // Canvas.FrameRect(R);
        end;
        if FUseFocus and Focused then
          BiDi_DrawFocusRect(Canvas, R);

        { NOT USED IN DOUBLE BUFFERED
          // fill background around bitmap (reduce flicker)
          if not ParentBackground then
          begin
          oldDC:=SaveDC(Canvas.Handle);
          ExcludeClipRect(Canvas.Handle, DstOffset.X, DstOffset.Y, DstSize.cx+DstOffset.X, DstSize.cy+DstOffset.Y);
          Canvas.Brush.Color := GetBkColor;
          Canvas.Brush.Style:=bsSolid;
          Canvas.FillRect(r);
          RestoreDC(Canvas.Handle,oldDC);
          end;
          }
        // show background text
        if FShowBackgroundText and (FBackgroundText <> '') then
        begin
          Canvas.Brush.Style:=bsClear;
          Canvas.Font.Assign(Font);
          Rt:=R;
          Canvas.TextRect(Rt, FBackgroundText, [tfCenter, tfWordBreak, tfCalcRect]);
          R.Top:=(R.Bottom - R.Top - Rt.Bottom + Rt.Top) div 2 + R.Top;
          Canvas.TextRect(R, FBackgroundText, [tfCenter, tfWordBreak]);
          // .TextOut((Width - Canvas.TextWidth(FBackgroundText)) div 2, (Height - Canvas.TextHeight(FBackgroundText)) div 2, FBackgroundText);
        end;

        { OnPaint event }
        if Assigned(FOnPaint) then
          FOnPaint(Self);
      end;
    finally
      FDrawing:=Save;
      Canvas.Brush.Style:=OldBrushStyle;
      Canvas.Brush.Color:=OldBrushColor;
    end;
  finally
    FreeAndNil(tempBmp);
  end;
end;

function TrImageZoom.DoPaletteChange: boolean;
var
  ParentForm: TCustomForm;
  Tmp: TGraphic;
begin
  Result:=false;
  Tmp:=Picture.Graphic;
  if Visible and (not(csLoading in ComponentState)) and (Tmp <> nil) and (Tmp.PaletteModified) then
  begin
    if (Tmp.Palette = 0) then
      Tmp.PaletteModified:=false
    else
    begin
      ParentForm:=GetParentForm(Self);
      if Assigned(ParentForm) and ParentForm.Active and ParentForm.HandleAllocated then
      begin
        if FDrawing then
          ParentForm.Perform(wm_QueryNewPalette, 0, 0)
        else
          PostMessage(ParentForm.Handle, wm_QueryNewPalette, 0, 0);
        Result:=true;
        Tmp.PaletteModified:=false;
      end;
    end;
  end;
end;

procedure TrImageZoom.Progress(Sender: TObject; Stage: TProgressStage; PercentDone: Byte; RedrawNow: boolean; const R: TRect; const Msg: string);
begin
  if FIncrementalDisplay and RedrawNow then
  begin
    if DoPaletteChange then
      Update
    else
      Paint;
  end;
  if Assigned(FOnProgress) then
    FOnProgress(Sender, Stage, PercentDone, RedrawNow, R, Msg);
end;

procedure TrImageZoom.SetTransparent(Value: boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent:=Value;
    PictureChanged(Self);
  end;
end;

procedure TrImageZoom.UpdateScrollBar;
var
  maxWidth, maxHeight: integer;
  ScrollInfo: TScrollInfo;
begin
  with ScaleInfo do
  begin
    maxWidth:=GetPictureWidth - SrcSize.cX;
    maxHeight:=GetPictureHeight - SrcSize.cY;
    FZoomOffset:=SrcOffset;
    if FZoomOffset.X < 0 then
      FZoomOffset.X:=0;
    if FZoomOffset.Y < 0 then
      FZoomOffset.Y:=0;
    if (maxWidth > 0) and (maxWidth < FZoomOffset.X) then
      FZoomOffset.X:=maxWidth;
    if (maxHeight > 0) and (maxHeight < FZoomOffset.Y) then
      FZoomOffset.Y:=maxHeight;

    if (HandleAllocated) and not(FDrawing) then
    begin
      if (FScrollBar) then
      begin
        case FTypeScrollBar of
          tsNone:
            begin
              // SetScrollRange(Handle, SB_BOTH, 0, 0, false);
            end;
          tsBoth:
            begin
              ScrollInfo.cbSize:=Sizeof(ScrollInfo);
              ScrollInfo.fMask:=SIF_RANGE;
              ScrollInfo.nMin:=0;
              ScrollInfo.nMax:=GetPictureWidth - 1;
              SetScrollInfo(Handle, SB_HORZ, ScrollInfo, false);
              ScrollInfo.nMin:=0;
              ScrollInfo.nMax:=GetPictureHeight - 1;
              SetScrollInfo(Handle, SB_VERT, ScrollInfo, false);
              { aggiorno valore scrollbar }
              ScrollInfo.fMask:=SIF_POS;
              ScrollInfo.nPos:=FZoomOffset.Y;
              SetScrollInfo(Handle, SB_VERT, ScrollInfo, false);
              ScrollInfo.nPos:=FZoomOffset.X;
              SetScrollInfo(Handle, SB_HORZ, ScrollInfo, false);
              { set page scrollbars }
              ScrollInfo.cbSize:=Sizeof(ScrollInfo);
              ScrollInfo.fMask:=SIF_PAGE;
              ScrollInfo.nPage:=SrcSize.cX;
              SetScrollInfo(Handle, SB_HORZ, ScrollInfo, true);
              ScrollInfo.cbSize:=Sizeof(ScrollInfo);
              ScrollInfo.fMask:=SIF_PAGE;
              ScrollInfo.nPage:=SrcSize.cY;
              SetScrollInfo(Handle, SB_VERT, ScrollInfo, true);
            end;
        end;
        {
          TForm(Owner).caption := format('zo:%d x %d, i:%d x %d, c:%d x %d, s:%d x %d, d:%d x %d',
          [FZoomOffset.X, FZoomOffset.Y,
          FPicture.Width, FPicture.Height,
          ClientWidth, ClientHeight,
          SrcSize.cx, SrcSize.cy,
          DstSize.cx, DstSize.cy]);
          }
      end
      else
        ShowScrollBar(Handle, SB_BOTH, false);
    end;
  end;
  if not FDrawing then
    Invalidate;
end;

procedure TrImageZoom.PictureChanged(Sender: TObject);
var
  G: TGraphic;
begin
  G:=FPicture.Graphic;
  if Assigned(G) then
    if DoPaletteChange and FDrawing then
      Update;
  SetInternalZoom;
end;

procedure TrImageZoom.Fit;
begin
  FZoomOffset:=Point(0, 0);
  if (GetPictureWidth > 0) and (GetPictureHeight > 0) then
  begin
    Zoom:=Min(Trunc((Width * 100) / GetPictureWidth), Trunc((Height * 100) / GetPictureHeight));
  end;
  // FReduceToFit := true;
  // FEnlargeToFit:=true;
end;

procedure TrImageZoom.ZoomChanged;
begin
  if Assigned(FOnChangeZoom) then
    FOnChangeZoom(Self);
end;

procedure TrImageZoom.DoResize;
begin
  SetInternalZoom;
  if Assigned(FOnResize) then
    FOnResize(Self);
end;

procedure TrImageZoom.WMEraseBkgnd(var Message: TMessage);
begin
  if ParentBackground then
    inherited
  else
    Message.Result:=0;
end;

procedure TrImageZoom.WMSize(var Message: TWMSize);
begin
  inherited;
  if not(csLoading in ComponentState) then
    DoResize;
end;

procedure TrImageZoom.WMHScroll(var Message: TWMHScroll);
var
  sp: integer;
begin
  sp:=ScrollMessage(SB_HORZ, Message, ClientHeight);
  if sp <> FZoomOffset.X then
  begin
    FZoomOffset:=Point(sp, FZoomOffset.Y);
    UpdateScrollBar;
  end;
end;

procedure TrImageZoom.WMVScroll(var Message: TWMVScroll);
var
  sp: integer;
begin
  sp:=ScrollMessage(SB_VERT, Message, ClientHeight);
  if sp <> FZoomOffset.Y then
  begin
    FZoomOffset:=Point(FZoomOffset.X, sp);
    UpdateScrollBar;
  end;
end;

function TrImageZoom.ScrollMessage(SbType: integer; var Msg: TWMScroll; ClientSize: integer): integer;
var
  ScrollPos, MinPos, MaxPos: integer;
begin
  with Msg do
  begin
    ScrollPos:=GetScrollPos(Handle, SbType);
    GetScrollRange(Handle, SbType, MinPos, MaxPos);
    case ScrollCode of
      SB_LINEUP:
        Dec(ScrollPos);
      SB_LINEDOWN:
        inc(ScrollPos);
      SB_PAGEUP:
        Dec(ScrollPos, (ClientSize div 2));
      SB_PAGEDOWN:
        inc(ScrollPos, (ClientSize div 2));
      SB_THUMBPOSITION, SB_THUMBTRACK:
        ScrollPos:=Pos;
      SB_TOP:
        ScrollPos:=MinPos;
      SB_BOTTOM:
        ScrollPos:=MaxPos;
    end;
    if ScrollPos < MinPos then
      ScrollPos:=MinPos;
    if ScrollPos > MaxPos then
      ScrollPos:=MaxPos;
  end;
  Result:=ScrollPos;
end;

procedure TrImageZoom.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
  if FDragAndMove and (Button = mbLeft) then
  begin
    BeginDrag(true);
    FDragStartPoint:=Point(X, Y);
    FDragStartOffset:=ZoomOffset;
    FIsDraging:=true;
  end;
end;

procedure TrImageZoom.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
  FIsDraging:=false;
end;

procedure TrImageZoom.DragOver(Source: TObject; X, Y: integer; State: TDragState; var Accept: boolean);
begin
  inherited;
  if FDragAndMove then
  begin
    Accept:=FCanDrag;
    if not FIsDraging then
      Exit;
    SetZoomOffset(Point(FDragStartOffset.X + Round((FDragStartPoint.X - X) * 100 / FInternalZoom), FDragStartOffset.Y + Round
          ((FDragStartPoint.Y - Y) * 100 / FInternalZoom)));
  end;
end;

procedure TrImageZoom.ZoomIn(Percent: integer = 25; Center: boolean = true);
begin
  Zoom:=Ceil(FInternalZoom * (100 + Percent) / 100);
  if Center then
    MoveToCenter;
end;

procedure TrImageZoom.ZoomOut(Percent: integer = 25; Center: boolean = true);
begin
  Zoom:=Floor(FInternalZoom * (100 - Percent) / 100);
  if Center then
    MoveToCenter;
end;

function TrImageZoom.CanModify: boolean;
begin
  Result:=not FReadOnly;
end;

function TrImageZoom.CutImageToClipboard: boolean;
begin
  if CanModify then
  begin
    Result:=Assigned(Picture);
    Clipboard.Assign(Picture);
    ImgGraphicType:=rpictool.gtNone;
    Picture.Graphic:=nil;
  end
  else
    Result:=false;
end;

function TrImageZoom.CopyImageToClipboard: boolean;
begin
  Result:=true;
  try
    Clipboard.Assign(Picture);
  except
    Result:=false;
  end;
end;

function TrImageZoom.PasteImageFromClipboard: boolean;
begin
  if CanModify then
  begin
    Result:=true;
    if Clipboard.HasFormat(CF_BITMAP) then
      Picture.Bitmap.Assign(Clipboard)
    else
      if Clipboard.HasFormat(CF_PICTURE) then
        Picture.Graphic.Assign(Clipboard)
      else
        Result:=false;
  end
  else
    Result:=false;
end;

function TrImageZoom.SaveAsFile(const FileName: string): boolean;
var
  GrType: TrGraphicType;
  oldCursor: TCursor;
begin
  oldCursor:=WCursor;
  try
    if ImgGraphicType = rpictool.gtNone then
      if FSaveInJPG then
        ImgGraphicType:=gtJPG
      else
        ImgGraphicType:=gtBMP;
    if FileName = '' then
      GrType:=SaveDlg_PictureToFile(Picture, '', ImgGraphicType)
    else
      GrType:=SavePictureToFile(Picture, FileName, rpictool.gtNone);
    Result:=not(GrType in [rpictool.gtNone, rpictool.gtError]);
    if Result then
    begin
      ImgGraphicType:=GrType;
      LastFileName:=RosiCompConfig.rDBToolCfg.rpictool_LastFileName;
    end;
  finally
    ChCursor(oldCursor);
  end;
end;

function TrImageZoom.LoadFromFile(const FileName: string): boolean;
var
  GrType: TrGraphicType;
  oldCursor: TCursor;
begin
  if CanModify then
  begin
    oldCursor:=WCursor;
    try
      if ImgGraphicType = rpictool.gtNone then
        if FSaveInJPG then
          ImgGraphicType:=gtJPG
        else
          ImgGraphicType:=gtBMP;
      if FileName = '' then
        GrType:=OpenDlg_PictureFromFile(Picture, '', ImgGraphicType)
      else
        GrType:=LoadPictureFromFile(Picture, FileName);
      Result:=not(GrType in [rpictool.gtNone, rpictool.gtError]);
      if Result then
      begin
        ImgGraphicType:=GrType;
        LastFileName:=RosiCompConfig.rDBToolCfg.rpictool_LastFileName;
        // !!! picture must be updated to save JPG picture = bug?
        Self.Update;
      end;
    finally
      ChCursor(oldCursor);
    end;
  end
  else
    Result:=false;
end;

function TrImageZoom.ResizePic(ratio: double): boolean;
var
  BmpImage: TBitmap;
  R: integer;
  oldCursor: TCursor;
begin
  Result:=false;
  if CanModify then
  begin
    if ratio = 0 then
    begin
      R:=50;
      if not rdlg.InputInt(sInput, sResizeRatioQ, R) then
        Exit;
      if (R < 1) or (R > 500) then
        Exit;
      ratio:=R;
    end;
    oldCursor:=WCursor;
    BmpImage:=TBitmap.Create;
    try
      BmpImage.Assign(Picture.Graphic);
      rpictool.SmoothResizeBitmap(BmpImage, Trunc(BmpImage.Width * ratio / 100), Trunc(BmpImage.Height * ratio / 100), true);
      Picture.Assign(BmpImage);
      Result:=true;
    finally
      FreeAndNil(BmpImage);
      ChCursor(oldCursor);
    end;
  end;
end;

function TrImageZoom.PictureInfo: string;
begin
  if (ImgGraphicType = rpictool.gtNone) and Assigned(Picture.Graphic) then
  begin
    if Pos('JPEG', Picture.Graphic.ClassName) > 0 then
      ImgGraphicType:=rpictool.gtJPG
    else
      ImgGraphicType:=rpictool.gtBMP;
  end;
  case ImgGraphicType of
    rpictool.gtNone:
      Result:=sPictureEmpty;
    gtError:
      Result:=sPictureCorrupted;
  else
    Result:=GraphicTypeLabel[ImgGraphicType] + ': ' + Format('%d x %d; %s: %d%%', [Picture.Width, Picture.Height, sZoom, Round(FInternalZoom)]);
  end;
end;

function TrImageZoom.ShowEditorWindow: boolean;
begin
  Result:=false;
  if rImageEdit.ShowImageEdit(nil, Picture, nil, FShowEditor in [ieShowModal, ieEditModal], (FShowEditor = ieEditModal) and CanModify) then
  begin
    // update
    Self.Picture.Assign(Picture);
    Result:=true;
  end;
end;

function TrImageZoom.GetPopupItem(Tag: integer): TMenuItem;
var
  i: integer;
begin
  Result:=nil;
  for i:=0 to FPopupMenu.Items.Count - 1 do
    if FPopupMenu.Items[i].Tag = Tag then
    begin
      Result:=FPopupMenu.Items[i];
      Break;
    end;
end;

procedure TrImageZoom.OnPopupMenu(Sender: TObject);
var
  cm, ex: boolean;
begin
  if FirstOpenMenu then
    GetPopupItem(cPicSaveInJpgFormat).Checked:=SaveInJPG;
  FirstOpenMenu:=false;
  FPopupMenu.Items[0].Caption:=PictureInfo;
  GetPopupItem(cPicZoomFit).Checked:=ReduceToFit;

  cm:=CanModify;
  ex:=not(ImgGraphicType in [rpictool.gtNone, gtError]);
  GetPopupItem(cPicCut).Enabled:=cm and ex;
  GetPopupItem(cPicCopy).Enabled:=ex;
  GetPopupItem(cPicPaste).Enabled:=cm;
  GetPopupItem(cPicLoadFromFile).Enabled:=cm;
  GetPopupItem(cPicSaveToFile).Enabled:=ex;
  GetPopupItem(cPicZoomIn).Enabled:=ex;
  GetPopupItem(cPicZoomOut).Enabled:=ex;
  GetPopupItem(cPicResize).Enabled:=cm and ex;
  GetPopupItem(cPicEdit).Enabled:=ex;
  GetPopupItem(cPicEdit).Visible:=(ShowEditor <> ieNone);
end;

procedure TrImageZoom.PopupClick(Sender: TObject);
begin
  case TMenuItem(Sender).Tag of
    cPicCut:
      CutImageToClipboard;
    cPicCopy:
      CopyImageToClipboard;
    cPicPaste:
      PasteImageFromClipboard;
    cPicLoadFromFile:
      LoadFromFile('');
    cPicSaveToFile:
      SaveAsFile('');
    cPicZoomFit:
      AutoFit:=not AutoFit;
    {
      if ReduceToFit then
      begin
      AutoFit:=false;
      Zoom:=100;
      end
      else
      begin
      Zoom:=100;
      ReduceToFit:=true;
      end;
      }
    cPicZoomIn:
      begin
        ZoomIn;
        AutoFit:=false;
      end;
    cPicZoomOut:
      begin
        ZoomOut;
        AutoFit:=false;
      end;
    cPicResize:
      ResizePic(0);
    cPicEdit:
      ShowEditorWindow;
    cPicSaveInJpgFormat:
      SaveInJPG:=GetPopupItem(cPicSaveInJpgFormat).Checked;
  end;
end;

procedure TrImageZoom.SetReadOnly(Value: boolean);
begin
  FReadOnly:=Value;
end;

end.
