unit rStyleHooks;

{<
  Style hook for TrPageControlEx and TrTabControlEx.
  Draw styled close tab button.
  Add to uses part of your application to use.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2006 - Jun 2013)
  @lastmod(Dec 2015)
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, rImprovedComps, StdCtrls, rHTMLLabel, Menus, ImgList,
  Vcl.Styles,
  Vcl.Themes;

type
  { Style hook for TrPageControlEx and TrTabControlEx }
  TrTabControlExStyleHook = class(TTabControlStyleHook)
  private
    TabPropsEx: TTabPropsEx;
  strict protected
    procedure DrawTab(Canvas: TCanvas; Index: Integer); override;
  public
    constructor Create(AControl: TWinControl); override;
  end;

implementation

constructor TrTabControlExStyleHook.Create(AControl: TWinControl);
begin
  inherited;
  if AControl is TrPageControlEx then
    TabPropsEx := TrPageControlEx(AControl).TabPropsEx
  else if AControl is TrTabControlEx then
    TabPropsEx := TrTabControlEx(AControl).TabPropsEx
  else
    TabPropsEx := nil;
end;

procedure TrTabControlExStyleHook.DrawTab(Canvas: TCanvas; Index: Integer);
var
  Details: TThemedElementDetails;
  ButtonR: TRect;
  FButtonState: TThemedWindow;
begin
  inherited;
  if not Assigned(TabPropsEx) then
    Exit;
  if TabPropsEx.CloseBtnShown(Index) then
  begin
    if Index=TabPropsEx.HotBtnIndex then
      FButtonState := twSmallCloseButtonHot
    else
      if Index = TabIndex then
        FButtonState := twSmallCloseButtonNormal
      else
        FButtonState := twSmallCloseButtonNormal; //twSmallCloseButtonDisabled;

    Details := StyleServices.GetElementDetails(FButtonState);
    ButtonR := TabPropsEx.CloseBtnRect(Index);
    if ButtonR.Bottom - ButtonR.Top > 0 then
      StyleServices.DrawElement(Canvas.Handle, Details, ButtonR);
  end;
end;

Initialization

  TStyleManager.Engine.RegisterStyleHook(TCustomTabControl, TrTabControlExStyleHook);
  TStyleManager.Engine.RegisterStyleHook(TTabControl, TrTabControlExStyleHook);

end.
