unit rTip;

{<
  rTip form provides "Tip of the day" dialog
  or OK/Cancel dialog with "Do not show next time" function.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2006)
  @lastmod(Jan 2013)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, rHTMLLabel, ExtCtrls, PropSaveMain, ImgList,
  pngimage, rdlg_baseform, rImprovedComps, RosiCompGlobalCfg;

{ show OK dialog, if "Do not show next time" has been checked already dialog is not shown,
  PropSaveFile and ID is used for saving of "Dont show next time" }
procedure ShowTip(FormatedInfo: string; Id: string; PropSaveFile: TPropSaveFile);

{ show OK dialog, if "Do not show next time" has been checked already dialog is not shown,
  PropSaveFile and ID is used for saving of "Dont show next time",
  OnAnchorClick is callback procedure used if user click to anchor in TrHTMLLabel
  AlwaysShow overwrite "Dont show next time" }
procedure ShowTipEx(FormatedInfo: string; Id: string; PropSaveFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; AlwaysShow: boolean);

{ show OK Cancel dialog, if "Do not show next time" has been checked already
  dialog is not shown, PropSaveFile and ID is used for saving of "Dont show next time" }
function ShowTipQ(FormatedQuestion: string; Id: string; PropSaveFile: TPropSaveFile): integer;

{ show OK Cancel dialog, if "Do not show next time" has been checked already
  dialog is not shown, PropSaveFile and ID is used for saving of "Dont show next time",
  OnAnchorClick is callback procedure used if user click to anchor in TrHTMLLabel
  AlwaysShow overwrite "Dont show next time" }
function ShowTipQEx(FormatedQuestion: string; Id: string; PropSaveFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; AlwaysShow: boolean): integer;

{ delete all "Do not show next time" settings in PropSaveFile }
procedure EnableAllTips(PropSaveFile: TPropSaveFile);

type
  { "Tip of the day" dialog or OK/Cancel dialog
    with "Do not show next time" function }
  TrTipForm = class(TrDlgBaseForm)
    Image1: TImage;
    lTip: TrHTMLLabel;
    Bevel1: TBevel;
    pQuest: TPanel;
    cbSuppresNextTime1: TCheckBox;
    cbSuppresNextTime2: TCheckBox;
    Panel1: TPanel;
    btOk2: TrButton;
    procedure DlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure DlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
  private
    FId: string;
  public
  end;

var
  rTipForm: TrTipForm;

implementation

{$R *.dfm}

uses rtool, rdlg;

function _ShowTip(Info: string; Id: string; PSFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; Q, A: boolean): integer; forward;

procedure ShowTip(FormatedInfo: string; Id: string; PropSaveFile: TPropSaveFile);
begin
  _ShowTip(FormatedInfo, Id, PropSaveFile, nil, false, false);
end;

procedure ShowTipEx(FormatedInfo: string; Id: string; PropSaveFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; AlwaysShow: boolean);
begin
  _ShowTip(FormatedInfo, Id, PropSaveFile, OnAnchorClick, false, AlwaysShow);
end;

function ShowTipQ(FormatedQuestion: string; Id: string; PropSaveFile: TPropSaveFile): integer;
begin
  Result:=_ShowTip(FormatedQuestion, Id, PropSaveFile, nil, true, false);
end;

function ShowTipQEx(FormatedQuestion: string; Id: string; PropSaveFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; AlwaysShow: boolean): integer;
begin
  Result:=_ShowTip(FormatedQuestion, Id, PropSaveFile, OnAnchorClick, true, AlwaysShow);
end;

function _ShowTip(Info: string; Id: string; PSFile: TPropSaveFile; OnAnchorClick: TAnchorClickEvent; Q, A: boolean): integer;
begin
  rTipForm:=TrTipForm.Create(GetActiveOwner);
  with rTipForm do
  begin
    FId:=Id;
    DlgPropSave.PropSaveFile:=PSFile;
    pBottom.Visible:=Q;
    pQuest.Visible:=not Q;
    if pQuest.Visible then
      ActiveControl:=btOk2;
    cbSuppresNextTime1.Enabled:=Assigned(PSFile);
    cbSuppresNextTime2.Enabled:=cbSuppresNextTime1.Enabled;
    DlgPropSave.ReadProperties;
    Result:=mrOK;
    if A or not cbSuppresNextTime1.Checked then
    begin
      lTip.Width:=ClientWidth - lTip.Left - 20;
      lTip.Caption:=Info;
      // Set window size according Label size
      if lTip.Height > lTip.Width - 10 then
        lTip.Width:=Round(lTip.Width * 1.5);
      if lTip.Width > Screen.Width then
        lTip.Width:=Screen.Width - 65;
      if lTip.Left + lTip.Width > ClientWidth then
        ClientWidth:=lTip.Left + lTip.Width + 20;
      if lTip.Top + lTip.Height > ClientHeight - pBottom.Height then
        ClientHeight:=lTip.Top + lTip.Height + pBottom.Height + 5;
      lTip.OnAnchorClick:=OnAnchorClick;
      Result:=ShowModal;
      if Q then
        cbSuppresNextTime1.Checked:=cbSuppresNextTime2.Checked;
      DlgPropSave.SaveProperties;
    end;
  end;
  FreeAndNil(rTipForm);
end;

procedure TrTipForm.DlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
begin
  cbSuppresNextTime1.Checked:=PSFile.ReadBool(Key, FId, false);
  cbSuppresNextTime2.Checked:=PSFile.ReadBool(Key, FId, false);
end;

procedure TrTipForm.DlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteBool(Key, FId, cbSuppresNextTime1.Checked);
end;

procedure EnableAllTips(PropSaveFile: TPropSaveFile);
begin
  if Assigned(PropSaveFile) then
    PropSaveFile.EraseSection('rTipForm');
end;

end.
