unit RTFRep_DialogEx;

{<
  RTFRep_DialogEx provides universal dialog before execution of report.
  User can choose report from list of available reports and some properties.
  List of reports is defined by array of TRTFRepInfoEx.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2008)
  @lastmod(Sep 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Buttons, ExtCtrls, DB,
  RTFReport2, AppDirCfg, ShlObj, ImgList, rHTMLLabel,
  rImprovedComps, PropSaveMain, rMapiMail, rdbdlg_baseform;

type
  TDlgAutoAction = (daaNone, daaShow, daaPrint, daaMail);

  { List of datasets for one report }
  TDataSetListEx = array of TDataSet;

  { Information record for each report (each item) -
    define all properties of report }
  TRTFRepInfoEx = record
    GroupName: string;
    Caption: string;
    FormattedInfo: string;
    Disabled: boolean;
    ReportFile: string;
    OutputFile: string;
    DataSets: TDataSetListEx;
    DefaultCharset: integer;
    FieldNotFoundValue: string;
    Variables: array of string;
    ShowDetail: boolean;
    ReportVersion: integer;
    PrintCopies: byte;
  end;

  { Information record for dialog, define dialog events and properties }
  TConfigPDFPrinter = function(OutputPDFFile: string; var PDFPrinterName: string): boolean;

  { define structure for reports definition info }
  TRTFRepDlgInfoClass = class
    OnBeforeEdit: TNotifyEvent;
    OnReadInfo: TNotifyEvent;
    OnAfterRun: TNotifyEvent;
    ShowMailBtn: boolean;
    ShowPDFBtn: boolean;
    OnConfigPDFPrinter: TConfigPDFPrinter;
    OnPDFSetup: TNotifyEvent;
    ShowPDFSaveAsDlg: boolean;
    UseMapi: boolean;
    MailTo: string;
    MailApp: string;
    SendMailDir: string;
    AutoAction: TDlgAutoAction;
    LangVersions: array of string;
    LangSelected: string;
    constructor Create;
  end;

procedure ShowRTFRepDialogEx(const RTFRepInfo: array of TRTFRepInfoEx; RTFRepDlgInfo: TRTFRepDlgInfoClass);
procedure SetRTFRepInfo(var RTFRepInfo: TRTFRepInfoEx; Group, Caption, Info, ReportFile, OutputFile: string; const DataSets: array of TDataSet);
function CreateDataSetListEx(const DataSets: array of TDataSet): TDataSetListEx;
procedure AddRTFRepVariableEx(var RTFRepInfo: TRTFRepInfoEx; Variable: string);
procedure ClearRTFRepVariableEx(var RTFRepInfo: TRTFRepInfoEx);

type
  { universal dialog before execution of RTFReport }
  TRTFRepDialogEx = class(TrDBDlgBaseForm)
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    RTFRep2: TRTFReport2;
    AppDirCfg: TAppDirCfg;
    TimerPDF: TTimer;
    pcRun: TPageControl;
    tsButton: TTabSheet;
    btPDFSetup: TrURLLabel;
    Label2: TLabel;
    chPDF: TCheckBox;
    btShow: TrButton;
    btPrint: TrButton;
    eCopies: TrEditNum;
    btMail: TrButton;
    btClose: TrButton;
    btGen: TrButton;
    tsCancel: TTabSheet;
    Label1: TLabel;
    lInfo2: TLabel;
    btAbort: TrButton;
    Label7: TLabel;
    Label8: TLabel;
    tv: TTreeView;
    ScrollBox1: TScrollBox;
    Info: TrHTMLLabel;
    gbRep: TGroupBox;
    fRep: TrHTMLLabel;
    btEdit: TrURLLabel;
    btOpen: TrButton;
    gbOut: TGroupBox;
    fOut: TrHTMLLabel;
    btDefFile: TrURLLabel;
    btSave: TrButton;
    gbDetail: TGroupBox;
    Label6: TLabel;
    FormatLabel1: TrHTMLLabel;
    FormatLabel2: TrHTMLLabel;
    eCharset: TrEditNum;
    eNotFound: TEdit;
    eVar: TMemo;
    cbLang: TComboBox;
    PdfSaveDlg: TSaveDialog;
    lblLang: TLabel;
    MapiEmail: TrMapiEmail;
    procedure FormShow(Sender: TObject);
    procedure btShowClick(Sender: TObject);
    procedure btAbortClick(Sender: TObject);
    procedure RTFRep1ReportProcess(Sender: TObject; RecNbr: integer; var Abort: boolean);
    procedure btEditClick(Sender: TObject);
    procedure btOpenClick(Sender: TObject);
    procedure btSaveClick(Sender: TObject);
    procedure tvChange(Sender: TObject; Node: TTreeNode);
    procedure tvChanging(Sender: TObject; Node: TTreeNode; var AllowChange: boolean);
    procedure btDefFileClick(Sender: TObject);
    procedure tvCustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: boolean);
    procedure FormActivate(Sender: TObject);
    procedure tvDblClick(Sender: TObject);
    procedure cbLangChange(Sender: TObject);
    procedure TimerPDFTimer(Sender: TObject);
    procedure btPDFSetupClick(Sender: TObject);
  private
    RepInfo: array of TRTFRepInfoEx;
    RepDlgInfo: TRTFRepDlgInfoClass;
    Aborted: boolean;
    PDFFile: string;
    PDFPrinterName2: string;
    PDFWaitCounter: integer;
    function RunRTF(DS: array of TDataSet): boolean;
    procedure SendMail(Attachment: string);
    procedure OnReadInfo(Sender: TObject);
    function WaitForPDFFile: boolean;
    function SetPDFProps: boolean;
    procedure UpdateInfo(Info: string);
  public
  end;

var
  RTFRepDialogEx: TRTFRepDialogEx;

implementation

uses RosiCompGlobalCfg, rstring, rtool, rdlg, rdbtool;

const
  PDFWaitSeconds2 = 10;
{$R *.dfm}

constructor TRTFRepDlgInfoClass.Create;
begin
  self.OnBeforeEdit:=nil;
  self.OnReadInfo:=nil;
  self.OnAfterRun:=nil;
  self.OnConfigPDFPrinter:=nil;
  self.OnPDFSetup:=nil;
end;

procedure SetRTFRepInfo(var RTFRepInfo: TRTFRepInfoEx; Group, Caption, Info, ReportFile, OutputFile: string; const DataSets: array of TDataSet);
begin
  RTFRepInfo.GroupName:=Group;
  RTFRepInfo.Caption:=Caption;
  RTFRepInfo.FormattedInfo:=Info;
  RTFRepInfo.Disabled:=false;
  RTFRepInfo.ReportFile:=ReportFile;
  RTFRepInfo.OutputFile:=OutputFile;
  RTFRepInfo.DataSets:=CreateDataSetListEx(DataSets);
  RTFRepInfo.DefaultCharset:=0;
  RTFRepInfo.FieldNotFoundValue:='';
  RTFRepInfo.ShowDetail:=true;
  RTFRepInfo.ReportVersion:=2;
  RTFRepInfo.PrintCopies:=1;
end;

function CreateDataSetListEx(const DataSets: array of TDataSet): TDataSetListEx;
var
  i: integer;
begin
  SetLength(Result, High(DataSets) + 1);
  for i:=0 to High(DataSets) do
    Result[i]:=DataSets[i];
end;

procedure AddRTFRepVariableEx(var RTFRepInfo: TRTFRepInfoEx; Variable: string);
var
  h: integer;
begin
  h:= High(RTFRepInfo.Variables);
  SetLength(RTFRepInfo.Variables, h + 2);
  RTFRepInfo.Variables[h + 1]:=Variable;
end;

procedure ClearRTFRepVariableEx(var RTFRepInfo: TRTFRepInfoEx);
begin
  SetLength(RTFRepInfo.Variables, 0);
end;

procedure ShowRTFRepDialogEx(const RTFRepInfo: array of TRTFRepInfoEx; RTFRepDlgInfo: TRTFRepDlgInfoClass);
var
  i: integer;
begin
  if not Assigned(RTFRepDlgInfo) then
    Exit;
  RTFRepDialogEx:=TRTFRepDialogEx.Create(GetActiveOwner);
  try
    SetLength(RTFRepDialogEx.RepInfo, High(RTFRepInfo) + 1);
    for i:=0 to High(RTFRepInfo) do
      RTFRepDialogEx.RepInfo[i]:=RTFRepInfo[i];
    RTFRepDialogEx.RepDlgInfo:=RTFRepDlgInfo;
    RTFRepDialogEx.btMail.Visible:=RTFRepDlgInfo.ShowMailBtn;
    RTFRepDialogEx.chPDF.Visible:=RTFRepDlgInfo.ShowPDFBtn;
    RTFRepDialogEx.chPDF.Enabled:=Assigned(RTFRepDlgInfo.OnConfigPDFPrinter);
    RTFRepDialogEx.btPDFSetup.Visible:=RTFRepDlgInfo.ShowPDFBtn and Assigned(RTFRepDlgInfo.OnPDFSetup);

    if RTFRepDlgInfo.AutoAction <> daaNone then
    begin
      RTFRepDialogEx.Show;
      if RTFRepDialogEx.btShow.Enabled then
      begin
        RTFRepDialogEx.Update;
        if RTFRepDlgInfo.AutoAction = daaShow then
          RTFRepDialogEx.btShow.Click;
        if RTFRepDlgInfo.AutoAction = daaPrint then
          RTFRepDialogEx.btPrint.Click;
        if RTFRepDlgInfo.AutoAction = daaMail then
          RTFRepDialogEx.btMail.Click;
      end;
    end
    else
      RTFRepDialogEx.ShowModal;
  finally
    FreeF(RTFRepDialogEx);
  end;
end;

procedure TRTFRepDialogEx.FormShow(Sender: TObject);
var
  i: integer;
  s: string;
  n: TTreeNode;
begin
  inherited;
  pcRun.ActivePage:=tsButton;
  OpenDlg.Filter:=sFileTypeRTF + '|' + sFileTypeAll;
  SaveDlg.Filter:=OpenDlg.Filter;
  PdfSaveDlg.Filter:=sFileTypePDF + '|' + sFileTypeAll;
  // proc langver
  cbLang.Items.Clear;
  for i:=0 to High(RepDlgInfo.LangVersions) do
  begin
    s:=RepDlgInfo.LangVersions[i];
    s:=rtool.FindStrPart(s, ';', 1, true);
    cbLang.Items.Add(s);
    if SameText(s, RepDlgInfo.LangSelected) then
      RosiCompConfig.RTFRepDlgCfg.LastLangVerIndex:=i;
  end;
  cbLang.Visible:=cbLang.Items.Count > 0;
  lblLang.Visible:=cbLang.Visible;
  if cbLang.Visible then
  begin
    if (RosiCompConfig.RTFRepDlgCfg.LastLangVerIndex >= 0) and (RosiCompConfig.RTFRepDlgCfg.LastLangVerIndex < cbLang.Items.Count) then
      cbLang.ItemIndex:=RosiCompConfig.RTFRepDlgCfg.LastLangVerIndex
    else
      cbLang.ItemIndex:=0;
  end
  else
  begin
    tv.Height:=gbDetail.BoundsRect.Bottom - tv.Top;
  end;
  // proc RepInfos
  for i:=0 to High(RepInfo) do
  begin
    if RepInfo[i].GroupName <> '' then
    begin
      // find or create group
      n:=tv.Items.GetFirstNode;
      repeat
        if n = nil then
        begin
          // create group
          n:=tv.Items.AddObject(nil, RepInfo[i].GroupName, Pointer(-1));
        end;
        if (n.Text = RepInfo[i].GroupName) and (integer(n.Data) = -1) then
        begin
          // add items
          tv.Items.AddChildObject(n, RepInfo[i].Caption, Pointer(i));
          Break;
        end;
        n:=n.GetNextSibling;
      until false;
    end
    else
      tv.Items.AddChildObject(nil, RepInfo[i].Caption, Pointer(i));

  end;
  if tv.Items.Count > 14 then
    tv.FullCollapse
  else
    tv.FullExpand;
  n:=tv.Items.GetFirstNode;
  if n <> nil then
    n.Selected:=true;
end;

procedure TRTFRepDialogEx.FormActivate(Sender: TObject);
begin ;
end;

procedure TRTFRepDialogEx.tvCustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: boolean);
var
  d: integer;
begin
  DefaultDraw:=true;
  if Node = nil then
    Exit;
  d:=integer(Node.Data);
  if (d <> -1) and (RepInfo[d].Disabled) then
    tv.Canvas.Font.Color:=clGrayText;
end;

procedure TRTFRepDialogEx.tvChanging(Sender: TObject; Node: TTreeNode; var AllowChange: boolean);
var
  d, i: integer;
begin
  if tv.Selected = nil then
    Exit;
  d:=integer(tv.Selected.Data);
  if d = -1 then
    Exit;
  with RepInfo[d] do
  begin
    // ReportFile:=fRep.Caption;
    OutputFile:=fOut.Caption;
    DefaultCharset:=eCharset.AsInteger;
    FieldNotFoundValue:=eNotFound.Text;
    SetLength(Variables, eVar.Lines.Count);
    for i:=0 to eVar.Lines.Count - 1 do
      Variables[i]:=eVar.Lines[i];
  end;
end;

procedure TRTFRepDialogEx.tvChange(Sender: TObject; Node: TTreeNode);
var
  d, i: integer;
  s: string;
begin
  d:=integer(Node.Data);
  btPrint.Enabled:=(d <> -1) and (not RepInfo[d].Disabled);
  btShow.Enabled:=btPrint.Enabled;
  btMail.Enabled:=btPrint.Enabled;
  btGen.Enabled:=btPrint.Enabled;
  if d = -1 then
  begin
    Info.Caption:='';
    fRep.Caption:='';
    fOut.Caption:='';
    gbDetail.Visible:=false;
  end
  else
  begin
    with RepInfo[d] do
    begin
      Info.Caption:=FormattedInfo;
      gbDetail.Visible:=ShowDetail;
      if cbLang.ItemIndex >= 0 then
        s:=self.RepDlgInfo.LangVersions[cbLang.ItemIndex]
      else
        s:='';
      s:=rtool.FindStrPart(s, ';', 2, true);
      s:=ExtractFilePath(ReportFile) + ReplaceStr('.rtf', s + '.rtf', ExtractFileName(ReportFile), false, false);
      fRep.Caption:=s;
      fOut.Caption:=OutputFile;
      eCharset.AsInteger:=DefaultCharset;
      eNotFound.Text:=FieldNotFoundValue;
      eVar.Lines.Clear;
      for i:=0 to High(Variables) do
        eVar.Lines.Add(Variables[i]);
      if PrintCopies = 0 then
        PrintCopies:=1;
      eCopies.AsInteger:=PrintCopies;
    end;
  end;
end;

procedure TRTFRepDialogEx.cbLangChange(Sender: TObject);
begin
  RosiCompConfig.RTFRepDlgCfg.LastLangVerIndex:=cbLang.ItemIndex;
  tvChange(nil, tv.Selected);
end;

procedure TRTFRepDialogEx.UpdateInfo(Info: string);
begin
  lInfo2.Caption:=Info;
  Update;
end;

function TRTFRepDialogEx.RunRTF(DS: array of TDataSet): boolean;
begin
  Result:=false;
  try
    RTFRep2.Execute(DS);
    Result:=true;
  except
    on E: Exception do
      ShowErr('', E, false);
  end;
end;

procedure TRTFRepDialogEx.SendMail(Attachment: string);
var
  path: string;
begin
  UpdateInfo(sMailGenInfo);
  if RepDlgInfo.UseMapi then
  begin
    if RepDlgInfo.SendMailDir = '' then
      path:=GetAppDir
    else
      path:=RepDlgInfo.SendMailDir;
    try
      MapiEmail.InitMail;
      MapiEmail.SendTo.Add(RepDlgInfo.MailTo);
      MapiEmail.Subject:=sReport + ': ' + ExtractFileName(fRep.Caption);
      MapiEmail.Attachments.Add(Attachment);
      MapiEmail.SendAutoSelect;
    except
      on E: Exception do
        DlgE(E.Message);
    end;
  end
  else
    RunFile(RepDlgInfo.MailApp, '"' + Attachment + '"', '', SW_SHOW);
  UpdateInfo('');
end;

function TRTFRepDialogEx.WaitForPDFFile: boolean;
begin
  Result:=false;
  UpdateInfo(Format(rstring.sFileGeneration, ['PDF']));
  PDFWaitCounter:=PDFWaitSeconds2 * 1000 div TimerPDF.Interval;
  TimerPDF.Enabled:=true;
  try
    repeat
      Application.ProcessMessages;
      if Aborted then
        Break;
      if PDFWaitCounter = -1 then
        Result:=true;
    until Result;
  finally
    TimerPDF.Enabled:=false;
    UpdateInfo('');
  end;
end;

procedure TRTFRepDialogEx.TimerPDFTimer(Sender: TObject);
begin
  Dec(PDFWaitCounter);
  if PDFWaitCounter = 0 then
  begin
    if DlgQ(sPDFWaitAgainQ) = mrYes then
      PDFWaitCounter:=PDFWaitSeconds2 * 1000 div TimerPDF.Interval
    else
      Aborted:=true;
  end;
  if FileExists(PDFFile) then
    PDFWaitCounter:=-1;
end;

function TRTFRepDialogEx.SetPDFProps: boolean;
begin
  Result:=false;
  if not Assigned(RepDlgInfo.OnConfigPDFPrinter) then
    Exit;
  UpdateInfo(sPDFConfiguration);
  try
    Result:=RepDlgInfo.OnConfigPDFPrinter(PDFFile, PDFPrinterName2);
    if not Result then
      DlgE(sPDFConfigurationErr);
  except
    on E: Exception do
      ShowErr(sPDFConfigurationErr, E, false);
  end;
  UpdateInfo('');
end;

procedure TRTFRepDialogEx.btShowClick(Sender: TObject);
var
  d: integer;
begin
  d:=integer(tv.Selected.Data);
  if d = -1 then
    Exit;

  Aborted:=false;
  pcRun.ActivePage:=tsCancel;
  UpdateInfo(sInitialize);
  try
    RTFRep2.ReportFile:=fRep.Caption;
    RTFRep2.OutputFile:=fOut.Caption;
    RTFRep2.DefaultCharset:=eCharset.AsInteger;
    RTFRep2.FieldNotFoundValue:=eNotFound.Text;
    RTFRep2.Variables.Assign(eVar.Lines);
    RTFRep2.OnReadInfo:=OnReadInfo;
    RTFRep2.PrintCopies:=eCopies.AsInteger;
    RTFRep2.RaiseUnknownField:=SameText(RTFRep2.FieldNotFoundValue, 'raiseerror');

    RTFRep2.ActionAfter:=aaNone;
    if chPDF.Checked then
    begin
      // generate PDF via virtual printer
      PDFFile:=ChangeFileExt(fOut.Caption, '.pdf');
      if RepDlgInfo.ShowPDFSaveAsDlg then
      begin
        PdfSaveDlg.FileName:=PDFFile;
        if not PdfSaveDlg.Execute then
          Exit;
        PDFFile:=PdfSaveDlg.FileName;
      end;
      if not SetPDFProps then
      begin
        if btPDFSetup.Visible then
          btPDFSetupClick(self);
        Exit;
      end;

      // delete old pdf file
      if FileExists(PDFFile) then
        DeleteFile(PDFFile);
      // check if deleted
      if FileExists(PDFFile) then
      begin
        DlgE(Format(sDeleteOldFileErr, ['PDF']));
        Exit;
      end;
      RTFRep2.ActionAfter:=aaPrint;
    end
    else
    begin
      if Sender = btShow then
        RTFRep2.ActionAfter:=aaShow;
      if Sender = btPrint then
        RTFRep2.ActionAfter:=aaPrint;
    end;
    if RunRTF(RepInfo[d].DataSets) then
    begin
      if chPDF.Checked then
      begin
        // use generated PDF file
        if WaitForPDFFile then
        begin
          if (Sender = btShow) then
            RunFile(PDFFile, '', '', SW_SHOW);
          if (Sender = btPrint) then
            PrintFile(PDFFile, '', '', SW_SHOW);
          if (Sender = btMail) then
            SendMail(PDFFile);
        end;
      end
      else
        if (Sender = btMail) then
          SendMail(fOut.Caption);
    end;
    if Assigned(RepDlgInfo.OnAfterRun) then
      RepDlgInfo.OnAfterRun(nil);
  finally
    UpdateInfo('');
    pcRun.ActivePage:=tsButton;
  end;
end;

procedure TRTFRepDialogEx.OnReadInfo(Sender: TObject);
var
  L: TStrings;
  s: string;
begin
  L:=RTFRep2.ExtendedInfo;
  if Assigned(L) then
  begin
    // check Printer
    s:=L.Values['PrinterName'];
    if (s <> '') then
      RTFRep2.PrinterName:=s;
  end;
  if Assigned(RepDlgInfo.OnReadInfo) then
    RepDlgInfo.OnReadInfo(RTFRep2);
end;

procedure TRTFRepDialogEx.RTFRep1ReportProcess(Sender: TObject; RecNbr: integer; var Abort: boolean);
begin
  if chPDF.Checked and (RTFRep2.PrinterName = '') then
    RTFRep2.PrinterName:=PDFPrinterName2;
  UpdateInfo(sRecord + ' ' + IntToStr(RecNbr));
  Abort:=Aborted;
end;

procedure TRTFRepDialogEx.btEditClick(Sender: TObject);
begin
  if fRep.Caption = '' then
    Exit;
  if Assigned(RepDlgInfo.OnBeforeEdit) then
    RepDlgInfo.OnBeforeEdit(nil);
  RunFile(fRep.Caption, '', '', SW_SHOW);
end;

procedure TRTFRepDialogEx.btDefFileClick(Sender: TObject);
begin
  if not btPrint.Enabled then
    Exit;
  fOut.Caption:=rtool.SetEndOfPath(AppDirCfg.GetSpecialDir(CSIDL_PERSONAL)) + 'report.rtf';
end;

procedure TRTFRepDialogEx.btAbortClick(Sender: TObject);
begin
  Aborted:=true;
end;

procedure TRTFRepDialogEx.btOpenClick(Sender: TObject);
begin
  OpenDlg.FileName:=fRep.Caption;
  if OpenDlg.Execute then
    fRep.Caption:=OpenDlg.FileName;
end;

procedure TRTFRepDialogEx.btSaveClick(Sender: TObject);
begin
  SaveDlg.FileName:=fOut.Caption;
  if SaveDlg.Execute then
    fOut.Caption:=SaveDlg.FileName;
end;

procedure TRTFRepDialogEx.tvDblClick(Sender: TObject);
begin
  btShow.Click;
end;

procedure TRTFRepDialogEx.btPDFSetupClick(Sender: TObject);
begin
  if Assigned(RepDlgInfo.OnPDFSetup) then
    RepDlgInfo.OnPDFSetup(self);
end;

end.
