unit rDBExport;

{<
  rDBExport provides universal dialog for exporting data from DBGrid or Dataset;
  user can select export type, columns and scope of data;
  dialog provides export to TXT, SLK, Excel and report in RTF/DOCX/ODT format.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, Buttons, ExtCtrls, DBGrids, db, ImgList,
  PropSaveMain, PropSaveGrids, rImprovedComps, rHTMLLabel,
  SLKExport, rXLSExport, rXMLReport, RTFReport2, TXTExport,
  rDBRecView, rDBTreeView, rStringGridEd,
  RosiCompGlobalCfg, rdbdlg_baseform;

{ Show  DBExport dialog for DBGrid, parameters define posibility to:
  set filter + set exported columns + export to RTF/DOCX/ODT report }
procedure DBExport(DBGrid: TCustomDBGrid; SetFilter: boolean = true; SetColumn: boolean = true; BlockReport: boolean = false;
  ExportStructureOnly: boolean = false); overload;
{ Show  DBExport dialog for Dataset, parameters define posibility to:
  set filter + set exported columns + export to RTF/DOCX/ODT report }
procedure DBExport(DataSet: TDataSet; SetFilter: boolean = true; SetColumn: boolean = true; BlockReport: boolean = false;
  ExportStructureOnly: boolean = false); overload;
{ Show  DBExport dialog for DBRecView, parameters define posibility to:
  set exported columns + export to RTF/DOCX/ODT report }
procedure DBExport(rDBRecView: TrDBRecView; BlockReport: boolean = false; ExportStructureOnly: boolean = false); overload;
{ Show  DBExport dialog for TreeView }
procedure DBExport(TreeView: TCustomTreeView); overload;
{ Show  DBExport dialog for StringGrid }
procedure DBExport(rStringGridEd: TrStringGridEd; ExportStructureOnly: boolean = false); overload;

type
  { universal dialog for exporting data from DBGrid or Dataset }
  TrDBExportForm = class(TrDBDlgBaseForm)
    cbAction_nostore: TComboBox;
    rg: TRadioGroup;
    btCancel: TrButton;
    btOk: TrButton;
    pg: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    gbMoznosti: TGroupBox;
    chFName1: TCheckBox;
    GroupBox1: TGroupBox;
    chFName2: TCheckBox;
    SLKExp: TSLKExport;
    chSum2: TCheckBox;
    chCol2: TCheckBox;
    lCharSet: TLabel;
    eCharSet: TrEditNum;
    btRepFile: TrButton;
    chCol1: TCheckBox;
    Image1: TImage;
    Image3: TImage;
    Label2: TLabel;
    TXTExp: TTXTExport;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    lRepFile: TrHTMLLabel;
    RTFRep: TRTFReport2;
    XLSExp: TrXLSExport;
    Image2a: TImage;
    Label1: TLabel;
    cbSeparator_nostore: TComboBox;
    chAdd1: TCheckBox;
    chAddSheet2: TCheckBox;
    chAdd2: TCheckBox;
    chFreezePanes2: TCheckBox;
    Image2b: TImage;
    XMLRep: TrXMLReport;
    gbData: TGroupBox;
    rbSelCol: TRadioButton;
    rbAllCol: TRadioButton;
    btFields: TrButton;
    btFilter: TrButton;
    chSelRecOnly: TCheckBox;
    chBorders2: TCheckBox;
    chFRealName2: TCheckBox;
    chFRealName1: TCheckBox;
    lblSetDefaultCol: TrHTMLLabel;
    DBGridsPS: TDBGridsPropSave;
    procedure rgClick(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btRepFileClick(Sender: TObject);
    procedure btFilterClick(Sender: TObject);
    procedure btFieldsClick(Sender: TObject);
    procedure chFName2Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure chFName1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure DBGridsPSGetKeyString(Comp: TComponent; var Key: string);
    procedure lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
  private
    SourceType: TSourceType;
    DataSet: TDataSet;
    OrigGr: TCustomDBGrid;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    SetFlt, SetCol, StructOnly: boolean;
    procedure AssignGridProps;
  public
  end;

var
  rDBExportForm: TrDBExportForm;

implementation

{$R *.dfm}

uses rtool, rdbtool, rdlg, rstring,
  rDBGrid, rDBGrid_MS, rDBFilter, rDBFields;

const
  FileFilter: array [0 .. 2] of string = (sFileTypeTXT, sFileTypeXLS, sFileTypeSLK);
  FileExt: array [0 .. 2] of string = ('txt', 'xls', 'slk');
  TXTIdx = 0;
  XLSIdx = 1;
  SLKIdx = 2;
  RepIdx = 3;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

procedure DBExportInt(SourceType: TSourceType; O: TObject; SetFilter, SetColumn, BlockReport, SelectedDataOnly, ExportStructureOnly: boolean); forward;

procedure DBExport(DBGrid: TCustomDBGrid; SetFilter: boolean = true; SetColumn: boolean = true; BlockReport: boolean = false;
  ExportStructureOnly: boolean = false); overload;
var
  SelData: boolean;
begin
  if DBGrid is TrDBGrid_MS then
    SelData:=TrDBGrid_MS(DBGrid).GetSelectedRecCount > 0
  else
    SelData:=THackCustomDBGrid(DBGrid).SelectedRows.Count > 0;
  DBExportInt(stGrid, DBGrid, SetFilter, SetColumn, BlockReport, SelData, ExportStructureOnly);
end;

procedure DBExport(DataSet: TDataSet; SetFilter: boolean = true; SetColumn: boolean = true; BlockReport: boolean = false;
  ExportStructureOnly: boolean = false); overload;
begin
  DBExportInt(stDataset, DataSet, SetFilter, SetColumn, BlockReport, false, ExportStructureOnly);
end;

procedure DBExport(rDBRecView: TrDBRecView; BlockReport: boolean = false; ExportStructureOnly: boolean = false);
begin
  DBExportInt(stRecView, rDBRecView, false, false, BlockReport, false, ExportStructureOnly);
end;

procedure DBExport(TreeView: TCustomTreeView); overload;
begin
  DBExportInt(stTreeView, TreeView, false, false, true, false, false);
end;

procedure DBExport(rStringGridEd: TrStringGridEd; ExportStructureOnly: boolean = false); overload;
begin
  DBExportInt(stStringGrid, rStringGridEd, false, false, true, false, ExportStructureOnly);
end;

procedure DBExportInt(SourceType: TSourceType; O: TObject; SetFilter, SetColumn, BlockReport, SelectedDataOnly, ExportStructureOnly: boolean);
var
  d: TDataSource;
  DataSet: TDataSet;
begin
  if not Assigned(O) then
    Exit;
  DataSet:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      d:=TCustomDBGrid(O).DataSource;
    stRecView:
      d:=(O as TrDBRecView).DataSource;
    stDataset:
      DataSet:=(O as TDataSet);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.DataSet) or not d.DataSet.Active then
      Exit;
    DataSet:=d.DataSet;
  end;

  rDBExportForm:=TrDBExportForm.Create(GetActiveOwner);
  try
    rDBExportForm.DataSet:=DataSet;
    rDBExportForm.SourceType:=SourceType;
    rDBExportForm.StructOnly:=ExportStructureOnly;
    rDBExportForm.SetFlt:=SetFilter;
    rDBExportForm.SetCol:=SetColumn;
    rDBExportForm.chSelRecOnly.Visible:=SelectedDataOnly;
    if BlockReport then
      rDBExportForm.rg.Items.Delete(3);

    case rDBExportForm.SourceType of
      stGrid:
        begin
          rDBExportForm.OrigGr:=TCustomDBGrid(O);
          if O is TrDBGrid then
            rDBExportForm.DBGrid:=TrDBGrid.Create(rDBExportForm)
          else
            rDBExportForm.DBGrid:=TCustomDBGrid.Create(rDBExportForm);
          rDBExportForm.DBGrid.Name:='gr';
          rDBExportForm.DBGrid.Visible:=false;
          rDBExportForm.DBGrid.Parent:=rDBExportForm.OrigGr.Parent;
          rDBExportForm.AssignGridProps;
          rDBExportForm.DBGridsPS.Components.Text:='gr';
          rDBExportForm.DBGridsPS.Disable:=RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields;
        end;
      stRecView:
        rDBExportForm.RecView:=TrDBRecView(O);
      stTreeView:
        rDBExportForm.TreeView:=TCustomTreeView(O);
      stStringGrid:
        rDBExportForm.StringGrid:=TrStringGridEd(O);
    end;
    rDBExportForm.ShowModal;
  finally
    FreeF(rDBExportForm);
  end;
end;

procedure TrDBExportForm.FormCreate(Sender: TObject);
begin
  inherited;
  // DBGrid:=TCustomDBGrid.Create(Self);
end;

procedure TrDBExportForm.FormDestroy(Sender: TObject);
begin
  // DBGrid.Free;
  inherited;
end;

procedure TrDBExportForm.FormShow(Sender: TObject);
begin
  inherited;
  lblSetDefaultCol.Visible:=SourceType = stGrid;
  chSum2.Visible:=SourceType <> stTreeView;
  rgClick(nil);
  chFName1Click(nil);
  chFName2Click(nil);
end;

procedure TrDBExportForm.AssignGridProps;
begin
  rdbtool.AssignGridProps(OrigGr, DBGrid);
  if OrigGr is TrDBGrid then
    TrDBGrid(DBGrid).Assign(TrDBGrid(OrigGr));
end;

procedure TrDBExportForm.rgClick(Sender: TObject);
begin
  case rg.ItemIndex of
    TXTIdx:
      pg.ActivePageIndex:=0;
    XLSIdx, SLKIdx:
      pg.ActivePageIndex:=1;
    RepIdx:
      pg.ActivePageIndex:=2;
  end;
  chBorders2.Visible:=rg.ItemIndex = XLSIdx;
  chAddSheet2.Visible:=rg.ItemIndex = XLSIdx;
  chAdd2.Visible:=rg.ItemIndex = XLSIdx;
  chFreezePanes2.Visible:=rg.ItemIndex = XLSIdx;
  Image2a.Visible:=rg.ItemIndex = XLSIdx;
  Image2b.Visible:=not Image2a.Visible;
  pg.Visible:=true;
  btFilter.Enabled:=SetFlt;
  btFields.Enabled:=SetCol and (rg.ItemIndex <> RepIdx);
  rbSelCol.Enabled:=btFields.Enabled;
  rbAllCol.Enabled:=btFields.Enabled;
  Self.chSelRecOnly.Enabled:=btFields.Enabled;
  btOk.Enabled:=(rg.ItemIndex <> RepIdx) or (RTFRep.ReportFile <> '');
  SaveDlg.FileName:='';
end;

procedure TrDBExportForm.btOkClick(Sender: TObject);
var
  s: string;
  noFileName: boolean;
begin
  noFileName:=(rg.ItemIndex = XLSIdx) and RosiCompConfig.RTFRepDlgCfg.ExportExcelNoFileName and (cbAction_nostore.ItemIndex > 0) and (not chAddSheet2.Checked)
    and (not chAdd2.Checked);
  if not noFileName then
  begin
    if rg.ItemIndex = RepIdx then
    begin
      s:=copy(ExtractFileExt(RTFRep.ReportFile), 2, 100);
      SaveDlg.Filter:='';
      if SameText(s, 'rtf') then
        SaveDlg.Filter:=sFileTypeRTF;
      if SameText(s, 'docx') then
        SaveDlg.Filter:=sFileTypeDOCX;
      if SameText(s, 'odt') then
        SaveDlg.Filter:=sFileTypeODT;
      if SaveDlg.Filter <> '' then
        SaveDlg.Filter:=SaveDlg.Filter + '|';
      SaveDlg.Filter:=SaveDlg.Filter + sFileTypeAll;
      SaveDlg.DefaultExt:=s;
    end
    else
    begin
      SaveDlg.Filter:=FileFilter[rg.ItemIndex] + '|' + sFileTypeAll;
      SaveDlg.DefaultExt:=FileExt[rg.ItemIndex];
    end;
    if not SaveDlg.Execute then
      Exit;
  end;
  Update;
  ShowInfo(sExporting + sOfData, Self);
  try
    case rg.ItemIndex of
      TXTIdx:
        begin
          TXTExp.UseDisableControl:=RosiCompConfig.rDBExportCfg.UseDisableControl;
          TXTExp.ExportFile:=SaveDlg.FileName;
          TXTExp.ExportFieldName:=chFName1.Checked;
          TXTExp.RealFieldNamesAsTitle:=chFRealName1.Checked;
          TXTExp.ExportStructureOnly:=StructOnly;
          TXTExp.ExportOnlySelectedRows:=chSelRecOnly.Visible and chSelRecOnly.Checked;
          if chCol1.Checked then
          begin
            TXTExp.TextApostrophe:='';
            TXTExp.FillChar:=' ';
          end
          else
          begin
            TXTExp.TextApostrophe:='"';
            TXTExp.FillChar:=#0;
          end;
          s:=cbSeparator_nostore.Text;
          TXTExp.FieldSeparator:=s;
          if Length(s) > 0 then
          begin
            if s[1] = '#' then
            begin
              s:=copy(s, 2, 3);
              TXTExp.FieldSeparator:=chr(StrToIntDef(Trim(s), 0));
            end
            else
              if Length(s) > 8 then
                TXTExp.FieldSeparator:=s[1];
          end;
          TXTExp.ExportOnlyVisibleFields:=rbSelCol.Checked;
          TXTExp.AppendToFile:=chAdd1.Checked;
          TXTExp.ActionAfter:=TActionAfter(cbAction_nostore.ItemIndex);
          case SourceType of
            stGrid:
              TXTExp.Execute(DBGrid);
            stDataset:
              TXTExp.Execute(DataSet);
            stRecView:
              TXTExp.Execute(RecView);
            stTreeView:
              TXTExp.Execute(TreeView);
            stStringGrid:
              TXTExp.Execute(StringGrid);
          end;
        end;
      XLSIdx:
        begin
          XLSExp.UseDisableControl:=RosiCompConfig.rDBExportCfg.UseDisableControl;
          if noFileName then
            XLSExp.ExportFile:=''
          else
            XLSExp.ExportFile:=SaveDlg.FileName;
          if chFName2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleExportTitles]
          else
            XLSExp.Options:=XLSExp.Options - [xleExportTitles];
          if chFRealName2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleRealFieldNamesAsTitle]
          else
            XLSExp.Options:=XLSExp.Options - [xleRealFieldNamesAsTitle];
          if chSum2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleAddSumRow]
          else
            XLSExp.Options:=XLSExp.Options - [xleAddSumRow];
          if chCol2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleSetColumnWidth]
          else
            XLSExp.Options:=XLSExp.Options - [xleSetColumnWidth];
          if chBorders2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleSetCellBorders]
          else
            XLSExp.Options:=XLSExp.Options - [xleSetCellBorders];
          if chAddSheet2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleAddAsNewSheet]
          else
            XLSExp.Options:=XLSExp.Options - [xleAddAsNewSheet];
          if chAdd2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleAddToExistingRows]
          else
            XLSExp.Options:=XLSExp.Options - [xleAddToExistingRows];
          if chFreezePanes2.Checked then
            XLSExp.Options:=XLSExp.Options + [xleFreezeTitles]
          else
            XLSExp.Options:=XLSExp.Options - [xleFreezeTitles];
          if rbSelCol.Checked then
            XLSExp.Options:=XLSExp.Options + [xleOnlyVisibleFields]
          else
            XLSExp.Options:=XLSExp.Options - [xleOnlyVisibleFields];
          if chSelRecOnly.Visible and chSelRecOnly.Checked then
            XLSExp.Options:=XLSExp.Options + [xleOnlySelectedRows]
          else
            XLSExp.Options:=XLSExp.Options - [xleOnlySelectedRows];
          XLSExp.ExportStructureOnly:=StructOnly;

          XLSExp.ActionAfter:=TActionAfter(cbAction_nostore.ItemIndex);
          XLSExp.StartRow:=1;
          XLSExp.StartCol:=1;
          case SourceType of
            stGrid:
              XLSExp.ExportDBTable(DBGrid);
            stDataset:
              XLSExp.ExportDBTable(DataSet);
            stRecView:
              XLSExp.ExportDBTable(RecView);
            stTreeView:
              XLSExp.ExportDBTable(TreeView);
            stStringGrid:
              XLSExp.ExportDBTable(StringGrid);
          end;
          XLSExp.Disconnect;
        end;
      SLKIdx:
        begin
          SLKExp.UseDisableControl:=RosiCompConfig.rDBExportCfg.UseDisableControl;
          SLKExp.ExportFile:=SaveDlg.FileName;
          SLKExp.ExportFieldName:=chFName2.Checked;
          SLKExp.RealFieldNamesAsTitle:=chFRealName2.Checked;
          SLKExp.ExportStructureOnly:=StructOnly;
          SLKExp.SetColWidth:=chCol2.Checked;
          SLKExp.MakeSumRow:=chSum2.Checked;
          SLKExp.ExportOnlyVisibleFields:=rbSelCol.Checked;
          SLKExp.ExportOnlySelectedRows:=chSelRecOnly.Visible and chSelRecOnly.Checked;
          SLKExp.ActionAfter:=TActionAfter(cbAction_nostore.ItemIndex);
          case SourceType of
            stGrid:
              SLKExp.Execute(DBGrid);
            stDataset:
              SLKExp.Execute(DataSet);
            stRecView:
              SLKExp.Execute(RecView);
            stTreeView:
              SLKExp.Execute(TreeView);
            stStringGrid:
              SLKExp.Execute(StringGrid);
          end;
        end;
      RepIdx:
        begin
          s:=copy(ExtractFileExt(RTFRep.ReportFile), 2, 100);
          if SameText(s, '.docx') or SameText(s, '.odt') then
          begin
            XMLRep.OutputFile:=SaveDlg.FileName;
            XMLRep.ActionAfter:=TActionAfter(cbAction_nostore.ItemIndex);
            XMLRep.Execute([DataSet]);
          end
          else
          begin
            RTFRep.OutputFile:=SaveDlg.FileName;
            RTFRep.DefaultCharset:=eCharSet.AsInteger;
            RTFRep.ActionAfter:=TActionAfter(cbAction_nostore.ItemIndex);
            RTFRep.Execute([DataSet]);
          end;
        end;
    end;
  except
    on E: Exception do
      ShowErr(sExportDataErr, E, false);
  end;
  ShowInfo('', nil);
end;

procedure TrDBExportForm.btRepFileClick(Sender: TObject);
begin
  OpenDlg.Filter:=sFileTypeRTF + '|' + sFileTypeDOCX + '|' + sFileTypeODT + '|' + sFileTypeAll;
  if not OpenDlg.Execute then
    Exit;
  lRepFile.Caption:=OpenDlg.FileName;
  lRepFile.Visible:=true;
  RTFRep.ReportFile:=OpenDlg.FileName;
  XMLRep.ReportFile:=OpenDlg.FileName;
  eCharSet.Enabled:=SameText(ExtractFileExt(OpenDlg.FileName), '.rtf');
  btOk.Enabled:=true;
end;

procedure TrDBExportForm.chFName1Click(Sender: TObject);
begin
  chFRealName1.Enabled:=chFName1.Checked;
end;

procedure TrDBExportForm.chFName2Click(Sender: TObject);
begin
  chFRealName2.Enabled:=chFName2.Checked;
end;

procedure TrDBExportForm.btFilterClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      begin
        FilterRec(DBGrid);
        if OrigGr is TrDBGrid then
          TrDBGrid(OrigGr).FilterChangedNotification;
      end;
    stDataset:
      FilterRec(DataSet);
  end;
end;

procedure TrDBExportForm.btFieldsClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      ChooseFields(DBGrid);
    stDataset:
      ChooseFields(DataSet);
  end;
end;

procedure TrDBExportForm.DBGridsPSGetKeyString(Comp: TComponent; var Key: string);
begin
  // Sets key in INI file according Form and DBGrid name
  Key:=Key + '.' + OrigGr.Owner.Name + '.' + OrigGr.Name;
end;

procedure TrDBExportForm.lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
begin
  AssignGridProps;
end;

end.
