unit rDBFields;

{<
  rDBFields provides universal dialog for user definition of visibility
  and properties of DBGrid columns or Dataset fields.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, CheckLst, Buttons, db, dbGrids, ExtCtrls,
  PropSaveMain, ImgList, rImprovedComps, rHTMLLabel,
{$IFDEF DXE2P} UITypes, {$ENDIF} rdbdlg_baseform;

{ Show DBFields dialog for DBGrid (Columns collection is used),
  parametr define possibility to set Fixed Column properties }
procedure ChooseFields(const DBGrid: TCustomDBGrid; SetFixedCol: integer = 0); overload;
{ Show DBFields dialog for Dataset (Fields collection is used),
  parametr define possibility to set Fixed Column properties }
procedure ChooseFields(const DataSet: TDataSet; SetFixedCol: integer = 0); overload;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  { universal universal dialog for user definition of visibility
    and properties of DBGrid columns or Dataset fields }
  TrDBFieldsForm = class(TrDBDlgBaseForm)
    lSeznam: TLabel;
    lbLabels: TrCheckListBox;
    btUp: TrButton;
    btDown: TrButton;
    btClose: TrButton;
    btAll: TrButton;
    gb: TGroupBox;
    rgAlign: TRadioGroup;
    Label3: TLabel;
    eCaption: TEdit;
    Label4: TLabel;
    lFormat: TLabel;
    enWidth: TrEditNum;
    lColor: TLabel;
    cbColor: TrColorBox;
    btFont: TrButton;
    Panel1: TPanel;
    lTest: TLabel;
    btApply: TrButton;
    FontDlg: TFontDialog;
    btDefault: TrURLLabel;
    cbFormat: TComboBox;
    mNumFormat: TrStringsHolder;
    mFldList: TrStringsHolder;
    lContent: TLabel;
    lContentInfo1: TrHTMLLabel;
    chGroupFields: TCheckBox;
    btGroupSettings: TrURLLabel;
    lContentInfo2: TrHTMLLabel;
    procedure lbLabelsDragOver(Sender, Source: TObject; X, Y: integer; State: TDragState; var Accept: Boolean);
    procedure lbLabelsClickCheck(Sender: TObject);
    procedure btUpClick(Sender: TObject);
    procedure btDownClick(Sender: TObject);
    procedure btAllClick(Sender: TObject);
    procedure lbLabelsKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure lbLabelsDblClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure lbLabelsClick(Sender: TObject);
    procedure btFontClick(Sender: TObject);
    procedure cbColorChange(Sender: TObject);
    procedure btApplyClick(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
    procedure btDefaultClick(Sender: TObject);
    procedure lbLabelsDrawItem(Control: TWinControl; Index: integer; Rect: TRect; State: TOwnerDrawState);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure chGroupFieldsClick(Sender: TObject);
    procedure eCaptionExit(Sender: TObject);
    procedure eCaptionChange(Sender: TObject);
    procedure eCaptionEnter(Sender: TObject);
    procedure btGroupSettingsClick(Sender: TObject);
  private
    GDS: TObject;
    DS: TDataSet;
    Gr: TCustomDBGrid;
    GrHack: THackCustomDBGrid;
    isDBGrid, isRDBGrid, UseFixedCol, SelFixedCol: Boolean;
    SetAll: Boolean;
    F: TField;
    C: TColumn;
    act, FixedNum: integer;
    procedure LoadAll;
    procedure MoveI(List: TCheckListBox; old, new: integer);
    procedure CheckI(act: integer);
    procedure RefState;
    procedure ReadProp;
    procedure SetProp;
    procedure SetTest;
    function CheckSave: Boolean;
  public
  end;

var
  rDBFieldsForm: TrDBFieldsForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rtool, rdlg, rdbtool, rstring, rDBGrid, rDBFieldsGroup;

var
  isLoading: Boolean;

procedure ChooseFieldsInt(const GridOrDataSet: TObject; SetFixedCol: integer); forward;

procedure ChooseFields(const DBGrid: TCustomDBGrid; SetFixedCol: integer = 0);
begin
  ChooseFieldsInt(DBGrid, SetFixedCol);
end;

procedure ChooseFields(const DataSet: TDataSet; SetFixedCol: integer = 0);
begin
  ChooseFieldsInt(DataSet, SetFixedCol);
end;

procedure ChooseFieldsInt(const GridOrDataSet: TObject; SetFixedCol: integer);
begin
  if not IsLinkedAndActive(GridOrDataSet) then
    Exit;
  isLoading:=true;
  rDBFieldsForm:=TrDBFieldsForm.Create(GetActiveOwner);
  try
    rDBFieldsForm.GDS:=GridOrDataSet;
    rDBFieldsForm.UseFixedCol:=RosiCompConfig.GetConfig_OnOffAuto(SetFixedCol, RosiCompConfig.rDBFieldsCfg.SetFixedCol);
    rDBFieldsForm.ShowModal;
  finally
    FreeF(rDBFieldsForm);
  end;
end;

procedure TrDBFieldsForm.FormShow(Sender: TObject);
begin
  inherited;
  btApply.Font.Style:=[fsBold];
  isDBGrid:=(GDS is TCustomDBGrid);
  isRDBGrid:=(GDS is TrDBGrid);
  DS:=GetDataSet(GDS);
  if isDBGrid then
  begin
    Gr:=TCustomDBGrid(GDS);
    GrHack:=THackCustomDBGrid(Gr);
  end;
  btDefault.Enabled:=isDBGrid;
  chGroupFields.Visible:=isRDBGrid and RosiCompConfig.GetConfig_OnOffAuto(RosiCompConfig.rDBGridPropsCfg.HideTitleGroups, TrDBGrid(Gr).FieldsGroups.Count > 0);
  chGroupFields.Checked:=chGroupFields.Visible and TrDBGrid(Gr).OptionsEx2.Groups.Active;
  btGroupSettings.Visible:=chGroupFields.Visible;
  LoadAll;
end;

procedure TrDBFieldsForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if isRDBGrid then
    TrDBGrid(Gr).FixedGroupedCol(false);
  inherited;
end;

procedure TrDBFieldsForm.LoadAll;
var
  i: integer;
begin
  // Load all labels
  GetFldList(GDS, lbLabels.Items, [fdGetLabels, fdAllField]);
  // Load visibility
  if isDBGrid then
  begin
    for i:=0 to GrHack.Columns.Count - 1 do
      lbLabels.Checked[i]:=GrHack.Columns[i].Visible
  end
  else
  begin
    for i:=0 to DS.FieldCount - 1 do
      lbLabels.Checked[i]:=DS.Fields[i].Visible;
  end;
  SetAll:=true;
  FixedNum:=0;
  if isDBGrid and UseFixedCol then
  begin
    FixedNum:=1;
    lbLabels.Items.Insert(0, sIndicatorColumnName);
    lbLabels.Checked[0]:=(dgIndicator in GrHack.Options);
    GetFldList(GDS, mFldList.Strings, [fdAllField]);
    for i:=0 to mFldList.Strings.Count - 1 do
      mFldList.Strings[i]:='<<' + mFldList.Strings[i] + '>>';
    mFldList.Strings.Insert(0, '<<#>>');
  end;

  // Select labels
  lbLabels.ItemIndex:=0;
  if isDBGrid then
    try
      lbLabels.ItemIndex:=Gr.SelectedIndex + FixedNum;
    except
    end;
  // Set prop enability
  btFont.Enabled:=isDBGrid;
  cbColor.Enabled:=isDBGrid;
  lColor.Enabled:=isDBGrid;
  isLoading:=false;
  RefState;
  ReadProp;
end;

procedure TrDBFieldsForm.MoveI(List: TCheckListBox; old, new: integer);
begin
  if (old < FixedNum) or (new < FixedNum) then
    Exit;
  if not MoveCheckListBox(List, old, new) then
    Exit;
  if isDBGrid then
    GrHack.Columns[old - FixedNum].Index:=new - FixedNum
  else
    DS.Fields[old - FixedNum].Index:=new - FixedNum;
  ReadProp;
  lbLabels.Invalidate;
end;

procedure TrDBFieldsForm.CheckI(act: integer);
begin
  if act < FixedNum then
  begin
    if lbLabels.Checked[act] then
      GrHack.Options:=GrHack.Options + [dgIndicator]
    else
      GrHack.Options:=GrHack.Options - [dgIndicator];
  end
  else
  begin
    if isDBGrid then
      GrHack.Columns[act - FixedNum].Visible:=lbLabels.Checked[act]
    else
      DS.Fields[act - FixedNum].Visible:=lbLabels.Checked[act];
    if isRDBGrid then
    begin
      if TrDBGrid(Gr).GetFieldGroupInfo(GrHack.Columns[act - FixedNum].Field).IsMaster then
      begin
        if not GrHack.Columns[act - FixedNum].Visible then
          TrDBGrid(Gr).CollapseGroup(GrHack.Columns[act - FixedNum].Field);
        RefState;
      end;
      if GrHack.Columns[act - FixedNum].Visible then
        TrDBGrid(Gr).SetOneColumnWidth(cwNone, GrHack.Columns[act - FixedNum]);
      if TrDBGrid(Gr).ColumnWidth in [cwSameWidth, cwAutoWidth] then
        TrDBGrid(Gr).SetColumnWidth(TrDBGrid(Gr).ColumnWidth);
    end;
  end;
end;

procedure TrDBFieldsForm.RefState;
var
  i: integer;
begin
  if isRDBGrid then
  begin
    for i:=0 to GrHack.Columns.Count - 1 do
    begin
      with (Gr as TrDBGrid).GetFieldGroupInfo(GrHack.Columns[i].Field) do
        if IsGrouped and not IsMaster then
          lbLabels.ItemEnabled[i + FixedNum]:=(Gr as TrDBGrid).IsFieldGroupMasterVisible(GrHack.Columns[i].Field)
        else
          lbLabels.ItemEnabled[i + FixedNum]:=true;
      if not lbLabels.ItemEnabled[i + FixedNum] then
        lbLabels.Checked[i + FixedNum]:=false;
    end;
    lbLabels.Invalidate;
  end;
end;

procedure TrDBFieldsForm.ReadProp;
begin
  act:=lbLabels.ItemIndex;
  gb.Caption:=' ' + lbLabels.Items[act] + ': ';
  SelFixedCol:=act < FixedNum;
  gb.Visible:=not SelFixedCol or isRDBGrid;
  cbColor.Visible:=not SelFixedCol;
  lColor.Visible:=cbColor.Visible;
  lFormat.Visible:=not SelFixedCol;
  lContent.Visible:=SelFixedCol;
  lContentInfo1.Visible:=SelFixedCol;
  lContentInfo2.Visible:=SelFixedCol;
  cbFormat.Enabled:=true;
  enWidth.Enabled:=true;

  if SelFixedCol then
  begin
    // indicator column
    C:=nil;
    F:=nil;
    cbFormat.Items.Assign(mFldList.Strings);
    if isRDBGrid then
      with (Gr as TrDBGrid) do
      begin
        FontDlg.Font.Assign(FixedColText.Font);
        rgAlign.ItemIndex:=ord(FixedColText.TextAlign);
        eCaption.Text:=FixedColText.TitleText;
        enWidth.AsInteger:=FixedColText.FixedColWidth;
        cbFormat.Text:=FixedColText.TextMask;
      end;
  end
  else
  begin
    // dataset field or grid column
    cbFormat.Items.Assign(mNumFormat.Strings);
    // Find selected Field
    if isDBGrid then
      F:=GrHack.Columns[act - FixedNum].Field
    else
      F:=DS.Fields[act - FixedNum];

    if isDBGrid then
    begin
      // Find selected Column
      C:=GrHack.Columns[act - FixedNum];
      // Load column prop
      cbColor.Selected:=C.Color;
      FontDlg.Font.Assign(C.Font);
      rgAlign.ItemIndex:=ord(C.Alignment);
      eCaption.Text:=C.Title.Caption;
    end
    else
    begin
      // Load field prop
      rgAlign.ItemIndex:=ord(F.Alignment);
      eCaption.Text:=F.DisplayLabel;
    end;

    cbFormat.Text:='';
    enWidth.Value:=0;
    if F <> nil then
    begin
      enWidth.Value:=F.DisplayWidth;
      cbFormat.Enabled:=(F is TNumericField) or (F is TDateTimeField);
      if (F is TNumericField) then
        cbFormat.Text:=(F as TNumericField).DisplayFormat;
      if (F is TDateTimeField) then
        cbFormat.Text:=(F as TDateTimeField).DisplayFormat;
    end
    else
    begin
      cbFormat.Enabled:=false;
      enWidth.Enabled:=false;
    end;
  end;

  if cbFormat.Enabled then
    cbFormat.Color:=clWindow
  else
    cbFormat.Color:=clBtnFace;
  if enWidth.Enabled then
    enWidth.Color:=clWindow
  else
    enWidth.Color:=clBtnFace;

  SetTest;
  btApply.Enabled:=false;
end;

procedure TrDBFieldsForm.SetProp;
begin
  if SelFixedCol then
  begin
    if isRDBGrid then
      with (Gr as TrDBGrid) do
      begin
        FixedColText.Font.Assign(FontDlg.Font);
        FixedColText.TextAlign:=TAlignment(rgAlign.ItemIndex);
        FixedColText.TitleText:=eCaption.Text;
        FixedColText.FixedColWidth:=enWidth.AsInteger;
        FixedColText.TextMask:=cbFormat.Text;
      end;
  end
  else
  begin
    if isDBGrid then
    begin
      // Set selected column
      C.Color:=cbColor.Selected;
      C.Font.Assign(FontDlg.Font);
      // C.Title.Font.Assign(C.Font);

      C.Alignment:=TAlignment(rgAlign.ItemIndex);
      C.Title.Alignment:=C.Alignment;
      C.Title.Caption:=eCaption.Text;
    end;
    if F <> nil then
    begin
      // Set selected field
      F.Alignment:=TAlignment(rgAlign.ItemIndex);
      F.DisplayLabel:=eCaption.Text;
      F.DisplayWidth:=enWidth.AsInteger;
      if (F is TNumericField) then (F as TNumericField)
        .DisplayFormat:=cbFormat.Text;
      if (F is TDateTimeField) then (F as TDateTimeField)
        .DisplayFormat:=cbFormat.Text;
    end;
  end;
  btApply.Enabled:=false;
end;

procedure TrDBFieldsForm.SetTest;
begin
  if cbColor.Visible then
    lTest.Color:=cbColor.Selected
  else
    lTest.Color:=clBtnFace;
  lTest.Font.Assign(FontDlg.Font);
  lTest.Alignment:=TAlignment(rgAlign.ItemIndex);
  btApply.Enabled:=true;
  // lTest.Caption:=GetFldData(F,'',true);
end;

function TrDBFieldsForm.CheckSave: Boolean;
var
  qRes: integer;
begin
  Result:=true;
  if btApply.Enabled then
  begin
    qRes:=ShowDlg(dtQ, btYesNoCancel, [sUseChangesQ]);
    lbLabels.EndDrag(false);
    case qRes of
      mrYes:
        SetProp;
      mrCancel:
        begin
          lbLabels.ItemIndex:=act;
          Result:=false;
        end;
      mrNo:
        ReadProp;
    end;
  end;
end;

procedure TrDBFieldsForm.btUpClick(Sender: TObject);
begin
  MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex - 1);
end;

procedure TrDBFieldsForm.btDownClick(Sender: TObject);
begin
  MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex + 1);
end;

procedure TrDBFieldsForm.lbLabelsDragOver(Sender, Source: TObject; X, Y: integer; State: TDragState; var Accept: Boolean);
var
  new: integer;
begin
  new:=Y div lbLabels.ItemHeight + lbLabels.TopIndex;
  Accept:=not((lbLabels.ItemIndex < FixedNum) or (new < FixedNum));
  MoveI(lbLabels, lbLabels.ItemIndex, new);
end;

procedure TrDBFieldsForm.lbLabelsDrawItem(Control: TWinControl; Index: integer; Rect: TRect; State: TOwnerDrawState);
begin
  with Control as TCheckListBox do
  begin
    Canvas.FillRect(Rect);
    if (Index < FixedNum) then
      Canvas.Font.Style:=Canvas.Font.Style + [fsItalic];
    if isRDBGrid and (Index >= FixedNum) and (Gr as TrDBGrid).GetFieldGroupInfo(GrHack.Columns[Index - FixedNum].Field).IsMaster then
      Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
    Canvas.TextOut(Rect.Left + 1, Rect.Top, Items[Index]);
  end;
end;

procedure TrDBFieldsForm.lbLabelsClickCheck(Sender: TObject);
begin
  CheckI(lbLabels.ItemIndex);
end;

procedure TrDBFieldsForm.lbLabelsDblClick(Sender: TObject);
begin
  if lbLabels.ItemEnabled[lbLabels.ItemIndex] then
  begin
    lbLabels.Checked[lbLabels.ItemIndex]:=not lbLabels.Checked[lbLabels.ItemIndex];
    CheckI(lbLabels.ItemIndex);
  end;
end;

procedure TrDBFieldsForm.btAllClick(Sender: TObject);
var
  a: integer;
begin
  for a:=0 to lbLabels.Items.Count - 1 do
  begin
    lbLabels.Checked[a]:=SetAll;
    CheckI(a);
  end;
  SetAll:=not SetAll;
end;

procedure TrDBFieldsForm.lbLabelsKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key = VK_UP) and (ssCtrl in Shift) then
  begin
    MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex - 1);
    Key:=0;
  end;

  if (Key = VK_DOWN) and (ssCtrl in Shift) then
  begin
    MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex + 1);
    Key:=0;
  end;
end;

procedure TrDBFieldsForm.lbLabelsClick(Sender: TObject);
begin
  if CheckSave then
    ReadProp;
end;

procedure TrDBFieldsForm.btFontClick(Sender: TObject);
begin
  if FontDlg.Execute then
    SetTest;
end;

procedure TrDBFieldsForm.cbColorChange(Sender: TObject);
begin
  SetTest;
end;

procedure TrDBFieldsForm.eCaptionChange(Sender: TObject);
begin
  eCaption.Tag:=1;
  SetTest;
end;

procedure TrDBFieldsForm.eCaptionEnter(Sender: TObject);
begin
  eCaption.Tag:=0;
end;

procedure TrDBFieldsForm.eCaptionExit(Sender: TObject);
var
  w: integer;
begin
  if SelFixedCol and (eCaption.Tag = 1) then
  begin
    w:=lTest.Canvas.TextWidth(eCaption.Text) + IndicatorWidth + 8;
    if w > enWidth.AsInteger then
      enWidth.AsInteger:=w;
  end;
  eCaption.Tag:=0;
end;

procedure TrDBFieldsForm.btApplyClick(Sender: TObject);
begin
  SetProp;
end;

procedure TrDBFieldsForm.btCloseClick(Sender: TObject);
begin
  if CheckSave then
    Close;
end;

procedure TrDBFieldsForm.chGroupFieldsClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  TrDBGrid(Gr).OptionsEx2.Groups.Active:=chGroupFields.Checked;
  RefState;
end;

procedure TrDBFieldsForm.btDefaultClick(Sender: TObject);
var
  i: integer;
  Comp: TComponent;
begin
  if isDBGrid and (DlgQ(sSetDefaultValues) = mrYes) then
  begin
    if Gr is TrDBGrid then
      TrDBGrid(Gr).RestoreColumnsWidth
    else
      GrHack.Columns.RestoreDefaults;
    // find parent form
    Comp:=Gr;
    while (Comp <> nil) and not(Comp is TForm) do
      Comp:=Comp.Owner;
    if Assigned(Comp) then
    begin
      // find all CustomPropSave and add grid to disable component list
      for i:=0 to Comp.ComponentCount - 1 do
        if Comp.Components[i] is TCustomPropSave then (Comp.Components[i] as TCustomPropSave)
          .DisableComponents.Add(GDS);
    end;
    // call global SetDefault event
    if Assigned(RosiCompConfig.rDBFieldsCfg.OnSetDefaultGridColumns) then
      RosiCompConfig.rDBFieldsCfg.OnSetDefaultGridColumns(GDS);
    Close;
  end;
end;

procedure TrDBFieldsForm.btGroupSettingsClick(Sender: TObject);
begin
  rDBFieldsGroup.SetFieldsGroup(TrDBGrid(Gr));
  chGroupFields.Checked:=TrDBGrid(Gr).OptionsEx2.Groups.Active;
  LoadAll;
end;

end.
