unit rDBFieldsGroup;

{<
  rDBFieldsGroup provides universal dialog for user definition of rDBGrid groups.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(May 2014)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, CheckLst, Buttons, db, dbGrids, rDBGrid, ExtCtrls, IniFiles,
  PropSaveMain, ImgList, rImprovedComps, rHTMLLabel, rdbdlg_baseform;

{ Show DBFieldsGroup dialog for rDBGrid }
procedure SetFieldsGroup(const rDBGrid: TrDBGrid);

type
  { universal dialog for user definition of rDBGrid groups }
  TrDBFieldsGroupForm = class(TrDBDlgBaseForm)
    btDefault: TrURLLabel;
    gbLeftGroups: TrGroupBox;
    Label29: TLabel;
    lbGroups: TrListBox;
    btNewGroup: TrButton;
    btDelGroup: TrButton;
    gbGroupBottom: TrGroupBox;
    Label31: TLabel;
    Label32: TLabel;
    Label35: TLabel;
    btNone1: TrURLLabel;
    btNone2: TrURLLabel;
    Label36: TLabel;
    rHTMLLabel3: TrHTMLLabel;
    eGroupTitle: TEdit;
    cgGroupTitleColor: TrColorBox;
    cgGroupLineColor: TrColorBox;
    eGroupMasterMask: TEdit;
    btGroupAddMask: TrButton;
    btGroupCollapse: TrButton;
    btGroupExpand: TrButton;
    gbRightGroups: TrGroupBox;
    Label33: TLabel;
    lbGroupFields: TrCheckListBox;
    rGroupBox1: TrGroupBox;
    Label34: TLabel;
    cbGroupMaster: TComboBox;
    btGroupUpdateAll: TrButton;
    lblLocked: TLabel;
    btOk: TrButton;
    btCancel: TrButton;
    procedure FormShow(Sender: TObject);
    procedure btDefaultClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lbGroupsClick(Sender: TObject);
    procedure eGroupTitleExit(Sender: TObject);
    procedure btGroupUpdateClick(Sender: TObject);
    procedure btNewGroupClick(Sender: TObject);
    procedure btDelGroupClick(Sender: TObject);
    procedure lbGroupFieldsClickCheck(Sender: TObject);
    procedure lbGroupFieldsDblClick(Sender: TObject);
    procedure cbGroupMasterChange(Sender: TObject);
    procedure btGroupUpdateAllClick(Sender: TObject);
    procedure btGroupAddMaskClick(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btGroupCollapseClick(Sender: TObject);
    procedure lbGroupFieldsClick(Sender: TObject);
  private
    Gr: TrDBGrid;
    FFieldsGroupsIni: TMemIniFile;
    Groups, Flds, Cols: TStrings;
    procedure LoadGroups(KeepSelected: Boolean);
    procedure LoadGroupFields;
    function GetSelGroupName: string;
  public
  end;

var
  rDBFieldsGroupForm: TrDBFieldsGroupForm;

implementation

{$R *.DFM}
{$I _rosi.inc}
uses RosiCompGlobalCfg, rtool, rdlg, rdbtool, rstring, rpictool;

var
  isLoading: Boolean;

procedure SetFieldsGroup(const rDBGrid: TrDBGrid);
begin
  isLoading:=true;
  rDBFieldsGroupForm:=TrDBFieldsGroupForm.Create(GetActiveOwner);
  try
    rDBFieldsGroupForm.Gr:=rDBGrid;
    rDBFieldsGroupForm.ShowModal;
  finally
    FreeF(rDBFieldsGroupForm);
  end;
end;

procedure TrDBFieldsGroupForm.FormCreate(Sender: TObject);
begin
  inherited;
  FFieldsGroupsIni:=TMemIniFile.Create('');
  Groups:=TStringList.Create;
  Flds:=TStringList.Create;
  Cols:=TStringList.Create;
end;

procedure TrDBFieldsGroupForm.FormDestroy(Sender: TObject);
begin
  FFieldsGroupsIni.Free;
  Groups.Free;
  Flds.Free;
  Cols.Free;
end;

procedure TrDBFieldsGroupForm.FormShow(Sender: TObject);
var
  i: integer;
begin
  inherited;
  Cols.Clear;
  Flds.Clear;
  for i:=0 to Gr.Columns.Count - 1 do
  begin
    Cols.Add(Gr.Columns[i].Title.Caption);
    Flds.Add(Gr.Columns[i].FieldName);
  end;
  FFieldsGroupsIni.SetStrings(Gr.FieldsGroups);
  LoadGroups(false);
  isLoading:=false;
end;

procedure TrDBFieldsGroupForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Gr.FixedGroupedCol(false);
  inherited;
end;

procedure TrDBFieldsGroupForm.btOkClick(Sender: TObject);
begin
  btGroupUpdateAllClick(nil);
  Close;
end;

procedure TrDBFieldsGroupForm.btGroupUpdateAllClick(Sender: TObject);
begin
  btGroupUpdateClick(nil);
  Gr.FieldsGroups.Clear;
  FFieldsGroupsIni.GetStrings(Gr.FieldsGroups);
  Gr.FixedGroupedCol(true);
  lbGroupsClick(nil);
end;

procedure TrDBFieldsGroupForm.btDefaultClick(Sender: TObject);
begin
  if (DlgQ(sSetDefaultValues) = mrYes) then
  begin
    Gr.FixedGroupedCol(true, true);
    FFieldsGroupsIni.SetStrings(Gr.FieldsGroups);
    LoadGroups(false);
  end;
end;

function TrDBFieldsGroupForm.GetSelGroupName: string;
begin
  if lbGroups.ItemIndex >= 0 then
    Result:=Groups[lbGroups.ItemIndex]
  else
    Result:='';
end;

procedure TrDBFieldsGroupForm.LoadGroups(KeepSelected: Boolean);
var
  i, idx: integer;
  s: string;
begin
  FFieldsGroupsIni.ReadSections(Groups);
  idx:=lbGroups.ItemIndex;
  lbGroups.Clear;
  for i:=0 to Groups.Count - 1 do
  begin
    s:=Groups[i];
    lbGroups.Items.Add(rtool.RemoveStrPart(s, GroupHeaderSep, true) + ' - ' + rtool.RemoveStrPart(s, GroupHeaderSep, true));
  end;
  lbGroups.ItemIndex:=-1;
  if KeepSelected and (idx >= 0) and (idx < lbGroups.Items.Count) then
    lbGroups.ItemIndex:=idx;
  lbGroupsClick(nil);
end;

procedure TrDBFieldsGroupForm.btNewGroupClick(Sender: TObject);
var
  ns: string;
  i, id, max: integer;
begin
  // find max used id
  max:=0;
  for i:=0 to Groups.Count - 1 do
  begin
    id:=StrToIntDef(rtool.FindStrPart(Groups[i], GroupHeaderSep, 1, true), -1);
    if id > max then
      max:=id;
  end;
  // generate new group
  ns:=IntToStr(max + 1) + GroupHeaderSep; // + GroupHeaderSep + GroupHeaderSep;
  // create section by temp value
  FFieldsGroupsIni.WriteString(ns, '_temp_', '0');
  FFieldsGroupsIni.DeleteKey(ns, '_temp_');
  LoadGroups(false);
end;

procedure TrDBFieldsGroupForm.btDelGroupClick(Sender: TObject);
begin
  if lbGroups.ItemIndex >= 0 then
  begin
    FFieldsGroupsIni.EraseSection(GetSelGroupName);
    LoadGroups(false);
  end;
end;

procedure TrDBFieldsGroupForm.lbGroupsClick(Sender: TObject);
var
  GroupId: integer;
  Title, Content: string;
  TitleColor, LineColor: TColor;
  Locked: boolean;
begin
  if lbGroups.ItemIndex >= 0 then
  begin
    if Gr.Fields_GetGroupSettings(GetSelGroupName, GroupId, Title, Content, TitleColor, LineColor, Locked, nil) then
    begin
      eGroupTitle.Text:=Title;
      eGroupMasterMask.Text:=Content;
      cgGroupTitleColor.Selected:=TitleColor;
      cgGroupLineColor.Selected:=LineColor;
      lblLocked.Visible:=Locked;
    end;
    lbGroups.Tag:=GroupId;
    LoadGroupFields;
  end
  else
  begin
    lbGroups.Tag:=0;
    eGroupTitle.Text:='';
    cgGroupTitleColor.Selected:=clNone;
    cgGroupLineColor.Selected:=clNone;
    lbGroupFields.CheckAll(cbUnchecked);
  end;

  gbGroupBottom.Enabled:=(lbGroups.ItemIndex >= 0) and not lblLocked.Visible;
  gbRightGroups.Enabled:=gbGroupBottom.Enabled;
  btDelGroup.Enabled:=gbGroupBottom.Enabled;
  btGroupCollapse.Enabled:=(lbGroups.ItemIndex >= 0);
  btGroupExpand.Enabled:=btGroupCollapse.Enabled;
  btGroupAddMask.Enabled:=false;
end;

procedure TrDBFieldsGroupForm.LoadGroupFields;
var
  i, idx: integer;
  FieldList, L: TStringList;
begin
  // load group columns list and reset state
  lbGroupFields.Items.Assign(Cols);
  lbGroupFields.CheckAll(cbUnchecked);
  cbGroupMaster.Items.Clear;
  FieldList:=TStringList.Create;
  try
    // load list of fields in the group
    L:=TStringList.Create;
    try
      FFieldsGroupsIni.ReadSectionValues(GetSelGroupName, L);
      FieldList.Clear;
      for i:=0 to L.Count - 1 do
        if SameText(L.ValueFromIndex[i], 'M') then
          FieldList.Insert(0, L.Names[i])
        else
          FieldList.Add(L.Names[i]);
    finally
      L.Free;
    end;
    // remove non existing fields from Group definition
    for i:=0 to FieldList.Count - 1 do
    begin
      if Flds.IndexOf(FieldList[i]) < 0 then
        FFieldsGroupsIni.DeleteKey(GetSelGroupName, FieldList[i]);
    end;
    // check grouped fields
    for i:=0 to Flds.Count - 1 do
    begin
      idx:=FieldList.IndexOf(Flds[i]);
      if idx >= 0 then
      begin
        lbGroupFields.Checked[i]:=true;
        cbGroupMaster.Items.Add(lbGroupFields.Items[i]);
        if idx = 0 then
          cbGroupMaster.ItemIndex:=cbGroupMaster.Items.IndexOf(lbGroupFields.Items[i]);
      end;
    end;
  finally
    FieldList.Free;
  end;
end;

procedure TrDBFieldsGroupForm.btGroupAddMaskClick(Sender: TObject);
begin
  if lbGroupFields.ItemIndex >= 0 then
  begin
    eGroupMasterMask.Text:=eGroupMasterMask.Text + '<<' + Flds[lbGroupFields.ItemIndex] + '>>';
    eGroupMasterMask.SetFocus;
  end;
end;

procedure TrDBFieldsGroupForm.btGroupCollapseClick(Sender: TObject);
var
  id: integer;
begin
  id:=StrToIntDef(rtool.FindStrPart(GetSelGroupName, GroupHeaderSep, 1, true), -1);
  if Sender = btGroupCollapse then
    Gr.CollapseGroup(id)
  else
    Gr.ExpandGroup(id);
end;

procedure TrDBFieldsGroupForm.btGroupUpdateClick(Sender: TObject);
var
  newsec: string;
  L: TStringList;
  i: integer;
begin
  if Sender = btNone1 then
    cgGroupTitleColor.Selected:=clNone;
  if Sender = btNone2 then
    cgGroupLineColor.Selected:=clNone;

  newsec:=Gr.Fields_SetGroupSettings(lbGroups.Tag, eGroupTitle.Text, eGroupMasterMask.Text, cgGroupTitleColor.Selected, cgGroupLineColor.Selected,
    lblLocked.Visible, nil);
  L:=TStringList.Create;
  try
    // update section name
    FFieldsGroupsIni.GetStrings(L);
    i:=L.IndexOf('[' + GetSelGroupName + ']');
    if i >= 0 then
      L[i]:='[' + newsec + ']';
    FFieldsGroupsIni.SetStrings(L);
    LoadGroups(true);
  finally
    L.Free;
  end;
end;

procedure TrDBFieldsGroupForm.cbGroupMasterChange(Sender: TObject);
var
  i: integer;
begin
  for i:=0 to lbGroupFields.Items.Count - 1 do
  begin
    if lbGroupFields.Checked[i] then
      FFieldsGroupsIni.WriteString(GetSelGroupName, Flds[i], '0');
  end;
  i:=lbGroupFields.Items.IndexOf(cbGroupMaster.Items[cbGroupMaster.ItemIndex]);
  if i >= 0 then
    FFieldsGroupsIni.WriteString(GetSelGroupName, Flds[i], 'M');
end;

procedure TrDBFieldsGroupForm.eGroupTitleExit(Sender: TObject);
begin
  btGroupUpdateClick(Sender);
end;

procedure TrDBFieldsGroupForm.lbGroupFieldsClick(Sender: TObject);
begin
  btGroupAddMask.Enabled:=true;
end;

procedure TrDBFieldsGroupForm.lbGroupFieldsClickCheck(Sender: TObject);
var
  i, idx: integer;
  s: string;
begin
  i:=lbGroupFields.ItemIndex;
  if lbGroupFields.Checked[i] then
  begin
    // add grouped field
    if cbGroupMaster.Items.Count = 0 then
      // first field as Master
      s:='M'
    else
      s:='0';
    FFieldsGroupsIni.WriteString(GetSelGroupName, Flds[i], s);
    cbGroupMaster.Items.Add(lbGroupFields.Items[i]);
    if cbGroupMaster.Items.Count = 1 then
      cbGroupMaster.ItemIndex:=0;
  end
  else
  begin
    // remove grouped field
    FFieldsGroupsIni.DeleteKey(GetSelGroupName, Flds[i]);
    idx:=cbGroupMaster.Items.IndexOf(lbGroupFields.Items[i]);
    if idx >= 0 then
      cbGroupMaster.Items.Delete(idx);
  end;
end;

procedure TrDBFieldsGroupForm.lbGroupFieldsDblClick(Sender: TObject);
begin
  lbGroupFields.Checked[lbGroupFields.ItemIndex]:=not lbGroupFields.Checked[lbGroupFields.ItemIndex];
  lbGroupFieldsClickCheck(nil);
end;

end.
