unit rDBFilter;

{<
  rDBFilter provides filtering functions and universal dialog
  for user definition of filter string.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Apr 2017)
}

// define compilation for TpFIBDataSet
{$DEFINE ShowRecCountInfoOnly}
{$DEFINE UseManualRecordCount}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, db, dbgrids, Grids, ExtCtrls, PropSaveMain,
  ImgList, rHTMLLabel, rdbtool, rdbdlg_baseform, rImprovedComps, ComCtrls,
  ToolWin, Menus, StrUtils, DateUtils, rStringGridEd;

{ Show DBFilter dialog and set filter for DBGrid (Columns collection is used) }
function FilterRec(const DBGrid: TCustomDBGrid; SortList: boolean = false): boolean; overload;
{ Show DBFilter dialog and set filter for Dataset (Fields collection is used) }
function FilterRec(const DataSet: TDataSet; SortList: boolean = false): boolean; overload;

{ Show DBFilter dialog and set FilterStr parameter for DBGrid (Columns collection is used),
  filter is not activated }
function CompoundFilterString(const DBGrid: TCustomDBGrid; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean; overload;
{ Show DBFilter dialog and set FilterStr parameter for Dataset (Fields collection is used),
  filter is not activated }
function CompoundFilterString(const DataSet: TDataSet; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean; overload;

{ only for backward compability }
function FilterRecSorted(const DBGrid: TCustomDBGrid): boolean; overload;
{ only for backward compability }
function FilterRecSorted(const DataSet: TDataSet): boolean; overload;

{ auxilary functions moved to rdbtool unit, keep here only for compability }
function SetFilterString(D: TDataSet; Filter: string; ShowError: boolean = true): boolean;
procedure ClearFilter(const DBGrid: TCustomDBGrid); overload;
procedure ClearFilter(const DataSet: TDataSet); overload;
function GetFilterCond(Field: TField; Value: string; IsNull: boolean): string;
function FilterByField(Field: TField; FilterJoin: TFilterJoin = fjNone): boolean;
function FilterBySel(const DBGrid: TCustomDBGrid; FilterJoin: TFilterJoin = fjNone): boolean;
function NegateFilter(const DBGrid: TCustomDBGrid): boolean; overload;
function NegateFilter(const DataSet: TDataSet): boolean; overload;

type
  { filtering functions and universal dialog for user definition of filter string }
  TrDBFilterForm = class(TrDBDlgBaseForm)
    btOk: TrButton;
    btCancel: TrButton;
    lCond: TLabel;
    eRes: TMemo;
    chCase: TCheckBox;
    btClear: TrButton;
    btTest: TrButton;
    lTest: TLabel;
    GroupBox1: TGroupBox;
    lFCap: TLabel;
    lFName: TLabel;
    lOper: TLabel;
    lbSignNum1: TListBox;
    sgPol: TrStringGridEd;
    sgSign: TrStringGridEd;
    lbSignNum2: TListBox;
    lValue: TLabel;
    btAdd: TrButton;
    rgAndOr: TRadioGroup;
    btLoad: TrButton;
    btSave: TrButton;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    btAddOk: TrButton;
    cbDelQuotes: TCheckBox;
    eValue: TComboBox;
    lblClearHist: TrHTMLLabel;
    chShowRecCount: TCheckBox;
    eValDate1: TDateTimePicker;
    eValDate2: TDateTimePicker;
    eValNum1: TrEditNum;
    eValNum2: TrEditNum;
    lValue1: TLabel;
    lValue2: TLabel;
    pmPeriod: TPopupMenu;
    mnDay: TMenuItem;
    rDay1: TMenuItem;
    rDay2: TMenuItem;
    rDay3: TMenuItem;
    N2: TMenuItem;
    aDay1: TMenuItem;
    aDay2: TMenuItem;
    mnWeek: TMenuItem;
    rWeek1: TMenuItem;
    rWeek2: TMenuItem;
    rWeek3: TMenuItem;
    mnMonth: TMenuItem;
    rMonth1: TMenuItem;
    rMonth2: TMenuItem;
    rMonth3: TMenuItem;
    N1: TMenuItem;
    aMonth1: TMenuItem;
    aMonth2: TMenuItem;
    aMonth3: TMenuItem;
    aMonth4: TMenuItem;
    aMonth5: TMenuItem;
    aMonth6: TMenuItem;
    aMonth7: TMenuItem;
    aMonth8: TMenuItem;
    aMonth9: TMenuItem;
    aMonth10: TMenuItem;
    aMonth11: TMenuItem;
    aMonth12: TMenuItem;
    mnQuarter: TMenuItem;
    rQuarter1: TMenuItem;
    rQuarter2: TMenuItem;
    rQuarter3: TMenuItem;
    N4: TMenuItem;
    aQuarter1: TMenuItem;
    aQuarter2: TMenuItem;
    aQuarter3: TMenuItem;
    aQuarter4: TMenuItem;
    mnYear: TMenuItem;
    rYear1: TMenuItem;
    rYear2: TMenuItem;
    rYear3: TMenuItem;
    N3: TMenuItem;
    aYear1: TMenuItem;
    aYear2: TMenuItem;
    aYear3: TMenuItem;
    aYear4: TMenuItem;
    aYear5: TMenuItem;
    aYear6: TMenuItem;
    aYear7: TMenuItem;
    aYear8: TMenuItem;
    aYear9: TMenuItem;
    aYear10: TMenuItem;
    aYear11: TMenuItem;
    ToolBarPeriod: TToolBar;
    btPeriod: TToolButton;
    mnOther: TMenuItem;
    rOther1: TMenuItem;
    rOther2: TMenuItem;
    rOther4: TMenuItem;
    N5: TMenuItem;
    rOther5: TMenuItem;
    rOther6: TMenuItem;
    rOther8: TMenuItem;
    mnHalfYear: TMenuItem;
    rHalf1: TMenuItem;
    rHalf2: TMenuItem;
    rHalf3: TMenuItem;
    N6: TMenuItem;
    aHalf1: TMenuItem;
    aHalf2: TMenuItem;
    rOther3: TMenuItem;
    rOther7: TMenuItem;
    rbValTrue: TRadioButton;
    rbValFalse: TRadioButton;
    lbSignStr1: TListBox;
    lbSignStr2: TListBox;
    procedure sgPolDblClick(Sender: TObject);
    procedure btAddClick(Sender: TObject);
    procedure btClearClick(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btTestClick(Sender: TObject);
    procedure eResChange(Sender: TObject);
    procedure eValueChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure sgSignDblClick(Sender: TObject);
    procedure btLoadClick(Sender: TObject);
    procedure btSaveClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure sgPolDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: boolean);
    procedure lblClearHistAnchorClick(Sender: TObject; LinkName: String);
    procedure rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure sgPolClick(Sender: TObject);
    procedure sgSignClick(Sender: TObject);
    procedure mnPeriodClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    GDS: TComponent;
    Compound: boolean;
    SQLFormat: boolean;
    Sorted: boolean;
    RecCount: Integer;
    SelFld: TField;
    SelFldType: Integer;
    procedure UpdateSpecCondition;
    procedure ScanCallBack(DataSet: TDataSet; RecIndex: Integer; var Abort: boolean);
    function GetPSKey: string;
    procedure InitMenu;
    procedure RefSelField;
    procedure RefSelCond;
    function GetSelFieldName: string;
    function SetFilter(test: boolean): boolean;
    procedure SetText(s: string);
    procedure SortList;
    function UpdateStringValue(EventID: Integer; const s: string): string;
  public
    procedure AddTextToFieldList(Caption, Name: string);
    procedure AddTextToValueList(Text: string);
  end;

var
  rDBFilterForm: TrDBFilterForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rtool, rdlg, rstring, rDBGrid;

const
  cSelField = 'SelectedField';
  cSelOperatorId = 'SelectedOperator';
  cBetweenRow = 6;
  cLikeMaxRow = 1;

function FilterRecInternal(const GridOrDataSet: TObject; Sorted: boolean): boolean; forward;
function CompoundFilterStringInt(const GridOrDataSet: TObject; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean; forward;

function FilterRec(const DBGrid: TCustomDBGrid; SortList: boolean = false): boolean; overload;
begin
  Result:=FilterRecInternal(DBGrid, SortList);
end;

function FilterRec(const DataSet: TDataSet; SortList: boolean = false): boolean; overload;
begin
  Result:=FilterRecInternal(DataSet, SortList);
end;

function FilterRecSorted(const DBGrid: TCustomDBGrid): boolean; overload;
begin
  Result:=FilterRecInternal(DBGrid, true);
end;

function FilterRecSorted(const DataSet: TDataSet): boolean; overload;
begin
  Result:=FilterRecInternal(DataSet, true);
end;

function CompoundFilterString(const DBGrid: TCustomDBGrid; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean; overload;
begin
  Result:=CompoundFilterStringInt(DBGrid, FilterStr, PSFile, SQLFormat);
end;

function CompoundFilterString(const DataSet: TDataSet; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean; overload;
begin
  Result:=CompoundFilterStringInt(DataSet, FilterStr, PSFile, SQLFormat);
end;

function FilterRecInternal(const GridOrDataSet: TObject; Sorted: boolean): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(GridOrDataSet) then
    Exit;
  rDBFilterForm:=TrDBFilterForm.Create(GetActiveOwner);
  try
    rDBFilterForm.GDS:=GridOrDataSet as TComponent;
    rDBFilterForm.eRes.Text:=GetDataSet(GridOrDataSet).Filter;
    rDBFilterForm.cbDelQuotes.Visible:=RosiCompConfig.rDBFilterCfg.ShowDelQuotes;
    if not RosiCompConfig.rDBFilterCfg.ShowDelQuotes then
      rDBFilterForm.cbDelQuotes.Checked:=false;
    rDBFilterForm.Sorted:=Sorted;
    Result:=rDBFilterForm.ShowModal = mrOK;
  finally
    FreeF(rDBFilterForm);
  end;
end;

function CompoundFilterStringInt(const GridOrDataSet: TObject; var FilterStr: string; PSFile: TPropSaveFile; SQLFormat: boolean): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(GridOrDataSet) then
    Exit;
  rDBFilterForm:=TrDBFilterForm.Create(GetActiveOwner);
  try
    rDBFilterForm.GDS:=GridOrDataSet as TComponent;
    rDBFilterForm.Compound:=true;
    rDBFilterForm.SQLFormat:=SQLFormat;
    rDBFilterForm.eRes.Text:=FilterStr;
    rDBFilterForm.rDlgPropSave.PropSaveFile:=PSFile;
    Result:=rDBFilterForm.ShowModal = mrOK;
    if Result then
      FilterStr:=rDBFilterForm.eRes.Text;
  finally
    FreeF(rDBFilterForm);
  end;
end;

procedure TrDBFilterForm.FormCreate(Sender: TObject);
begin
  inherited;
  Compound:=false;
  SQLFormat:=false;
  Sorted:=false;
  SelFldType:=0;
end;

procedure TrDBFilterForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  rDlgPropSave.SaveProperties;
  inherited;
end;

procedure TrDBFilterForm.FormShow(Sender: TObject);
var
  poc: Integer;
  L: TStrings;
begin
  inherited;
  sgPol.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  sgSign.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  { if RosiCompConfig.rDBFilterCfg.HideFieldNames then
    begin
    sgPol.ColWidths[1]:=0;
    sgPol.ColWidths[0]:=sgPol.ClientWidth-20;
    end
    else
    if sgPol.ColWidths[1]=0 then
    begin
    sgPol.ColWidths[0]:=(sgPol.ClientWidth-20) div 2;
    sgPol.ColWidths[1]:=sgPol.ColWidths[0];
    end;
    { }
  sgPol.Columns[1].Visible:=not RosiCompConfig.rDBFilterCfg.HideFieldNames;
  lFName.Visible:=not RosiCompConfig.rDBFilterCfg.HideFieldNames;
  lCond.Font.Style:=[fsBold];
  L:=TStringList.Create;
  try
    poc:=GetFldList(GDS, L, [fdGetLabels, fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
    sgPol.RowCount:=poc;
    sgPol.Cols[0].Assign(L);
    GetFldList(GDS, L, [fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
    sgPol.Cols[1].Assign(L);
    lTest.Caption:='';
    btTest.Visible:=not Compound;
    eRes.SelStart:=Length(eRes.Text);
    chCase.Checked:=not(foCaseInsensitive in GetDataSet(GDS).FilterOptions);
    chCase.Visible:=RosiCompConfig.rDBFilterCfg.ShowCaseSensitive;
    OpenDlg.Filter:=OpenDlg.Filter + '|' + sFileTypeAll;
    SaveDlg.Filter:=OpenDlg.Filter;
  finally
    L.Free;
  end;
  if Sorted then
    SortList;
  InitMenu;
  SelFld:=nil;
  self.eValDate1.DateTime:=Trunc(Date);
  self.eValDate2.DateTime:=Trunc(Date);
  rDlgPropSave.ReadProperties;
  eValue.Text:='';
  sgPolClick(nil);
end;

procedure TrDBFilterForm.InitMenu;
var
  i, j: Integer;
  mi: TMenuItem;
begin
  for i:=0 to self.pmPeriod.Items.Count - 1 do
    for j:=0 to pmPeriod.Items[i].Count - 1 do
    begin
      mi:=pmPeriod.Items[i].Items[j];
      mi.OnClick:=mnPeriodClick;
      if PosEx('aYear', mi.Name) = 1 then
        mi.Caption:=IntToStr(DateUtils.YearOf(Date) + mi.Tag);
      if PosEx('aMonth', mi.Name) = 1 then
      begin
{$IFDEF DXEP}
        mi.Caption:=FormatSettings.LongMonthNames[mi.Tag];
{$ELSE}
        mi.Caption:=LongMonthNames[mi.Tag];
{$ENDIF}
      end;
    end;
end;

procedure TrDBFilterForm.UpdateSpecCondition;
begin
  if IsAdvantageDatabaseDataset(GetDataSet(GDS)) then
  begin
    sgSign.Cells[1, 0]:='$';
    sgSign.Cells[1, 1]:='$';
  end;
end;

procedure TrDBFilterForm.mnPeriodClick(Sender: TObject);
var
  mi: TMenuItem;
  act: TDateTime;
begin
  mi:=Sender as TMenuItem;
  act:=Trunc(DateUtils.Today);
  if PosEx('rDay', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=act + mi.Tag;
    eValDate2.DateTime:=eValDate1.DateTime;
    Exit;
  end;
  if PosEx('aDay', mi.Name) = 1 then
  begin
    if mi.Tag = 1 then
      eValDate1.DateTime:=DateUtils.StartOfTheMonth(act)
    else
      eValDate1.DateTime:=DateUtils.EndOfTheMonth(act);
    eValDate2.DateTime:=eValDate1.DateTime;
    Exit;
  end;
  if PosEx('rWeek', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=DateUtils.StartOfTheWeek(act) + mi.Tag * 7;
    eValDate2.DateTime:=eValDate1.DateTime + 6;
    Exit;
  end;
  if PosEx('rMonth', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(DateUtils.StartOfTheMonth(act), mi.Tag);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('aMonth', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), mi.Tag);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('rQuarter', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 3 * 3 + 1)), mi.Tag * 3);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 2));
    Exit;
  end;
  if PosEx('aQuarter', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(YearOf(act), (mi.Tag - 1) * 3 + 1, 1);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 2));
    Exit;
  end;
  if PosEx('rHalf', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncMonth(DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 6 * 6 + 1)), mi.Tag * 6);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 5));
    Exit;
  end;
  if PosEx('aHalf', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(YearOf(act), (mi.Tag - 1) * 6 + 1, 1);
    eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(eValDate1.DateTime, 5));
    Exit;
  end;
  if PosEx('rYear', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=IncYear(DateUtils.StartOfTheYear(act), mi.Tag);
    eValDate2.DateTime:=DateUtils.EndOfTheYear(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('aYear', mi.Name) = 1 then
  begin
    eValDate1.DateTime:=EncodeDate(DateUtils.YearOf(act) + mi.Tag, 1, 1);
    eValDate2.DateTime:=DateUtils.EndOfTheYear(eValDate1.DateTime);
    Exit;
  end;
  if PosEx('rOther', mi.Name) = 1 then
  begin
    case mi.Tag of
      - 1:
        eValDate1.DateTime:=DateUtils.StartOfTheMonth(act);
      -2:
        eValDate1.DateTime:=DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 3 * 3 + 1));
      -3:
        eValDate1.DateTime:=DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 6 * 6 + 1));
      -4:
        eValDate1.DateTime:=DateUtils.StartOfTheYear(act);
      1:
        eValDate2.DateTime:=DateUtils.EndOfTheMonth(act);
      2:
        eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 3 * 3 + 1))
              , 2));
      3:
        eValDate2.DateTime:=DateUtils.EndOfTheMonth(IncMonth(DateUtils.RecodeMonth(DateUtils.StartOfTheMonth(act), ((DateUtils.MonthOf(act) - 1) div 6 * 6 + 1))
              , 5));
      4:
        eValDate2.DateTime:=DateUtils.EndOfTheYear(act);
    end;
    if mi.Tag < 0 then
      eValDate2.DateTime:=act
    else
      eValDate1.DateTime:=act;
    Exit;
  end;
end;

function TrDBFilterForm.GetPSKey;
begin
  Result:=rDBFilterForm.Name + '.' + rDBFilterForm.GDS.Owner.Name + '.' + rDBFilterForm.GDS.Name;
end;

procedure TrDBFilterForm.RefSelField;
var
  FldType: Integer;
begin
  FldType:=1;
  if Assigned(SelFld) then
  begin
    if ((SelFld is TNumericField) or IsDateTimeField(SelFld)) then
      FldType:=2
    else
      if (SelFld is TBooleanField) then
        FldType:=3;
  end;
  if SelFldType <> FldType then
  begin
    if FldType > 1 then
    begin
      sgSign.RowCount:=lbSignNum1.Items.Count;
      sgSign.Cols[0].Assign(lbSignNum1.Items);
      sgSign.Cols[1].Assign(lbSignNum2.Items);
      if FldType=3 then
        sgSign.RowCount:=2;
    end
    else
    begin
      sgSign.RowCount:=lbSignStr1.Items.Count;
      sgSign.Cols[0].Assign(lbSignStr1.Items);
      sgSign.Cols[1].Assign(lbSignStr2.Items);
      UpdateSpecCondition;
    end;
    SelFldType:=FldType;
    sgSign.Row:=0;
  end;
  sgSignClick(nil);
end;

procedure TrDBFilterForm.RefSelCond;
begin
  // set visibility according selected field
  if Assigned(SelFld) then
  begin
    eValNum1.Visible:=SelFld is TNumericField;
    eValDate1.Visible:=IsDateTimeField(SelFld);
    rbValTrue.Visible:=SelFld is TBooleanField;
  end
  else
  begin
    eValNum1.Visible:=false;
    eValDate1.Visible:=false;
    rbValTrue.Visible:=false;
  end;

  // set visibility according selected field and sign
  eValNum2.Visible:=eValNum1.Visible and (sgSign.Row = cBetweenRow);
  eValDate2.Visible:=eValDate1.Visible and (sgSign.Row = cBetweenRow);
  rbValFalse.Visible:=rbValTrue.Visible;
  eValue.Visible:=not eValNum1.Visible and not eValDate1.Visible and not rbValTrue.Visible;
  ToolBarPeriod.Visible:=eValDate2.Visible;

  lValue1.Visible:=eValNum2.Visible or eValDate2.Visible;
  lValue2.Visible:=lValue1.Visible;
  lValue.Visible:=not lValue1.Visible;
  lblClearHist.Visible:=eValue.Visible;
end;

function TrDBFilterForm.GetSelFieldName: string;
begin
  Result:=GetQuotedFieldName(sgPol.Cells[1, sgPol.Row]);
end;

procedure TrDBFilterForm.SetText(s: string);
begin
  eRes.SetSelTextBuf(PChar(s));
end;

procedure TrDBFilterForm.sgPolClick(Sender: TObject);
begin
  SelFld:=GetDataSet(GDS).FindField(sgPol.Cells[1, sgPol.Row]);
  RefSelField;
end;

procedure TrDBFilterForm.sgPolDblClick(Sender: TObject);
begin
  SetText(GetSelFieldName);
end;

procedure TrDBFilterForm.sgSignClick(Sender: TObject);
begin
  RefSelCond;
end;

procedure TrDBFilterForm.sgSignDblClick(Sender: TObject);
begin
  SetText(sgSign.Cells[1, sgSign.Row]);
end;

function TrDBFilterForm.UpdateStringValue(EventID: Integer; const s: string): string;
var
  vparam: variant;
begin
  if Assigned(RosiCompConfig.Global_OnRosiCompGeneralEvent) then
  begin
    vparam:=s;
    RosiCompConfig.Global_OnRosiCompGeneralEvent(self, EventID, SelFld, nil, vparam);
    Result:=vparam;
  end
  else
    Result:=s;
end;

procedure TrDBFilterForm.btAddClick(Sender: TObject);
var
  sep, sval, sadd, snot, sOper, LikeOper: string;
  LikeCh: Char;
  L: TStringList;
begin
  // add text to history
  if eValue.Visible and (Trim(eValue.Text) <> '') and (eValue.Items.IndexOf(eValue.Text) < 0) then
  begin
    L:=TStringList.Create;
    try
      L.Assign(eValue.Items);
      L.Add(eValue.Text);
      L.Sort;
      eValue.Items.Assign(L);
    finally
      L.Free;
    end;
  end;

  if IsApostrField(SelFld) then
    sep:=''''
  else
    sep:='';
  LikeCh:=rdbtool.GetFilterWildCardChar(SelFld.DataSet, SQLFormat);
  LikeOper:=rdbtool.GetFilterLikeOperator(SelFld.DataSet, SQLFormat);
  if Sender = btAddOk then
    eRes.Text:='';
  if Trim(eRes.Text) <> '' then
  begin
    if rgAndOr.ItemIndex = 0 then
      SetText(' AND ')
    else
      SetText(' OR ');
  end;

  sadd:='';
  sval:='';
  if SelFld is TNumericField then
  begin
    if self.sgSign.Row = cBetweenRow then
      sadd:=MakeFilterBetweenCond(GetSelFieldName, UpdateStringValue(rDBFilter_SetValueTextEventID, GetNumericFieldFilterValue(eValNum1.Value)),
        UpdateStringValue(rDBFilter_SetValueTextEventID, GetNumericFieldFilterValue(eValNum2.Value)))
    else
      sval:=GetNumericFieldFilterValue(eValNum1.Value);
  end;
  if IsDateTimeField(SelFld) then
  begin
    if self.sgSign.Row = cBetweenRow then
      sadd:=MakeFilterBetweenCond(GetSelFieldName, UpdateStringValue(rDBFilter_SetValueTextEventID, GetDateFieldFilterValue(eValDate1.Date)), UpdateStringValue
          (rDBFilter_SetValueTextEventID, GetDateFieldFilterValue(eValDate2.Date)), '''')
    else
      sval:=GetDateFieldFilterValue(eValDate1.Date);
  end;
  if SelFld is TBooleanField then
  begin
    sval:=RosiCompConfig.rDBFilterCfg.BooleanString[self.rbValTrue.Checked];
  end;

  if sadd = '' then
  begin
    if (SelFldType = 1) and (sgSign.Row <= cLikeMaxRow) and IsAdvantageDatabaseDataset(SelFld.DataSet) then
    begin
      // use 'text' $ Fieldname for AdvantageDatabase instead of LIKE
      sadd:='(' + sep + UpdateStringValue(rDBFilter_SetValueTextEventID, eValue.Text) + sep + ' $ ' + GetSelFieldName + ')';
    end
    else
    begin
      if sval = '' then
      begin
        sval:=eValue.Text;
        if (sgSign.Row = 0) then
          sval:=LikeCh + sval + LikeCh;
        if (sgSign.Row = 1) then
          sval:=sval + LikeCh;
      end;
      // find correct sing
      if (SelFldType = 1) and (sgSign.Row <= cLikeMaxRow) then
        sOper:=LikeOper
      else
        sOper:=sgSign.Cells[1, sgSign.Row];
      snot:='';
      if rdbtool.IsClientDataset(SelFld.DataSet) and IsMemoField(SelFld) then
      begin
        // use "LIKE" or "NOT FieldName LIKE" instead of "=" or "<>" for ClientDataSet and memo field
        if (sgSign.Row = 2) or (sgSign.Row = 3) then
          sOper:='LIKE';
        if (sgSign.Row = 3) then
          snot:=' NOT ';
      end;
      sadd:=MakeFilterCond(snot + GetSelFieldName, sOper, UpdateStringValue(rDBFilter_SetValueTextEventID, sval), sep);
    end;
  end;
  SetText(UpdateStringValue(rDBFilter_SetFilterTextEventID, sadd));
  eValue.Text:='';
  sgPol.SetFocus;
  if Sender = btAddOk then
    btOk.Click;
end;

procedure TrDBFilterForm.btClearClick(Sender: TObject);
begin
  eRes.Text:='';
end;

procedure TrDBFilterForm.btOkClick(Sender: TObject);
begin
  if Compound or SetFilter(false) then
    ModalResult:=mrOK;
end;

procedure TrDBFilterForm.ScanCallBack(DataSet: TDataSet; RecIndex: Integer; var Abort: boolean);
begin
  Inc(RecCount);
end;

function TrDBFilterForm.SetFilter(test: boolean): boolean;
var
  D: TDataSet;
  s: string;
begin
  D:=GetDataSet(GDS);
  try
    if chCase.Checked then
      D.FilterOptions:=[]
    else
      D.FilterOptions:=[foCaseInsensitive];
  except
  end;
  s:=eRes.Text;
  if cbDelQuotes.Checked then
    s:=ClearStr('"', s, true);
  Result:=SetFilterString(D, s, true);
  if s = '' then
    D.Filter:='';
  if (not Result) or test then
    Exit;
  if chShowRecCount.Checked then
  begin
{$IFDEF UseManualRecordCount}
    RecCount:=0;
    rdbtool.ScanAllRecords(D, self.ScanCallBack);
{$ELSE}
    RecCount:=D.RecordCount;
{$ENDIF}
{$IFDEF ShowRecCountInfoOnly}
    DlgI(Format(sSelRecords, [RecCount]));
{$ELSE}
    Result:=DlgQ2(Format(sSelRecords, [RecCount]), sUseFilterQ) = mrYes;
{$ENDIF}
  end;
  if Result and (GDS is TrDBGrid) then
    TrDBGrid(GDS).FilterChangedNotification;
end;

procedure TrDBFilterForm.btTestClick(Sender: TObject);
var
  oldOpt: TFilterOptions;
  oldFilter: string;
  oldFiltered: boolean;
begin
  with GetDataSet(GDS) do
  begin
    oldOpt:=FilterOptions;
    oldFilter:=Filter;
    oldFiltered:=Filtered;
    if SetFilter(true) then
    begin
      lTest.Caption:=sAllOk + ' ' + Format(sSelRecords, [RecordCount]);
      btTest.Enabled:=false;
    end
    else
      lTest.Caption:='';
    FilterOptions:=oldOpt;
    Filter:=oldFilter;
    Filtered:=oldFiltered;
  end;
end;

procedure TrDBFilterForm.eResChange(Sender: TObject);
begin
  btTest.Enabled:=true;
  lTest.Caption:='';
end;

procedure TrDBFilterForm.eValueChange(Sender: TObject);
begin
  btAdd.Default:=eValue.Text <> '';
  btOk.Default:=not btAdd.Default;
end;

procedure TrDBFilterForm.btLoadClick(Sender: TObject);
begin
  if OpenDlg.Execute then
    eRes.Lines.LoadFromFile(OpenDlg.FileName);
end;

procedure TrDBFilterForm.btSaveClick(Sender: TObject);
begin
  if SaveDlg.Execute then
    eRes.Lines.SaveToFile(SaveDlg.FileName);
end;

procedure TrDBFilterForm.rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
var
  fn: string;
  i: Integer;
begin
  if RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields then
    fn:=''
  else
    fn:=PSFile.ReadString(GetPSKey, cSelField, '');
  try
    i:=sgPol.Cols[1].IndexOf(fn);
    if i < 0 then
    begin
      if (GDS is TCustomDBGrid) and Assigned(TCustomDBGrid(GDS).SelectedField) then
        i:=sgPol.Cols[1].IndexOf(TCustomDBGrid(GDS).SelectedField.FieldName);
      if i < 0 then
        i:=0;
    end;
    sgPol.Row:=i;
    sgSign.Row:=PSFile.ReadInteger(GetPSKey, cSelOperatorId, 0);
    if eValue.Visible then
      eValue.SetFocus
    else
      if eValNum1.Visible then
        eValNum1.SetFocus
      else
        if eValDate1.Visible then
          eValDate1.SetFocus;
  except
  end;
end;

procedure TrDBFilterForm.rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  if not RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields then
    PSFile.WriteString(GetPSKey, cSelField, sgPol.Cells[1, sgPol.Row]);
  PSFile.WriteInteger(GetPSKey, cSelOperatorId, sgSign.Row);
end;

procedure TrDBFilterForm.SortList;
var
  i, j, min: Integer;
  L1, L2: TStringList;
begin
  L1:=TStringList.Create;
  L2:=TStringList.Create;
  try
    L1.Assign(sgPol.Cols[0]);
    L2.Assign(sgPol.Cols[1]);
    for i:=0 to L1.Count - 2 do
    begin
      min:=i;
      for j:=i + 1 to L1.Count - 1 do
      begin
        if AnsiCompareText(L1[j], L1[min]) < 0 then
          min:=j;
      end;
      if min <> i then
      begin
        L1.Move(min, i);
        L2.Move(min, i);
      end;
    end;
    sgPol.Cols[0].Assign(L1);
    sgPol.Cols[1].Assign(L2);
    sgPol.Row:=0;
  finally
    L1.Free;
    L2.Free;
  end;
end;

procedure TrDBFilterForm.sgPolDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: boolean);
begin
  Accept:=true;
end;

procedure TrDBFilterForm.lblClearHistAnchorClick(Sender: TObject; LinkName: String);
begin
  eValue.Items.Clear;
end;

procedure TrDBFilterForm.AddTextToFieldList(Caption: string; Name: string);
begin
  sgPol.RowCount:=sgPol.RowCount + 1;
  sgPol.Cells[0, sgPol.RowCount - 1]:=Caption;
  sgPol.Cells[1, sgPol.RowCount - 1]:=Name;
end;

procedure TrDBFilterForm.AddTextToValueList(Text: string);
begin
  if eValue.Items.IndexOf(Text) < 0 then
    eValue.Items.Add(Text);
end;

// Auxilary functions

function SetFilterString(D: TDataSet; Filter: string; ShowError: boolean = true): boolean;
begin
  Result:=rdbtool.SetFilterString(D, Filter, ShowError);
end;

procedure ClearFilter(const DBGrid: TCustomDBGrid);
begin
  rdbtool.ClearFilter(DBGrid);
end;

procedure ClearFilter(const DataSet: TDataSet);
begin
  rdbtool.ClearFilter(DataSet);
end;

function GetFilterCond(Field: TField; Value: string; IsNull: boolean): string;
begin
  Result:=rdbtool.GetFilterCond(Field, Value, IsNull);
end;

function FilterByField(Field: TField; FilterJoin: TFilterJoin = fjNone): boolean;
begin
  Result:=rdbtool.FilterByField(Field, FilterJoin);
end;

function FilterBySel(const DBGrid: TCustomDBGrid; FilterJoin: TFilterJoin = fjNone): boolean;
begin
  Result:=rdbtool.FilterBySel(DBGrid, FilterJoin);
end;

function NegateFilter(const DBGrid: TCustomDBGrid): boolean; overload;
begin
  Result:=rdbtool.NegateFilter(DBGrid);
end;

function NegateFilter(const DataSet: TDataSet): boolean; overload;
begin
  Result:=rdbtool.NegateFilter(DataSet);
end;

end.
