unit rDBFilterSaver;

{<
  rDBFilterSaver provides universal dialog for saving and loading
  of filter strings.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Feb 2003)
  @lastmod(Sep 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, rDBGrid, PropSaveMain, ExtCtrls, StdCtrls,
  Buttons, ImgList, PropSaveGrids, rStringGridEd, rImprovedComps,
  rdbdlg_baseform, DB;

{ Show dialog for saving of current filter - user can edit filter name and save }
procedure SaveFilter(rDBGrid: TrDBGrid);
{ Show DBFilterSaver dialog with list of stored filters - user can select and activate filter }
function LoadFilter(rDBGrid: TrDBGrid): boolean;
{ Load list of stored filters names - used for dropdown menu }
function LoadFilterNames(rDBGrid: TrDBGrid; L: TStrings): boolean;
{ Activate stored filter }
function ActiveFilterName(rDBGrid: TrDBGrid; FilterName: string): boolean;

type
  { universal dialog for saving and loading
    of filter strings. Dialog uses PropSaveMain component }
  TrDBFilterSaverForm = class(TrDBDlgBaseForm)
    GridsPropSave: TGridsPropSave;
    pBottom: TPanel;
    btDel: TrButton;
    btUse: TrButton;
    Panel1: TPanel;
    btClose: TrButton;
    sgFilter: TrStringGridEd;
    procedure GridsPropSaveGetKeyString(Comp: TComponent; var Key: String);
    procedure FormActivate(Sender: TObject);
    procedure btDelClick(Sender: TObject);
    procedure btUseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    G: TrDBGrid;
    mRes: integer;
    function SaveNewFilter: boolean;
    procedure UpdateGrid;
  public
  end;

var
  rDBFilterSaverForm: TrDBFilterSaverForm;

implementation

{$R *.dfm}

uses RosiCompGlobalCfg, rstring, rtool, rdbtool, rdlg, rDBFilter;

function ShowFilterSaverInt(G: TrDBGrid; Load: boolean): boolean; forward;

procedure SaveFilter(rDBGrid: TrDBGrid);
begin
  ShowFilterSaverInt(rDBGrid, false);
end;

function LoadFilter(rDBGrid: TrDBGrid): boolean;
begin
  Result:=ShowFilterSaverInt(rDBGrid, true);
end;

function LoadFilterNames(rDBGrid: TrDBGrid; L: TStrings): boolean;
var
  i: integer;
  s: string;
  fcreated: boolean;
begin
  Result:=false;
  L.Clear;
  if not Assigned(rDBGrid) or (PropSaveMain.GetGlobal_PropSaveFile = nil) then
    Exit;
  fcreated:=Assigned(rDBFilterSaverForm);
  if not fcreated then
    rDBFilterSaverForm:=TrDBFilterSaverForm.Create(GetActiveOwner);
  rDBFilterSaverForm.G:=rDBGrid;
  rDBFilterSaverForm.rDlgPropSave.ReadProperties;
  for i:=1 to rDBFilterSaverForm.sgFilter.RowCount - 1 do
  begin
    s:=rDBFilterSaverForm.sgFilter.Cells[1, i];
    if Trim(s) <> '' then
      L.Add(s);
  end;
  if not fcreated then
    FreeF(rDBFilterSaverForm);
  Result:=true;
end;

function ActiveFilterName(rDBGrid: TrDBGrid; FilterName: string): boolean;
var
  i: integer;
  s: string;
  fcreated: boolean;
begin
  Result:=false;
  if not Assigned(rDBGrid) or (PropSaveMain.GetGlobal_PropSaveFile = nil) then
    Exit;
  fcreated:=Assigned(rDBFilterSaverForm);
  if not fcreated then
    rDBFilterSaverForm:=TrDBFilterSaverForm.Create(GetActiveOwner);
  rDBFilterSaverForm.G:=rDBGrid;
  rDBFilterSaverForm.rDlgPropSave.ReadProperties;
  for i:=1 to rDBFilterSaverForm.sgFilter.RowCount - 1 do
  begin
    s:=rDBFilterSaverForm.sgFilter.Cells[1, i];
    if SameText(s, FilterName) then
      Result:=SetFilterString(rDBGrid.DataSource.DataSet, rDBFilterSaverForm.sgFilter.Cells[2, i], true);
  end;
  if not fcreated then
    FreeF(rDBFilterSaverForm);
end;

function ShowFilterSaverInt(G: TrDBGrid; Load: boolean): boolean;
begin
  rDBFilterSaverForm:=TrDBFilterSaverForm.Create(GetActiveOwner);
  try
    rDBFilterSaverForm.G:=G;
    if Load then
      Result:=rDBFilterSaverForm.ShowModal = mrOk
    else
      Result:=rDBFilterSaverForm.SaveNewFilter;
  finally
    FreeF(rDBFilterSaverForm);
  end;
end;

procedure TrDBFilterSaverForm.FormActivate(Sender: TObject);
begin
  rDBFilterSaverForm.sgFilter.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  UpdateGrid;
  mRes:=mrCancel;
end;

procedure TrDBFilterSaverForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  inherited;
  ModalResult:=mRes;
end;

function TrDBFilterSaverForm.SaveNewFilter: boolean;
  function InSep(s: string): boolean;
  var
    i, p: integer;
  begin
    p:=0;
    for i:=1 to Length(s) do
      if CharInSet(s[i], ['''']) then
        Inc(p);
    Result:=p mod 2 <> 0;
  end;

  function PresetFilterName: string;
  const
    SpecChars = [' ', '=', '<', '>'];
  var
    i, p, rc: integer;
    s, fn: string;
    found: boolean;
  begin
    // replace fieldname by displaylabel
    s:=G.DataSource.DataSet.Filter;
    s:=ReplaceStr('(', '', s, true, true);
    s:=ReplaceStr(')', '', s, true, true);
    rc:=100;
    repeat
      found:=false;
      for i:=0 to G.DataSource.DataSet.FieldCount - 1 do
      begin
        fn:=G.DataSource.DataSet.Fields[i].FieldName;
        if fn = G.DataSource.DataSet.Fields[i].DisplayLabel then
          Continue;
        p:=AnsiPosEx(fn, s, 1, false);
        if (p = 0) or InSep(copy(s, 1, p - 1)) then
          Continue;
        if ((p = 1) or CharInSet(s[p - 1], SpecChars)) and ((p + Length(fn) - 1 = Length(s)) or CharInSet(s[p + Length(fn)], SpecChars)) then
        begin
          s:=copy(s, 1, p - 1) + G.DataSource.DataSet.Fields[i].DisplayLabel + copy(s, p + Length(fn), Length(s));
          found:=true;
        end;
      end;
      Dec(rc);
    until not found or (rc < 0);
    Result:=s;
  end;

var
  name: string;
  r: integer;
begin
  name:=PresetFilterName;
  Result:=InputStr(sInput, sFilterName + ':', name);
  if Result then
  begin
    rDBFilterSaverForm.rDlgPropSave.ReadProperties;
    // Add new filter
    if (sgFilter.RowCount > 2) or (Trim(sgFilter.Cells[1, 1]) <> '') or (Trim(sgFilter.Cells[2, 1]) <> '') then
      sgFilter.RowCount:=sgFilter.RowCount + 1;
    r:=sgFilter.RowCount - 1;
    sgFilter.Cells[1, r]:=name;
    sgFilter.Cells[2, r]:=G.DataSource.DataSet.Filter;
    rDBFilterSaverForm.rDlgPropSave.SaveProperties;
  end;
end;

procedure TrDBFilterSaverForm.UpdateGrid;
var
  i: integer;
begin
  for i:=1 to sgFilter.RowCount - 1 do
  begin
    if (sgFilter.Cells[2, i] = G.DataSource.DataSet.Filter) then
      sgFilter.Cells[0, i]:='+'
    else
      sgFilter.Cells[0, i]:='';
  end;
end;

procedure TrDBFilterSaverForm.GridsPropSaveGetKeyString(Comp: TComponent; var Key: String);
begin
  // Sets key in INI file according Form and DBGrid name
  if Comp = sgFilter then
    Key:=Key + '.' + G.Owner.Name + '.' + G.Name;
end;

procedure TrDBFilterSaverForm.btUseClick(Sender: TObject);
var
  s: string;
begin
  s:=sgFilter.Cells[2, sgFilter.Row];
  SetFilterString(G.DataSource.DataSet, s, true);
  UpdateGrid;
  mRes:=mrOk;
  if Sender = btUse then
    Close;
end;

procedure TrDBFilterSaverForm.btDelClick(Sender: TObject);
var
  r: integer;
  s: TStringList;
begin
  if DlgQ(sDelRecordQ) <> mrYes then
    Exit;
  s:=TStringList.Create;
  try
    r:=sgFilter.Row;
    s.Assign(sgFilter.Cols[1]);
    s.Delete(r);
    s.Add('');
    sgFilter.Cols[1].Assign(s);
    s.Assign(sgFilter.Cols[2]);
    s.Delete(r);
    s.Add('');
    sgFilter.Cols[2].Assign(s);
    if sgFilter.RowCount > 2 then
      sgFilter.RowCount:=sgFilter.RowCount - 1;
    if r >= sgFilter.RowCount then
      Dec(r);
    sgFilter.Row:=r;
  finally
    s.Free;
  end;
  UpdateGrid;
end;

end.
