unit rDBGridCellParams;

{<
  rDBGridCellParams provides universal dialog for editing of
  conditions for cell highlighting regarding data in records.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jun 2012)
  @lastmod(March 2018)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, rStringGridEd, PropSaveMain, PropSaveGrids, rDBGrid,
  StdCtrls, ImgList, rImprovedComps, ExtCtrls, DB, rdbdlg_baseform, rHTMLLabel;

{ Show DBGridCellParams dialog for grid }
function DBGridCellParams(rDBGrid: TrDBGrid): boolean;

type
  { provides universal dialog for editing of
  conditions for cell highlighting regarding data in records }
  TrDBGridCellParamsForm = class(TrDBDlgBaseForm)
    grCond: TrStringGridEd;
    GridsPropSave: TGridsPropSave;
    FldNames: TrStringsHolder;
    pButtons: TPanel;
    btUp: TrButton;
    btDown: TrButton;
    btDel: TrButton;
    btDelAll: TrButton;
    lbl2: TLabel;
    pRight: TPanel;
    btOk: TrButton;
    btCancel: TrButton;
    lbl3: TLabel;
    btUse: TrButton;
    FldLabels: TrStringsHolder;
    str: TrStringsHolder;
    pTop: TPanel;
    lbl1: TLabel;
    lblTextHint: TrHTMLLabel;
    procedure grCondFinishCellEditing(Sender: TObject; Col, Row: Integer);
    procedure grCondCanEdit(Sender: TObject; Col, Row: Integer; var CanEdit: boolean);
    procedure grCondGetCellParams(Sender: TObject; Col, Row: Integer; AFont: TFont; var Background: TColor; var Highlight: boolean);
    procedure FormShow(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btUseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btDelClick(Sender: TObject);
    procedure btDelAllClick(Sender: TObject);
    procedure btUpClick(Sender: TObject);
    procedure btDownClick(Sender: TObject);
    procedure lbValuesClick(Sender: TObject);
    procedure grCondEllipsisClicked(Sender: TObject; Col, Row: Integer);
  private
    rDBGrid: TrDBGrid;
    GridID: string;
    L: TStringList;
    FldCond: TField;
    procedure LoadCond;
    procedure StoreCond;
    function ActionIsSet(R: Integer): boolean;
    procedure ScanCallBack(Dataset: TDataSet; RecIndex: Integer; var Abort: boolean);
  public
  end;

var
  rDBGridCellParamsForm: TrDBGridCellParamsForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rstring, rtool, rdbtool, rdlg, rpictool,
  rDBGridCellParamsValues;

const
  cField = 1;
  cOperator = 2;
  cValue = 3;
  cAction = 4;
  cFontColor = 5;
  cBkColor = 6;
  cBold = 7;
  cWholeRow = 8;
  cGlyphIdx = 9;
  cOwnText = 10;
  cStopIfTrue = 11;
  cDeactive = 12;

  cCellParamsCondOperators: array [TCellParamsCondOperators] of string = ('=', '<', '<=', '>', '>=', '<>', sCellParamsCondStart, sCellParamsCondNotStart,
    sCellParamsCondContain, sCellParamsCondNotContain, sCellParamsCondAllways);
  cCellParamsCondActions: array [TCellParamsCondActions] of string = (sCellParamsActSet, sCellParamsActAND, sCellParamsActOR);
  cCellParamsCondBool: array [boolean] of string = ('0', '1');

function DBGridCellParams(rDBGrid: TrDBGrid): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(rDBGrid) then
    Exit;
  rDBGridCellParamsForm:=TrDBGridCellParamsForm.Create(GetActiveOwner);
  try
    rDBGridCellParamsForm.rDBGrid:=rDBGrid;
    rDBGridCellParamsForm.GridID:=rDBGridCellParamsForm.rDBGrid.Owner.Name + '.' + rDBGridCellParamsForm.rDBGrid.Name;
    Result:=rDBGridCellParamsForm.ShowModal = mrOk;
  finally
    FreeF(rDBGridCellParamsForm);
  end;
end;

procedure TrDBGridCellParamsForm.FormShow(Sender: TObject);
var
  i: Integer;
begin
  inherited;
  rDlgPropSave.ReadProperties;
  grCond.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  // load field list
  rdbtool.GetFldList(rDBGrid, FldNames.Strings, [fdSimpleField, fdMemoField]);
  rdbtool.GetFldList(rDBGrid, FldLabels.Strings, [fdSimpleField, fdMemoField, fdGetLabels]);
  grCond.Columns[cField].PickList.Assign(FldLabels.Strings);
  // load picklist
  str.Strings.Clear;
  for i:=0 to ord( High(TCellParamsCondOperators)) do
    str.Strings.Add(cCellParamsCondOperators[TCellParamsCondOperators(i)]);
  grCond.Columns[cOperator].PickList.Assign(str.Strings);
  str.Strings.Clear;
  for i:=0 to ord( High(TCellParamsCondActions)) do
    str.Strings.Add(cCellParamsCondActions[TCellParamsCondActions(i)]);
  grCond.Columns[cAction].PickList.Assign(str.Strings);
  grCond.Columns[cGlyphIdx].DataImageList:=rDBGrid.DataImageList;
  // load current conditions
  LoadCond;
end;

procedure TrDBGridCellParamsForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  rDlgPropSave.SaveProperties;
  inherited;
end;

procedure TrDBGridCellParamsForm.LoadCond;
var
  i, idx: Integer;
  cp: TCellParamsCond;
begin
  grCond.RowCount:=2;
  grCond.RowCount:=rDBGrid.CellParamsCondList.Count + 2;
  for i:=0 to rDBGrid.CellParamsCondList.Count - 1 do
  begin
    cp:=rDBGrid.CellParamsCondList.Items[i];
    idx:=FldNames.Strings.IndexOf(cp.FieldName);
    if (idx >= 0) and (idx < FldLabels.Strings.Count) then
      grCond.Cells[cField, i + 1]:=FldLabels.Strings[idx]
    else
      grCond.Cells[cField, i + 1]:='';
    grCond.Cells[cOperator, i + 1]:=cCellParamsCondOperators[cp.CondOperator];
    grCond.Cells[cValue, i + 1]:=cp.Value;
    grCond.Cells[cAction, i + 1]:=cCellParamsCondActions[cp.Action];
    if cp.FontColor <> clNone then
      grCond.Cells[cFontColor, i + 1]:=rpictool.ColorToString(cp.FontColor, true);
    if cp.BkColor <> clNone then
      grCond.Cells[cBkColor, i + 1]:=rpictool.ColorToString(cp.BkColor, true);
    grCond.Cells[cBold, i + 1]:=cCellParamsCondBool[cp.Bold];
    grCond.Cells[cWholeRow, i + 1]:=cCellParamsCondBool[cp.WholeRow];
    // ImageDef
    if cp.ImageDef.UseDataImage then
    begin
      if cp.ImageDef.ImageIndex >= 0 then
        grCond.IntCells[cGlyphIdx, i + 1]:=cp.ImageDef.ImageIndex;
      grCond.Cells[cOwnText, i + 1]:=cp.ImageDef.ShowText;
    end;

    grCond.Cells[cStopIfTrue, i + 1]:=cCellParamsCondBool[cp.StopIfTrue];
    grCond.Cells[cDeactive, i + 1]:=cCellParamsCondBool[cp.Deactive];
  end;
end;

procedure TrDBGridCellParamsForm.StoreCond;
var
  i, j, idx: Integer;
  cp: TCellParamsCond;
begin
  rDBGrid.CellParamsCondList.Clear;
  for i:=1 to grCond.RowCount - 2 do
  begin
    cp:=TCellParamsCond.Create;
    try
      idx:=FldLabels.Strings.IndexOf(grCond.Cells[cField, i]);
      if (idx >= 0) and (idx < FldNames.Strings.Count) then
        cp.FieldName:=FldNames.Strings[idx]
      else
        cp.FieldName:='';
      idx:=0;
      for j:=0 to ord( High(TCellParamsCondOperators)) do
        if SameText(cCellParamsCondOperators[TCellParamsCondOperators(j)], grCond.Cells[cOperator, i]) then
          idx:=j;
      cp.CondOperator:=TCellParamsCondOperators(idx);
      cp.Value:=grCond.Cells[cValue, i];
      idx:=0;
      for j:=0 to ord( High(TCellParamsCondActions)) do
        if SameText(cCellParamsCondActions[TCellParamsCondActions(j)], grCond.Cells[cAction, i]) then
          idx:=j;
      cp.Action:=TCellParamsCondActions(idx);
      cp.FontColor:=rpictool.StringToColor(grCond.Cells[cFontColor, i]);
      cp.BkColor:=rpictool.StringToColor(grCond.Cells[cBkColor, i]);
      cp.Bold:=grCond.Cells[cBold, i] = '1';
      cp.WholeRow:=grCond.Cells[cWholeRow, i] = '1';
      // ImageDef
      cp.ImageDef.UseDataImage:=(grCond.Cells[cGlyphIdx, i] <> '') or (grCond.Cells[cOwnText, i] <> '');
      cp.ImageDef.ImageIndex:=StrToIntDef(grCond.Cells[cGlyphIdx, i], -1);
      cp.ImageDef.ShowText:=grCond.Cells[cOwnText, i];
      if cp.ImageDef.ShowText <> '' then
        cp.ImageDef.ImageAlignment:='A';

      cp.StopIfTrue:=grCond.Cells[cStopIfTrue, i] = '1';
      cp.Deactive:=grCond.Cells[cDeactive, i] = '1';
      rDBGrid.CellParamsCondList.Add(cp);
    except
      on E: exception do
      begin
        cp.Free;
        raise ;
      end;
    end;
  end;
  rDBGrid.OptionsEx2.CellCondition.Active:=true;
  rDBGrid.RefreshCellParamsCond;
end;

function TrDBGridCellParamsForm.ActionIsSet(R: Integer): boolean;
begin
  try
    Result:=(grCond.Columns[cAction].PickList.Count > 0) and (grCond.Cells[cAction, R] = grCond.Columns[cAction].PickList[0]);
  except
    Result:=false;
  end;
end;

procedure TrDBGridCellParamsForm.grCondCanEdit(Sender: TObject; Col, Row: Integer; var CanEdit: boolean);
begin
  if not ActionIsSet(Row) then
  begin
    if Col > cAction then
      CanEdit:=false;
  end;
end;

procedure TrDBGridCellParamsForm.grCondFinishCellEditing(Sender: TObject; Col, Row: Integer);
begin
  if (Col = cField) and (grCond.Cells[cField, Row] <> '') then
  begin
    if grCond.Cells[cOperator, Row] = '' then
      grCond.Cells[cOperator, Row]:='=';
    if grCond.Cells[cAction, Row] = '' then
      grCond.Cells[cAction, Row]:=grCond.Columns[cAction].PickList[0];
  end;
  if (Col = cAction) and not ActionIsSet(Row) then
  begin
    grCond.Cells[cFontColor, Row]:='';
    grCond.Cells[cBkColor, Row]:='';
    grCond.Cells[cBkColor, Row]:='';
    grCond.Cells[cBold, Row]:='';
    grCond.Cells[cWholeRow, Row]:='';
    grCond.Cells[cStopIfTrue, Row]:='';
  end;
  if (Col = cAction) then
    grCond.Invalidate;
end;

procedure TrDBGridCellParamsForm.grCondGetCellParams(Sender: TObject; Col, Row: Integer; AFont: TFont; var Background: TColor; var Highlight: boolean);
begin
  if not ActionIsSet(Row) then
  begin
    if Col > cAction then
      Background:=clBtnFace;
  end;
end;

procedure TrDBGridCellParamsForm.lbValuesClick(Sender: TObject);
begin
  grCond.SetFocus;
end;

procedure TrDBGridCellParamsForm.btDelAllClick(Sender: TObject);
var
  i: Integer;
begin
  str.Strings.Clear;
  for i:=0 to grCond.RowCount - 1 do
    str.Strings.Add('');
  for i:=0 to grCond.ColCount - 1 do
  begin
    grCond.Cols[i].Assign(str.Strings);
  end;
  grCond.RowCount:=2;
end;

procedure TrDBGridCellParamsForm.btDelClick(Sender: TObject);
var
  i: Integer;
begin
  if grCond.Row = grCond.RowCount - 1 then
    Exit;
  for i:=0 to grCond.ColCount - 1 do
  begin
    str.Strings.Assign(grCond.Cols[i]);
    str.Strings.Delete(grCond.Row);
    str.Strings.Add('');
    grCond.Cols[i].Assign(str.Strings);
  end;
  if grCond.RowCount > 2 then
    grCond.RowCount:=grCond.RowCount - 1;
end;

procedure TrDBGridCellParamsForm.btOkClick(Sender: TObject);
begin
  StoreCond;
  Close;
end;

procedure TrDBGridCellParamsForm.btUpClick(Sender: TObject);
var
  i: Integer;
begin
  if grCond.Row < 2 then
    Exit;
  if grCond.Row < grCond.RowCount - 1 then
    for i:=0 to grCond.ColCount - 1 do
    begin
      str.Strings.Assign(grCond.Cols[i]);
      str.Strings.Move(grCond.Row, grCond.Row - 1);
      grCond.Cols[i].Assign(str.Strings);
    end;
  grCond.Row:=grCond.Row - 1;
end;

procedure TrDBGridCellParamsForm.btDownClick(Sender: TObject);
var
  i: Integer;
begin
  if grCond.Row >= grCond.RowCount - 2 then
    Exit;
  for i:=0 to grCond.ColCount - 1 do
  begin
    str.Strings.Assign(grCond.Cols[i]);
    str.Strings.Move(grCond.Row, grCond.Row + 1);
    grCond.Cols[i].Assign(str.Strings);
  end;
  grCond.Row:=grCond.Row + 1;
end;

procedure TrDBGridCellParamsForm.btUseClick(Sender: TObject);
begin
  StoreCond;
  LoadCond;
end;

procedure TrDBGridCellParamsForm.ScanCallBack(Dataset: TDataSet; RecIndex: Integer; var Abort: boolean);
begin
  if L.IndexOf(FldCond.AsString) < 0 then
    L.Add(FldCond.AsString);
end;

procedure TrDBGridCellParamsForm.grCondEllipsisClicked(Sender: TObject; Col, Row: Integer);
var
  s, fld: string;
  idx: Integer;
begin
  idx:=FldLabels.Strings.IndexOf(grCond.Cells[cField, Row]);
  if idx < 0 then
    Exit;
  FldCond:=rDBGrid.DataSource.Dataset.FindField(FldNames.Strings[idx]);
  if not Assigned(FldCond) then
    Exit;
  rDBGridCellParamsValuesForm:=TrDBGridCellParamsValuesForm.Create(GetActiveOwner);
  try
    rDBGridCellParamsValuesForm.cond:=grCond.Cells[cField, Row] + ' ' + grCond.Cells[cOperator, Row];
    // load field values
    if FldCond is TBooleanField then
    begin
      rDBGridCellParamsValuesForm.cbFix.Items.Add(sYes);
      rDBGridCellParamsValuesForm.cbFix.Items.Add(sNo);
    end
    else
    begin
      L:=TStringList.Create;
      try
        rdbtool.ScanAllRecords(FldCond.Dataset, Self.ScanCallBack);
        // sort items
        if FldCond is TDateTimeField then
          L.CustomSort(SortDate)
        else
          if FldCond is TNumericField then
            L.CustomSort(SortNum)
          else
            L.Sort;
        rDBGridCellParamsValuesForm.cbFix.Items.Assign(L);
      finally
        L.Free;
      end;
    end;
    // assign field list and set enability
    rDBGridCellParamsValuesForm.cbField.Items.Assign(FldNames.Strings);
    rDBGridCellParamsValuesForm.rbDate.Enabled:=FldCond is TDateTimeField;
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_Sum);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_Min);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_Max);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_AVG);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_CountVal);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_CountNull);
    rDBGridCellParamsValuesForm.cbSumType.Items.Add(cSummaryTag_CountAll);
    // select current state
    s:=grCond.Cells[cValue, Row];
    if copy(s, 1, 1) <> '=' then
    begin
      rDBGridCellParamsValuesForm.cbFix.Text:=s
    end
    else
    begin
      s:=copy(s, 2, 255);
      if s = rActualDateAbr then
      begin
        rDBGridCellParamsValuesForm.rbDate.Checked;
        rDBGridCellParamsValuesForm.chDate.Checked:=true;
      end
      else
      begin
        rDBGridCellParamsValuesForm.rbField.Checked:=true;
        fld:=rtool.FindStrFromTo('(', ')', s, false, true);
        if fld='' then
          rDBGridCellParamsValuesForm.cbField.Text:=s
        else
        begin
          rDBGridCellParamsValuesForm.cbField.Text:=fld;
          rDBGridCellParamsValuesForm.cbSumType.Text:=rtool.FindStrPart(s, '(', 1);
        end;
      end;
    end;

    if rDBGridCellParamsValuesForm.ShowModal = mrOk then
    begin
      if rDBGridCellParamsValuesForm.rbFix.Checked then
        s:=rDBGridCellParamsValuesForm.cbFix.Text;
      if rDBGridCellParamsValuesForm.rbDate.Checked then
        if rDBGridCellParamsValuesForm.chDate.Checked then
          s:='=' + rActualDateAbr
        else
          s:=DateToStr(rDBGridCellParamsValuesForm.eDate.Date);
      if rDBGridCellParamsValuesForm.rbField.Checked then
      begin
        s:=rDBGridCellParamsValuesForm.cbField.Text;
        if rDBGridCellParamsValuesForm.cbSumType.Text<>'' then
          s:=rDBGridCellParamsValuesForm.cbSumType.Text+'('+s+')';
        s:='=' + s;
      end;
      grCond.Cells[cValue, Row]:=s;
    end;
  finally
    FreeF(rDBGridCellParamsValuesForm);
  end;
end;

end.
