unit rDBGridProps;

{<
  rDBGridProps provides universal dialog for editing of
  rDBGrid apperance and properties.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Feb 2003)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ExtCtrls, Grids, DBGrids, rDBGrid,
  ImgList, rImprovedComps, rHTMLLabel, PropSaveMain, rdbdlg_baseform;

{ Show DBGridProps dialog for rDBGrid }
procedure DBGridProperties(rDBGrid: TrDBGrid; SetColumns: boolean = true; SetCellCond: boolean = true; SetFilterButtons: boolean = true;
  SetGridStyle: boolean = true);

type
  { universal dialog for editing of rDBGrid apperance properties }
  TrDBGridPropsForm = class(TrDBDlgBaseForm)
    gbProps: TGroupBox;
    chRowLines: TCheckBox;
    chColoredOdd: TCheckBox;
    cbColorOdd: TrColorBox;
    Label1: TLabel;
    eTitleLines: TrEditNum;
    gbColumns: TGroupBox;
    btColumn: TrButton;
    btWidthDef: TrButton;
    Label2: TLabel;
    btWidthData: TrButton;
    cbAutoWidth: TCheckBox;
    cbSameWidth: TCheckBox;
    btClose: TrButton;
    btFontTitle: TrButton;
    FontDlg: TFontDialog;
    btFont: TrButton;
    btDefault: TrURLLabel;
    chGradient: TCheckBox;
    chColorActiveRow: TCheckBox;
    cbColorActiveRow: TrColorBox;
    pGradFrom: TPanel;
    lblFromTo: TLabel;
    pGradTo: TPanel;
    ColorDlg: TColorDialog;
    chFontFiltered: TCheckBox;
    cbColorFiltered: TrColorBox;
    chColorRec: TCheckBox;
    btColorRec: TrURLLabel;
    chTitleFilterBtn: TCheckBox;
    chGroupFields: TCheckBox;
    lblRowLines: TLabel;
    eRowLines: TrEditNum;
    btGroupSettings: TrURLLabel;
    lblStyle: TLabel;
    cbStyle: TComboBox;
    chTitleFilterBar: TCheckBox;
    cbColorDataLine: TrColorBox;
    chRowLineDot: TCheckBox;
    Label3: TLabel;
    eFixedCols: TrEditNum;
    Label4: TLabel;
    cbColorFixedCol: TrColorBox;
    Label5: TLabel;
    procedure FormShow(Sender: TObject);
    procedure chColoredOddClick(Sender: TObject);
    procedure btColumnClick(Sender: TObject);
    procedure btWidthDefClick(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
    procedure chRowLinesClick(Sender: TObject);
    procedure eTitleLinesChange(Sender: TObject);
    procedure cbAutoWidthClick(Sender: TObject);
    procedure btWidthDataClick(Sender: TObject);
    procedure cbSameWidthClick(Sender: TObject);
    procedure btFontTitleClick(Sender: TObject);
    procedure btFontClick(Sender: TObject);
    procedure btDefaultClick(Sender: TObject);
    procedure chColorActiveRowClick(Sender: TObject);
    procedure chGradientClick(Sender: TObject);
    procedure pGradFromClick(Sender: TObject);
    procedure pGradToClick(Sender: TObject);
    procedure chFontFilteredClick(Sender: TObject);
    procedure chColorRecClick(Sender: TObject);
    procedure btColorRecClick(Sender: TObject);
    procedure chTitleFilterBtnClick(Sender: TObject);
    procedure chGroupFieldsClick(Sender: TObject);
    procedure eRowLinesChange(Sender: TObject);
    procedure btGroupSettingsClick(Sender: TObject);
    procedure cbStyleChange(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure chTitleFilterBarClick(Sender: TObject);
    procedure chRowLineDotClick(Sender: TObject);
    procedure eFixedColsChange(Sender: TObject);
    procedure cbColorFixedColChange(Sender: TObject);
  private
    Gr: TrDBGrid;
    procedure LoadColors;
    procedure SetAll;
  public
  end;

var
  rDBGridPropsForm: TrDBGridPropsForm;

implementation

{$R *.dfm}

uses RosiCompGlobalCfg, rstring, rtool, rdbtool, rdlg, rDBFields, rDBGridCellParams, rDBFieldsGroup;

const
  stdLineColor = $F0F0F0;

var
  isLoading: boolean;

procedure DBGridProperties(rDBGrid: TrDBGrid; SetColumns: boolean = true; SetCellCond: boolean = true; SetFilterButtons: boolean = true;
  SetGridStyle: boolean = true);
begin
  isLoading:=true;
  rDBGridPropsForm:=TrDBGridPropsForm.Create(GetActiveOwner);
  try
    rDBGridPropsForm.Gr:=rDBGrid;
    rDBGridPropsForm.gbColumns.Visible:=SetColumns;
    rDBGridPropsForm.chColorRec.Visible:=not RosiCompConfig.rDBGridPropsCfg.HideCellParams and SetCellCond;
    rDBGridPropsForm.btColorRec.Visible:=rDBGridPropsForm.chColorRec.Visible;
    rDBGridPropsForm.chTitleFilterBtn.Visible:=not RosiCompConfig.rDBGridPropsCfg.HideTitleFilterBtn and SetFilterButtons;
    rDBGridPropsForm.chTitleFilterBar.Visible:=not RosiCompConfig.rDBGridPropsCfg.HideTitleFilterText and SetFilterButtons;
    rDBGridPropsForm.chGroupFields.Visible:=RosiCompConfig.GetConfig_OnOffAuto(RosiCompConfig.rDBGridPropsCfg.HideTitleGroups,rDBGrid.FieldsGroups.Count > 0);
    rDBGridPropsForm.btGroupSettings.Visible:=rDBGridPropsForm.chGroupFields.Visible;
    rDBGridPropsForm.eRowLines.Visible:=rDBGrid.OptionsEx2.Editors.ShowMemoContent or rDBGrid.OptionsEx2.Editors.ShowGraphicContent;
    rDBGridPropsForm.lblRowLines.Visible:=rDBGridPropsForm.eRowLines.Visible;
    rDBGridPropsForm.cbStyle.Visible:=not RosiCompConfig.rDBGridPropsCfg.HideGridStyles and SetGridStyle;
    rDBGridPropsForm.lblStyle.Visible:=rDBGridPropsForm.cbStyle.Visible;
    rDBGridPropsForm.ShowModal;
  finally
    FreeF(rDBGridPropsForm);
  end;
end;

procedure TrDBGridPropsForm.FormResize(Sender: TObject);
var
  x, w: integer;
begin
  inherited;
  x:=pGradFrom.Left + pGradFrom.Width;
  w:=pGradTo.Left - x;
  lblFromTo.Left:=x + (w - lblFromTo.Width) div 2;
end;

procedure TrDBGridPropsForm.FormShow(Sender: TObject);
begin
  inherited;
  LoadGridLookStyleList(cbStyle.Items);
  cbStyle.ItemIndex:=ord(Gr.GridLookStyle);
  chRowLines.Checked:=(dgRowLines in Gr.Options);
  chRowLineDot.Checked:=Gr.OptionsEx2.Appearance.DataRowLineStyle = psDot;
  chColorRec.Checked:=Gr.OptionsEx2.CellCondition.Active;
  eRowLines.AsInteger:=Gr.RowLines;
  eTitleLines.AsInteger:=Gr.TitleLines;
  eFixedCols.AsInteger:=Gr.OptionsEx2.FixedCell.FixedCols;
  chGroupFields.Checked:=Gr.OptionsEx2.Groups.Active;
  LoadColors;
  SetAll;
  isLoading:=false;
end;

procedure TrDBGridPropsForm.LoadColors;
begin
  isLoading:=true;
  if Gr.OptionsEx2.Appearance.DataRowLineColor = clNone then
    cbColorDataLine.Selected:=stdLineColor
  else
    cbColorDataLine.Selected:=Gr.OptionsEx2.Appearance.DataRowLineColor;
  chColorActiveRow.Checked:=Gr.ColorActiveRow <> clNone;
  cbColorActiveRow.Selected:=Gr.ColorActiveRow;
  chColoredOdd.Checked:=Gr.OptionsEx2.Appearance.DrawColoredRow;
  cbColorOdd.Selected:=Gr.ColorOddRow;
  chFontFiltered.Checked:=Gr.FontColorFiltered <> clNone;
  cbColorFiltered.Selected:=Gr.FontColorFiltered;
  chGradient.Checked:=Gr.DrawingStyle = gdsGradient;
  pGradFrom.Color:=Gr.GradientStartColor;
  pGradTo.Color:=Gr.GradientEndColor;
  chTitleFilterBtn.Checked:=Gr.OptionsEx2.Filters.FunnelBtn;
  chTitleFilterBar.Checked:=Gr.OptionsEx2.Filters.TextBar;
  cbColorFixedCol.Selected:=Gr.OptionsEx2.FixedCell.FixedColColor;
  isLoading:=false;
end;

procedure TrDBGridPropsForm.SetAll;
begin
  isLoading:=true;
  cbColorActiveRow.Enabled:=chColorActiveRow.Checked;
  cbColorOdd.Enabled:=chColoredOdd.Checked;
  cbAutoWidth.Checked:=Gr.ColumnWidth = cwAutoWidth;
  cbSameWidth.Checked:=Gr.ColumnWidth = cwSameWidth;
  isLoading:=false;
end;

procedure TrDBGridPropsForm.chRowLinesClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  if chRowLines.Checked then
    Gr.Options:=Gr.Options + [dgRowLines]
  else
    Gr.Options:=Gr.Options - [dgRowLines];
  Gr.Invalidate;
end;

procedure TrDBGridPropsForm.chRowLineDotClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  if chRowLineDot.Checked then
    Gr.OptionsEx2.Appearance.DataRowLineStyle:=psDot
  else
    Gr.OptionsEx2.Appearance.DataRowLineStyle:=psSolid;

  if (Gr.OptionsEx2.Appearance.DataRowLineStyle = psSolid) and (Gr.OptionsEx2.Appearance.DataRowLineColor = stdLineColor) then
    Gr.OptionsEx2.Appearance.DataRowLineColor:=clNone
  else
    Gr.OptionsEx2.Appearance.DataRowLineColor:=cbColorDataLine.Selected;
end;

procedure TrDBGridPropsForm.chTitleFilterBarClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.Filters.TextBar:=chTitleFilterBar.Checked;
end;

procedure TrDBGridPropsForm.chTitleFilterBtnClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.Filters.FunnelBtn:=chTitleFilterBtn.Checked;
end;

procedure TrDBGridPropsForm.chGradientClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  if chGradient.Checked then
    Gr.DrawingStyle:=gdsGradient
  else
    Gr.DrawingStyle:=gdsThemed;
end;

procedure TrDBGridPropsForm.chGroupFieldsClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.Groups.Active:=chGroupFields.Checked;
end;

procedure TrDBGridPropsForm.chColorActiveRowClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  SetAll;
  if not chColorActiveRow.Checked then
    cbColorActiveRow.Selected:=clNone
  else
    if cbColorActiveRow.Selected = clNone then
      cbColorActiveRow.Selected:=clInfoBk;
  Gr.ColorActiveRow:=cbColorActiveRow.Selected;
end;

procedure TrDBGridPropsForm.chColoredOddClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  SetAll;
  Gr.OptionsEx2.Appearance.DrawColoredRow:=chColoredOdd.Checked;
  Gr.ColorOddRow:=cbColorOdd.Selected;
end;

procedure TrDBGridPropsForm.chColorRecClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.CellCondition.Active:=chColorRec.Checked;
end;

procedure TrDBGridPropsForm.chFontFilteredClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  SetAll;
  if not chFontFiltered.Checked then
    cbColorFiltered.Selected:=clNone
  else
    if cbColorFiltered.Selected = clNone then
      cbColorFiltered.Selected:=clNavy;
  Gr.FontColorFiltered:=cbColorFiltered.Selected;
end;

procedure TrDBGridPropsForm.eFixedColsChange(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.FixedCell.FixedCols:=eFixedCols.AsInteger;
end;

procedure TrDBGridPropsForm.cbColorFixedColChange(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.OptionsEx2.FixedCell.FixedColColor:=cbColorFixedCol.Selected;
end;

procedure TrDBGridPropsForm.eRowLinesChange(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.RowLines:=eRowLines.AsInteger;
end;

procedure TrDBGridPropsForm.eTitleLinesChange(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.TitleLines:=eTitleLines.AsInteger;
end;

procedure TrDBGridPropsForm.cbAutoWidthClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  if cbAutoWidth.Checked then
    Gr.ColumnWidth:=cwAutoWidth
  else
    Gr.ColumnWidth:=cwNone;
  SetAll;
end;

procedure TrDBGridPropsForm.cbSameWidthClick(Sender: TObject);
begin
  if isLoading then
    Exit;
  if cbSameWidth.Checked then
    Gr.ColumnWidth:=cwSameWidth
  else
    Gr.ColumnWidth:=cwNone;
  SetAll;
end;

procedure TrDBGridPropsForm.cbStyleChange(Sender: TObject);
begin
  if isLoading then
    Exit;
  Gr.DrawingStyle:=gdsGradient;
  Gr.GridLookStyle:=TGridLookStyles(cbStyle.ItemIndex);
  LoadColors;
end;

procedure TrDBGridPropsForm.btWidthDataClick(Sender: TObject);
begin
  Gr.ColumnWidth:=cwDataLabelWidth;
  Gr.ColumnWidth:=cwNone;
  SetAll;
end;

procedure TrDBGridPropsForm.btWidthDefClick(Sender: TObject);
begin
  Gr.ColumnWidth:=cwDefaultWidth;
  Gr.ColumnWidth:=cwNone;
  Gr.RestoreColumnsWidth(false, false);
  SetAll;
end;

procedure TrDBGridPropsForm.btColumnClick(Sender: TObject);
begin
  ChooseFields(Gr);
end;

procedure TrDBGridPropsForm.btCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TrDBGridPropsForm.btFontTitleClick(Sender: TObject);
begin
  FontDlg.Font.Assign(Gr.TitleFont);
  if FontDlg.Execute then
    Gr.TitleFont.Assign(FontDlg.Font);
end;

procedure TrDBGridPropsForm.btFontClick(Sender: TObject);
var
  i: integer;
begin
  FontDlg.Font.Assign(Gr.Font);
  if FontDlg.Execute then
  begin
    Gr.Font.Assign(FontDlg.Font);
    for i:=0 to Gr.Columns.Count - 1 do
      Gr.Columns[i].Font.Assign(FontDlg.Font);
  end;
end;

procedure TrDBGridPropsForm.pGradFromClick(Sender: TObject);
begin
  ColorDlg.Color:=pGradFrom.Color;
  if ColorDlg.Execute then
  begin
    pGradFrom.Color:=ColorDlg.Color;
    Gr.GradientStartColor:=ColorDlg.Color;
    // gr.Invalidate;
  end;
end;

procedure TrDBGridPropsForm.pGradToClick(Sender: TObject);
begin
  ColorDlg.Color:=pGradTo.Color;
  if ColorDlg.Execute then
  begin
    pGradTo.Color:=ColorDlg.Color;
    Gr.GradientEndColor:=ColorDlg.Color;
    // gr.Invalidate;
  end;
end;

procedure TrDBGridPropsForm.btColorRecClick(Sender: TObject);
begin
  rDBGridCellParams.DBGridCellParams(Gr);
  chColorRec.Checked:=Gr.OptionsEx2.CellCondition.Active;
end;

procedure TrDBGridPropsForm.btDefaultClick(Sender: TObject);
var
  i: integer;
  Comp: TComponent;
begin
  if (DlgQ(sSetDefaultValues) = mrYes) then
  begin
    // Gr.Columns.RestoreDefaults;
    Gr.RestoreColumnsWidth;
    // find parent form
    Comp:=Gr;
    while (Comp <> nil) and not(Comp is TForm) do
      Comp:=Comp.Owner;
    if Assigned(Comp) then
    begin
      // find all CustomPropSave and add grid to disable component list
      for i:=0 to Comp.ComponentCount - 1 do
        if Comp.Components[i] is TCustomPropSave then (Comp.Components[i] as TCustomPropSave)
          .DisableComponents.Add(Gr);
    end;
    // call global SetDefault event
    if Assigned(RosiCompConfig.rDBFieldsCfg.OnSetDefaultGridColumns) then
      RosiCompConfig.rDBFieldsCfg.OnSetDefaultGridColumns(Gr);
    Close;
  end;
end;

procedure TrDBGridPropsForm.btGroupSettingsClick(Sender: TObject);
begin
  rDBFieldsGroup.SetFieldsGroup(Gr);
  chGroupFields.Checked:=Gr.OptionsEx2.Groups.Active;
end;

end.
