unit rDBImport;

{<
  rDBImport provides universal dialog for import of data from TXT, XLS, DB
  and other sources to any DB table.

  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2006)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}
{$DEFINE UseADO}
{$DEFINE noUseZeos}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, Grids, DBGrids, rDBGrid, Contnrs,
  ExtCtrls, ComCtrls, DB, ADODB, ADOConWizard,
{$IFDEF UseZeos} ZAbstractDataset, ZDataset, {$ENDIF}
  Menus, PropSaveMain, rImprovedComps, rHTMLLabel, ImgList, rdbdlg_baseform;

const
  cImportLogInfo = 1;
  cImportLogProgress = 2;
  cImportLogResult = 5;
  cImportLogTotalResult = 6;
  cImportLogWarning = 9;
  cImportLogError = 10;

type
  { callback event for creating of backup }
  TrDBImport_BackupCallBack = procedure(var Continue: boolean) of object;
  TrDBImport_LogCallBack = procedure(LogStr: string; Level: integer; var Aborted: boolean) of object;

  TrDBImport_Setting = record
    AutoRun_ShowForm: boolean;
    AutoRun_ShowProgress: boolean;
    AutoRun_DoCheck: boolean;
    AutoRun_OwnerForm: TWinControl;
    StandardChars: string;
    LocalChars: string;
    SpecialChars: string;
    MaxImportedRecords: integer;
    CopyNumericFieldAsFloat: boolean;
    ResultInTableForm: boolean;
    ResultValueSeparator: string;
  end;

  { provides universal dialog for import of data from TXT, XLS, DB
    and other sources to any DB table }
  TrDBImportForm = class(TrDBDlgBaseForm)
    ImpCon: TADOConnection;
    dsImp: TDataSource;
    pc: TPageControl;
    ts0: TTabSheet;
    ts1: TTabSheet;
    pBottom: TPanel;
    OpenDlg: TOpenDialog;
    p2Client: TPanel;
    p2Bot: TPanel;
    lblFilter: TrHTMLLabel;
    p2Top: TPanel;
    lbl4: TrHTMLLabel;
    grData: TrDBGrid;
    grImpData2: TrDBGrid;
    dsData: TDataSource;
    pcFields: TPageControl;
    tsField1: TTabSheet;
    tsField2: TTabSheet;
    sb1: TScrollBox;
    lInfo1: TrHTMLLabel;
    rgTyp: TRadioGroup;
    sb2: TScrollBox;
    lInfo2: TrHTMLLabel;
    ts2: TTabSheet;
    btPrev: TrButton;
    btNext: TrButton;
    Panel8: TPanel;
    btClose: TrButton;
    lbl6: TrHTMLLabel;
    btRun: TrButton;
    lbl7: TrHTMLLabel;
    ts3: TTabSheet;
    mErr: TMemo;
    lbl8: TrHTMLLabel;
    Panel5: TPanel;
    btSaveErr: TrButton;
    pm: TPopupMenu;
    mnFind: TMenuItem;
    mnFields: TMenuItem;
    mnGrid: TMenuItem;
    mnFilterSel: TMenuItem;
    mnFilterMan: TMenuItem;
    N1: TMenuItem;
    mnFilterOnOff: TMenuItem;
    Panel1: TPanel;
    Splitter1: TSplitter;
    PSFile: TPropSaveFile;
    PSMain: TPropSaveMain;
    PSComps: TrStringsHolder;
    btSettings: TrHTMLLabel;
    OpenDlgSet: TOpenDialog;
    SaveDlgSet: TSaveDialog;
    LookupSQLStr: TrStringsHolder;
    chBackup: TCheckBox;
    gbPage1: TrGroupBox;
    pcCon: TPageControl;
    tsCon: TTabSheet;
    pPg1Bottom: TPanel;
    btConExp: TrButton;
    mConInfo: TMemo;
    tsData: TTabSheet;
    grImpData1: TrDBGrid;
    btMoveItems: TrHTMLLabel;
    mFields: TrStringsHolder;
    pInfoBefore: TPanel;
    lblRes1: TrHTMLLabel;
    lblRes2: TrHTMLLabel;
    lblRes3: TrHTMLLabel;
    lblRes4: TrHTMLLabel;
    lblRes5: TrHTMLLabel;
    lblRes5b: TrHTMLLabel;
    lblRes4b: TrHTMLLabel;
    lblRes3b: TrHTMLLabel;
    lblRes2b: TrHTMLLabel;
    lblShowErr1: TrURLLabel;
    pInfoPost: TPanel;
    lblRes11: TrHTMLLabel;
    lblRes12: TrHTMLLabel;
    lblRes13: TrHTMLLabel;
    lblRes14: TrHTMLLabel;
    lblRes15: TrHTMLLabel;
    lblRes15b: TrHTMLLabel;
    lblRes14b: TrHTMLLabel;
    lblRes13b: TrHTMLLabel;
    lblRes12b: TrHTMLLabel;
    lblShowErr2: TrURLLabel;
    lblRes16: TrHTMLLabel;
    lblRes16b: TrHTMLLabel;
    lblRes17: TrHTMLLabel;
    lblRes17b: TrHTMLLabel;
    gbTop1: TrGroupBox;
    lblOpen1: TrHTMLLabel;
    lblOpen2: TrHTMLLabel;
    lbl1: TrHTMLLabel;
    lFile: TrHTMLLabel;
    lbl2: TrHTMLLabel;
    cbTable: TComboBox;
    btNewCon: TrButton;
    btSelFile: TrButton;
    gbTop2: TrGroupBox;
    lbl3: TrHTMLLabel;
    btSaveAutoDesign: TrButton;
    qImp: TADOQuery;
    lblSelect: TrHTMLLabel;
    eSelectCmd: TEdit;
    lblLoadData: TrURLLabel;
    lblExpStructure: TrHTMLLabel;
    pcValues: TPageControl;
    tsValues: TTabSheet;
    v: TTabSheet;
    lbl5: TrHTMLLabel;
    mVal: TMemo;
    chUseValAddOnly: TCheckBox;
    lblCalc: TrHTMLLabel;
    mCalc: TMemo;
    mCalcHelp: TrStringsHolder;
    mWarn: TrStringsHolder;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ImpConWizInfo(Sender: TObject; Info: String);
    procedure cbTableChange(Sender: TObject);
    procedure grImpData1ChangeSort(Sender: TObject; DataSet: TDataSet; SortFieldName: String; SortDesc: boolean; var Accept: boolean);
    procedure grImpData1RefreshData(Sender: TObject);
    procedure btSelFileClick(Sender: TObject);
    procedure rgTypClick(Sender: TObject);
    procedure lInfoBeforeGetParameter(Sender: TObject; var ParamStr: String);
    procedure btCloseClick(Sender: TObject);
    procedure btPrevClick(Sender: TObject);
    procedure btNextClick(Sender: TObject);
    procedure btRunClick(Sender: TObject);
    procedure lInfoPostGetParameter(Sender: TObject; var ParamStr: String);
    procedure btNewConClick(Sender: TObject);
    procedure btConExpClick(Sender: TObject);
    procedure btSaveErrClick(Sender: TObject);
    procedure mnFindClick(Sender: TObject);
    procedure grImpData1Enter(Sender: TObject);
    procedure mnFieldsClick(Sender: TObject);
    procedure mnGridClick(Sender: TObject);
    procedure mnFilterSelClick(Sender: TObject);
    procedure mnFilterManClick(Sender: TObject);
    procedure mnFilterOnOffClick(Sender: TObject);
    procedure grImpData1TitleExClick(Sender: TObject);
    procedure OnAbort(Sender: TObject);
    procedure btSettingsAnchorClick(Sender: TObject; LinkName: String);
    procedure lFileSel2AnchorClick(Sender: TObject; LinkName: String);
    procedure lblFilterAnchorClick(Sender: TObject; LinkName: String);
    procedure gbPage1DropFiles(Sender: TObject; X, Y: integer);
    procedure PSMainSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure PSMainReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure btMoveItemsAnchorClick(Sender: TObject; LinkName: string);
    procedure lFileSel1AnchorClick(Sender: TObject; LinkName: string);
    procedure lblShowErr1Click(Sender: TObject);
    procedure lblExpStructureAnchorClick(Sender: TObject; LinkName: string);
    procedure btSaveAutoDesignClick(Sender: TObject);
    procedure lblLoadDataClick(Sender: TObject);
    procedure eSelectCmdChange(Sender: TObject);
    procedure lblCalcAnchorClick(Sender: TObject; LinkName: string);
  private
    ImpConWiz: TADOConWizard;
    ImpDS: TDataSet;
    AutoMode: integer;
    FldCount: integer;
    sum, add, edit, skip: integer;
    sum2, add2, edit2, skip2, err2, war2: integer;
    actGrid: TrDBGrid;
    DoInit, WasAborted: boolean;
    LookupDSList: TObjectList;
    LookupDS: TDataSet;
    AutoSetting: TStrings;
    lastLogBlock: string;
    lastLogBlockLevel: integer;
    BackupCallBack: TrDBImport_BackupCallBack;
    LogCallBack: TrDBImport_LogCallBack;
    CalcList: TObjectList;
    function GetCBName(Owner: TWinControl; Idx: integer): string;
    procedure Discon;
    function ConnectFile(FileName: string): boolean;
    function CreateConnectString(ConStr: string): boolean;
    function GetSelectCmd: string;
    procedure OpenData;
    procedure LoadLookupSQL;
    procedure SetLookupDS(Idx: integer);
    procedure LoadTables(const FileName: string; SetSelect: boolean);
    procedure LoadSel;
    procedure LoadCheck;
    function InitCalcList(Check: boolean): boolean;
    function UseCalc(Fld: TField; val: string): boolean;
    function FindDataRec: boolean;
    procedure DoAutoImport(SourceDataSet: TDataSet);
    procedure DefLogBlock(const LogBlock: string; Level: integer = 0);
    procedure AddLog(const LogStr: string; Level: integer = 0);
    procedure AddLogTime(const LogStr: string; Level: integer = 0);
    procedure AddWarn(const LogStr: string);
    function FormatFieldForLog(Fld: TField): string;
    function FieldTextForLog(Fld: TField): string;
    procedure ProcProgress(const msg: string; proc: double = -1);
    procedure LoadAutoSetting;
    procedure SaveAutoSetting;
  public
    Imported: boolean;
  end;

  { run manual import to TargetDataSet, source is selected by dialog }
function DBImport(TargetDataSet: TDataSet; LookupSQL: TStrings; UDLPath: string; PSFile: TPropSaveFile; BackupCallBack: TrDBImport_BackupCallBack): boolean;
{ run or prepare manual import to TargetDataSet with possibility to store the AutoSetting }
function DBImportAutoDesign(TargetDataSet: TDataSet; LookupSQL, AutoSetting: TStrings; UDLPath: string; PSFile: TPropSaveFile;
  BackupCallBack: TrDBImport_BackupCallBack): boolean;
{ run automatic import to TargetDataSet driven by AutoSetting, source can be defined by SourceDataSet or by AutoSetting }
function DBImportAutoRun(TargetDataSet, SourceDataSet: TDataSet; LookupSQL, AutoSetting: TStrings; LogCallBack: TrDBImport_LogCallBack): boolean;

var
  rDBImportForm: TrDBImportForm;
  rDBImport_Setting: TrDBImport_Setting;

implementation

{$R *.dfm}

uses RosiCompGlobalCfg, rstring, rdlg, rtool, rdbtool, rdbtool_ADO,
  ADOConEd, rDBImport_strings, rDBFind, rDBFilter, rDBFields, rDBGridProps, rDBExport;

const
  cManual = 0;
  cAutoDesign = 1;
  cAutoRun = 2;

  cSourceSection = 'SourceData';
  cSourceConString = 'ConnectionString';
  cSourceTable = 'TableName';
  cSourceSelectCmd = 'SelectCmd';

  cCalcTrim = 1;
  cCalcClean = 2;
  cCalcRemoveChar = 3;
  cCalcLeft = 11;
  cCalcRight = 12;
  cCalcMiddle = 13;
  cCalcPart = 15;
  cCalcFromTo = 16;
  cCalcReplace = 21;

type
  TCalc = class(TObject)
  private
    FieldName: string;
    Field: TField;
    CalcType: integer;
    Params: string;
    Param1: string;
    Param2: string;
    AllowedChars: string;
    ExcludedChars: string;
  public
    constructor Create;
  end;

var
  fn, v1: array of variant;
  v2: variant;

function DBImport_Int(AutoMode: integer; TargetDataSet, SourceDataSet: TDataSet; LookupSQL, AutoSetting: TStrings; UDLPath: string; PSFile: TPropSaveFile;
  BackupCallBack: TrDBImport_BackupCallBack; LogCallBack: TrDBImport_LogCallBack): boolean;
begin
  Result:=false;
  if not(Assigned(TargetDataSet) and TargetDataSet.Active) then
    Exit;
  rDBImportForm:=TrDBImportForm.Create(GetActiveOwner);
  try
    rDBImportForm.AutoMode:=AutoMode;
    rDBImportForm.rDlgPropSave.PropSaveFile:=PSFile;
    if Assigned(LookupSQL) then
      rDBImportForm.LookupSQLStr.Strings.Assign(LookupSQL);
    rDBImportForm.AutoSetting:=AutoSetting;
    rDBImportForm.ImpConWiz.UDLFilePath:=SetEndOfPath(UDLPath);
    // target
    rDBImportForm.dsData.DataSet:=TargetDataSet;
    // backup
    rDBImportForm.BackupCallBack:=BackupCallBack;
    rDBImportForm.chBackup.Enabled:=Assigned(BackupCallBack);
    rDBImportForm.chBackup.Checked:=rDBImportForm.chBackup.Enabled;
    // log
    rDBImportForm.LogCallBack:=LogCallBack;
    // autosetting
    rDBImportForm.btSaveAutoDesign.Visible:=AutoMode = cAutoDesign;
    rDBImportForm.DoInit:=(AutoMode = cAutoDesign);
    if AutoMode <> cAutoRun then
      rDBImportForm.ShowModal
    else
      rDBImportForm.DoAutoImport(SourceDataSet);
    Result:=rDBImportForm.Imported;
  finally
    FreeF(rDBImportForm);
  end;
end;

function DBImport(TargetDataSet: TDataSet; LookupSQL: TStrings; UDLPath: string; PSFile: TPropSaveFile; BackupCallBack: TrDBImport_BackupCallBack): boolean;
begin
  Result:=DBImport_Int(cManual, TargetDataSet, nil, LookupSQL, nil, UDLPath, PSFile, BackupCallBack, nil);
end;

function DBImportAutoDesign(TargetDataSet: TDataSet; LookupSQL, AutoSetting: TStrings; UDLPath: string; PSFile: TPropSaveFile;
  BackupCallBack: TrDBImport_BackupCallBack): boolean;
begin
  Result:=DBImport_Int(cAutoDesign, TargetDataSet, nil, LookupSQL, AutoSetting, UDLPath, PSFile, BackupCallBack, nil);
end;

function DBImportAutoRun(TargetDataSet, SourceDataSet: TDataSet; LookupSQL, AutoSetting: TStrings; LogCallBack: TrDBImport_LogCallBack): boolean;
begin
  Result:=DBImport_Int(cAutoRun, TargetDataSet, SourceDataSet, LookupSQL, AutoSetting, '', nil, nil, LogCallBack);
end;

procedure TrDBImportForm.FormCreate(Sender: TObject);
begin
  inherited;
  ImpConWiz:=TADOConWizard.Create(Self);
  ImpConWiz.ADOConnection:=ImpCon;
  ImpConWiz.OnInfo:=ImpConWizInfo;
  ImpConWiz.LogDetails:=true;
  ImpConWiz.ErrDialogShow:=[AfterConnecting, AfterLogin, AfterOpening];
  LookupDSList:=TObjectList.Create;
  CalcList:=TObjectList.Create(true);
  ImpDS:=qImp;
  AutoMode:=cManual;
  ts0.TabVisible:=false;
  ts1.TabVisible:=false;
  ts2.TabVisible:=false;
  ts3.TabVisible:=false;
  pc.ActivePageIndex:=0;
  pcCon.ActivePageIndex:=0;
  pcFields.ActivePageIndex:=0;
  ImpConWiz.UDLFilePath:=GetAppDir;
  OpenDlg.InitialDir:=GetAppDir;
  OpenDlg.Filter:=sFileTypeAll + '|' + sFileTypeMDB + '|' + sFileTypeXLS + '|' + sFileTypeDBF;
  Imported:=false;
  WasAborted:=false;
  pInfoPost.Visible:=false;
  lastLogBlock:='';
end;

procedure TrDBImportForm.FormActivate(Sender: TObject);
var
  s: string;
begin
  inherited;
  if not DoInit then
    Exit;
  DoInit:=false;
  if Assigned(AutoSetting) then
  begin
    s:=AutoSetting.Values[cSourceConString];
    eSelectCmd.Text:=AutoSetting.Values[cSourceSelectCmd];
    if s <> '' then
      if CreateConnectString(s) then
      begin
        LoadTables('', false);
        OpenData;
      end;
  end;
end;

procedure TrDBImportForm.FormDestroy(Sender: TObject);
begin
  PSFile.DisableAll:=true;
  ImpConWiz.Free;
  LookupDSList.Clear;
  LookupDSList.Free;
  CalcList.Clear;
  CalcList.Free;
end;

procedure TrDBImportForm.FormResize(Sender: TObject);
begin
  p2Top.Height:=p2Client.Height div 2;
end;

procedure TrDBImportForm.ImpConWizInfo(Sender: TObject; Info: String);
begin
  mConInfo.Lines.add(Info);
end;

procedure TrDBImportForm.Discon;
begin
  pcCon.ActivePageIndex:=0;
  mConInfo.Lines.Clear;
  ImpConWiz.Disconnect;
  ImpCon.ConnectionString:='';
  cbTable.Items.Clear;
end;

function TrDBImportForm.ConnectFile(FileName: string): boolean;
var
  ext: string;
begin
  Discon;
  lFile.Caption:=FileName;
  Update;
  ext:=ExtractFileExt(FileName);
  ext:=copy(ext, 2, 255);
  ImpConWiz.UDLFileMask:='*_' + ext + '.udl';
  Result:=ImpConWiz.ConnectToFile(FileName, true);
end;

function TrDBImportForm.CreateConnectString(ConStr: string): boolean;
begin
  Result:=false;
  Discon;
  if (ConStr <> '') then
    ImpCon.ConnectionString:=ConStr
  else
    if not EditConnectionString(ImpCon) then
      Exit;
  try
    Self.mConInfo.Lines.add(ImpCon.ConnectionString);
    ImpCon.Open;
    Result:=true;
  except
    on E: Exception do
      ShowErr('', E, false);
  end;
end;

procedure TrDBImportForm.LoadTables(const FileName: string; SetSelect: boolean);
var
  i: integer;
  s: string;
begin
  if SameText('.dbf', ExtractFileExt(FileName)) or SameText('.txt', ExtractFileExt(FileName)) or SameText('.csv', ExtractFileExt(FileName)) then
    cbTable.Items.add(ExtractFileName(FileName))
  else
    ImpCon.GetTableNames(cbTable.Items, SameText('.xls', ExtractFileExt(ImpConWiz.ConnectedFile)));
  if cbTable.Items.Count = 0 then
    Exit;
  pcCon.ActivePageIndex:=1;
  s:=ExtractFileName(FileName);
  s:=copy(s, 1, Length(s) - Length(ExtractFileExt(FileName)));
  i:=cbTable.Items.IndexOf(s);
  if i < 0 then
    i:=0;
  // try to find table in AutoSetting
  if Assigned(AutoSetting) then
  begin
    s:=AutoSetting.Values[cSourceTable];
    if cbTable.Items.IndexOf(s) >= 0 then
      i:=cbTable.Items.IndexOf(s);
  end;
  cbTable.ItemIndex:=i;
  if SetSelect then
    cbTableChange(Self);
end;

function TrDBImportForm.GetSelectCmd: string;
var
  s: string;
begin
  s:=cbTable.Text;
  if Pos('$', s) = Length(s) then
    s:='[' + s + ']'
  else
    s:=rdbtool.GetQuotedTableName(s);
  Result:=Format('SELECT * FROM %s WHERE 1=1', [s]);
end;

procedure TrDBImportForm.OpenData;
begin
  qImp.Close;
  qImp.Filter:='';
  qImp.Filtered:=false;
  qImp.SQL.Text:=eSelectCmd.Text;
  qImp.Open;
  grImpData1.SetColumnWidth(cwDefaultWidth);
  lFile.Caption:=ImpConWiz.ConnectedFile;
end;

procedure TrDBImportForm.lblLoadDataClick(Sender: TObject);
begin
  if ImpCon.Connected then
  begin
    try
      OpenData;
    except
      on E: Exception do
        ShowErr('', E, false);
    end;
  end;
end;

procedure TrDBImportForm.cbTableChange(Sender: TObject);
begin
  eSelectCmd.Text:=GetSelectCmd;
  try
    OpenData;
  except
    on E: Exception do
      ShowErr('', E, false);
  end;
end;

procedure TrDBImportForm.LoadLookupSQL;
var
  i: integer;
begin
  ProcProgress(sLoading + sOfData);
  for i:=LookupDSList.Count to LookupSQLStr.Strings.Count - 1 do
  begin
    try
{$IFDEF UseADO}
      LookupDS:=TADOQuery.Create(Self);
      LookupDSList.add(LookupDS);
      with LookupDS as TADOQuery do
      begin
        Connection:=(dsData.DataSet as TCustomADODataSet).Connection;
        LockType:=ltReadOnly;
        SQL.Text:=LookupSQLStr.Strings[i];
      end;
      LookupDS.Open;
{$ENDIF}
{$IFDEF UseZeos}
      LookupDS:=TZQuery.Create(Self);
      LookupDSList.add(LookupDS);
      with LookupDS as TZQuery do
      begin
        Connection:=(dsData.DataSet as TZAbstractDataSet).Connection;
        ReadOnly:=true;
        SQL.Text:=LookupSQLStr.Strings[i];
      end;
      LookupDS.Open;
{$ENDIF}
    except
      on E: Exception do
        ShowErr(sOpenDataErr, E, false);
    end;
  end;
  ProcProgress('');
end;

procedure TrDBImportForm.SetLookupDS(Idx: integer);
begin
  LookupDS:=LookupDSList.Items[Idx] as TDataSet;
end;

function TrDBImportForm.GetCBName(Owner: TWinControl; Idx: integer): string;
begin
  Result:=Owner.Name + 'cb' + IntToStr(Idx) + RosiCompConfig.PropSaveCfg.Suffix_BlockSaveItems;
end;

constructor TCalc.Create;
begin
  FieldName:='';
  Field:=nil;
  CalcType:=0;
  Params:='';
  Param1:='';
  Param2:='';
  AllowedChars:='';
  ExcludedChars:='';
end;

function TrDBImportForm.InitCalcList(Check: boolean): boolean;
var
  i: integer;
  s, cName, err: string;
  c: TCalc;

  procedure AddExcludedChars(s: string);
  var
    X: integer;
  begin
    if copy(s, 1, 1) = '#' then
    begin
      X:=StrToIntDef(copy(s, 2, 255), 0);
      if X > 0 then
        c.ExcludedChars:=c.ExcludedChars + WideChar(X);
    end
    else
      c.ExcludedChars:=c.ExcludedChars + s;
  end;

begin
  CalcList.Clear;
  err:='';
  for i:=0 to mCalc.Lines.Count - 1 do
  begin
    s:=Trim(mCalc.Lines[i]);
    if s = '' then
      Continue;
    c:=TCalc.Create;
    CalcList.add(c);
    // 1st part as fieldname
    c.FieldName:=Trim(rtool.RemoveStrPart(s, '='));
    // 2nd part as CalcName(param1,param2,....)
    s:=Trim(s);
    // CalcName = all till (
    cName:=rtool.RemoveStrPart(s, '(');
    // find last )
    if copy(s, Length(s), 1) = ')' then
      s:=copy(s, 1, Length(s) - 1);
    // params
    c.Params:=s;
    c.Param1:=rtool.RemoveStrPart(s, ',');
    c.Param2:=rtool.RemoveStrPart(s, ',');
    // CalcTypes
    if SameText(cName, 'Trim') then
      c.CalcType:=cCalcTrim;
    if SameText(cName, 'Clean') then
    begin
      c.CalcType:=cCalcClean;
      c.AllowedChars:=rDBImport_Setting.StandardChars;
      if not(SameText(c.Param1, 'All') or SameText(c.Param1, 'Local')) then
        c.AllowedChars:=c.AllowedChars + rDBImport_Setting.LocalChars;
      if not(SameText(c.Param1, 'All') or SameText(c.Param1, 'Spec')) then
        c.AllowedChars:=c.AllowedChars + rDBImport_Setting.SpecialChars;
      c.AllowedChars:=UpperCase(c.AllowedChars);
    end;
    if SameText(cName, 'RemoveChar') then
    begin
      c.CalcType:=cCalcRemoveChar;
      s:=c.Params;
      while s <> '' do
        AddExcludedChars(rtool.RemoveStrPart(s, ','));
    end;
    if SameText(cName, 'Left') then
      c.CalcType:=cCalcLeft;
    if SameText(cName, 'Right') then
      c.CalcType:=cCalcRight;
    if SameText(cName, 'Middle') then
      c.CalcType:=cCalcMiddle;
    if SameText(cName, 'Part') then
      c.CalcType:=cCalcPart;
    if SameText(cName, 'FromTo') then
      c.CalcType:=cCalcFromTo;
    if SameText(cName, 'Replace') then
      c.CalcType:=cCalcReplace;
    c.Field:=nil;
    if c.CalcType = 0 then
      err:=err + Format(sCalcUnknownCmd, [cName, i + 1]) + #13#10
    else
    begin
      c.Field:=dsData.DataSet.FindField(c.FieldName);
      if c.Field = nil then
        err:=err + Format(sCalcUnknownField, [c.FieldName, i + 1]) + #13#10;
    end;
  end;
  Result:=CalcList.Count > 0;
  if Check then
  begin
    if (err <> '') then
      DlgW2(sCalcContainErrorsW, err)
    else
      if mCalc.Lines.Count > 0 then
        DlgI(sOk);
  end;
end;

procedure TrDBImportForm.LoadSel;
var
  i, j, p, aTop, aHeight, aBorder, aWidth: integer;
  E: TEdit;
  cb: TComboBox;
  f: TField;
  ds: TDataSet;

  procedure AddEdit(i: integer; FldName: string; p: TWinControl; Lookup: boolean = false);
  begin
    E:=TEdit.Create(Self);
    E.Name:=p.Name + 'e' + IntToStr(i);
    E.Parent:=p;
    E.Text:=FldName;
    E.Top:=aHeight * i + aTop;
    E.Left:=aBorder;
    E.Width:=aWidth;
    E.ReadOnly:=true;
    E.Color:=clBtnFace;
    // if not Lookup then E.Font.Style:=[fsBold];
    if Lookup then
      E.Font.Style:=[fsItalic];
  end;

  procedure AddCB(i, tag: integer; FldName: string; p: TWinControl);
  begin
    cb:=TComboBox.Create(Self);
    cb.Name:=GetCBName(p, i);
    cb.Parent:=p;
    cb.tag:=tag;
    cb.Hint:=FldName;
    cb.Top:=aHeight * i + aTop;
    cb.Left:=aBorder * 2 + aWidth;
    cb.Width:=aWidth;
    cb.Style:=csDropDown;
    cb.DropDownCount:=15;
    cb.Items.Assign(mFields.Strings);
    cb.Text:='';
    PSMain.StdCtrl.Components.add(cb.Name);
  end;

begin
  // remove old controls
  for i:=sb1.ControlCount - 1 downto 0 do
    if sb1.Controls[i].Name <> 'lInfo1' then
      sb1.Controls[i].Free;
  for i:=sb2.ControlCount - 1 downto 0 do
    if sb2.Controls[i].Name <> 'lInfo2' then
      sb2.Controls[i].Free;
  // clear propsave
  PSMain.StdCtrl.Components.Clear;
  PSMain.StdCtrl.Components.Assign(PSComps.Strings);
  // load import fields
  mFields.Strings.Clear;
  mFields.Strings.add('');
  for i:=0 to ImpDS.FieldCount - 1 do
    mFields.Strings.add(ImpDS.Fields[i].FieldName);

  // create controls for data fields
  aTop:=lInfo1.Height;
  aHeight:=cbTable.Height + 4;
  aBorder:=Round(6 * gbTop1.ScaleRatio);
  aWidth:=(sb1.Width - 3 * aBorder - GetSystemMetrics(SM_CXVSCROLL)) div 2;
  lInfo1.TabGridWidth:=aWidth + 2 * aBorder;
  lInfo2.TabGridWidth:=lInfo1.TabGridWidth;
  FldCount:=0;
  for i:=0 to dsData.DataSet.FieldCount - 1 do
  begin
    f:=dsData.DataSet.Fields[i];
    // if not f.Visible then Continue;
    AddEdit(FldCount, f.FieldName, sb1);
    AddEdit(FldCount, f.FieldName, sb2);

    AddCB(FldCount, 0, f.FieldName, sb1);
    cb.Enabled:=(not f.ReadOnly) and ((f.FieldKind = fkLookup) or (f.FieldKind = fkData));
    if cb.Enabled then
    begin
      p:=cb.Items.IndexOf(f.FieldName);
      if p >= 0 then
        cb.ItemIndex:=p;
    end;
    AddCB(FldCount, 0, f.FieldName, sb2);
    Inc(FldCount);
  end;

  // create lookup controls
  for i:=0 to LookupDSList.Count - 1 do
  begin
    if not(LookupDSList[i] is TDataSet) then
      Continue;
    ds:=(LookupDSList[i] as TDataSet);
    if not ds.Active then
      Continue;
    // first field is keyfield, others are lookup fields
    for j:=1 to ds.FieldCount - 1 do
    begin
      f:=ds.Fields[j];
      AddEdit(FldCount, f.FieldName, sb1, true);
      AddEdit(FldCount, f.FieldName, sb2, true);

      AddCB(FldCount, i + 1000, f.FieldName, sb1);
      p:=cb.Items.IndexOf(f.FieldName);
      if p >= 0 then
        cb.ItemIndex:=p;
      AddCB(FldCount, i + 1000, f.FieldName, sb2);
      Inc(FldCount);
    end;
  end;

  grData.SetColumnWidth(cwDefaultWidth);
  grImpData2.SetColumnWidth(cwDefaultWidth);
end;

function TrDBImportForm.FindDataRec: boolean;
var
  i, p: integer;
  f: string;
  cb: TComboBox;
  rec, FindLookup: boolean;
begin
  f:='';
  p:=0;
  Result:=true;
  FindLookup:=false;
  SetLength(v1, dsData.DataSet.FieldCount);
  SetLength(fn, dsData.DataSet.FieldCount);
  for i:=0 to FldCount - 1 do
  begin
    cb:=TComboBox(sb2.FindChildControl(GetCBName(sb2, i)));
    if (cb.Text <> '') then
    begin
      if VarIsNull(ImpDS.FieldValues[cb.Text]) then
        Result:=false;
      if cb.tag < 1000 then
      begin
        if dsData.DataSet.FieldByName(cb.Hint).FieldKind = fkLookup then
          FindLookup:=true;
        v1[p]:=ImpDS.FieldValues[cb.Text];
        fn[p]:=cb.Hint;
      end
      else
      begin
        Self.SetLookupDS(cb.tag - 1000);
        if LocateRecord(Self.LookupDS.FieldByName(cb.Hint), ImpDS.FieldValues[cb.Text]) and not LookupDS.Fields[0].IsNull and not
          (Trim(LookupDS.Fields[0].AsString) = '') then
        begin
          v1[p]:=Self.LookupDS.Fields[0].AsVariant;
          fn[p]:=Self.LookupDS.Fields[0].FieldName;
        end
        else
        begin
          Result:=false;
        end;
      end;
      if f <> '' then
        f:=f + ';';
      f:=f + fn[p];
      Inc(p);
    end;
  end;
  if p = 0 then
    raise Exception.Create(sNoCompareFields);
  if not Result then
    Exit;
  if p = 1 then
    v2:=v1[0]
  else
  begin
    v2:=VarArrayCreate([0, p - 1], varVariant);
    for i:=0 to p - 1 do
      v2[i]:=v1[i];
  end;
  if not FindLookup then
    Result:=dsData.DataSet.Locate(f, v2, [loCaseInsensitive])
  else
  begin
    rec:=false;
    dsData.DataSet.First;
    while not dsData.DataSet.Eof do
    begin
      rec:=true;
      for i:=0 to p - 1 do
        rec:=rec and (dsData.DataSet.FieldValues[fn[i]] = v1[i]);
      if rec then
        Break;
      dsData.DataSet.Next;
    end;
    Result:=rec;
  end;
  v2:=null;
end;

procedure TrDBImportForm.LoadCheck;
var
  s: string;
  i: integer;
  rec: integer;
begin
  WasAborted:=false;
  ProcProgress(sChecking + sOfData);
  dsData.DataSet.DisableControls;
  mErr.Lines.Clear;
  AddLogTime(sCheckData, cImportLogInfo);
  DefLogBlock(sNotFoundRec, cImportLogWarning);
  if rDBImport_Setting.ResultInTableForm then
  begin
    s:='RecNo' + rDBImport_Setting.ResultValueSeparator;
    for i:=0 to ImpDS.FieldCount - 1 do
      s:=s + ImpDS.Fields[i].FieldName + rDBImport_Setting.ResultValueSeparator;
    AddLog(s, cImportLogWarning);
  end;
  try
    add:=0; edit:=0; skip:=0;
    sum:=ImpDS.RecordCount;
    if rgTyp.ItemIndex = 0 then
      add:=sum
    else
    begin
      rec:=0;
      ImpDS.First;
      while not ImpDS.Eof do
      begin
        if FindDataRec then
        begin
          case rgTyp.ItemIndex of
            1, 2:
              Inc(edit);
            3:
              Inc(skip);
          end;
        end
        else
        begin
          case rgTyp.ItemIndex of
            1:
              Inc(skip);
            2, 3:
              Inc(add);
          end;
          if rDBImport_Setting.ResultInTableForm then
          begin
            s:=IntToStr(ImpDS.RecNo) + rDBImport_Setting.ResultValueSeparator;
            for i:=0 to ImpDS.FieldCount - 1 do
              s:=s + FieldTextForLog(ImpDS.Fields[i]) + rDBImport_Setting.ResultValueSeparator;
          end
          else
          begin
            s:='Record ' + IntToStr(ImpDS.RecNo) + ':';
            for i:=0 to ImpDS.FieldCount - 1 do
              s:=s + FormatFieldForLog(ImpDS.Fields[i]);
          end;
          AddLog(s, cImportLogWarning);
        end;
        ImpDS.Next;
        Inc(rec);
        if rec mod 10 = 0 then
          ProcProgress(sChecking + sOfData, 100 * rec / sum);
        if WasAborted then
        begin
          AddLogTime(sAborted, cImportLogError);
          if AutoMode <> cAutoRun then
            DlgI(sAborted);
          Break;
        end;
      end;
    end;
  finally
    DefLogBlock('');
    ProcProgress('');
    dsData.DataSet.EnableControls;
  end;
  // log check result
  AddLog('');
  AddLogTime(sCheckResult, cImportLogResult);
  AddLog(Format(sCheckResultCount, [sum, add, edit, skip]), cImportLogResult);
end;

procedure TrDBImportForm.lFileSel1AnchorClick(Sender: TObject; LinkName: string);
begin
  btSelFileClick(nil);
end;

procedure TrDBImportForm.lFileSel2AnchorClick(Sender: TObject; LinkName: String);
begin
  btNewConClick(nil);
end;

procedure TrDBImportForm.btSelFileClick(Sender: TObject);
begin
  if OpenDlg.Execute then
    if ConnectFile(OpenDlg.FileName) then
      LoadTables(OpenDlg.FileName, true);
end;

procedure TrDBImportForm.btNewConClick(Sender: TObject);
begin
  if CreateConnectString('') then
    LoadTables('', true);
end;

procedure TrDBImportForm.rgTypClick(Sender: TObject);
begin
  sb2.Visible:=rgTyp.ItemIndex > 0;
end;

procedure TrDBImportForm.lblCalcAnchorClick(Sender: TObject; LinkName: string);
var
  fn: string;
begin
  if LinkName = 'Check' then
    InitCalcList(true);
  if LinkName = 'Help' then
  begin
    fn:=ImpConWiz.UDLFilePath + 'Import_calc_help.html';
    if not FileExists(fn) then
      Self.mCalcHelp.Strings.SaveToFile(fn, TEncoding.UTF8);
    if FileExists(fn) then
      RunFileSimple(fn);
  end;
end;

procedure TrDBImportForm.lblExpStructureAnchorClick(Sender: TObject; LinkName: string);
begin
  rDBExport.DBExport(Self.grData, true, true, true, SameText(LinkName, 'ExpStruct'));
end;

procedure TrDBImportForm.lblShowErr1Click(Sender: TObject);
begin
  pc.ActivePageIndex:=3;
end;

procedure TrDBImportForm.btMoveItemsAnchorClick(Sender: TObject; LinkName: string);
var
  c: TComponent;
  s: string;
  i, p: integer;
begin
  if LinkName = 'DelAll' then
  begin
    for i:=0 to FldCount - 1 do
      TComboBox(sb1.FindChildControl(GetCBName(sb1, i))).Text:='';
    Exit;
  end;
  c:=Self.ActiveControl;
  if not Assigned(c) or (copy(c.Name, 1, 5) <> 'sb1cb') then
  begin
    DlgW(sImportMoveItemsW);
    Exit;
  end;
  s:=c.Name;
  s:=copy(s, 6, 4);
  s:=rtool.FindStrPart(s, '_', 1, true);
  p:=StrToInt(s);

  if LinkName = 'Down' then
  begin
    for i:=FldCount - 1 downto p + 1 do
      TComboBox(sb1.FindChildControl(GetCBName(sb1, i))).Text:=TComboBox(sb1.FindChildControl(GetCBName(sb1, i - 1))).Text;
    TComboBox(sb1.FindChildControl(GetCBName(sb1, p))).Text:='';
  end;
  if LinkName = 'Up' then
  begin
    for i:=p to FldCount - 1 do
      TComboBox(sb1.FindChildControl(GetCBName(sb1, i - 1))).Text:=TComboBox(sb1.FindChildControl(GetCBName(sb1, i))).Text;
    TComboBox(sb1.FindChildControl(GetCBName(sb1, FldCount - 1))).Text:='';
  end;
  DlgI(sOk);
end;

procedure TrDBImportForm.btCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TrDBImportForm.btPrevClick(Sender: TObject);
begin
  pc.ActivePageIndex:=pc.ActivePageIndex - 1;
  btPrev.Enabled:=pc.ActivePageIndex > 0;
  btNext.Enabled:=pc.ActivePageIndex < 2;
  btSettings.Visible:=(pc.ActivePageIndex > 0) and (AutoMode = cManual);
end;

procedure TrDBImportForm.btNextClick(Sender: TObject);
var
  i: integer;
  empty: boolean;
begin
  try
    case pc.ActivePageIndex of
      0:
        begin
          if not ImpDS.Active then
            raise Exception.Create(sNoTableOpen);
          LoadLookupSQL;
          LoadSel;
          LoadAutoSetting;
        end;
      1:
        begin
          empty:=true;
          for i:=0 to FldCount - 1 do
            if TComboBox(sb1.FindChildControl(GetCBName(sb1, i))).Text <> '' then
              empty:=false;
          if empty then
            raise Exception.Create(sNoImportFields);
          InitCalcList(false);
          if (AutoMode <> cAutoRun) or rDBImport_Setting.AutoRun_DoCheck then
            LoadCheck;
          pInfoPost.Visible:=false;
          btSaveAutoDesign.Enabled:=Assigned(AutoSetting);
        end;
    end;
    pc.ActivePageIndex:=pc.ActivePageIndex + 1;
    btPrev.Enabled:=pc.ActivePageIndex > 0;
    btNext.Enabled:=pc.ActivePageIndex < 2;
    btSettings.Visible:=(pc.ActivePageIndex > 0) and (AutoMode = cManual);
  except
    on E: Exception do
      if AutoMode = cAutoRun then
        raise
      else
        ShowErr('', E, false);
  end;
end;

function TrDBImportForm.UseCalc(Fld: TField; val: string): boolean;
var
  i, p1, p2: integer;
  c: TCalc;

  procedure DoClean(excl: boolean);
  var
    Idx, aPos: integer;
    inUStr: string;
    use: boolean;
  begin
    aPos:=1;
    inUStr:=UpperCase(val);
    for Idx:=1 to Length(inUStr) do
    begin
      if excl then
        use:=AnsiPos(inUStr[Idx], c.ExcludedChars) = 0
      else
        use:=AnsiPos(inUStr[Idx], c.AllowedChars) > 0;
      if use then
      begin
        val[aPos]:=val[Idx];
        Inc(aPos);
      end;
    end;
    SetLength(val, aPos - 1);
  end;

begin
  Result:=false;
  for i:=0 to CalcList.Count - 1 do
  begin
    c:=TCalc(CalcList[i]);
    if c.Field = Fld then
    begin
      p1:=StrToIntDef(c.Param1, Length(val));
      p2:=StrToIntDef(c.Param2, Length(val));
      case c.CalcType of
        cCalcTrim:
          val:=Trim(val);
        cCalcClean:
          DoClean(false);
        cCalcRemoveChar:
          DoClean(true);
        cCalcLeft:
          val:=copy(val, 1, p1);
        cCalcRight:
          val:=copy(val, Length(val) - p1 + 1, p1);
        cCalcMiddle:
          val:=copy(val, p1, p2);
        cCalcPart:
          val:=rtool.FindStrPart(val, c.Param1, p2);
        cCalcFromTo:
          val:=rtool.FindStrFromTo(c.Param1, c.Param2, val, false, true);
        cCalcReplace:
          val:=rtool.ReplaceStr(c.Param1, c.Param2, val, true);
      end;
      Result:=true;
    end;
  end;
  if (Fld is TStringField) and (Fld.Size < Length(val)) then
  begin
    val:=copy(val, 1, Fld.Size);
    Result:=true;
  end;
  if Result then
    Fld.AsString:=Trim(val);
end;

procedure TrDBImportForm.btRunClick(Sender: TObject);
var
  cont, frec: boolean;
  cb: TComboBox;
  s: string;
  f: TField;
  rec: integer;

  procedure CopyData(f, fs: TField);
  begin
    if rDBImport_Setting.CopyNumericFieldAsFloat and (f is TNumericField) and (fs is TNumericField) then
      f.AsFloat:=fs.AsFloat
    else
      f.AsVariant:=fs.AsVariant;
  end;

  procedure ProcRec(add: boolean);
  var
    i, j: integer;
    fs: string;
  begin
    try
      if add then
        dsData.DataSet.Append
      else
        dsData.DataSet.edit;
      {
        if add then
        // clear not used field
        for i:=0 to dsData.DataSet.FieldCount-1 do
        begin
        f:=dsData.DataSet.Fields[i];
        cb:=sb1.FindChildControl('sb1cb'+IntToStr(num)) as TComboBox;
        if cb.Enabled and (cb.Text='') then f.Clear;
        end;
        }

      // set used data
      for i:=0 to FldCount - 1 do
      begin
        cb:=TComboBox(sb1.FindChildControl(GetCBName(sb1, i)));
        if cb.Text <> '' then
        begin
          if cb.tag < 1000 then
          begin
            f:=dsData.DataSet.FieldByName(cb.Hint);
            // dataset data field
            if f.FieldKind = fkData then
            begin
              fs:=ImpDS.FieldByName(cb.Text).AsString;
              if fs = '' then
                f.Clear
              else
                if f is TBooleanField then
                  f.AsBoolean:=(fs = '1') or (SameText(fs, 'true')) or (SameText(fs, 'ano')) or (SameText(fs, 'a')) or (SameText(fs, 'yes')) or
                    (SameText(fs, 'y'))
                else
                  if not UseCalc(f, fs) then
                    CopyData(f, ImpDS.FieldByName(cb.Text));
              // f.AsVariant:=ImpDS.FieldValues[cb.Text];
            end;
            // dataset lookup field - use lookup defined by field
            if f.FieldKind = fkLookup then
              try
                if Assigned(f.LookupDataSet) and (f.LookupDataSet.Active) then
                  if LocateRecord(f.LookupDataSet.FieldByName(f.LookupResultField), ImpDS.FieldValues[cb.Text]) then
                    CopyData(f.DataSet.FieldByName(f.KeyFields), f.LookupDataSet.FieldByName(f.LookupKeyFields))
                    // f.DataSet.FieldByName(f.KeyFields).AsVariant:=f.LookupDataSet.FieldByName(f.LookupKeyFields).AsVariant
                  else
                  begin
                    Inc(war2);
                    s:='';
                    for j:=0 to ImpDS.FieldCount - 1 do
                      s:=s + FormatFieldForLog(ImpDS.Fields[j]);
                    AddWarn(s);
                    AddWarn(Format(sLookupWarning, [FieldTextForLog(ImpDS.FieldByName(cb.Text)), cb.Text]));
                  end;
              except
              end;
          end
          else
          begin
            // external lookup dataset
            if ImpDS.FieldByName(cb.Text).AsString <> '' then
            begin
              SetLookupDS(cb.tag - 1000);
              f:=dsData.DataSet.FieldByName(Self.LookupDS.Fields[0].FieldName);
              if LocateRecord(LookupDS.FieldByName(cb.Hint), ImpDS.FieldValues[cb.Text]) then
                CopyData(f, Self.LookupDS.Fields[0])
                // f.AsVariant:=Self.LookupDS.Fields[0].AsVariant
              else
              begin
                f.Clear;
                Inc(war2);
                s:='';
                for j:=0 to ImpDS.FieldCount - 1 do
                  s:=s + FormatFieldForLog(ImpDS.Fields[j]);
                AddWarn(s);
                AddWarn(Format(sLookupWarning, [FieldTextForLog(ImpDS.FieldByName(cb.Text)), cb.Text]));
              end;
            end;
          end;
        end;
      end;

      // fixed values
      for i:=0 to mVal.Lines.Count - 1 do
      begin
        s:=mVal.Lines.Names[i];
        if s = '' then
          Continue;
        f:=dsData.DataSet.FieldByName(s);
        s:=mVal.Lines.Values[s];
        if (s <> '') and (add or not chUseValAddOnly.Checked) then
          if s = 'null' then
            f.Clear
          else
            f.AsVariant:=s;
      end;

      dsData.DataSet.Post;
      if add then
        Inc(add2)
      else
        Inc(edit2);
    except
      on E: Exception do
      begin
        if rDBImport_Setting.ResultInTableForm then
        begin
          if err2 = 0 then
          begin
            AddLog('');
            AddLog(sImportDataErr, cImportLogError);
            // add field list for first error
            for i:=0 to dsData.DataSet.FieldCount - 1 do
              s:=s + dsData.DataSet.Fields[i].FieldName + rDBImport_Setting.ResultValueSeparator;
            AddLog(s, cImportLogError);
          end;
          Inc(err2);
          s:='';
          for i:=0 to ImpDS.FieldCount - 1 do
            s:=s + FieldTextForLog(ImpDS.Fields[i]) + rDBImport_Setting.ResultValueSeparator;
          AddLog(s + E.Message, cImportLogError);
        end
        else
        begin
          Inc(err2);
          AddLog(sImportDataErr, cImportLogError);
          s:='';
          for i:=0 to ImpDS.FieldCount - 1 do
            s:=s + FormatFieldForLog(ImpDS.Fields[i]);
          AddLog(s, cImportLogError);
          AddLog(sImportedErrorData, cImportLogError);
          s:='';
          for i:=0 to dsData.DataSet.FieldCount - 1 do
            s:=s + FormatFieldForLog(dsData.DataSet.Fields[i]);
          AddLog(s, cImportLogError);
          AddLog(E.Message, cImportLogError);
        end;
        dsData.DataSet.Cancel;
      end;
    end;
  end;

begin
  cont:=true;
  if Self.chBackup.Checked then
    Self.BackupCallBack(cont);
  if not cont then
    Exit;
  pInfoPost.Visible:=false;
  WasAborted:=false;
  ProcProgress(sImporting + sOfData);
  dsData.DataSet.DisableControls;
  try
    mErr.Lines.Clear;
    mWarn.Strings.Clear;
    AddLogTime(sImportData, cImportLogInfo);
    DefLogBlock(sNotFoundRec, cImportLogWarning);
    add2:=0; edit2:=0; skip2:=0; err2:=0; war2:=0;
    sum2:=ImpDS.RecordCount;
    rec:=0;
    ImpDS.First;
    while not ImpDS.Eof do
    begin
      if rgTyp.ItemIndex = 0 then
        ProcRec(true)
      else
      begin
        frec:=FindDataRec;
        case rgTyp.ItemIndex of
          1:
            if frec then
              ProcRec(false)
            else
              Inc(skip2);
          2:
            if frec then
              ProcRec(false)
            else
              ProcRec(true);
          3:
            if frec then
              Inc(skip2)
            else
              ProcRec(true);
        end;
      end;
      ImpDS.Next;
      Inc(rec);
      if (rec mod 10) = 0 then
      begin
        ProcProgress(sImporting + sOfData + #13 + Format(sImportProcInfo, [rec, sum2, err2, war2]), 100 * rec / sum2);
{$IFDEF DEMO}
        if (rDBImport_Setting.MaxImportedRecords < 0) or (rDBImport_Setting.MaxImportedRecords > 500) then
          rDBImport_Setting.MaxImportedRecords:=500;
{$ENDIF}
        if (rDBImport_Setting.MaxImportedRecords >= 0) and (add2 + edit2 + skip2 + err2 >= rDBImport_Setting.MaxImportedRecords) then
        begin
          DlgW(Format('You are using limited version of importing tool. Maximum number of imported records is %d.', [rDBImport_Setting.MaxImportedRecords]));
          Break;
        end;
      end;
      if WasAborted then
      begin
        AddLogTime(sAborted, cImportLogError);
        if AutoMode <> cAutoRun then
          DlgI(sAborted);
        Break;
      end;
    end;
    pInfoPost.Visible:=true;
  finally
    DefLogBlock('');
    ProcProgress('');
    dsData.DataSet.EnableControls;
  end;
  Imported:=true;
  // add warnings
  if mWarn.Strings.Count > 0 then
  begin
    AddLog('');
    AddLog(rtool.FindStrPart(sLookupWarning, ' ', 1));
    mErr.Lines.AddStrings(mWarn.Strings);
  end;
  // log import result
  AddLog('');
  AddLogTime(sImportResult, cImportLogResult);
  AddLog(Format(sImportResultCount, [sum2, add2, edit2, skip2, err2, war2]), cImportLogResult);
  // log result as special string
  if Assigned(LogCallBack) then
    LogCallBack(Format('#RESULT#: Total=%d;Add=%d;Edit=%d;Skip=%d;Error=%d;Warning=%d;', [sum2, add2, edit2, skip2, err2, war2]), cImportLogTotalResult,
      WasAborted);
end;

procedure TrDBImportForm.lInfoBeforeGetParameter(Sender: TObject; var ParamStr: String);
begin
  if ParamStr = 'sum' then
    ParamStr:=IntToStr(sum);
  if ParamStr = 'add' then
    ParamStr:=IntToStr(add);
  if ParamStr = 'edit' then
    ParamStr:=IntToStr(edit);
  if ParamStr = 'skip' then
    ParamStr:=IntToStr(skip);
end;

procedure TrDBImportForm.lInfoPostGetParameter(Sender: TObject; var ParamStr: String);
begin
  if ParamStr = 'sum' then
    ParamStr:=IntToStr(sum2);
  if ParamStr = 'add' then
    ParamStr:=IntToStr(add2);
  if ParamStr = 'edit' then
    ParamStr:=IntToStr(edit2);
  if ParamStr = 'skip' then
    ParamStr:=IntToStr(skip2);
  if ParamStr = 'err' then
    ParamStr:=IntToStr(err2);
  if ParamStr = 'war' then
    ParamStr:=IntToStr(war2);
end;

procedure TrDBImportForm.btConExpClick(Sender: TObject);
var
  f: string;
begin
  f:=ImpConWiz.UDLFilePath + 'Connection_log.txt';
  mConInfo.Lines.SaveToFile(f);
  DlgI(Format(sFileSaved, [f]));
end;

procedure TrDBImportForm.btSaveErrClick(Sender: TObject);
var
  f: string;
begin
  f:=ImpConWiz.UDLFilePath + 'Import_err.txt';
  mErr.Lines.SaveToFile(f);
  DlgI(Format(sFileSaved, [f]));
end;

procedure TrDBImportForm.gbPage1DropFiles(Sender: TObject; X, Y: integer);
var
  fn: string;
begin
  fn:=gbPage1.DropFileNames[0];
  if ConnectFile(fn) then
    LoadTables(fn, true);
end;

procedure TrDBImportForm.grImpData1ChangeSort(Sender: TObject; DataSet: TDataSet; SortFieldName: String; SortDesc: boolean; var Accept: boolean);
begin
  ChangeSortADO(DataSet as TCustomADODataSet, SortFieldName, SortDesc, Accept);
end;

procedure TrDBImportForm.grImpData1RefreshData(Sender: TObject);
begin
  RequeryGridADO(Sender as TrDBGrid);
end;

procedure TrDBImportForm.grImpData1Enter(Sender: TObject);
begin
  actGrid:=Sender as TrDBGrid;
end;

procedure TrDBImportForm.mnFindClick(Sender: TObject);
begin
  rDBFind.FindRec(actGrid);
end;

procedure TrDBImportForm.mnFieldsClick(Sender: TObject);
begin
  rDBFields.ChooseFields(actGrid);
end;

procedure TrDBImportForm.mnGridClick(Sender: TObject);
begin
  rDBGridProps.DBGridProperties(actGrid, true);
end;

procedure TrDBImportForm.PSMainReadProperties(PSFile: TPropSaveFile; Key: string);
begin
  ImpDS.Filter:=PSFile.ReadString(Key, 'Filter', ImpDS.Filter);
  ImpDS.Filtered:=PSFile.ReadBool(Key, 'Filtered', ImpDS.Filtered);
end;

procedure TrDBImportForm.PSMainSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteString(Key, 'Filter', ImpDS.Filter);
  PSFile.WriteBool(Key, 'Filtered', ImpDS.Filtered);
end;

procedure TrDBImportForm.mnFilterSelClick(Sender: TObject);
begin
  rDBFilter.FilterBySel(actGrid, fjNone);
end;

procedure TrDBImportForm.mnFilterManClick(Sender: TObject);
begin
  rDBFilter.FilterRec(actGrid);
end;

procedure TrDBImportForm.mnFilterOnOffClick(Sender: TObject);
begin
  try
    if actGrid.DataSource.DataSet.Filtered then
    begin
      actGrid.DataSource.DataSet.Filtered:=false;
    end
    else
      SetFilterString(actGrid.DataSource.DataSet, actGrid.DataSource.DataSet.Filter, true);
  except
  end;
end;

procedure TrDBImportForm.grImpData1TitleExClick(Sender: TObject);
begin
  rdbtool.ShowRecordCount(Sender as TDBGrid);
end;

procedure TrDBImportForm.OnAbort(Sender: TObject);
begin
  WasAborted:=true;
end;

procedure TrDBImportForm.btSettingsAnchorClick(Sender: TObject; LinkName: String);
begin
  if SameText(LinkName, 'Load') then
  begin
    OpenDlgSet.InitialDir:=ImpConWiz.UDLFilePath;
    if not OpenDlgSet.Execute then
      Exit;
    PSFile.LoadFile(OpenDlgSet.FileName);
    PSMain.ReadProperties;
  end;
  if SameText(LinkName, 'Save') then
  begin
    SaveDlgSet.InitialDir:=ImpConWiz.UDLFilePath;
    SaveDlgSet.FileName:=OpenDlgSet.FileName;
    if not SaveDlgSet.Execute then
      Exit;
    PSMain.SaveProperties;
    PSFile.SaveFile(SaveDlgSet.FileName);
  end;
end;

procedure TrDBImportForm.lblFilterAnchorClick(Sender: TObject; LinkName: String);
begin
  if LinkName = 'Set' then
    rDBFilter.FilterRec(grImpData2);
  if LinkName = 'Reset' then
    grImpData2.DataSource.DataSet.Filtered:=false;
end;

procedure TrDBImportForm.DoAutoImport(SourceDataSet: TDataSet);
var
  s, err: string;
begin
  pc.Enabled:=false;
  pBottom.Enabled:=false;
  try
    if rDBImport_Setting.AutoRun_ShowForm then
      Show;
    // use SourceDataSet if defined
    if Assigned(SourceDataSet) then
    begin
      ImpDS:=SourceDataSet;
      dsImp.DataSet:=ImpDS;
    end
    else
    begin
      // use source connection
      s:=AutoSetting.Values[cSourceConString];
      err:=sSourceConnectionErr + s;
      ImpCon.ConnectionString:=s;
      ImpCon.Open;
      // use source table
      s:=AutoSetting.Values[cSourceTable];
      err:=sSourceTableErr + s;
      eSelectCmd.Text:=AutoSetting.Values[cSourceSelectCmd];
      OpenData;
    end;
    err:=sImportDataErr;
    // step 1
    if pc.ActivePageIndex = 0 then
      btNext.Click;
    // step 2
    if pc.ActivePageIndex = 1 then
    begin
      LoadAutoSetting;
      btNext.Click;
    end;
    // step3 = run
    if pc.ActivePageIndex = 2 then
      btRun.Click;
  except
    on E: Exception do
    begin
      AddLogTime(err, cImportLogError);
      AddLog(E.Message, cImportLogError);
    end;
  end;
end;

procedure TrDBImportForm.eSelectCmdChange(Sender: TObject);
begin
  qImp.Close;
end;

procedure TrDBImportForm.LoadAutoSetting;
begin
  if not Assigned(AutoSetting) then
    Exit;
  PSFile.SetStrings(AutoSetting);
  PSMain.ReadProperties;
end;

procedure TrDBImportForm.SaveAutoSetting;
begin
  if not Assigned(AutoSetting) then
    Exit;
  AutoSetting.Clear;
  PSMain.SaveProperties;
  PSFile.WriteString(cSourceSection, cSourceConString, ImpCon.ConnectionString);
  PSFile.WriteString(cSourceSection, cSourceTable, cbTable.Text);
  PSFile.WriteString(cSourceSection, cSourceSelectCmd, eSelectCmd.Text);
  PSFile.GetStrings(AutoSetting);
end;

procedure TrDBImportForm.btSaveAutoDesignClick(Sender: TObject);
begin
  SaveAutoSetting;
  btSaveAutoDesign.Enabled:=false;
end;

procedure TrDBImportForm.DefLogBlock(const LogBlock: string; Level: integer = 0);
begin
  lastLogBlock:=LogBlock;
  lastLogBlockLevel:=Level;
end;

procedure TrDBImportForm.AddLog(const LogStr: string; Level: integer = 0);
begin
  if (lastLogBlock <> '') and (Level <> cImportLogError) then
  begin
    mErr.Lines.add(lastLogBlock);
    mErr.Lines.add('');
    if Assigned(LogCallBack) then
      LogCallBack(lastLogBlock, lastLogBlockLevel, WasAborted);
    lastLogBlock:='';
  end;
  mErr.Lines.add(LogStr);
  if Assigned(LogCallBack) then
    LogCallBack(LogStr, Level, WasAborted);
end;

procedure TrDBImportForm.AddLogTime(const LogStr: string; Level: integer = 0);
begin
  AddLog(DateTimeToStr(Now) + ': ' + LogStr, Level);
end;

procedure TrDBImportForm.AddWarn(const LogStr: string);
begin
  mWarn.Strings.add(LogStr);
end;

function TrDBImportForm.FormatFieldForLog(Fld: TField): string;
begin
  Result:=Format('%s = %s ; ', [Fld.FieldName, FieldTextForLog(Fld)]);
end;

function TrDBImportForm.FieldTextForLog(Fld: TField): string;
begin
  Result:=Fld.AsString;
  if Length(Result) > 200 then
    Result:=copy(Result, 1, 200) + '...';
  Result:=rtool.ReplaceStr(#13#10, ' | ', Result, true, true);
  Result:=rtool.ReplaceStr(#13, ' | ', Result, true, true);
  Result:=rtool.ReplaceStr(#10, ' | ', Result, true, true);
  Result:=rtool.ReplaceStr(#0, ' | ', Result, true, true);
end;

procedure TrDBImportForm.ProcProgress(const msg: string; proc: double);
var
  own: TWinControl;
begin
  if (AutoMode <> cAutoRun) or rDBImport_Setting.AutoRun_ShowProgress then
  begin
    if Self.Visible then
      own:=Self
    else
      own:=rDBImport_Setting.AutoRun_OwnerForm;
    if proc < 0 then
      ShowInfoEx(msg, own, OnAbort)
    else
      ShowInfoProgressBar(msg, own, OnAbort, Round(proc));
  end;
  if proc < 0 then
    proc:=0;
  if Assigned(LogCallBack) then
    LogCallBack(IntToStr(Round(proc)), cImportLogProgress, WasAborted);
  Application.ProcessMessages;
end;

initialization

rDBImport_Setting.AutoRun_ShowForm:=false;
rDBImport_Setting.AutoRun_ShowProgress:=false;
rDBImport_Setting.AutoRun_DoCheck:=false;
rDBImport_Setting.AutoRun_OwnerForm:=nil;
rDBImport_Setting.StandardChars:=' 0123456789abcdefghchijklmnopqrstuvwxyz';
rDBImport_Setting.LocalChars:='؊ݎ';
rDBImport_Setting.SpecialChars:='.,;:+-*/()?!#&@{}\%=<>$';
rDBImport_Setting.MaxImportedRecords:=-1;
rDBImport_Setting.CopyNumericFieldAsFloat:=true;
rDBImport_Setting.ResultInTableForm:=false;
rDBImport_Setting.ResultValueSeparator:='; ';

end.
