unit rDBPrint;

{<
  rDBPrint provides universal dialog for printing of data from
  DBGrid, rDBRecView and TreeView/rDBTreeView.
  Unit uses RTFReport2 component and predefined RTF templates to print data.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(May 2006)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  ImgList, PropSaveMain, PropSaveGrids, Dialogs, RTFReport2, ComCtrls, StdCtrls,
  rImprovedComps, DBGrids, rStringGridEd, rDBRecView, rDBTreeView,
  RosiCompGlobalCfg, db, rHTMLLabel, ExtCtrls,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} rdbdlg_baseform;

{ Show rDBPrint dialog for DBGrid and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrint(DBGrid: TCustomDBGrid; SetColumns, SetFilter: boolean; RepCaption: string); overload;
{ Show rDBPrint dialog for Dataset and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrint(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string); overload;
{ Show rDBPrint dialog for rDBRecView and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrint(rDBRecView: TrDBRecView; RepCaption: string); overload;
{ Show rDBPrint dialog for TreeView/rDBTreeView and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrint(TreeView: TCustomTreeView; RepCaption: string); overload;
{ Show rDBPrint dialog for rStringGridEd and let user to define page orientation,
  font size etc... }
procedure DBPrint(rStringGridEd: TrStringGridEd; RepCaption: string); overload;

{ Show rDBPrint dialog for Dataset and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrintDataset(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string); deprecated 'use DBPrint instead';
{ Show rDBPrint dialog for rDBRecView and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrintRecView(rDBRecView: TrDBRecView; RepCaption: string); deprecated 'use DBPrint instead';
{ Show rDBPrint dialog for TreeView/rDBTreeView and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrintTreeView(TreeView: TCustomTreeView; RepCaption: string); deprecated 'use DBPrint instead';
{ Show rDBPrint dialog for rStringGridEd and let user to define page orientation,
  font size, columns to print etc... }
procedure DBPrintStringGrid(rStringGridEd: TrStringGridEd; RepCaption: string); deprecated 'use DBPrint instead';

type
  { provides universal dialog for printing of data from DBGrid, rDBRecView and TreeView/rDBTreeView }
  TrDBPrintForm = class(TrDBDlgBaseForm)
    gbData1: TGroupBox;
    rbSelCol: TRadioButton;
    rbAllCol: TRadioButton;
    btFields: TrButton;
    btFilter: TrButton;
    gbFormat: TGroupBox;
    Label1: TLabel;
    cbPaperType_nostore: TComboBox;
    Label2: TLabel;
    cbFont_nostore: TComboBox;
    chAutoformat: TCheckBox;
    chCut: TCheckBox;
    lblCut: TLabel;
    lRec: TLabel;
    RTFRep: TRTFReport2;
    Label5: TLabel;
    eSep: TrEditNum;
    Label6: TLabel;
    eCaption: TEdit;
    pc: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    btPrint: TrButton;
    btShow: TrButton;
    btCancel: TrButton;
    PrinterDlg: TPrinterSetupDialog;
    btAbort: TrButton;
    lInfo: TLabel;
    Label9: TLabel;
    DBGridsPS: TDBGridsPropSave;
    Label11: TLabel;
    cbPaperOrient_nostore: TComboBox;
    chSum: TCheckBox;
    lblWrap: TLabel;
    lblSetDefaultCol: TrHTMLLabel;
    Bevel1: TBevel;
    GroupBox1: TGroupBox;
    cbPrinter: TComboBox;
    btPrintSetup: TrButton;
    PBar: TProgressBar;
    Label10: TLabel;
    lPages: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btPrintClick(Sender: TObject);
    procedure btFieldsClick(Sender: TObject);
    procedure btFilterClick(Sender: TObject);
    procedure rbSelColClick(Sender: TObject);
    procedure RTFRepUserCalc(Sender: TObject; var CalcStr: String; var Replace: boolean);
    procedure btShowClick(Sender: TObject);
    procedure cbFont_nostoreClick(Sender: TObject);
    procedure btPrintSetupClick(Sender: TObject);
    procedure btAbortClick(Sender: TObject);
    procedure RTFRepReadInfo(Sender: TObject);
    procedure DBGridsPSGetKeyString(Comp: TComponent; var Key: String);
    procedure lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
    procedure DBGridsPSProcComponents(Sender: TCustomPropSave; psAction: TPropSaveAction);
  private
    SourceType: TSourceType;
    Dataset: TDataset;
    OrigGr: TCustomDBGrid;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    Loaded, Aborted: boolean;
    wFldName: array of string;
    wMax, wCol, wSumAdd: array of Integer;
    Sum: array of double;
    tvMax: Integer;
    RepCapt: string;
    RecIdx, RecCount: Integer;
    ScanType: Integer;
    ScanResult: string;
    procedure ScanCallBackAnalyze(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
    procedure ScanCallBackPrint(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
    procedure AnalyzeData;
    function GetWidthSum(what: byte): Integer;
    function GetActWidth(Numeric, Count: boolean): Integer;
    procedure SetAutoformat;
    procedure SetCut;
    procedure CalcColWidth;
    function GetDBRecLine(Typ: Integer): string;
    function GetContent(Typ: Integer): string;
    procedure RunExport;
    procedure ResetSum;
    function GetTreeVItemCount: Integer;
    function GetStringGridRowCount: Integer;
  public
  end;

var
  rDBPrintForm: TrDBPrintForm;

implementation

{$R *.dfm}
{$R printdata.res}

uses rstring, rdlg, rtool, rdbtool, rDBGrid, rDBFields, rDBFilter;

const
  rtfcount = 7;
  rtfw: array [1 .. 6, 1 .. rtfcount] of Integer = ((138, 118, 103, 92, 82, 75, 68), (203, 174, 152, 135, 121, 110, 101), (138, 118, 103, 92, 82, 75, 68),
    (187, 160, 142, 126, 117, 107, 95), (138, 118, 103, 92, 82, 75, 68), (245, 208, 185, 160, 144, 131, 120));
  rtfh: array [1 .. 6, 1 .. rtfcount] of Integer = ((105, 90, 79, 70, 64, 57, 49), (70, 60, 52, 46, 42, 38, 35), (98, 86, 75, 67, 62, 56, 48),
    (70, 60, 52, 46, 42, 38, 35), (128, 111, 97, 87, 80, 72, 64), (70, 60, 52, 46, 42, 38, 35));

  cSepChar = ' ';
  cCutChar = '~';
  cSepLine = '\par ';

  cOutFile = 'printdata%d.rtf';
  cTempFile = 'printdata_template%dc.rtf';
  cRCName = 'printtemp%d';

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

procedure DBPrintInt(SourceType: TSourceType; O: TObject; SetColumns, SetFilter: boolean; RepCaption: string); forward;

procedure DBPrint(DBGrid: TCustomDBGrid; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintInt(stGrid, DBGrid, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrint(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintInt(stDataset, Dataset, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrintDataset(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintInt(stDataset, Dataset, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrint(rDBRecView: TrDBRecView; RepCaption: string);
begin
  DBPrintInt(stRecView, rDBRecView, false, false, RepCaption);
end;

procedure DBPrintRecView(rDBRecView: TrDBRecView; RepCaption: string);
begin
  DBPrintInt(stRecView, rDBRecView, false, false, RepCaption);
end;

procedure DBPrint(TreeView: TCustomTreeView; RepCaption: string);
begin
  DBPrintInt(stTreeView, TreeView, false, false, RepCaption);
end;

procedure DBPrintTreeView(TreeView: TCustomTreeView; RepCaption: string);
begin
  DBPrintInt(stTreeView, TreeView, false, false, RepCaption);
end;

procedure DBPrint(rStringGridEd: TrStringGridEd; RepCaption: string);
begin
  DBPrintInt(stStringGrid, rStringGridEd, false, false, RepCaption);
end;

procedure DBPrintStringGrid(rStringGridEd: TrStringGridEd; RepCaption: string);
begin
  DBPrintInt(stStringGrid, rStringGridEd, false, false, RepCaption);
end;

procedure DBPrintInt(SourceType: TSourceType; O: TObject; SetColumns, SetFilter: boolean; RepCaption: string);
var
  d: TDataSource;
  Dataset: TDataset;
begin
  if not Assigned(O) then
    Exit;
  Dataset:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      d:=TCustomDBGrid(O).DataSource;
    stRecView:
      d:=TrDBRecView(O).DataSource;
    stDataset:
      Dataset:=TDataset(O);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.Dataset) or not d.Dataset.Active then
      Exit;
    Dataset:=d.Dataset;
  end;

  rDBPrintForm:=TrDBPrintForm.Create(GetActiveOwner);
  try
    rDBPrintForm.RepCapt:=RepCaption;
    rDBPrintForm.Dataset:=Dataset;
    rDBPrintForm.SourceType:=SourceType;
    rDBPrintForm.btFilter.Enabled:=SetFilter;
    rDBPrintForm.btFields.Enabled:=SetColumns;
    rDBPrintForm.OrigGr:=nil;
    case rDBPrintForm.SourceType of
      stGrid:
        begin
          rDBPrintForm.OrigGr:=TCustomDBGrid(O);
          rDBPrintForm.DBGrid.Name:='gr';
          rDBPrintForm.DBGrid.Visible:=false;
          rDBPrintForm.DBGrid.Parent:=rDBPrintForm.OrigGr.Parent;
          AssignGridProps(rDBPrintForm.OrigGr, rDBPrintForm.DBGrid);
          rDBPrintForm.DBGridsPS.Components.Text:=rDBPrintForm.DBGrid.Name;
          rDBPrintForm.DBGridsPS.Disable:=RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields;
        end;
      stDataset:
        ;
      // rDBPrintForm.DBGridsPS.Components.Text:=DataSet.Name;
      stRecView:
        rDBPrintForm.RecView:=TrDBRecView(O);
      stTreeView:
        rDBPrintForm.TreeView:=TCustomTreeView(O);
      stStringGrid:
        rDBPrintForm.StringGrid:=TrStringGridEd(O);
    end;
    rDBPrintForm.ShowModal;
  finally
    FreeF(rDBPrintForm);
  end;
end;

procedure TrDBPrintForm.FormCreate(Sender: TObject);
begin
  inherited;
  Loaded:=false;
  DBGrid:=TCustomDBGrid.Create(Self);
  rtool.GetPrinterNames(cbPrinter.Items);
  cbPrinter.ItemIndex:=cbPrinter.Items.IndexOf(rtool.GetDefaultPrinterName);
  pc.ActivePageIndex:=0;
end;

procedure TrDBPrintForm.FormShow(Sender: TObject);
begin
  inherited;
  lblCut.Visible:=SourceType in [stGrid, stDataset, stStringGrid];
  lblWrap.Visible:=not lblCut.Visible;
  rbSelCol.Visible:=SourceType in [stGrid, stDataset, stRecView, stTreeView];
  rbAllCol.Visible:=rbSelCol.Visible;
  chSum.Visible:=SourceType in [stGrid, stDataset, stStringGrid];
  btFilter.Visible:=SourceType in [stGrid, stDataset];
  btFields.Visible:=btFilter.Visible;
  lblSetDefaultCol.Visible:=SourceType = stGrid;
end;

procedure TrDBPrintForm.FormActivate(Sender: TObject);
begin
  inherited;
  Update;
  if Loaded then
    Exit;
  Loaded:=true;
  if RepCapt <> '' then
    eCaption.Text:=RepCapt;
  AnalyzeData;
  rbSelColClick(nil);
  // SetAutoformat;
end;

procedure TrDBPrintForm.FormDestroy(Sender: TObject);
begin
  DBGrid.Free;
  inherited;
end;

procedure TrDBPrintForm.ScanCallBackAnalyze(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
var
  i, w: Integer;
  F: TField;
begin
  RecIdx:=RecIndex;
  for i:=0 to High(wMax) do
  begin
    F:=Dataset.FindField(wFldName[i]);
    if (F = nil) or (not IsSimpleField(F)) then
      Continue;
    w:=Length(F.DisplayText);
    if w > wMax[i] then
      wMax[i]:=w;
  end;
  if RecIdx mod 100 = 0 then
    SetProgressBar(PBar, RecIdx, RecCount);
  Application.ProcessMessages;
  if Aborted then
    Abort:=true;
end;

procedure TrDBPrintForm.AnalyzeData;

  procedure ProcDataset;
  var
    j: Integer;
    F: TField;
  begin
    // load max widths
    lInfo.Caption:=sChecking + sOfData;
    Aborted:=false;
    RecCount:=Dataset.RecordCount;
    SetProgressBar(PBar, 0, RecCount);
    pc.ActivePageIndex:=1;
    Update;
    try
      rdbtool.ScanAllRecords(Dataset, Self.ScanCallBackAnalyze);
    finally
      pc.ActivePageIndex:=0;
    end;
    // set column space for Sum row
    for j:=0 to High(wMax) do
    begin
      F:=Dataset.FindField(wFldName[j]);
      if (F is TNumericField) then
        wSumAdd[j]:=2
      else
        wSumAdd[j]:=0;
    end;
  end;

var
  i, r, w: Integer;
begin
  case SourceType of
    stGrid:
      with THackCustomDBGrid(DBGrid) do
      begin
        SetLength(wFldName, Columns.Count);
        SetLength(wMax, Columns.Count);
        SetLength(wSumAdd, Columns.Count);
        SetLength(Sum, Columns.Count);
        // load name of fields and title widths
        for i:=0 to High(wFldName) do
        begin
          wFldName[i]:=Columns[i].FieldName;
          wMax[i]:=Length(Columns[i].Title.Caption);
        end;
        ProcDataset;
      end;
    stDataset:
      begin
        SetLength(wFldName, Dataset.FieldCount);
        SetLength(wMax, Dataset.FieldCount);
        SetLength(wSumAdd, Dataset.FieldCount);
        SetLength(Sum, Dataset.FieldCount);
        // load name of fields and title widths
        for i:=0 to High(wFldName) do
        begin
          wFldName[i]:=Dataset.Fields[i].FieldName;
          wMax[i]:=Length(Dataset.Fields[i].DisplayLabel);
        end;
        ProcDataset;
      end;
    stStringGrid:
      begin
        SetLength(wFldName, StringGrid.ColCount);
        SetLength(wMax, StringGrid.ColCount);
        SetLength(wSumAdd, StringGrid.Columns.Count);
        SetLength(Sum, StringGrid.Columns.Count);
        // load name of fields and title widths
        for i:=0 to High(wFldName) do
        begin
          wFldName[i]:=StringGrid.Columns[i].TitleCaption;
          wMax[i]:=Length(StringGrid.Columns[i].TitleCaption);
        end;
        // load max widths
        lInfo.Caption:=sChecking + sOfData;
        Aborted:=false;
        r:=0;
        RecCount:=GetStringGridRowCount;
        SetProgressBar(PBar, r, RecCount);
        pc.ActivePageIndex:=1;
        Update;
        try
          for r:=1 to GetStringGridRowCount - 1 do
          begin
            for i:=0 to High(wMax) do
            begin
              w:=Length(StringGrid.Cells[i, r]);
              if w > wMax[i] then
                wMax[i]:=w;
            end;
            if r mod 100 = 0 then
              SetProgressBar(PBar, r, RecCount);
            Application.ProcessMessages;
            if Aborted then
              Break;
          end;
          // set column space for Sum row
          for i:=0 to High(wMax) do
            if not(StringGrid.Columns[i].DataType in [sgdText]) then
              wSumAdd[i]:=2
            else
              wSumAdd[i]:=0;
        finally
          pc.ActivePageIndex:=0;
        end;
      end;
    stRecView:
      begin
        SetLength(wMax, 2);
        wMax[0]:=0;
        wMax[1]:=0;
        // load name of fields and title widths
        for i:=0 to RecView.GetFieldCount - 1 do
        begin
          w:=Length(RecView.GetCaption(i));
          if w > wMax[0] then
            wMax[0]:=w;
          w:=Length(RecView.GetValue(i));
          if w > wMax[1] then
            wMax[1]:=w;
        end;
      end;
  end;
end;

// what: bit 1..0=all columns, 1=only visible columns,
// bit 2..0=all columns, 1=only columns with numeric field
// bit 4: 0=calc Sum of width, 1=calc count of col
function TrDBPrintForm.GetWidthSum(what: byte): Integer;

  procedure ProcField(idx: Integer; F: TField; Visible: boolean);
  var
    j, x: Integer;
  begin
    if (F = nil) or (not IsSimpleField(F)) then
      Exit;
    x:=-1;
    // find index in wMax
    for j:=0 to High(wFldName) do
      if SameText(wFldName[j], F.FieldName) then
        x:=j;
    // calc width or count
    if x >= 0 then
    begin
      if (what and 1 = 0) or Visible then
        if (what and 2 = 0) or (F is TNumericField) or (F is TDateTimeField) then
          if (what and 4 = 0) then
          begin
            Inc(Result, wMax[idx] + eSep.AsInteger);
            if chSum.Checked and (F is TNumericField) then
              Inc(Result, wSumAdd[idx]);
          end
          else
            Inc(Result, 1);
    end;
  end;

var
  i: Integer;
begin
  Result:=0;
  case SourceType of
    stGrid:
      with THackCustomDBGrid(DBGrid) do
      begin
        for i:=0 to Columns.Count - 1 do
          ProcField(i, Columns[i].Field, Columns[i].Visible);
      end;
    stDataset:
      begin
        for i:=0 to Dataset.FieldCount - 1 do
          ProcField(i, Dataset.Fields[i], Dataset.Fields[i].Visible);
      end;
    stStringGrid:
      begin
        for i:=0 to StringGrid.ColCount - 1 do
        begin
          if (what and 2 = 0) or not(StringGrid.Columns[i].DataType in [sgdText]) then
            if (what and 4 = 0) then
            begin
              Inc(Result, wMax[i] + eSep.AsInteger);
              if chSum.Checked and not(StringGrid.Columns[i].DataType in [sgdText]) then
                Inc(Result, wSumAdd[i]);
            end
            else
              Inc(Result, 1);
        end;
      end;
    stRecView:
      Result:=wMax[0] + wMax[1] + eSep.AsInteger + 1;
    stTreeView:
      begin
        GetTreeVItemCount;
        Result:=tvMax;
      end;
  end;
end;

function TrDBPrintForm.GetActWidth(Numeric, Count: boolean): Integer;
begin
  Result:=GetWidthSum(ord(rbSelCol.Checked) + ord(Numeric) * 2 + ord(Count) * 4);
end;

procedure TrDBPrintForm.SetAutoformat;
var
  i, x, w: Integer;
begin
  w:=GetActWidth(false, false);
  x:=-1;
  // find font size
  for i:=1 to rtfcount do
    if rtfw[cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1, i] >= w then
      x:=i;
  if x < 0 then
  begin
    if cbPaperOrient_nostore.ItemIndex = 0 then
    begin
      // switch paper to landscape
      cbPaperOrient_nostore.ItemIndex:=1;
      SetAutoformat;
    end
    else
      // choose smallest font
      cbFont_nostore.ItemIndex:=0
  end
  else
    cbFont_nostore.ItemIndex:=x - 1;
  SetCut;
end;

procedure TrDBPrintForm.SetCut;
var
  i: Integer;
begin
  i:=0;
  case SourceType of
    stGrid, stDataset:
      i:=Dataset.RecordCount;
    stStringGrid:
      i:=GetStringGridRowCount;
    stRecView:
      i:=RecView.GetFieldCount;
    stTreeView:
      i:=GetTreeVItemCount;
  end;
  lRec.Caption:=IntToStr(i);
  lPages.Caption:=IntToStr(i div rtfh[cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1, cbFont_nostore.ItemIndex + 1] + 1);
  chCut.Checked:=rtfw[cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1, cbFont_nostore.ItemIndex + 1] < GetActWidth(false, false);
end;

function TrDBPrintForm.GetTreeVItemCount: Integer;
  function ProcTV(tn: TTreeNode; level: Integer): Integer;
  var
    w: Integer;
  begin
    Result:=0;
    while Assigned(tn) do
    begin
      Inc(Result);
      w:=level * 2 + Length(tn.Text);
      if w > tvMax then
        tvMax:=w;
      if tn.Expanded or Self.rbAllCol.Checked then
        Result:=Result + ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  tvMax:=0;
  Result:=ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
end;

function TrDBPrintForm.GetStringGridRowCount;
begin
  Result:=StringGrid.RowCount;
  if StringGrid.AutoAddNewRow then
    Dec(Result);
end;

procedure TrDBPrintForm.rbSelColClick(Sender: TObject);
begin
  if not Loaded then
    Exit;
  if chAutoformat.Checked then
    SetAutoformat
  else
    SetCut;
end;

procedure TrDBPrintForm.cbFont_nostoreClick(Sender: TObject);
begin
  if not Loaded then
    Exit;
  chAutoformat.Checked:=false;
  SetCut;
end;

procedure TrDBPrintForm.btFieldsClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      ChooseFields(DBGrid);
    stDataset:
      ChooseFields(Dataset);
  end;
  rbSelColClick(nil);
end;

procedure TrDBPrintForm.btFilterClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      begin
        FilterRec(DBGrid);
        if OrigGr is TrDBGrid then
          TrDBGrid(OrigGr).FilterChangedNotification;
      end;
    stDataset:
      FilterRec(Dataset);
  end;
  SetCut;
end;

procedure TrDBPrintForm.CalcColWidth;

  procedure ProcField(idx: Integer; F: TField; Visible: boolean; koef: double);
  var
    j, x: Integer;
  begin
    // not visible or not simple field ?
    if rbSelCol.Checked and not Visible then
      Exit;
    if (F = nil) or (not IsSimpleField(F)) then
      Exit;
    x:=-1;
    // find index in wMax
    for j:=0 to High(wFldName) do
      if SameText(wFldName[j], F.FieldName) then
        x:=j;
    if x >= 0 then
    begin
      if (F is TNumericField) or (F is TDateTimeField) then
      begin
        wCol[idx]:=wMax[x];
        if chSum.Checked and (F is TNumericField) then
          Inc(wCol[idx], wSumAdd[x]);
      end
      else
        wCol[idx]:=Trunc(wMax[x] * koef);
    end;
  end;

var
  i, wsum, wnumsum, wothersep, wpaper: Integer;
  koef: double;
begin
  // calc column width for export
  case SourceType of
    stGrid:
      SetLength(wCol, THackCustomDBGrid(DBGrid).Columns.Count);
    stDataset:
      SetLength(wCol, Dataset.FieldCount);
    stStringGrid:
      SetLength(wCol, StringGrid.ColCount);
  end;
  for i:=0 to High(wCol) do
    wCol[i]:=0;
  wpaper:=rtfw[cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1, cbFont_nostore.ItemIndex + 1];
  wsum:=GetActWidth(false, false);
  wnumsum:=GetActWidth(true, false);
  wothersep:=(GetActWidth(false, true) - GetActWidth(true, true)) * eSep.AsInteger;
  if (wnumsum > wpaper - wothersep) then
    raise Exception.Create(sPrintToomuchCols);
  // calc koef for cutting strings
  if wsum <= wpaper then
    koef:=1
  else
    koef:=(wpaper - wnumsum - wothersep) / (wsum - wnumsum - wothersep);
  if (koef < 0.2) then
    raise Exception.Create(sPrintToomuchCols);

  case SourceType of
    stGrid:
      with THackCustomDBGrid(DBGrid) do
      begin
        for i:=0 to High(wCol) do
          ProcField(i, Columns[i].Field, Columns[i].Visible, koef);
      end;
    stDataset:
      begin
        for i:=0 to High(wCol) do
          ProcField(i, Dataset.Fields[i], Dataset.Fields[i].Visible, koef);
      end;
    stStringGrid:
      for i:=0 to High(wCol) do
      begin
        if not(StringGrid.Columns[i].DataType in [sgdText]) then
        begin
          wCol[i]:=wMax[i];
          if chSum.Checked then
            Inc(wCol[i], wSumAdd[i]);
        end
        else
          wCol[i]:=Trunc(wMax[i] * koef);
      end;
  end;

  // check
  wsum:=0;
  for i:=0 to High(wCol) do
    if wCol[i] > 0 then
      wsum:=wsum + wCol[i] + eSep.AsInteger;
  if (wsum > wpaper) then
    raise Exception.Create(sPrintToomuchCols);
end;

function TrDBPrintForm.GetDBRecLine(Typ: Integer): string;
var
  s: string;

  function ProcField(idx: Integer; F: TField; Capt: string; Visible: boolean): boolean;
  begin
    Result:=false;
    if rbSelCol.Checked and not Visible then
      Exit;
    if (F = nil) or (not IsSimpleField(F)) then
      Exit;
    // find data string
    case Typ of
      0:
        s:=Capt;
      1:
        begin
          s:=F.DisplayText;
          if F is TNumericField then
            Sum[idx]:=Sum[idx] + F.AsFloat;
        end;
      2:
        if F is TNumericField then
        begin
          if (F as TNumericField).DisplayFormat <> '' then
            s:=FormatFloat((F as TNumericField).DisplayFormat, Sum[idx])
          else
            s:=FloatToStr(Sum[idx]);
        end
        else
          s:='';
    end;
    // cut?
    if Length(s) > wCol[idx] then
      s:=copy(s, 1, wCol[idx] - 1) + cCutChar;
    // align string
    case F.Alignment of
      taRightJustify:
        s:=FillStrL(s, cSepChar, wCol[idx]);
      taCenter:
        begin
          s:=FillStrL(s, cSepChar, Length(s) + (wCol[idx] - Length(s)) div 2);
          s:=FillStrR(s, cSepChar, wCol[idx]);
        end;
    else
      s:=FillStrR(s, cSepChar, wCol[idx]);
    end;
    Result:=true;
  end;

var
  i: Integer;
begin
  Result:='';
  case SourceType of
    stGrid:
      with THackCustomDBGrid(DBGrid) do
        for i:=0 to Columns.Count - 1 do
        begin
          if not ProcField(i, Columns[i].Field, Columns[i].Title.Caption, Columns[i].Visible) then
            Continue;
          // set font
          if Typ > 0 then
          begin
            if (Typ = 2) or (fsBold in Columns[i].Font.Style) then
              s:='\b ' + s + '\b0 ';
            if fsItalic in Columns[i].Font.Style then
              s:='\i ' + s + '\i0 ';
            if fsUnderline in Columns[i].Font.Style then
              s:='\ul ' + s + '\ul0 ';
          end;
          // add string and separator
          Result:=Result + s + StringOfChar(cSepChar, eSep.AsInteger);
        end;
    stDataset:
      for i:=0 to Dataset.FieldCount - 1 do
      begin
        if not ProcField(i, Dataset.Fields[i], Dataset.Fields[i].DisplayLabel, Dataset.Fields[i].Visible) then
          Continue;
        // add string and separator
        Result:=Result + s + StringOfChar(cSepChar, eSep.AsInteger);
      end;
  end;
end;

procedure TrDBPrintForm.ScanCallBackPrint(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
begin
  RecIdx:=RecIndex;
  if ScanResult <> '' then
    ScanResult:=ScanResult + cSepLine;
  ScanResult:=ScanResult + GetDBRecLine(ScanType);
  if RecIdx mod 100 = 0 then
    SetProgressBar(PBar, RecIdx, RecCount);
  Application.ProcessMessages;
  if Aborted then
    Abort:=true;
end;

function TrDBPrintForm.GetContent(Typ: Integer): string;

  function GetDBAllRec: string;
  begin
    ScanResult:='';
    ScanType:=Typ;
    RecCount:=Dataset.RecordCount;
    SetProgressBar(PBar, 0, RecCount);
    rdbtool.ScanAllRecords(Dataset, Self.ScanCallBackPrint);
    Result:=ScanResult;
  end;

  function GetGridRow(Row: Integer): string;
  var
    i: Integer;
    s: string;
  begin
    Result:='';
    for i:=0 to StringGrid.Columns.Count - 1 do
    begin
      // find data string
      case Typ of
        0:
          s:=StringGrid.Columns[i].TitleCaption;
        1:
          begin
            s:=StringGrid.Cells[i, Row];
            if StringGrid.Columns[i].DataType in [sgdNumber, sgdInteger] then
              if StringGrid.IsValidData(i, Row) = sdsOk then
                Sum[i]:=Sum[i] + StringGrid.NumCells[i, Row];
          end;
        2:
          if StringGrid.Columns[i].DataType in [sgdNumber, sgdInteger] then
          begin
            if StringGrid.Columns[i].NumFormat <> '' then
              s:=FormatFloat(StringGrid.Columns[i].NumFormat, Sum[i])
            else
              s:=FloatToStr(Sum[i]);
          end
          else
            s:='';
      end;
      // cut?
      if Length(s) > wCol[i] then
        s:=copy(s, 1, wCol[i] - 1) + cCutChar;
      // align string
      case StringGrid.GetDataAlignment(i, Row) of
        taRightJustify:
          s:=FillStrL(s, cSepChar, wCol[i]);
        taCenter:
          begin
            s:=FillStrL(s, cSepChar, Length(s) + (wCol[i] - Length(s)) div 2);
            s:=FillStrR(s, cSepChar, wCol[i]);
          end;
      else
        s:=FillStrR(s, cSepChar, wCol[i]);
      end;
      // set font
      if Typ > 0 then
      begin
        if (Typ = 2) or (fsBold in StringGrid.Columns[i].Font.Style) then
          s:='\b ' + s + '\b0 ';
        if fsItalic in StringGrid.Columns[i].Font.Style then
          s:='\i ' + s + '\i0 ';
        if fsUnderline in StringGrid.Columns[i].Font.Style then
          s:='\ul ' + s + '\ul0 ';
      end;
      // add string and separator
      Result:=Result + s + StringOfChar(cSepChar, eSep.AsInteger);
    end;
  end;

  function GetGridAllRow: string;
  var
    i, max: Integer;
  begin
    Result:='';
    max:=GetStringGridRowCount;
    SetProgressBar(PBar, 0, max);
    for i:=1 to GetStringGridRowCount - 1 do
    begin
      if Result <> '' then
        Result:=Result + cSepLine;
      Result:=Result + GetGridRow(i);
      if i mod 100 = 0 then
        SetProgressBar(PBar, i, max);
      Application.ProcessMessages;
      if Aborted then
        Break;
    end;
  end;

  function GetRecViewAll: string;
  var
    s, v: string;
    i: Integer;
  begin
    Result:='';
    for i:=0 to RecView.GetFieldCount - 1 do
    begin
      if i > 0 then
        Result:=Result + cSepLine;
      s:=FillStrR(RecView.GetCaption(i), cSepChar, wMax[0]);
      v:=RecView.GetValue(i);
      v:=ReplaceStr(#13#10, cSepLine + StringOfChar(cSepChar, wMax[0] + eSep.AsInteger + 1), v, true, true);
      Result:=Result + s + ':' + StringOfChar(cSepChar, eSep.AsInteger) + v;
    end;
  end;

  function GetTreeViewAll: string;
    procedure ProcTV(tn: TTreeNode; level: Integer);
    var
      i: Integer;
      s, t: string;
      no: TNodeObject;
    begin
      while Assigned(tn) do
      begin
        s:='';
        if Result <> '' then
          s:=cSepLine;
        for i:=0 to level - 1 do
          s:=s + '  ';
        t:=tn.Text;
        if Assigned(tn.Data) then
        begin
          no:=TNodeObject(tn.Data);
          if fsBold in no.Style then
            t:='\b ' + t + '\b0 ';
          if fsItalic in no.Style then
            t:='\i ' + t + '\i0 ';
          if fsUnderline in no.Style then
            t:='\ul ' + t + '\ul0 ';
        end;
        Result:=Result + s + t;
        if tn.Expanded or Self.rbAllCol.Checked then
          ProcTV(tn.GetFirstChild, level + 1);
        tn:=tn.GetNextSibling;
      end;
    end;

  begin
    Result:='';
    ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
  end;

begin
  Result:='';
  begin
    case Typ of
      0, 2:
        begin
          case SourceType of
            stGrid, stDataset:
              Result:=GetDBRecLine(Typ);
            stStringGrid:
              Result:=GetGridRow(0);
          end;
        end;
      1:
        begin
          case SourceType of
            stGrid, stDataset:
              Result:=GetDBAllRec;
            stRecView:
              Result:=GetRecViewAll;
            stTreeView:
              Result:=GetTreeViewAll;
            stStringGrid:
              Result:=GetGridAllRow;
          end;
        end;
    end;
  end;
end;

procedure TrDBPrintForm.ResetSum;
var
  i: Integer;
begin
  for i:=0 to High(Sum) do
    Sum[i]:=0;
end;

procedure TrDBPrintForm.RunExport;
var
  dir: string;
  rStream: TResourceStream;
  fStream: TFileStream;
  fname: string;
  i: Integer;
begin
  Aborted:=false;
  try
    // set file names
    dir:=rtool.SetEndOfPath(RosiCompConfig.rDBPrintCfg.PrintData_Dir);
    if dir = '' then
      dir:=GetAppDir
    else
      ForceDirectories(dir);
    fname:=dir + Format(cTempFile, [cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1]);
    // load file from resource if not exist
    if not FileExists(fname) then
    begin
      rStream:=TResourceStream.Create(hInstance, Format(cRCName, [cbPaperType_nostore.ItemIndex * 2 + cbPaperOrient_nostore.ItemIndex + 1]), RT_RCDATA);
      try
        fStream:=TFileStream.Create(fname, fmCreate);
        try
          fStream.CopyFrom(rStream, 0);
        finally
          fStream.Free;
        end;
      finally
        rStream.Free;
      end;
    end;
    RTFRep.ReportFile:=fname;
    // output file
    i:=0;
    repeat
      Inc(i);
      fname:=dir + Format(cOutFile, [i]);
    until not FileExists(fname);
    RTFRep.OutputFile:=fname;

    // set font size
    RTFRep.Variables.Clear;
    RTFRep.Variables.Add('FS=' + cbFont_nostore.Text);

    lInfo.Caption:=sCreatingReport;
    pc.ActivePageIndex:=1;
    try
      if SourceType in [stGrid, stDataset, stStringGrid] then
        CalcColWidth;
      ResetSum;
      RTFRep.Execute(Dataset);
    finally
      pc.ActivePageIndex:=0;
    end;
  except
    on E: Exception do
      ShowErr(sCreatingReportErr, E, false);
  end;
end;

procedure TrDBPrintForm.RTFRepUserCalc(Sender: TObject; var CalcStr: String; var Replace: boolean);
begin
  Replace:=true;
  if SameText(CalcStr, 'ReportName') then
    CalcStr:=eCaption.Text
  else
    if SameText(CalcStr, 'Date') then
      CalcStr:=DateToStr(Date)
    else
      if SameText(CalcStr, 'Header') then
        CalcStr:=GetContent(0)
      else
        if SameText(CalcStr, 'Line') then
          CalcStr:=GetContent(1)
        else
          if SameText(CalcStr, 'Footer') then
            if chSum.Checked then
              CalcStr:=GetContent(2)
            else
              CalcStr:=''
            else
              Replace:=false;
end;

procedure TrDBPrintForm.RTFRepReadInfo(Sender: TObject);
begin
  if cbPrinter.Text = rtool.GetDefaultPrinterName then
    RTFRep.PrinterName:=''
  else
    RTFRep.PrinterName:=cbPrinter.Text;
end;

procedure TrDBPrintForm.btAbortClick(Sender: TObject);
begin
  Aborted:=true;
end;

procedure TrDBPrintForm.btPrintClick(Sender: TObject);
begin
  RTFRep.ActionAfter:=aaPrint;
  RunExport;
end;

procedure TrDBPrintForm.btShowClick(Sender: TObject);
begin
  RTFRep.ActionAfter:=aaShow;
  RunExport;
end;

procedure TrDBPrintForm.btPrintSetupClick(Sender: TObject);
begin
  PrinterDlg.Execute;
end;

procedure DelFiles;
var
  i: Integer;
  dir: string;
begin
  dir:=rtool.SetEndOfPath(RosiCompConfig.rDBPrintCfg.PrintData_Dir);
  if dir = '' then
    dir:=GetAppDir;
  for i:=0 to 100 do
    DeleteFile(dir + Format(cOutFile, [i]));
end;

procedure TrDBPrintForm.DBGridsPSGetKeyString(Comp: TComponent; var Key: String);
begin
  // Sets key in INI file according Form and DBGrid name
  case SourceType of
    stGrid:
      Key:=Key + '.' + OrigGr.Owner.Name + '.' + OrigGr.Name;
    stDataset:
      Key:=Key + '.' + Dataset.Owner.Name + '.' + Dataset.Name;
    stRecView:
      ;
    stTreeView:
      ;
    stStringGrid:
      ;
  end;
end;

procedure TrDBPrintForm.DBGridsPSProcComponents(Sender: TCustomPropSave; psAction: TPropSaveAction);
begin
  if SourceType = stDataset then
    Sender.ProcComp(psAction, Dataset, false);
end;

procedure TrDBPrintForm.lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
begin
  AssignGridProps(OrigGr, DBGrid);
  AnalyzeData;
  SetAutoformat;
end;

initialization

finalization

DelFiles;

end.
