unit rDBPrintDirect;

{<
  rDBPrintDirect provides universal dialog for printing of data from
  DBGrid, rDBRecView and TreeView/rDBTreeView.
  Unit uses TPrinter to print data directy to printer without RTFReport2.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2012)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  ImgList, Dialogs, ComCtrls, StdCtrls, Math, PropSaveMain, PropSaveGrids,
  rImprovedComps, DBGrids, rDBGrid, rStringGridEd, rDBRecView, rDBTreeView, db,
  RosiCompGlobalCfg, rHTMLLabel, Printers, ExtCtrls,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} rdbdlg_baseform;

{ Show rDBPrintDirect dialog for DBGrid and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirect(DBGrid: TCustomDBGrid; SetColumns, SetFilter: boolean; RepCaption: string); overload;
{ Show rDBPrintDirect dialog for Dataset and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirect(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string); overload;
{ Show rDBPrintDirect dialog for rDBRecView and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirect(rDBRecView: TrDBRecView; RepCaption: string); overload;
{ Show rDBPrintDirect dialog for TreeView/rDBTreeView and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirect(TreeView: TCustomTreeView; RepCaption: string); overload;
{ Show rDBPrintDirect dialog for rStringGridEd and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirect(rStringGridEd: TrStringGridEd; RepCaption: string); overload;

{ Show rDBPrintDirect dialog for Dataset and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirectDataset(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string); deprecated 'use DBPrintDirect instead';
{ Show rDBPrintDirect dialog for rDBRecView and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirectRecView(rDBRecView: TrDBRecView; RepCaption: string); deprecated 'use DBPrintDirect instead';
{ Show rDBPrintDirect dialog for TreeView/rDBTreeView and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirectTreeView(TreeView: TCustomTreeView; RepCaption: string); deprecated 'use DBPrintDirect instead';
{ Show rDBPrintDirect dialog for rStringGridEd and let user to define
  page orientation, font size, columns to print etc... }
procedure DBPrintDirectStringGrid(rStringGridEd: TrStringGridEd; RepCaption: string); deprecated 'use DBPrintDirect instead';

type
  { @exclude }
  TPageProps = record
    ppix, ppiy, mmx, mmy, PageWidth, PageHeight: Integer;
    Hmargin, Vmargin: Integer;
    CurLine, CurPage: Integer;
    CurX, CurY: Integer;
    LineH, LineW, LineHSpace, ColVSpace: Integer;
  end;

  { provides universal dialog for printing of data from DBGrid, rDBRecView and TreeView/rDBTreeView }
  TrDBPrintDirectForm = class(TrDBDlgBaseForm)
    gbData1: TGroupBox;
    rbSelCol: TRadioButton;
    rbAllCol: TRadioButton;
    btFields: TrButton;
    btFilter: TrButton;
    gbFormat: TGroupBox;
    Label2: TLabel;
    cbFont_nostore: TComboBox;
    lRec1: TLabel;
    Label6: TLabel;
    eCaption: TEdit;
    pc: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    btPrint: TrButton;
    btPreview: TrButton;
    btCancel: TrButton;
    PrinterDlg: TPrinterSetupDialog;
    btAbort: TrButton;
    lInfo: TLabel;
    Label9: TLabel;
    Label11: TLabel;
    cbPaperOrient_nostore: TComboBox;
    chSum: TCheckBox;
    FontDlg: TFontDialog;
    Label1: TLabel;
    lblFont: TLabel;
    btFont: TrURLLabel;
    chUseColors: TCheckBox;
    chPrintImg: TCheckBox;
    chPrintFooter: TCheckBox;
    Label4: TLabel;
    lPages1: TLabel;
    chTrimCol: TCheckBox;
    DBGridsPS: TDBGridsPropSave;
    lblSetDefaultCol: TrHTMLLabel;
    chPrintRecCount: TCheckBox;
    Bevel1: TBevel;
    GroupBox1: TGroupBox;
    cbPrinter: TComboBox;
    btPrintSetup: TrButton;
    GroupBox2: TGroupBox;
    Label8: TLabel;
    ePageHMargin: TrEditNum;
    Label12: TLabel;
    ePageVMargin: TrEditNum;
    Label13: TLabel;
    cbLineSpace_nostore: TComboBox;
    PBar: TProgressBar;
    chPrintGrid: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btPrintClick(Sender: TObject);
    procedure btFieldsClick(Sender: TObject);
    procedure btFilterClick(Sender: TObject);
    procedure rbSelColClick(Sender: TObject);
    procedure btPrintSetupClick(Sender: TObject);
    procedure btAbortClick(Sender: TObject);
    procedure btFontClick(Sender: TObject);
    procedure cbFont_nostoreChange(Sender: TObject);
    procedure btPreviewClick(Sender: TObject);
    procedure cbPrinterClick(Sender: TObject);
    procedure ePageHMarginChange(Sender: TObject);
    procedure ePageVMarginChange(Sender: TObject);
    procedure cbLineSpace_nostoreClick(Sender: TObject);
    procedure DBGridsPSGetKeyString(Comp: TComponent; var Key: string);
    procedure lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
    procedure rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure FormShow(Sender: TObject);
    procedure DBGridsPSProcComponents(Sender: TCustomPropSave; psAction: TPropSaveAction);
  private
    SourceType: TSourceType;
    Dataset: TDataset;
    OrigGr: TCustomDBGrid;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    Loaded, Aborted: boolean;
    ColW: array of Integer;
    sum: array of double;
    PgProps: TPageProps;
    ACanvas: TCanvas;
    IsPreview: boolean;
    PreviewBmpList: TList;
    ABmp: TBitmap;
    RepCapt: string;
    RecIdx, RecCount: Integer;
    procedure ClearPreviewList;
    procedure InitCanvas;
    procedure PreviewNewPage;
    procedure UpdateFont;
    procedure SetNumPages;
    function GetTreeVItemCount: Integer;
    function GetStringGridRowCount: Integer;
    procedure ResetSum;
    procedure RunPrint;
    procedure InitDoc(StartDoc: boolean);
    procedure FinalizeDoc(Abort: boolean);
    procedure NewPage;
    procedure PrintHeader;
    procedure PrintFooter;
    procedure PrintBitmap(B: TBitmap; X, Y, W, H: Integer);
    function CheckSpaceForLine(ARect: TRect): boolean;
    procedure MoveToNextLine(ARect: TRect);
    procedure PrintText(var ARect: TRect; LeftMargin: Integer; s: string; WordWrap, MovePos, PrintBk, PrintGrid: boolean; Frm: TTextFormat = []);
    procedure PrintRecCount(n: Integer);
    procedure PrintSepLine(ShowTitle, ShowSum, Sep: boolean; H: Integer);
    procedure PrintDBGridRow(ShowTitle, ShowSum: boolean);
    procedure PrintStringGridRow(ShowTitle, ShowSum: boolean; Row: Integer);
    procedure ScanCallBack(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
    procedure PrintGrid;
    procedure PrintStringGrid;
    procedure PrintRecView;
    procedure PrintTreeView;
  public
  end;

var
  rDBPrintDirectForm: TrDBPrintDirectForm;

implementation

{$R *.dfm}

uses rstring, rdlg, rtool, rdbtool, rpictool,
  rDBFields, rDBFilter, rDBPrintDirectPreview;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

procedure DBPrintDirectInt(SourceType: TSourceType; O: TObject; SetColumns, SetFilter: boolean; RepCaption: string); forward;

procedure DBPrintDirect(DBGrid: TCustomDBGrid; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintDirectInt(stGrid, DBGrid, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrintDirect(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintDirectInt(stDataset, Dataset, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrintDirectDataset(Dataset: TDataset; SetColumns, SetFilter: boolean; RepCaption: string);
begin
  DBPrintDirectInt(stDataset, Dataset, SetColumns, SetFilter, RepCaption);
end;

procedure DBPrintDirect(rDBRecView: TrDBRecView; RepCaption: string);
begin
  DBPrintDirectInt(stRecView, rDBRecView, false, false, RepCaption);
end;

procedure DBPrintDirectRecView(rDBRecView: TrDBRecView; RepCaption: string);
begin
  DBPrintDirectInt(stRecView, rDBRecView, false, false, RepCaption);
end;

procedure DBPrintDirect(TreeView: TCustomTreeView; RepCaption: string);
begin
  DBPrintDirectInt(stTreeView, TreeView, false, false, RepCaption);
end;

procedure DBPrintDirectTreeView(TreeView: TCustomTreeView; RepCaption: string);
begin
  DBPrintDirectInt(stTreeView, TreeView, false, false, RepCaption);
end;

procedure DBPrintDirect(rStringGridEd: TrStringGridEd; RepCaption: string);
begin
  DBPrintDirectInt(stStringGrid, rStringGridEd, false, false, RepCaption);
end;

procedure DBPrintDirectStringGrid(rStringGridEd: TrStringGridEd; RepCaption: string);
begin
  DBPrintDirectInt(stStringGrid, rStringGridEd, false, false, RepCaption);
end;

procedure DBPrintDirectInt(SourceType: TSourceType; O: TObject; SetColumns, SetFilter: boolean; RepCaption: string);
var
  d: TDataSource;
  Dataset: TDataset;
begin
  if not Assigned(O) then
    Exit;
  Dataset:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      d:=TCustomDBGrid(O).DataSource;
    stRecView:
      d:=TrDBRecView(O).DataSource;
    stDataset:
      Dataset:=TDataset(O);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.Dataset) or not d.Dataset.Active then
      Exit;
    Dataset:=d.Dataset;
  end;

  rDBPrintDirectForm:=TrDBPrintDirectForm.Create(GetActiveOwner);
  try
    // init form vars
    rDBPrintDirectForm.RepCapt:=RepCaption;
    rDBPrintDirectForm.Dataset:=Dataset;
    rDBPrintDirectForm.SourceType:=SourceType;
    rDBPrintDirectForm.btFilter.Enabled:=SetFilter;
    rDBPrintDirectForm.btFields.Enabled:=SetColumns;
    rDBPrintDirectForm.OrigGr:=nil;
    case rDBPrintDirectForm.SourceType of
      stGrid:
        begin
          rDBPrintDirectForm.OrigGr:=TCustomDBGrid(O);
          rDBPrintDirectForm.DBGrid.Name:='gr';
          rDBPrintDirectForm.DBGrid.Visible:=false;
          rDBPrintDirectForm.DBGrid.Parent:=rDBPrintDirectForm.OrigGr.Parent;
          AssignGridProps(rDBPrintDirectForm.OrigGr, rDBPrintDirectForm.DBGrid);
          rDBPrintDirectForm.DBGridsPS.Components.Text:=rDBPrintDirectForm.DBGrid.Name;
          rDBPrintDirectForm.DBGridsPS.Disable:=RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields;
        end;
      stDataset:
        ;
      // rDBPrintDirectForm.DBGridsPS.Components.Text:=DataSet.Name;
      stRecView:
        rDBPrintDirectForm.RecView:=TrDBRecView(O);
      stTreeView:
        begin
          rDBPrintDirectForm.TreeView:=TCustomTreeView(O);
          rDBPrintDirectForm.FontDlg.Font.Assign(THackCustomTreeView(rDBPrintDirectForm.TreeView).Font);
        end;
      stStringGrid:
        rDBPrintDirectForm.StringGrid:=TrStringGridEd(O);
    end;
    rDBPrintDirectForm.ShowModal;
  finally
    FreeF(rDBPrintDirectForm);
  end;
end;

procedure TrDBPrintDirectForm.FormCreate(Sender: TObject);
begin
  inherited;
  Loaded:=false;
  DBGrid:=TCustomDBGrid.Create(Self);
  rtool.GetPrinterNames(cbPrinter.Items);
  cbPrinter.ItemIndex:=cbPrinter.Items.IndexOf(rtool.GetDefaultPrinterName);
  pc.ActivePageIndex:=0;
  IsPreview:=false;
  PreviewBmpList:=TList.Create;
  lInfo.Caption:=sCreatingReport;
end;

procedure TrDBPrintDirectForm.FormDestroy(Sender: TObject);
begin
  ClearPreviewList;
  PreviewBmpList.Free;
  DBGrid.Free;
end;

procedure TrDBPrintDirectForm.FormShow(Sender: TObject);
begin
  inherited;
  chTrimCol.Visible:=SourceType in [stGrid, stDataset, stStringGrid];
  rbSelCol.Visible:=SourceType in [stGrid, stDataset, stRecView, stTreeView];
  rbAllCol.Visible:=rbSelCol.Visible;
  chSum.Visible:=SourceType in [stGrid, stDataset, stStringGrid];
  btFilter.Visible:=SourceType in [stGrid, stDataset];
  btFields.Visible:=btFilter.Visible;
  lblSetDefaultCol.Visible:=SourceType = stGrid;
  chPrintRecCount.Visible:=SourceType <> stRecView;
  chUseColors.Visible:=SourceType <> stDataset;
end;

procedure TrDBPrintDirectForm.FormActivate(Sender: TObject);
begin
  inherited;
  Update;
  if Loaded then
    Exit;
  Loaded:=true;
  if RepCapt <> '' then
    eCaption.Text:=RepCapt;
  if SourceType in [stGrid, stDataset, stStringGrid] then
  begin
    chPrintImg.Enabled:=false;
    // chPrintImg.Checked:=false;
  end;
  UpdateFont;
  SetNumPages;
end;

procedure TrDBPrintDirectForm.ClearPreviewList;
var
  i: Integer;
begin
  // free all preview bitmaps and clear list
  for i:=0 to PreviewBmpList.Count - 1 do
  begin
    ABmp:=PreviewBmpList.Items[i];
    ABmp.Free;
  end;
  PreviewBmpList.Clear;
end;

procedure TrDBPrintDirectForm.cbLineSpace_nostoreClick(Sender: TObject);
var
  s: string;
begin
  s:=cbLineSpace_nostore.Items[cbLineSpace_nostore.ItemIndex];
  s:=ReplaceStr(',', GetDecimalSeparator, s, false, true);
  s:=ReplaceStr('.', GetDecimalSeparator, s, false, true);
  RosiCompConfig.rDBPrintDirectCfg.LineSpaceRatio:=StrToFloatDef(s, 1) - 1;
end;

procedure TrDBPrintDirectForm.DBGridsPSGetKeyString(Comp: TComponent; var Key: string);
begin
  // Sets key in INI file according Form and DBGrid name
  case SourceType of
    stGrid:
      Key:=Key + '.' + OrigGr.Owner.Name + '.' + OrigGr.Name;
    stDataset:
      Key:=Key + '.' + Dataset.Owner.Name + '.' + Dataset.Name;
    stRecView:
      ;
    stTreeView:
      ;
    stStringGrid:
      ;
  end;
end;

procedure TrDBPrintDirectForm.DBGridsPSProcComponents(Sender: TCustomPropSave; psAction: TPropSaveAction);
begin
  if SourceType = stDataset then
    Sender.ProcComp(psAction, Dataset, false);
end;

procedure TrDBPrintDirectForm.UpdateFont;
begin
  lblFont.Caption:=FontDlg.Font.Name;
  lblFont.Font.Name:=FontDlg.Font.Name;
  cbFont_nostore.Text:=IntToStr(FontDlg.Font.Size);
end;

procedure TrDBPrintDirectForm.rbSelColClick(Sender: TObject);
begin
  if not Loaded then
    Exit;
  SetNumPages;
end;

procedure TrDBPrintDirectForm.rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.ReadFont(Key, FontDlg.Font);
end;

procedure TrDBPrintDirectForm.rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteFont(Key, FontDlg.Font);
end;

procedure TrDBPrintDirectForm.ePageHMarginChange(Sender: TObject);
begin
  RosiCompConfig.rDBPrintDirectCfg.HMarginMM:=Self.ePageHMargin.AsInteger;
end;

procedure TrDBPrintDirectForm.ePageVMarginChange(Sender: TObject);
begin
  RosiCompConfig.rDBPrintDirectCfg.VMarginMM:=Self.ePageVMargin.AsInteger;
end;

procedure TrDBPrintDirectForm.cbFont_nostoreChange(Sender: TObject);
begin
  if not Loaded then
    Exit;
  FontDlg.Font.Size:=StrToIntDef(cbFont_nostore.Text, 8);
  SetNumPages;
end;

procedure TrDBPrintDirectForm.btFieldsClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      ChooseFields(DBGrid);
    stDataset:
      ChooseFields(Dataset);
  end;
  rbSelColClick(nil);
end;

procedure TrDBPrintDirectForm.btFilterClick(Sender: TObject);
begin
  case SourceType of
    stGrid:
      begin
        FilterRec(DBGrid);
        if OrigGr is TrDBGrid then
          TrDBGrid(OrigGr).FilterChangedNotification;
      end;
    stDataset:
      FilterRec(Dataset);
  end;
  SetNumPages;
end;

procedure TrDBPrintDirectForm.btFontClick(Sender: TObject);
begin
  if FontDlg.Execute then
  begin
    FontDlg.Font.Style:=[];
    UpdateFont;
    SetNumPages;
  end;
end;

procedure TrDBPrintDirectForm.SetNumPages;
var
  lnp, n, p: Integer;
begin
  Printer.Orientation:=TPrinterOrientation(cbPaperOrient_nostore.ItemIndex);
  InitDoc(false);
  lnp:=Trunc((PgProps.PageHeight - 2 * PgProps.Vmargin) / (PgProps.LineH * (1 + RosiCompConfig.rDBPrintDirectCfg.LineSpaceRatio))) - 1;
  n:=0;
  case SourceType of
    stGrid, stDataset:
      n:=Dataset.RecordCount;
    stStringGrid:
      n:=GetStringGridRowCount;
    stRecView:
      n:=RecView.GetFieldCount;
    stTreeView:
      n:=GetTreeVItemCount;
  end;
  lRec1.Caption:=IntToStr(n);
  p:=(n div lnp) + 1;
  lPages1.Caption:=IntToStr(p);
end;

procedure TrDBPrintDirectForm.InitCanvas;
begin
  ACanvas.Font.Assign(FontDlg.Font);
  ACanvas.Font.PixelsPerInch:=PgProps.ppiy; { BUG! }
  ACanvas.Font.Size:=FontDlg.Font.Size;
  PgProps.LineH:=Round(ACanvas.TextHeight('Wg'));
  PgProps.LineHSpace:=Round(PgProps.LineH * RosiCompConfig.rDBPrintDirectCfg.LineSpaceRatio);
  PgProps.ColVSpace:=Round(PgProps.mmx * (1 + RosiCompConfig.rDBPrintDirectCfg.LineSpaceRatio));
end;

procedure TrDBPrintDirectForm.PreviewNewPage;
begin
  ABmp:=TBitmap.Create;
  PreviewBmpList.Add(ABmp);
  ABmp.PixelFormat:=pf24bit;
  ABmp.SetSize(PgProps.PageWidth, PgProps.PageHeight);
  ACanvas:=ABmp.Canvas;
  InitCanvas;
end;

procedure TrDBPrintDirectForm.InitDoc(StartDoc: boolean);
var
  ratio: double;
  oldCursor: TCursor;
begin
  oldCursor:=WCursor;
  try
    PgProps.ppix:=GetDeviceCaps(Printer.Handle, LOGPIXELSX); { Pixels per inch, x direction }
    PgProps.ppiy:=GetDeviceCaps(Printer.Handle, LOGPIXELSY); { "			"  y direction }
    PgProps.mmx:=Round(PgProps.ppix / 25.4); { Pixels per mm, x direction }
    PgProps.mmy:=Round(PgProps.ppiy / 25.4); { "	"	 y direction }
    PgProps.Hmargin:=PgProps.mmx * RosiCompConfig.rDBPrintDirectCfg.HMarginMM; { Horizontal (left hand) margin }
    PgProps.Vmargin:=PgProps.mmy * RosiCompConfig.rDBPrintDirectCfg.VMarginMM; { Vertical (top) margin }
    PgProps.PageWidth:=Printer.PageWidth;
    PgProps.PageHeight:=Printer.PageHeight;

    if IsPreview then
    begin
      // update PgProps record according preview density
      ratio:=Printer.PageWidth / RosiCompConfig.rDBPrintDirectCfg.PreviewMaxWidth;
      if ratio < 1 then
        ratio:=1;
      PgProps.ppix:=Round(PgProps.ppix / ratio);
      PgProps.ppiy:=Round(PgProps.ppiy / ratio);
      PgProps.mmx:=Round(PgProps.mmx / ratio);
      PgProps.mmy:=Round(PgProps.mmy / ratio);
      PgProps.Hmargin:=Round(PgProps.Hmargin / ratio);
      PgProps.Vmargin:=Round(PgProps.Vmargin / ratio);
      PgProps.PageWidth:=Round(Printer.PageWidth / ratio);
      PgProps.PageHeight:=Round(Printer.PageHeight / ratio);
      PreviewNewPage;
    end
    else
    begin
      ACanvas:=Printer.Canvas;
      InitCanvas;
    end;

    // init pos
    PgProps.CurPage:=0;
    PgProps.CurLine:=0;
    PgProps.CurX:=PgProps.Hmargin;
    PgProps.CurY:=PgProps.Vmargin;

    if StartDoc then
    begin
      if not IsPreview then
        Printer.BeginDoc;
      PrintHeader;
    end;
  finally
    ChCursor(oldCursor);
  end;
end;

procedure TrDBPrintDirectForm.FinalizeDoc(Abort: boolean);
begin
  PrintFooter;
  if IsPreview then
    Exit;
  if Abort then
    Printer.Abort
  else
    Printer.EndDoc;
end;

procedure TrDBPrintDirectForm.NewPage;
begin
  PrintFooter;
  if IsPreview then
  begin
    PreviewNewPage
  end
  else
    Printer.NewPage;
  Inc(PgProps.CurPage);
  PgProps.CurLine:=0;
  PgProps.CurX:=PgProps.Hmargin;
  PgProps.CurY:=PgProps.Vmargin;
  PrintHeader;
end;

procedure TrDBPrintDirectForm.PrintHeader;
var
  R1: TRect;
  s: string;
begin
  if eCaption.Text <> '' then
  begin
    R1.Left:=PgProps.Hmargin;
    R1.Right:=PgProps.PageWidth - PgProps.Hmargin;
    R1.Top:=PgProps.Vmargin - Round(PgProps.LineH * 1.5);
    R1.Bottom:=R1.Top + PgProps.LineH;
    s:=eCaption.Text;
    ACanvas.TextRect(R1, s, [tfCenter, tfNoPrefix]);
    ACanvas.Pen.Color:=clBlack;
    ACanvas.Pen.Width:=2;
    ACanvas.MoveTo(R1.Left, R1.Bottom);
    ACanvas.LineTo(R1.Right, R1.Bottom);
  end;
  // print title row for grid
  case SourceType of
    stGrid, stDataset:
      PrintDBGridRow(true, false);
    stStringGrid:
      PrintStringGridRow(true, false, 0);
  end;
end;

procedure TrDBPrintDirectForm.PrintFooter;
var
  R1: TRect;
  s: string;
begin
  if chPrintFooter.Checked then
  begin
    R1.Left:=PgProps.Hmargin;
    R1.Right:=PgProps.PageWidth - PgProps.Hmargin;
    R1.Top:=PgProps.PageHeight - PgProps.Vmargin + Round(PgProps.LineH * 0.5);
    R1.Bottom:=R1.Top + Round(PgProps.LineH * 1.1);
    s:=IntToStr(PgProps.CurPage + 1);
    ACanvas.TextRect(R1, s, [tfRight, tfBottom, tfNoPrefix]);
    ACanvas.Pen.Color:=clBlack;
    ACanvas.Pen.Width:=2;
    ACanvas.MoveTo(R1.Left, R1.Top);
    ACanvas.LineTo(R1.Right, R1.Top);
  end;
end;

function TrDBPrintDirectForm.CheckSpaceForLine(ARect: TRect): boolean;
begin
  Result:=PgProps.CurY + (ARect.Bottom - ARect.Top) <= PgProps.PageHeight - PgProps.Vmargin;
end;

procedure TrDBPrintDirectForm.MoveToNextLine(ARect: TRect);
begin
  Inc(PgProps.CurY, ARect.Bottom - ARect.Top); // + PgProps.LineHIncSpace);
  Inc(PgProps.CurLine);
end;

procedure TrDBPrintDirectForm.PrintText(var ARect: TRect; LeftMargin: Integer; s: string; WordWrap, MovePos, PrintBk, PrintGrid: boolean;
  Frm: TTextFormat = []);
var
  TextR: TRect;
  TF: TTextFormat;
  th, recth: Integer;
  oldC: TColor;
begin
  // store orig height - set as minimal height
  recth:=ARect.Bottom - ARect.Top;
  // calc text rect
  TextR:=ARect;
  TextR.Left:=TextR.Left + LeftMargin;
  if WordWrap then
    TF:=[tfWordBreak, tfNoPrefix]
  else
    TF:=[tfEndEllipsis, tfNoPrefix];
  ACanvas.TextRect(TextR, s, TF + [tfCalcRect]);
  // store text height
  th:=TextR.Bottom - TextR.Top;
  // set text Rect
  TextR.Left:=ARect.Left + LeftMargin + PgProps.ColVSpace div 2;
  TextR.Right:=ARect.Right - PgProps.ColVSpace div 2;
  TextR.Top:=PgProps.CurY + PgProps.LineHSpace div 2;
  TextR.Bottom:=TextR.Top + th;
  // set Rect size
  ARect.Top:=PgProps.CurY;
  ARect.Bottom:=ARect.Top + Max(th + PgProps.LineHSpace, recth);
  if not CheckSpaceForLine(ARect) then
    NewPage;
  // print
  if PrintBk then
    ACanvas.FillRect(ARect);
  ACanvas.TextRect(TextR, s, TF + Frm);
  if PrintGrid then
  begin
    oldC:=ACanvas.Brush.Color;
    ACanvas.Brush.Color:=clBlack;
    // inc per 1 as brush width
    TextR:=ARect;
    TextR.Right:=TextR.Right + 1;
    TextR.Bottom:=TextR.Bottom + 1;
    ACanvas.FrameRect(TextR);
    ACanvas.Brush.Color:=oldC;
  end;
  if MovePos then
    MoveToNextLine(ARect);
end;

procedure TrDBPrintDirectForm.PrintBitmap(B: TBitmap; X, Y, W, H: Integer);
var
  Info: PBitmapInfo;
  InfoSize: DWORD;
  Image: Pointer;
  ImageSize: DWORD;
  Bits: HBITMAP;
  DIBWidth, DIBHeight: LongInt;
begin
  ACanvas.Lock;
  try
    begin
      Bits:=B.Handle;
      GetDIBSizes(Bits, InfoSize, ImageSize);
      Info:=AllocMem(InfoSize);
      try
        Image:=AllocMem(ImageSize);
        try
          GetDIB(Bits, 0, Info^, Image^);
          with Info^.bmiHeader do
          begin
            DIBWidth:=biWidth;
            DIBHeight:=biHeight;
          end;
          StretchDIBits(ACanvas.Handle, X, Y, W, H, 0, 0, DIBWidth, DIBHeight, Image, Info^, DIB_RGB_COLORS, SRCCOPY);
        finally
          FreeMem(Image, ImageSize);
        end;
      finally
        FreeMem(Info, InfoSize);
      end;
    end;
  finally
    ACanvas.Unlock;
  end;
end;

procedure TrDBPrintDirectForm.ResetSum;
var
  i: Integer;
begin
  for i:=0 to High(sum) do
    sum[i]:=0;
end;

procedure TrDBPrintDirectForm.PrintRecCount(n: Integer);
var
  R1: TRect;
begin
  Inc(PgProps.CurY, PgProps.LineH);
  R1:=Rect(PgProps.Hmargin, 0, PgProps.PageWidth - PgProps.Hmargin, 0);
  // PrintText(R1, ' ', false, true, false, false, []);
  PrintText(R1, 0, Format(sPrintRecCount, [n]), false, true, false, false, []);
end;

procedure TrDBPrintDirectForm.PrintSepLine(ShowTitle, ShowSum, Sep: boolean; H: Integer);
begin
  if ShowTitle or Sep or ShowSum then
  begin
    ACanvas.Pen.Color:=clBlack;
    ACanvas.Pen.Width:=1;
    if Sep then
      ACanvas.Pen.Style:=psDot;
    if ShowTitle or ShowSum then
    begin
      ACanvas.Pen.Style:=psSolid;
      ACanvas.Pen.Width:=2;
    end;
    Inc(H, ACanvas.Pen.Width + 1);
    ACanvas.MoveTo(PgProps.Hmargin, H);
    ACanvas.LineTo(PgProps.PageWidth - PgProps.Hmargin, H);
    Inc(PgProps.CurY, 2 * ACanvas.Pen.Width + 1);
  end;
end;

procedure TrDBPrintDirectForm.PrintDBGridRow(ShowTitle, ShowSum: boolean);
var
  L: Integer;
  R1: TRect;
  UseSep: boolean;

  procedure SetStyle(c: Integer);
  var
    F: TFont;
  begin
    if ShowTitle then
      F:=THackCustomDBGrid(DBGrid).Columns[c].Title.Font
    else
      F:=THackCustomDBGrid(DBGrid).Columns[c].Font;
    ACanvas.Font.Style:=F.Style;

    if chUseColors.Checked then
    begin
      ACanvas.Font.Color:=F.Color;
      if ShowTitle then
        ACanvas.Brush.Color:=THackCustomDBGrid(DBGrid).Columns[c].Title.Color
      else
        ACanvas.Brush.Color:=THackCustomDBGrid(DBGrid).Columns[c].Color;
    end;
  end;

  function GetFormat(i: Integer): TTextFormat;
  var
    A: TAlignment;
  begin
    case SourceType of
      stGrid:
        if ShowTitle then
          A:=THackCustomDBGrid(DBGrid).Columns[i].Title.Alignment
        else
          A:=THackCustomDBGrid(DBGrid).Columns[i].Alignment;
    else
      A:=Dataset.Fields[i].Alignment;
    end;
    case A of
      taRightJustify:
        Result:=[tfRight];
      taCenter:
        Result:=[tfCenter];
    else
      Result:=[];
    end;
  end;

  function UseField(F: TField; Visible: boolean): boolean;
  begin
    Result:=Assigned(F) and IsSimpleField(F) and (Self.rbAllCol.Checked or Visible);
  end;

  procedure PrintRow(idx: Integer; F: TField; Capt: string; Visible: boolean);
  var
    s: string;
    dx: Integer;
    IsBool, BoolValue: boolean;
    CheckBmp: TBitmap;
  begin
    if not UseField(F, Visible) then
      Exit;
    // find text to print
    if ShowTitle then
    begin
      s:=Capt;
      if OrigGr is TrDBGrid then
        s:=TrDBGrid(OrigGr).GetTitleText(THackCustomDBGrid(DBGrid).Columns[idx]);
    end
    else
      if ShowSum then
      begin
        if F is TNumericField then
        begin
          if (F as TNumericField).DisplayFormat <> '' then
            s:=FormatFloat((F as TNumericField).DisplayFormat, sum[idx])
          else
            s:=FloatToStr(sum[idx]);
        end
        else
          s:='';
      end
      else
      begin
        if OrigGr is TrDBGrid then
          s:=TrDBGrid(OrigGr).GetCellValue(F)
        else
          s:=F.DisplayText;
        // increment sum for record
        if F is TNumericField then
          sum[idx]:=sum[idx] + F.AsFloat;
      end;
    if s = '' then
      s:=' ';

    // check col width on page
    if (L > PgProps.Hmargin) and (L + ColW[idx] > PgProps.PageWidth - PgProps.Hmargin) then
    begin
      MoveToNextLine(R1);
      L:=PgProps.Hmargin;
      UseSep:=true;
    end;
    R1:=Rect(L, 0, L + ColW[idx], 0);
    if SourceType = stGrid then
      SetStyle(idx);
    if OrigGr is TrDBGrid then
      IsBool:=TrDBGrid(OrigGr).IsFieldAsBoolean(F)
    else
      IsBool:=F is TBooleanField;
    if not ShowTitle and RosiCompConfig.rDBPrintDirectCfg.BooleanAsCheckBox and IsBool then
    begin
      PrintText(R1, 0, ' ', false, false, chUseColors.Checked, chPrintGrid.Checked);
      // show checkbox
      if OrigGr is TrDBGrid then
      begin
        BoolValue:=TrDBGrid(OrigGr).FieldAsBooleanValue(F);
        CheckBmp:=rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(BoolValue), TrDBGrid(OrigGr).ScaleRatio, TrDBGrid(OrigGr).CustomGlyphs);
      end
      else
      begin
        BoolValue:=F.AsBoolean;
        CheckBmp:=rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(BoolValue), 1, nil);
      end;
      {
        case DBGrid.Columns[i].Alignment of
        taCenter:
        dx:=(R1.Right - R1.Left - PgProps.LineH - 1) div 2;
        taRightJustify:
        dx:=R1.Right - R1.Left - PgProps.LineH - 1 - PgProps.ColVSpace div 2;
        else
        dx:=PgProps.ColVSpace div 2;
        end;
        }
      // always center
      dx:=(R1.Right - R1.Left - PgProps.LineH - 1) div 2;
      dx:=Max(1, dx);
      PrintBitmap(CheckBmp, R1.Left + dx, R1.Top + PgProps.LineHSpace div 2 + 1, PgProps.LineH - 2, PgProps.LineH - 2);
    end
    else
    begin
      PrintText(R1, 0, s, false, false, chUseColors.Checked, chPrintGrid.Checked, GetFormat(idx));
    end;
    L:=L + ColW[idx]; // + sepw;
  end;

var
  i, sumw, c: Integer;
  oldF: TFontRecall;
  oldB: TBrushRecall;

begin
  if High(ColW) < 0 then
  begin
    // init col width array
    case SourceType of
      stGrid:
        with THackCustomDBGrid(DBGrid) do
        begin
          SetLength(ColW, Columns.Count);
          for i:=0 to Columns.Count - 1 do
          begin
            c:=Columns[i].Width;
            if c < 5 then
              c:=Columns[i].DefaultWidth;
            ColW[i]:=Round(c / Canvas.TextWidth('W') * ACanvas.TextWidth('W'));
          end;
        end;
      stDataset:
        begin
          SetLength(ColW, Dataset.FieldCount);
          for i:=0 to Dataset.FieldCount - 1 do
          begin
            c:=Dataset.Fields[i].DisplayWidth;
            if c < 5 then
              c:=5;
            if c > 50 then
              c:=50;
            ColW[i]:=c * ACanvas.TextWidth('W');
          end;
        end;
    end;
    // init sum array
    SetLength(sum, Length(ColW));
    ResetSum;

    if chTrimCol.Checked then
    begin
      // calculate sum of cols width
      sumw:=0;
      case SourceType of
        stGrid:
          with THackCustomDBGrid(DBGrid) do
            for i:=0 to Columns.Count - 1 do
              if UseField(Columns[i].Field, Columns[i].Visible) then
                sumw:=sumw + ColW[i];
        stDataset:
          for i:=0 to Dataset.FieldCount - 1 do
            if UseField(Dataset.Fields[i], Dataset.Fields[i].Visible) then
              sumw:=sumw + ColW[i];
      end;
      // update width for each column (dec spaces for each column)
      for i:=0 to High(ColW) do
        ColW[i]:=Floor(ColW[i] * (PgProps.PageWidth - 2 * PgProps.Hmargin) / sumw);
    end;
  end;

  // save original font and init
  oldF:=TFontRecall.Create(ACanvas.Font);
  oldB:=TBrushRecall.Create(ACanvas.Brush);
  UseSep:=false;
  L:=PgProps.Hmargin;
  if ShowSum then
  begin
    PrintSepLine(ShowTitle, ShowSum, true, PgProps.CurY); // - Round(PgProps.LineHIncSpace / 2));
    // Inc(PgProps.CurY, 1);
  end;
  try
    case SourceType of
      stGrid:
        with THackCustomDBGrid(DBGrid) do
          for i:=0 to Columns.Count - 1 do
            PrintRow(i, Columns[i].Field, Columns[i].Title.Caption, Columns[i].Visible);
      stDataset:
        for i:=0 to Dataset.FieldCount - 1 do
          PrintRow(i, Dataset.Fields[i], Dataset.Fields[i].DisplayLabel, Dataset.Fields[i].Visible);
    end;

    if not chPrintGrid.Checked then
      PrintSepLine(ShowTitle, ShowSum, UseSep, R1.Bottom); // + Round(PgProps.LineHIncSpace / 2));
    MoveToNextLine(R1);
  finally
    oldF.Free;
    oldB.Free;
  end;
end;

procedure TrDBPrintDirectForm.ScanCallBack(Dataset: TDataset; RecIndex: Integer; var Abort: boolean);
begin
  RecIdx:=RecIndex;
  PrintDBGridRow(false, false);
  if RecIndex mod 100 = 0 then
    SetProgressBar(PBar, RecIdx, RecCount);
  Application.ProcessMessages;
  if Aborted then
    Abort:=true;
  if PreviewBmpList.Count >= 60 then
  begin
    DlgW(sPrintPreviewLimit);
    Abort:=true;
  end;
end;

procedure TrDBPrintDirectForm.PrintGrid;
begin
  ResetSum;
  RecCount:=Dataset.RecordCount;
  SetProgressBar(PBar, 0, RecCount);
  // ProcLine(true) - title is show in PreviewNewPage already
  rdbtool.ScanAllRecords(Dataset, Self.ScanCallBack);
  if chSum.Checked then
    PrintDBGridRow(false, true);
  if chPrintRecCount.Checked then
    PrintRecCount(RecIdx);
end;

procedure TrDBPrintDirectForm.PrintStringGridRow(ShowTitle, ShowSum: boolean; Row: Integer);

  procedure SetStyle(c, r: Integer);
  begin
    StringGrid.GetCellStyle(c, r, ACanvas);
    if not chUseColors.Checked then
    begin
      ACanvas.Font.Color:=clBlack;
      ACanvas.Brush.Color:=clWhite;
    end;
  end;

  function GetFormat(c, r: Integer): TTextFormat;
  var
    A: TAlignment;
  begin
    if ShowTitle then
      A:=StringGrid.Columns[c].TitleAlignment
    else
      A:=StringGrid.GetDataAlignment(c, r);
    case A of
      taRightJustify:
        Result:=[tfRight];
      taCenter:
        Result:=[tfCenter];
    else
      Result:=[];
    end;
  end;

var
  s: string;
  i, L, sumw, c, dx: Integer;
  R1: TRect;
  UseSep: boolean;
  oldF: TFontRecall;
  oldB: TBrushRecall;
  CheckBmp: TBitmap;
begin
  if High(ColW) < 0 then
  begin
    // init col width array
    SetLength(ColW, StringGrid.Columns.Count);
    for i:=0 to StringGrid.Columns.Count - 1 do
    begin
      c:=StringGrid.ColWidths[i];
      ColW[i]:=Round(c / StringGrid.Canvas.TextWidth('W') * ACanvas.TextWidth('W'));
    end;
    // init sum array
    SetLength(sum, StringGrid.Columns.Count);
    ResetSum;

    if chTrimCol.Checked then
    begin
      // calculate sum of cols width
      sumw:=0;
      for i:=0 to StringGrid.Columns.Count - 1 do
        sumw:=sumw + ColW[i];
      // update width for each column (dec spaces for each column)
      for i:=0 to StringGrid.Columns.Count - 1 do
        ColW[i]:=Floor(ColW[i] * (PgProps.PageWidth - 2 * PgProps.Hmargin) / sumw);
    end;
  end;

  // save original font and init
  oldF:=TFontRecall.Create(ACanvas.Font);
  oldB:=TBrushRecall.Create(ACanvas.Brush);
  UseSep:=false;
  L:=PgProps.Hmargin;
  if ShowSum then
  begin
    PrintSepLine(ShowTitle, ShowSum, true, PgProps.CurY); // - Round(PgProps.LineHIncSpace / 2));
    // Inc(PgProps.CurY, 1);
  end;
  try
    for i:=0 to StringGrid.Columns.Count - 1 do
    begin
      // find text to print
      if ShowTitle then
        s:=StringGrid.Columns[i].TitleCaption
      else
        if ShowSum then
        begin
          if StringGrid.GetDataType(i, Row) in [sgdNumber, sgdInteger] then
          begin
            if StringGrid.Columns[i].NumFormat <> '' then
              s:=FormatFloat(StringGrid.Columns[i].NumFormat, sum[i])
            else
              s:=FloatToStr(sum[i]);
          end
          else
            s:='';
        end
        else
        begin
          s:=StringGrid.Cells[i, Row];
          if StringGrid.GetDataType(i, Row) in [sgdNumber, sgdInteger] then
            if StringGrid.IsValidData(i, Row) = sdsOk then
              sum[i]:=sum[i] + StringGrid.NumCells[i, Row];
        end;
      if s = '' then
        s:=' ';

      // check col width on page
      if (L > PgProps.Hmargin) and (L + ColW[i] > PgProps.PageWidth - PgProps.Hmargin) then
      begin
        MoveToNextLine(R1);
        L:=PgProps.Hmargin;
        UseSep:=true;
      end;
      R1:=Rect(L, 0, L + ColW[i], 0);
      SetStyle(i, Row);
      if not ShowTitle and RosiCompConfig.rDBPrintDirectCfg.BooleanAsCheckBox and (StringGrid.GetEditStyleEx(i, Row) = sgbCheckbox) then
      begin
        PrintText(R1, 0, ' ', false, false, chUseColors.Checked, chPrintGrid.Checked);
        // show checkbox
        CheckBmp:=rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(StringGrid.BoolCells[i, Row]), StringGrid.ScaleRatio, StringGrid.CustomGlyphs);
        case StringGrid.Columns[i].CheckboxAlignment of
          taCenter:
            dx:=(R1.Right - R1.Left - PgProps.LineH - 1) div 2;
          taRightJustify:
            dx:=R1.Right - R1.Left - PgProps.LineH - 1 - PgProps.ColVSpace div 2;
        else
          dx:=PgProps.ColVSpace div 2;
        end;
        dx:=Max(1, dx);
        PrintBitmap(CheckBmp, R1.Left + dx, R1.Top + PgProps.LineHSpace div 2 + 1, PgProps.LineH - 2, PgProps.LineH - 2);
      end
      else
      begin
        PrintText(R1, 0, s, false, false, chUseColors.Checked, chPrintGrid.Checked, GetFormat(i, Row));
      end;
      L:=L + ColW[i]; // + sepw;
    end;
    if not chPrintGrid.Checked then
      PrintSepLine(ShowTitle, ShowSum, UseSep, R1.Bottom); // + Round(PgProps.LineHIncSpace / 2));
    MoveToNextLine(R1);
  finally
    oldF.Free;
    oldB.Free;
  end;
end;

procedure TrDBPrintDirectForm.PrintStringGrid;
var
  i, rec, Max: Integer;
begin
  ResetSum;
  rec:=0;
  Max:=GetStringGridRowCount;
  SetProgressBar(PBar, rec, Max);
  // ProcLine(true) - title is show in PreviewNewPage already
  for i:=1 to GetStringGridRowCount - 1 do
  begin
    PrintStringGridRow(false, false, i);
    Inc(rec);
    if rec mod 100 = 0 then
      SetProgressBar(PBar, rec, Max);
    Application.ProcessMessages;
    if Aborted then
      Break;
    if PreviewBmpList.Count >= 60 then
    begin
      DlgW(sPrintPreviewLimit);
      Break;
    end;
  end;
  if chSum.Checked then
    PrintStringGridRow(false, true, 0);
  if chPrintRecCount.Checked then
    PrintRecCount(rec);
end;

procedure TrDBPrintDirectForm.PrintRecView;
var
  s: string;
  i, W, maxw1, maxw2, imgh, aL, aR, maxR: Integer;
  R1, R2: TRect;
  F: TField;
  B: TBitmap;
begin
  // check field caption max width
  maxw1:=0;
  for i:=0 to RecView.GetFieldCount - 1 do
  begin
    s:=RecView.GetCaption(i);
    R1:=Rect(PgProps.Hmargin, 0, PgProps.PageWidth - PgProps.Hmargin, 0);
    ACanvas.TextRect(R1, s, [tfNoPrefix, tfCalcRect]);
    W:=R1.Right - R1.Left;
    if W > maxw1 then
      maxw1:=W;
  end;
  // check field value max width
  maxw2:=0;
  for i:=0 to RecView.GetFieldCount - 1 do
  begin
    s:=RecView.GetValue(i);
    R1:=Rect(PgProps.Hmargin, 0, PgProps.PageWidth - PgProps.Hmargin, 0);
    ACanvas.TextRect(R1, s, [tfNoPrefix, tfCalcRect]);
    W:=R1.Right - R1.Left;
    if W > maxw2 then
      maxw2:=W;
  end;
  // set borders for field values
  aL:=PgProps.Hmargin + maxw1 + PgProps.LineH + 4;
  maxR:=PgProps.PageWidth - PgProps.Hmargin;
  aR:=aL + maxw2 + 4;
  if aR > maxR then
    aR:=maxR;

  // print all
  for i:=0 to RecView.GetFieldCount - 1 do
  begin
    // print value
    R2:=Rect(aL, 0, maxR, 0);
    F:=RecView.GetField(i);
    if chPrintImg.Checked and IsGraphicField(F) then
    begin
      if F.IsNull then
        PrintText(R2, 2, ' ', true, false, false, chPrintGrid.Checked)
      else
      begin
        B:=TBitmap.Create;
        try
          try
            ReadBitmapFromField(F, B);
            // check size of img
            imgh:=B.Height;
            if B.Width > PgProps.PageWidth - PgProps.Hmargin - aL then
              imgh:=Round((PgProps.PageWidth - PgProps.Hmargin - aL) / B.Width * imgh);
            if imgh > PgProps.PageHeight - PgProps.Vmargin - PgProps.CurY then
              NewPage;
            rpictool.SmoothResizeBitmap(B, PgProps.PageWidth - PgProps.Hmargin - aL, PgProps.PageHeight - PgProps.Vmargin - PgProps.CurY, false);
            // draw frame only
            R2.Bottom:=R2.Top + B.Height;
            PrintText(R2, 0, ' ', true, false, false, chPrintGrid.Checked);
            // draw image
            R2:=Rect(aL, PgProps.CurY, aL + B.Width, PgProps.CurY + B.Height);
            PrintBitmap(B, R2.Left, R2.Top, B.Width, B.Height);
          except
            PrintText(R2, 2, sPictureCorrupted, true, false, false, chPrintGrid.Checked);
          end;
        finally
          B.Free;
        end;
      end;
    end
    else
    begin
      s:=RecView.GetValue(i);
      if s = '' then
        s:=' ';
      if chUseColors.Checked then
        RecView.GetFontStyle(i, ACanvas.Font, true);
      if RecView.GetDataAlignment(i) = DT_RIGHT then
      begin
        R2:=Rect(aL, 0, aR, 0);
        PrintText(R2, 2, s, true, false, false, chPrintGrid.Checked, [tfRight]);
      end
      else
        PrintText(R2, 2, s, true, false, false, chPrintGrid.Checked);
    end;

    // print caption
    s:=RecView.GetCaption(i) + ':';
    R1:=Rect(PgProps.Hmargin, 0, aL, R2.Bottom - R2.Top);
    ACanvas.Font.Assign(FontDlg.Font);
    PrintText(R1, 2, s, false, false, false, chPrintGrid.Checked);

    MoveToNextLine(R2);
  end;
end;

function TrDBPrintDirectForm.GetTreeVItemCount: Integer;

  function ProcTV(tn: TTreeNode; level: Integer): Integer;
  begin
    Result:=0;
    while Assigned(tn) do
    begin
      Inc(Result);
      if tn.Expanded or Self.rbAllCol.Checked then
        Result:=Result + ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  Result:=ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
end;

function TrDBPrintDirectForm.GetStringGridRowCount;
begin
  Result:=StringGrid.RowCount;
  if StringGrid.AutoAddNewRow then
    Dec(Result);
end;

procedure TrDBPrintDirectForm.PrintTreeView;
var
  B: TBitmap;
  num: Integer;

  procedure ProcTV(tn: TTreeNode; level: Integer);
  var
    L, dx, dy: Integer;
    R1: TRect;
    no: TNodeObject;
    s: string;
  begin
    while Assigned(tn) do
    begin
      Inc(num);
      // if num mod 20 = 0 then lNum.Caption:=IntToStr(PgProps.CurPage + 1);
      Application.ProcessMessages;
      if Aborted then
        Break;

      if Assigned(tn.Data) then
      begin
        no:=TNodeObject(tn.Data);
        ACanvas.Font.Style:=no.Style;
        if chUseColors.Checked then
        begin
          if no.FontColor = clNone then
            ACanvas.Font.Color:=clBlack
          else
            ACanvas.Font.Color:=no.FontColor;
          if no.BkColor = clNone then
            ACanvas.Brush.Color:=clWhite
          else
            ACanvas.Brush.Color:=no.BkColor;
        end;
      end;

      s:=tn.Text;
      if s = '' then
        s:=' ';
      L:=PgProps.Hmargin + PgProps.LineH * level;
      R1:=Rect(L, 0, PgProps.PageWidth - PgProps.Hmargin, 0);
      dx:=0;
      if chPrintImg.Checked and (tn.ImageIndex >= 0) then
        dx:=Round(PgProps.LineH * 1.5);
      PrintText(R1, dx + 4, s, true, true, chUseColors.Checked, chPrintGrid.Checked);

      if chPrintImg.Checked and (tn.ImageIndex >= 0) then
        if THackCustomTreeView(TreeView).Images.GetBitmap(tn.ImageIndex, B) then
        begin
          dy:=Max(0, (R1.Bottom - R1.Top - B.Height) div 2);
          PrintBitmap(B, L + 4, R1.Top + dy, PgProps.LineH, PgProps.LineH);
        end;

      if tn.Expanded or rbAllCol.Checked then
        ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  num:=0;
  B:=TBitmap.Create;
  try
    ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
    if chPrintRecCount.Checked then
      PrintRecCount(num);
  finally
    B.Free;
  end;
end;

procedure TrDBPrintDirectForm.RunPrint;
begin
  ClearPreviewList;
  SetLength(ColW, 0);

  Aborted:=false;
  try
    pc.ActivePageIndex:=1;
    SetProgressBar(PBar, 0, 100);
    Application.ProcessMessages;
    try
      InitDoc(true);

      case SourceType of
        stGrid, stDataset:
          PrintGrid;
        stStringGrid:
          PrintStringGrid;
        stRecView:
          PrintRecView;
        stTreeView:
          PrintTreeView;
      end;

      FinalizeDoc(Aborted);
      if Aborted then
        DlgW(rstring.sAborted);
    finally
      pc.ActivePageIndex:=0;
    end;
  except
    on E: Exception do
      ShowErr(sCreatingReportErr, E, false);
  end;
end;

procedure TrDBPrintDirectForm.btAbortClick(Sender: TObject);
begin
  Aborted:=true;
end;

procedure TrDBPrintDirectForm.btPreviewClick(Sender: TObject);
var
  pr: boolean;
begin
  IsPreview:=true;
  RunPrint;
  if not Aborted then
  begin
    rDBPrintDirectPreviewForm:=TrDBPrintDirectPreviewForm.Create(GetActiveOwner);
    try
      rDBPrintDirectPreviewForm.SetBmpList(PreviewBmpList);
      pr:=rDBPrintDirectPreviewForm.ShowModal = mrOk;
    finally
      FreeF(rDBPrintDirectPreviewForm);
    end;
    ClearPreviewList;
    if pr then
      btPrint.Click;
  end;
  ClearPreviewList;
end;

procedure TrDBPrintDirectForm.btPrintClick(Sender: TObject);
begin
  IsPreview:=false;
  RunPrint;
end;

procedure TrDBPrintDirectForm.cbPrinterClick(Sender: TObject);
begin
  Printer.PrinterIndex:=cbPrinter.ItemIndex;
  SetNumPages;
end;

procedure TrDBPrintDirectForm.btPrintSetupClick(Sender: TObject);
begin
  PrinterDlg.Execute;
end;

procedure TrDBPrintDirectForm.lblSetDefaultColAnchorClick(Sender: TObject; LinkName: string);
begin
  AssignGridProps(OrigGr, DBGrid);
  SetNumPages;
end;

end.
