unit rDBPrintDirectPreview;

{<
  rDBPrintDirectPreview provides preview dialog for rDBPrintDirect unit.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2012)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  StdCtrls, ExtCtrls, Dialogs, ImgList, PropSaveMain, rImageZoom, rImprovedComps,
  rdbdlg_baseform;

type
  { provides preview dialog for rDBPrintDirect unit }
  TrDBPrintDirectPreviewForm = class(TrDBDlgBaseForm)
    pBottom: TPanel;
    Img: TrImageZoom;
    btPrev: TrButton;
    btNext: TrButton;
    pBotRight: TPanel;
    btCancel: TrButton;
    btPrint: TrButton;
    btFit: TrButton;
    btZoomIn: TrButton;
    btZoomOut: TrButton;
    Label1: TLabel;
    ePage: TrEditNum;
    lPage: TLabel;
    procedure btPrevClick(Sender: TObject);
    procedure btNextClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btFitClick(Sender: TObject);
    procedure btZoomInClick(Sender: TObject);
    procedure btZoomOutClick(Sender: TObject);
    procedure ePageChange(Sender: TObject);
    procedure rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
  private
    aIdx: integer;
    BmpList: TList;
    IsLoaded: boolean;
  public
    procedure SetBmpList(List: TList);
    procedure ShowBmp(Idx: integer);
  end;

var
  rDBPrintDirectPreviewForm: TrDBPrintDirectPreviewForm;

implementation

{$R *.dfm}
uses rstring, rdlg, rtool, rdbtool, rpictool;

procedure TrDBPrintDirectPreviewForm.FormCreate(Sender: TObject);
begin
  inherited;
  IsLoaded:=false;
  btFit.Hint:=rstring.sZoomAutoFit;
  btZoomIn.Hint:=rstring.sZoomIn;
  btZoomOut.Hint:=rstring.sZoomOut;
end;

procedure TrDBPrintDirectPreviewForm.rDlgPropSaveReadProperties(
  PSFile: TPropSaveFile; Key: string);
begin
  Img.Zoom:=PSFile.ReadInteger(Key, 'Zoom', Img.Zoom);
end;

procedure TrDBPrintDirectPreviewForm.rDlgPropSaveSaveProperties(
  PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteInteger(Key, 'Zoom', Img.Zoom);
end;

procedure TrDBPrintDirectPreviewForm.SetBmpList(List: TList);
begin
  BmpList:=List;
  lPage.Caption:='/ ' + IntToStr(BmpList.Count);
  ePage.MaxValue:=BmpList.Count;
  ShowBmp(0);
  Img.Fit;
  IsLoaded:=true;
end;

procedure TrDBPrintDirectPreviewForm.ShowBmp(Idx: integer);
begin
  if Idx in [0 .. BmpList.Count - 1] then
  begin
    Img.Picture.Assign(BmpList[Idx]);
    aIdx:=Idx;
    ePage.AsInteger:=Idx + 1;
    btPrev.Enabled:=aIdx > 0;
    btNext.Enabled:=aIdx < BmpList.Count - 1;
  end;
end;

procedure TrDBPrintDirectPreviewForm.btZoomInClick(Sender: TObject);
begin
  Img.ZoomIn;
end;

procedure TrDBPrintDirectPreviewForm.btZoomOutClick(Sender: TObject);
begin
  Img.ZoomOut;
end;

procedure TrDBPrintDirectPreviewForm.btFitClick(Sender: TObject);
begin
  Img.Fit;
end;

procedure TrDBPrintDirectPreviewForm.btPrevClick(Sender: TObject);
begin
  ShowBmp(aIdx - 1);
end;

procedure TrDBPrintDirectPreviewForm.btNextClick(Sender: TObject);
begin
  ShowBmp(aIdx + 1);
end;

procedure TrDBPrintDirectPreviewForm.ePageChange(Sender: TObject);
begin
  if IsLoaded then
    ShowBmp(ePage.AsInteger - 1);
end;

end.
