unit rDBRecViewFields;

{<
  rDBRecViewFields provides universal dialog for user definition of visibility
  and properties of DBRecordView rows.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2012)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, CheckLst, Buttons, db, dbGrids, ExtCtrls, {$IFDEF DXE2P} UITypes, {$ENDIF}
  PropSaveMain, ImgList, rImprovedComps, rHTMLLabel, rDBRecView, rdbdlg_baseform;

{ Show DBRecViewFields dialog for rDBRecordView }
procedure ChooseRecViewFields(const rDBRecView: TrDBRecView);

type
  { universal universal dialog for user definition of visibility
    and properties of grid or dataset fields }
  TrDBRecViewFieldsForm = class(TrDBDlgBaseForm)
    lSeznam: TLabel;
    lbLabels: TrCheckListBox;
    btUp: TrButton;
    btDown: TrButton;
    btClose: TrButton;
    btAll: TrButton;
    gb: TGroupBox;
    Label3: TLabel;
    eCaption: TEdit;
    Panel1: TPanel;
    lblExample: TLabel;
    btApply: TrButton;
    FontDlg: TFontDialog;
    btDefault: TrURLLabel;
    GroupBox1: TGroupBox;
    chBold: TCheckBox;
    chItalic: TCheckBox;
    chUnderline: TCheckBox;
    Label1: TLabel;
    cbColor: TrColorBox;
    btFont: TrButton;
    lblTest: TLabel;
    procedure lbLabelsDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure lbLabelsClickCheck(Sender: TObject);
    procedure btUpClick(Sender: TObject);
    procedure btDownClick(Sender: TObject);
    procedure btAllClick(Sender: TObject);
    procedure lbLabelsKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure lbLabelsDblClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure lbLabelsClick(Sender: TObject);
    procedure cbColorChange(Sender: TObject);
    procedure btApplyClick(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
    procedure btDefaultClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btFontClick(Sender: TObject);
  private
    RV: TrDBRecView;
    DS: TDataSet;
    SetAll: Boolean;
    F: TField;
    act: Integer;
    procedure MoveI(List: TCheckListBox; old, new: Integer);
    procedure CheckI(act: Integer);
    procedure ReadProp;
    procedure SetProp;
    procedure SetTest;
    function CheckSave: Boolean;
    procedure SaveAll;
  public
  end;

var
  rDBRecViewFieldsForm: TrDBRecViewFieldsForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rtool, rdlg, rdbtool, rpictool, rstring;

type
  THackWinControl = class(TWinControl)
  end;

procedure ChooseRecViewFields(const rDBRecView: TrDBRecView);
begin
  if not Assigned(rDBRecView) or not Assigned(rDBRecView.DataSource) or not Assigned(rDBRecView.DataSource.DataSet)
    or not rDBRecView.DataSource.DataSet.Active then
    Exit;
  rDBRecViewFieldsForm:=TrDBRecViewFieldsForm.Create(GetActiveOwner);
  try
    rDBRecViewFieldsForm.RV:=rDBRecView;
    rDBRecViewFieldsForm.ShowModal;
  finally
    FreeF(rDBRecViewFieldsForm);
  end;
end;

procedure TrDBRecViewFieldsForm.FormShow(Sender: TObject);
var
  i: Integer;
  fn, fl: string;
begin
  inherited;
  btApply.Font.Style:=[fsBold];
  DS:=RV.DataSource.DataSet;
  // Load all labels
  lbLabels.Items.Clear;
  // load selected fields for RV
  for i:=0 to RV.Fields.Count - 1 do
  begin
    fn:=RV.Fields.Names[i];
    fl:=RV.Fields.ValueFromIndex[i];
    F:=DS.FindField(fn);
    if Assigned(F) then
    begin
      lbLabels.Items.AddObject(fl, F);
      lbLabels.Checked[lbLabels.Items.Count - 1]:=true;
    end;
  end;
  // load all another fields from Dataset
  for i:=0 to DS.FieldCount - 1 do
  begin
    F:=DS.Fields[i];
    if lbLabels.Items.IndexOfObject(F) < 0 then
    begin
      lbLabels.Items.AddObject(F.DisplayLabel, F);
      if RV.Fields.Count = 0 then
        lbLabels.Checked[lbLabels.Items.Count - 1]:=F.Visible;
    end;
  end;
  SetAll:=true;
  // Select labels
  lbLabels.ItemIndex:=0;
  ReadProp;
end;

procedure TrDBRecViewFieldsForm.SaveAll;
var
  i: Integer;
begin
  RV.Fields.Clear;
  for i:=0 to self.lbLabels.Items.Count - 1 do
  begin
    if lbLabels.Checked[i] then
    begin
      F:=TField(lbLabels.Items.Objects[i]);
      RV.Fields.Add(F.FieldName + '=' + lbLabels.Items[i]);
    end;
  end;
  // RV.SetRows;
end;

procedure TrDBRecViewFieldsForm.MoveI(List: TCheckListBox; old, new: Integer);
begin
  if not MoveCheckListBox(List, old, new) then
    Exit;
  ReadProp;
  SaveAll;
end;

procedure TrDBRecViewFieldsForm.CheckI(act: Integer);
begin
  SaveAll;
  RV.SetRows;
end;

procedure TrDBRecViewFieldsForm.ReadProp;
var
  s, frm: string;
  i: Integer;
begin
  // find field
  act:=lbLabels.ItemIndex;
  if act < 0 then
    Exit;
  F:=TField(lbLabels.Items.Objects[act]);

  gb.Caption:=' ' + lbLabels.Items[act] + ': ';
  eCaption.Text:=lbLabels.Items[act];

  // load saved format for field
  i:=RV.SpecFields.IndexOfName(F.FieldName);
  if i < 0 then
    frm:=''
  else
    frm:=RV.SpecFields.ValueFromIndex[i];

  // set controls according format
  chBold.Checked:=false;
  chItalic.Checked:=false;
  chUnderline.Checked:=false;
  cbColor.Selected:=RV.Font.Color;

  lblTest.Font.Style:=[];
  lblTest.Font.Color:=clNone;
  while frm <> '' do
  begin
    s:=AnsiUpperCase(Trim(rtool.RemoveStrPart(frm, ';')));
    // if s='SF' then chStandardFont.Checked:=true
    // else
    // if s='N' then chNormal.Checked:=true
    // else
    rtool.UseFontFormatParams(lblTest.Font, s);
  end;
  chBold.Checked:=fsBold in lblTest.Font.Style;
  chItalic.Checked:=fsItalic in lblTest.Font.Style;
  chUnderline.Checked:=fsUnderline in lblTest.Font.Style;
  if lblTest.Font.Color <> clNone then
    cbColor.Selected:=lblTest.Font.Color;

  SetTest;
  btApply.Enabled:=false;
end;

procedure TrDBRecViewFieldsForm.SetProp;
var
  i: Integer;
  frm: string;
begin
  // find field
  if act < 0 then
    Exit;
  F:=TField(lbLabels.Items.Objects[act]);

  // save props
  if eCaption.Text <> '' then
    lbLabels.Items[act]:=eCaption.Text;

  frm:='';
  // if chStandardFont.Checked then frm:=frm+'SF;';
  // if chNormal.Checked then frm:=frm+'N;';
  if chBold.Checked then
    frm:=frm + 'B;';
  if chItalic.Checked then
    frm:=frm + 'I;';
  if chUnderline.Checked then
    frm:=frm + 'U;';
  if cbColor.Selected <> clNone then
    frm:=frm + rpictool.ColorToString(cbColor.Selected, false) + ';';
  if frm = '' then
    frm:='0';
  i:=RV.SpecFields.IndexOfName(F.FieldName);
  if i < 0 then
    RV.SpecFields.Add(F.FieldName + '=' + frm)
  else
    RV.SpecFields.Values[F.FieldName]:=frm;

  btApply.Enabled:=false;
  SaveAll;
  RV.Invalidate;
end;

procedure TrDBRecViewFieldsForm.SetTest;
begin
  lblExample.Font.Assign(RV.Font);
  lblExample.Color:=RV.Color;
  lblExample.Font.Color:=cbColor.Selected;
  lblExample.Font.Style:=[];
  if chBold.Checked then
    lblExample.Font.Style:=lblExample.Font.Style + [fsBold];
  if chItalic.Checked then
    lblExample.Font.Style:=lblExample.Font.Style + [fsItalic];
  if chUnderline.Checked then
    lblExample.Font.Style:=lblExample.Font.Style + [fsUnderline];
  btApply.Enabled:=true;
end;

function TrDBRecViewFieldsForm.CheckSave: Boolean;
var
  qRes: Integer;
begin
  Result:=true;
  if btApply.Enabled then
  begin
    qRes:=ShowDlg(dtQ, btYesNoCancel, [sUseChangesQ]);
    lbLabels.EndDrag(false);
    case qRes of
      mrYes:
        SetProp;
      mrCancel:
        begin
          lbLabels.ItemIndex:=act;
          Result:=false;
        end;
      mrNo:
        ReadProp;
    end;
  end;
end;

procedure TrDBRecViewFieldsForm.btUpClick(Sender: TObject);
begin
  MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex - 1);
end;

procedure TrDBRecViewFieldsForm.btDownClick(Sender: TObject);
begin
  MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex + 1);
end;

procedure TrDBRecViewFieldsForm.btFontClick(Sender: TObject);
begin
  FontDlg.Font.Assign(RV.Font);
  if FontDlg.Execute then
  begin
    RV.Font.Assign(FontDlg.Font);
    RV.SpecFont.Assign(FontDlg.Font);
    SetTest;
    btApply.Click;
    // RV.Invalidate;
  end;
end;

procedure TrDBRecViewFieldsForm.lbLabelsDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept:=true;
  MoveI(lbLabels, lbLabels.ItemIndex, Y div lbLabels.ItemHeight + lbLabels.TopIndex);
end;

procedure TrDBRecViewFieldsForm.lbLabelsClickCheck(Sender: TObject);
begin
  CheckI(lbLabels.ItemIndex);
end;

procedure TrDBRecViewFieldsForm.lbLabelsDblClick(Sender: TObject);
begin
  lbLabels.Checked[lbLabels.ItemIndex]:=not lbLabels.Checked[lbLabels.ItemIndex];
  CheckI(lbLabels.ItemIndex);
end;

procedure TrDBRecViewFieldsForm.btAllClick(Sender: TObject);
var
  a: Integer;
begin
  for a:=0 to lbLabels.Items.Count - 1 do
  begin
    lbLabels.Checked[a]:=SetAll;
    CheckI(a);
  end;
  SetAll:=not SetAll;
end;

procedure TrDBRecViewFieldsForm.lbLabelsKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key = VK_UP) and (ssCtrl in Shift) then
  begin
    MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex - 1);
    Key:=0;
  end;

  if (Key = VK_DOWN) and (ssCtrl in Shift) then
  begin
    MoveI(lbLabels, lbLabels.ItemIndex, lbLabels.ItemIndex + 1);
    Key:=0;
  end;
end;

procedure TrDBRecViewFieldsForm.lbLabelsClick(Sender: TObject);
begin
  if CheckSave then
    ReadProp;
end;

procedure TrDBRecViewFieldsForm.cbColorChange(Sender: TObject);
begin
  SetTest;
end;

procedure TrDBRecViewFieldsForm.btApplyClick(Sender: TObject);
begin
  SetProp;
  RV.SetRows;
end;

procedure TrDBRecViewFieldsForm.btCloseClick(Sender: TObject);
begin
  if CheckSave then
    Close;
end;

procedure TrDBRecViewFieldsForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  RV.SetRows;
  inherited;
end;

procedure TrDBRecViewFieldsForm.btDefaultClick(Sender: TObject);
begin
  RV.Fields.Clear;
  RV.SpecFields.Clear;
  try
    RV.Font.Assign(THackWinControl(RV.Parent).Font);
    RV.SpecFont.Assign(RV.Font);
  except
  end;
  Close;
end;

end.
