unit rDBSort;

{<
  rDBSort provides universal dialog for sorting of any grid.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(May 2009)
  @lastmod(Sep 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, db, Grids, rStringGridEd, rDBGrid, ImgList,
  PropSaveMain, PropSaveGrids, rImprovedComps, rdbdlg_baseform;

{ Show DBSort dialog for grid }
function SortRec(rDBGrid: TrDBGrid): boolean;

type
  { universal dialog for sorting of any grid }
  TrDBSortForm = class(TrDBDlgBaseForm)
    Label1: TLabel;
    grSort: TrStringGridEd;
    str: TrStringsHolder;
    btOk: TrButton;
    btCancel: TrButton;
    PSFile: TPropSaveFile;
    PSStore: TPropSaveMain;
    GridsPropSave1: TGridsPropSave;
    btSave: TrButton;
    btLoad: TrButton;
    btLoadPick: TSpeedButton;
    lbPickList: TrListBox;
    btUp: TrButton;
    btDown: TrButton;
    btDel: TrButton;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    btDelAll: TrButton;
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure grSortSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
    procedure btUpClick(Sender: TObject);
    procedure btDelClick(Sender: TObject);
    procedure btDownClick(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btSaveClick(Sender: TObject);
    procedure btLoadClick(Sender: TObject);
    procedure btLoadPickClick(Sender: TObject);
    procedure lbPickListClick(Sender: TObject);
    procedure btDelAllClick(Sender: TObject);
  private
    rDBGrid: TrDBGrid;
    GridID: string;
    FN: TStrings;
    FL: TStrings;
    function TrySort: boolean;
    procedure SortList;
    procedure RefGrid;
    procedure LoadPicklist;
    procedure LoadSort(FileName: string);
  public
  end;

var
  rDBSortForm: TrDBSortForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rtool, rdbtool, rstring, rdlg;

const
  sGridID = 'GridID';

function SortRec(rDBGrid: TrDBGrid): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(rDBGrid) then
    Exit;
  rDBSortForm:=TrDBSortForm.Create(GetActiveOwner);
  try
    rDBSortForm.rDBGrid:=rDBGrid;
    rDBSortForm.GridID:=rDBSortForm.rDBGrid.Owner.Name + '.' + rDBSortForm.rDBGrid.Name;
    Result:=rDBSortForm.ShowModal = mrOk;
  finally
    FreeF(rDBSortForm);
  end;
end;

procedure TrDBSortForm.FormCreate(Sender: TObject);
begin
  inherited;
  FN:=TStringList.Create;
  FL:=TStringList.Create;
end;

procedure TrDBSortForm.FormShow(Sender: TObject);
var
  i, fidx: Integer;
  SortFieldsArr: TSortFieldNames;
  SortDirsArr: TSortDirections;
  flbl: string;
begin
  inherited;
  rDlgPropSave.ReadProperties;
  grSort.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  // load fields and labels
  GetFldList(rDBGrid, FN, [fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  GetFldList(rDBGrid, FL, [fdGetLabels, fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  if RosiCompConfig.rDBSortCfg.SortFieldList then
    SortList;
  // fill cells
  grSort.Columns[1].PickList.Assign(FL);
  SortFieldsArr:=rDBGrid.SortFieldNamesArr;
  SortDirsArr:=rDBGrid.SortDirectionsArr;
  for i:=0 to High(SortFieldsArr) do
  begin
    grSort.RowCount:=i + 3;
    fidx:=FN.IndexOf(SortFieldsArr[i]);
    if fidx >= 0 then
      flbl:=FL[fidx]
    else
      flbl:=SortFieldsArr[i];
    grSort.Cells[1, i + 1]:=flbl;
    if i <= High(SortDirsArr) then
      if SortDirsArr[i] then
        grSort.Cells[2, i + 1]:='1';
  end;
  RefGrid;
  LoadPicklist;
  OpenDlg.Filter:=OpenDlg.Filter + '|' + sFileTypeAll;
  SaveDlg.Filter:=OpenDlg.Filter;
  OpenDlg.InitialDir:=RosiCompConfig.rDBSortCfg.DefaultPath;
  SaveDlg.InitialDir:=OpenDlg.InitialDir;
end;

procedure TrDBSortForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  rDlgPropSave.SaveProperties;
  inherited;
end;

procedure TrDBSortForm.FormDestroy(Sender: TObject);
begin
  FN.Free;
  FL.Free;
end;

function TrDBSortForm.TrySort: boolean;
const
  FieldSep = ',';
var
  s, d: string;
  i, fidx: Integer;
begin
  s:=''; d:='';
  for i:=1 to grSort.RowCount - 2 do
    if grSort.Cells[1, i] <> '' then
    begin
      fidx:=FL.IndexOf(grSort.Cells[1, i]);
      if fidx >= 0 then
      begin
        if s <> '' then
          s:=s + FieldSep;
        s:=s + FN[fidx];
        if d <> '' then
          d:=d + FieldSep;
        if grSort.Cells[2, i] = '1' then
          d:=d + 'D'
        else
          d:=d + 'A';
      end;
    end;
  Result:=rDBGrid.ChangeSortEx(s, d);
end;

procedure TrDBSortForm.SortList;
var
  i, j, min: Integer;
begin
  for i:=0 to FL.Count - 2 do
  begin
    min:=i;
    for j:=i + 1 to FL.Count - 1 do
    begin
      if AnsiCompareText(FL[j], FL[min]) < 0 then
        min:=j;
    end;
    if min <> i then
    begin
      FL.Move(min, i);
      FN.Move(min, i);
    end;
  end;
end;

procedure TrDBSortForm.RefGrid;
var
  i: Integer;
begin
  for i:=1 to grSort.RowCount - 2 do
    grSort.Cells[0, i]:=IntToStr(i);
  grSort.Cells[0, grSort.RowCount - 1]:='*';
end;

procedure TrDBSortForm.LoadPicklist;
var
  sr: TSearchRec;
begin
  // load list of saved sort files with correct GridID
  lbPickList.Items.Clear;
  if FindFirst(rtool.SetEndOfPath(RosiCompConfig.rDBSortCfg.DefaultPath) + '*.sort', faAnyFile, sr) = 0 then
  begin
    repeat
      if (sr.Attr and faDirectory) = faDirectory then
        Continue;
      PSFile.LoadFile(sr.Name);
      if PSFile.ReadString(Self.Name, sGridID, '') = GridID then
        lbPickList.Items.Add(sr.Name);
    until FindNext(sr) <> 0;
    FindClose(sr);
  end;
  btLoadPick.Enabled:=lbPickList.Items.Count > 0;
end;

procedure TrDBSortForm.LoadSort(FileName: string);
begin
  PSFile.LoadFile(FileName);
  PSStore.ReadProperties;
end;

procedure TrDBSortForm.grSortSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
begin
  if (ARow = grSort.RowCount - 1) and (ACol = 1) and (Trim(grSort.Cells[ACol, ARow]) <> '') then
  begin
    // add new empty row
    grSort.RowCount:=grSort.RowCount + 1;
    RefGrid;
  end;
end;

procedure TrDBSortForm.btUpClick(Sender: TObject);
begin
  if grSort.Row < 2 then
    Exit;
  if grSort.Row < grSort.RowCount - 1 then
  begin
    str.Strings.Assign(grSort.Cols[1]);
    str.Strings.Move(grSort.Row, grSort.Row - 1);
    grSort.Cols[1].Assign(str.Strings);
    str.Strings.Assign(grSort.Cols[2]);
    str.Strings.Move(grSort.Row, grSort.Row - 1);
    grSort.Cols[2].Assign(str.Strings);
  end;
  grSort.Row:=grSort.Row - 1;
end;

procedure TrDBSortForm.btDownClick(Sender: TObject);
begin
  if grSort.Row >= grSort.RowCount - 2 then
    Exit;
  str.Strings.Assign(grSort.Cols[1]);
  str.Strings.Move(grSort.Row, grSort.Row + 1);
  grSort.Cols[1].Assign(str.Strings);
  str.Strings.Assign(grSort.Cols[2]);
  str.Strings.Move(grSort.Row, grSort.Row + 1);
  grSort.Cols[2].Assign(str.Strings);
  grSort.Row:=grSort.Row + 1;
end;

procedure TrDBSortForm.btDelClick(Sender: TObject);
begin
  if grSort.Row = grSort.RowCount - 1 then
    Exit;
  str.Strings.Assign(grSort.Cols[1]);
  str.Strings.Delete(grSort.Row);
  str.Strings.Add('');
  grSort.Cols[1].Assign(str.Strings);
  str.Strings.Assign(grSort.Cols[2]);
  str.Strings.Delete(grSort.Row);
  str.Strings.Add('');
  grSort.Cols[2].Assign(str.Strings);
  if grSort.RowCount > 2 then
    grSort.RowCount:=grSort.RowCount - 1;
  RefGrid;
end;

procedure TrDBSortForm.btDelAllClick(Sender: TObject);
begin
  str.Strings.Clear;
  str.Strings.Add('');
  str.Strings.Add('');
  grSort.Cols[1].Assign(str.Strings);
  grSort.Cols[2].Assign(str.Strings);
  grSort.RowCount:=2;
  RefGrid;
end;

procedure TrDBSortForm.btOkClick(Sender: TObject);
begin
  if TrySort then
    ModalResult:=mrOk
  else
    DlgW(sSortDataErr);
end;

procedure TrDBSortForm.btSaveClick(Sender: TObject);
begin
  if SaveDlg.Execute then
  begin
    DeleteFile(SaveDlg.FileName);
    PSFile.LoadFile(SaveDlg.FileName);
    PSFile.WriteString(Self.Name, sGridID, GridID);
    PSStore.SaveProperties;
    PSFile.SaveFile(SaveDlg.FileName);
  end;
end;

procedure TrDBSortForm.btLoadClick(Sender: TObject);
begin
  if OpenDlg.Execute then
    LoadSort(OpenDlg.FileName);
end;

procedure TrDBSortForm.btLoadPickClick(Sender: TObject);
begin
  lbPickList.Visible:=not lbPickList.Visible;
end;

procedure TrDBSortForm.lbPickListClick(Sender: TObject);
begin
  LoadSort(lbPickList.Items[lbPickList.ItemIndex]);
  lbPickList.Visible:=false;
end;

end.
