unit rDBSort2;

{<
  rDBSort2 provides universal dialog for sorting of any grid.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2011)
  @lastmod(Sep 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, db, Grids, rStringGridEd, rDBGrid, ImgList,
  PropSaveMain, PropSaveGrids, rImprovedComps, rdbdlg_baseform;

{ Show DBSort dialog for grid }
function SortRec(rDBGrid: TrDBGrid): boolean;
{ Load list of stored sorting names }
function SortRec_LoadSortNames(rDBGrid: TrDBGrid; L: TStrings): boolean;
{ Check whether sorting index is active currently }
function SortRec_ActiveSortIdx(rDBGrid: TrDBGrid; Idx: integer): boolean;
{ Check whether SortName is active currently }
function SortRec_ActiveSortName(rDBGrid: TrDBGrid; SortName: string): boolean;

type
  { universal dialog for sorting of any grid }
  TrDBSortForm2 = class(TrDBDlgBaseForm)
    Label1: TLabel;
    grSort: TrStringGridEd;
    str: TrStringsHolder;
    btOk: TrButton;
    btCancel: TrButton;
    GridsPropSave1: TGridsPropSave;
    btSave: TrButton;
    lbSavedSort: TrListBox;
    btUp: TrButton;
    btDown: TrButton;
    btDel: TrButton;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    btDelAll: TrButton;
    PSSaveSort: TPropSaveMain;
    Label2: TLabel;
    btDelSort: TrButton;
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure grSortSetEditText(Sender: TObject; ACol, ARow: integer; const Value: String);
    procedure btUpClick(Sender: TObject);
    procedure btDelClick(Sender: TObject);
    procedure btDownClick(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure btSaveClick(Sender: TObject);
    procedure lbSavedSortClick(Sender: TObject);
    procedure btDelAllClick(Sender: TObject);
    procedure GridsPropSave1GetKeyString(Comp: TComponent; var Key: string);
    procedure PSSaveSortSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure btDelSortClick(Sender: TObject);
  private
    rDBGrid: TrDBGrid;
    SaveIdx: integer;
    SaveSortName: string;
    FN: TStrings;
    FL: TStrings;
    function TrySort: boolean;
    procedure SortList;
    procedure RefGrid;
    procedure LoadSortList;
  public
  end;

var
  rDBSortForm2: TrDBSortForm2;

implementation

{$R *.DFM}
{$I _rosi.inc}
uses RosiCompGlobalCfg, rtool, rdbtool, rstring, rdlg;

const
  sSortDlgKeyId = 'SortDlg2';
  sSortNameID = 'SortName';

function GetSaveSectionId(G: TrDBGrid; Idx: integer): string;
begin
  // Sets key in INI file according Form and DBGrid name
  Result:=sSortDlgKeyId + '.' + G.Owner.Name + '.' + G.Name + '.' + IntToStr(Idx);
end;

function GetSaveName(G: TrDBGrid; Idx: integer): string;
begin
  Result:='';
  if not Assigned(G) or (PropSaveMain.GetGlobal_PropSaveFile=nil) then
    Exit;
  Result:=PropSaveMain.GetGlobal_PropSaveFile.ReadString(GetSaveSectionId(G, Idx), sSortNameID, '');
end;

function SortRec(rDBGrid: TrDBGrid): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(rDBGrid) then
    Exit;
  rDBSortForm2:=TrDBSortForm2.Create(GetActiveOwner);
  try
    rDBSortForm2.rDBGrid:=rDBGrid;
    Result:=rDBSortForm2.ShowModal = mrOk;
  finally
    FreeF(rDBSortForm2);
  end;
end;

function SortRec_LoadSortNames(rDBGrid: TrDBGrid; L: TStrings): boolean;
var
  Idx: integer;
  s: string;
  LS: TStringList;
begin
  Result:=false;
  L.Clear;
  if not Assigned(rDBGrid) or (PropSaveMain.GetGlobal_PropSaveFile=nil) then
    Exit;
  Idx:=0;
  repeat
    s:=GetSaveName(rDBGrid, Idx);
    if s <> '' then
      L.AddObject(s, TObject(Idx));
    Inc(Idx);
  until Idx >= 20;
  LS:=TStringList.Create;
  try
    LS.Assign(L);
    LS.Sort;
    L.Assign(LS);
  finally
    LS.Free;
  end;
  Result:=true;
end;

function SortRec_ActiveSortIdx(rDBGrid: TrDBGrid; Idx: integer): boolean;
var
  fcreated: boolean;
begin
  Result:=false;
  if not Assigned(rDBGrid) or (PropSaveMain.GetGlobal_PropSaveFile=nil) then
    Exit;
  fcreated:=Assigned(rDBSortForm2);
  if not fcreated then
    rDBSortForm2:=TrDBSortForm2.Create(GetActiveOwner);
  rDBSortForm2.rDBGrid:=rDBGrid;
  rDBSortForm2.SaveIdx:=Idx;
  GetFldList(rDBGrid, rDBSortForm2.FN, [fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  GetFldList(rDBGrid, rDBSortForm2.FL, [fdGetLabels, fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  rDBSortForm2.PSSaveSort.ReadProperties;
  Result:=rDBSortForm2.TrySort;
  if not fcreated then
    FreeF(rDBSortForm2);
end;

function SortRec_ActiveSortName(rDBGrid: TrDBGrid; SortName: string): boolean;
var
  Idx: integer;
  L: TStrings;
begin
  Result:=false;
  L:=TStringList.Create;
  try
    SortRec_LoadSortNames(rDBGrid, L);
    Idx:=L.IndexOf(SortName);
    if Idx < 0 then
      Exit;
    Result:=SortRec_ActiveSortIdx(rDBGrid, Idx);
  finally
    L.Free;
  end;
end;

procedure TrDBSortForm2.FormCreate(Sender: TObject);
begin
  inherited;
  FN:=TStringList.Create;
  FL:=TStringList.Create;
  PSSaveSort.PropSaveFile:=PropSaveMain.GetGlobal_PropSaveFile;
  if not Assigned(rDlgPropSave.PropSaveFile) then
  begin
    btSave.Enabled:=false;
    lbSavedSort.Enabled:=false;
  end;
end;

procedure TrDBSortForm2.FormShow(Sender: TObject);
var
  i, fidx: integer;
  SortFieldsArr: TSortFieldNames;
  SortDirsArr: TSortDirections;
  flbl: string;
begin
  inherited;
  rDlgPropSave.ReadProperties;
  grSort.DrawingStyle:=RosiCompConfig.rDBToolCfg.rDlg_GridDrawingStyle;
  // load fields and labels
  GetFldList(rDBGrid, FN, [fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  GetFldList(rDBGrid, FL, [fdGetLabels, fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  if RosiCompConfig.rDBSortCfg.SortFieldList then
    SortList;
  // fill cells
  grSort.Columns[1].PickList.Assign(FL);
  SortFieldsArr:=rDBGrid.SortFieldNamesArr;
  SortDirsArr:=rDBGrid.SortDirectionsArr;
  for i:=0 to High(SortFieldsArr) do
  begin
    grSort.RowCount:=i + 3;
    fidx:=FN.IndexOf(SortFieldsArr[i]);
    if fidx >= 0 then
      flbl:=FL[fidx]
    else
      flbl:=SortFieldsArr[i];
    grSort.Cells[1, i + 1]:=flbl;
    if i <= High(SortDirsArr) then
      if SortDirsArr[i] then
        grSort.Cells[2, i + 1]:='1';
  end;
  RefGrid;
  LoadSortList;
  OpenDlg.Filter:=OpenDlg.Filter + '|' + sFileTypeAll;
  SaveDlg.Filter:=OpenDlg.Filter;
  OpenDlg.InitialDir:=RosiCompConfig.rDBSortCfg.DefaultPath;
  SaveDlg.InitialDir:=OpenDlg.InitialDir;
end;

procedure TrDBSortForm2.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  rDlgPropSave.SaveProperties;
  inherited;
end;

procedure TrDBSortForm2.FormDestroy(Sender: TObject);
begin
  FN.Free;
  FL.Free;
end;

function TrDBSortForm2.TrySort: boolean;
const
  FieldSep = ',';
var
  s, d: string;
  i, fidx: integer;
begin
  s:=''; d:='';
  for i:=1 to grSort.RowCount - 2 do
    if grSort.Cells[1, i] <> '' then
    begin
      fidx:=FL.IndexOf(grSort.Cells[1, i]);
      if fidx >= 0 then
      begin
        if s <> '' then
          s:=s + FieldSep;
        s:=s + FN[fidx];
        if d <> '' then
          d:=d + FieldSep;
        if grSort.Cells[2, i] = '1' then
          d:=d + 'D'
        else
          d:=d + 'A';
      end;
    end;
  Result:=rDBGrid.ChangeSortEx(s, d);
end;

procedure TrDBSortForm2.SortList;
var
  i, j, min: integer;
begin
  for i:=0 to FL.Count - 2 do
  begin
    min:=i;
    for j:=i + 1 to FL.Count - 1 do
    begin
      if AnsiCompareText(FL[j], FL[min]) < 0 then
        min:=j;
    end;
    if min <> i then
    begin
      FL.Move(min, i);
      FN.Move(min, i);
    end;
  end;
end;

procedure TrDBSortForm2.RefGrid;
var
  i: integer;
begin
  for i:=1 to grSort.RowCount - 2 do
    grSort.Cells[0, i]:=IntToStr(i);
  grSort.Cells[0, grSort.RowCount - 1]:='*';
end;

procedure TrDBSortForm2.grSortSetEditText(Sender: TObject; ACol, ARow: integer; const Value: String);
begin
  if (ARow = grSort.RowCount - 1) and (ACol = 1) and (Trim(grSort.Cells[ACol, ARow]) <> '') then
  begin
    // add new empty row
    grSort.RowCount:=grSort.RowCount + 1;
    RefGrid;
  end;
end;

procedure TrDBSortForm2.btUpClick(Sender: TObject);
begin
  if grSort.Row < 2 then
    Exit;
  if grSort.Row < grSort.RowCount - 1 then
  begin
    str.Strings.Assign(grSort.Cols[1]);
    str.Strings.Move(grSort.Row, grSort.Row - 1);
    grSort.Cols[1].Assign(str.Strings);
    str.Strings.Assign(grSort.Cols[2]);
    str.Strings.Move(grSort.Row, grSort.Row - 1);
    grSort.Cols[2].Assign(str.Strings);
  end;
  grSort.Row:=grSort.Row - 1;
end;

procedure TrDBSortForm2.btDownClick(Sender: TObject);
begin
  if grSort.Row >= grSort.RowCount - 2 then
    Exit;
  str.Strings.Assign(grSort.Cols[1]);
  str.Strings.Move(grSort.Row, grSort.Row + 1);
  grSort.Cols[1].Assign(str.Strings);
  str.Strings.Assign(grSort.Cols[2]);
  str.Strings.Move(grSort.Row, grSort.Row + 1);
  grSort.Cols[2].Assign(str.Strings);
  grSort.Row:=grSort.Row + 1;
end;

procedure TrDBSortForm2.btDelClick(Sender: TObject);
begin
  if grSort.Row = grSort.RowCount - 1 then
    Exit;
  str.Strings.Assign(grSort.Cols[1]);
  str.Strings.Delete(grSort.Row);
  str.Strings.Add('');
  grSort.Cols[1].Assign(str.Strings);
  str.Strings.Assign(grSort.Cols[2]);
  str.Strings.Delete(grSort.Row);
  str.Strings.Add('');
  grSort.Cols[2].Assign(str.Strings);
  if grSort.RowCount > 2 then
    grSort.RowCount:=grSort.RowCount - 1;
  RefGrid;
end;

procedure TrDBSortForm2.btDelAllClick(Sender: TObject);
begin
  str.Strings.Clear;
  str.Strings.Add('');
  str.Strings.Add('');
  grSort.Cols[1].Assign(str.Strings);
  grSort.Cols[2].Assign(str.Strings);
  grSort.RowCount:=2;
  RefGrid;
end;

procedure TrDBSortForm2.btOkClick(Sender: TObject);
begin
  if TrySort then
    ModalResult:=mrOk
  else
    DlgW(sSortDataErr);
end;

procedure TrDBSortForm2.LoadSortList;
begin
  SortRec_LoadSortNames(rDBGrid, lbSavedSort.Items);
  btDelSort.Enabled:=(lbSavedSort.Items.Count > 0) and (lbSavedSort.ItemIndex >= 0);
end;

procedure TrDBSortForm2.GridsPropSave1GetKeyString(Comp: TComponent; var Key: string);
begin
  Key:=GetSaveSectionId(rDBGrid, SaveIdx);
end;

procedure TrDBSortForm2.PSSaveSortSaveProperties(PSFile: TPropSaveFile; Key: string);
begin
  PSFile.WriteString(GetSaveSectionId(rDBGrid, SaveIdx), sSortNameID, SaveSortName);
end;

procedure TrDBSortForm2.btSaveClick(Sender: TObject);
var
  s: string;
  r: integer;
begin
  s:='';
  for r:=1 to grSort.RowCount - 1 do
    if grSort.Cells[1, r] <> '' then
    begin
      if s <> '' then
        s:=s + ', ';
      s:=s + grSort.Cells[1, r];
    end;

  if lbSavedSort.ItemIndex >= 0 then
    s:=lbSavedSort.Items[lbSavedSort.ItemIndex];
  if not rdlg.InputStr(sInput, sInputItemName, s) then
    Exit;
  if s = '' then
    Exit;
  SaveSortName:=s;
  SaveIdx:=lbSavedSort.Items.IndexOf(s);
  if SaveIdx < 0 then
  begin
    SaveIdx:=-1;
    repeat
      Inc(SaveIdx);
      s:=GetSaveName(rDBGrid, SaveIdx);
    until s = '';
  end;
  PSSaveSort.SaveProperties;
  LoadSortList;
  SaveIdx:=lbSavedSort.Items.IndexOf(SaveSortName);
  if SaveIdx >= 0 then
    lbSavedSort.ItemIndex:=SaveIdx;
  btDelSort.Enabled:=(lbSavedSort.Items.Count > 0) and (lbSavedSort.ItemIndex >= 0);
end;

procedure TrDBSortForm2.lbSavedSortClick(Sender: TObject);
var
  i: integer;
begin
  i:=lbSavedSort.ItemIndex;
  if i < 0 then
    Exit;
  SaveIdx:=integer(lbSavedSort.Items.Objects[i]);
  PSSaveSort.ReadProperties;
  btDelSort.Enabled:=(lbSavedSort.Items.Count > 0) and (lbSavedSort.ItemIndex >= 0);
end;

procedure TrDBSortForm2.btDelSortClick(Sender: TObject);
var
  i: integer;
  s: string;
begin
  i:=lbSavedSort.ItemIndex;
  if i < 0 then
    Exit;
  SaveIdx:=integer(lbSavedSort.Items.Objects[i]);
  s:=GetSaveSectionId(rDBGrid, SaveIdx);
  PSSaveSort.PropSaveFile.EraseSection(s);
  PSSaveSort.PropSaveFile.EraseSection(s + '.Column0');
  PSSaveSort.PropSaveFile.EraseSection(s + '.Column1');
  PSSaveSort.PropSaveFile.EraseSection(s + '.Column2');
  LoadSortList;
end;

end.
