unit rdbdlg_baseform;

{<
  base form for all DB shared dialogs

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2015)
  @lastmod(Sep 2015)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, Buttons, ExtCtrls, ImgList, PropSaveMain,
  rImprovedComps, db;

type
  { base form for all DB shared dialogs }
  TrDBDlgBaseForm = class(TForm)
    rDlgImgEna: TrImageListEx;
    rDlgPropSave: TPropSaveMain;
    rDlgImgDis: TrImageListEx;
    procedure FormCreate(Sender: TObject);
    function FormHelp1(Command: Word; Data: Integer; var CallHelp: boolean): boolean;
    function FormHelp2(Command: Word; Data: NativeInt; var CallHelp: Boolean): Boolean;
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormActivate(Sender: TObject);
  private
  public
  end;

var
  rDBDlgBaseForm: TrDBDlgBaseForm;

function IsLinkedAndActive(const GridOrDataSet: TObject): boolean;

implementation

{$R *.dfm}
uses RosiCompGlobalCfg, rtool, rdbtool;

procedure TrDBDlgBaseForm.FormCreate(Sender: TObject);
begin
  Font.Name:=RosiCompConfig.rDBToolCfg.rDlg_DefaultFont;
  Font.Size:=RosiCompConfig.rDBToolCfg.rDlg_DefaultSize;
  DPI_SetFontHeight(Font);

  rDlgPropSave.PropSaveFile:=PropSaveMain.GetGlobal_PropSaveFile;
  if rDlgPropSave.FormIsSaved then
    Position:=poDesigned;
  if Assigned(RosiCompConfig.Global_ProcOnHelp) then
  begin
    {$IFDEF DXE2P}
    self.OnHelp:=FormHelp2;
    {$ELSE}
    self.OnHelp:=FormHelp1;
    {$ENDIF}
    BorderIcons:=BorderIcons + [biHelp];
    HelpContext:=1;
  end;
end;

procedure TrDBDlgBaseForm.FormShow(Sender: TObject);
begin
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBForms_FormShowEventID);
end;

procedure TrDBDlgBaseForm.FormActivate(Sender: TObject);
begin
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBForms_FormActivateEventID);
end;

procedure TrDBDlgBaseForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBForms_FormCloseEventID);
end;

function TrDBDlgBaseForm.FormHelp1(Command: Word; Data: Integer; var CallHelp: boolean): boolean;
begin
  CallHelp:=false;
  Result:=true;
  if (Command = HELP_CONTEXTPOPUP) then
    RosiCompConfig.DoProcOnHelp(Self, Data);
end;

function TrDBDlgBaseForm.FormHelp2(Command: Word; Data: NativeInt;
  var CallHelp: Boolean): Boolean;
begin
  CallHelp:=false;
  Result:=true;
  if (Command = HELP_CONTEXTPOPUP) then
    RosiCompConfig.DoProcOnHelp(Self, Data);
end;

function IsLinkedAndActive(const GridOrDataSet: TObject): boolean;
var
  ds: TDataSet;
begin
  ds:=GetDataSet(GridOrDataSet);
  Result:=Assigned(ds) and ds.Active;
end;

end.
